// SP62: a program for the formation of multiple alignments in the ICMAUS framework.

// This program is released to the public domain without any restrictions.
// You can redistribute it and/or modify it as you wish.

// In any publication that refers to any version of this program, please
// acknowledge J G Wolff as the originator of the program. 

// A description of this program may be found in "Syntax, parsing and
// production of natural language in a framework of information compression
// by multiple alignment, unification and search", J G Wolff, Journal of Universal
// Computer Science 6(8), 781-829, 2000. See also
// www.cognitionresearch.org.uk/sp.htm.

// Please retain this notice on all versions.

// Dr J G Wolff, www.CognitionResearch.org.uk. Email: jgw@cognitionresearch.org.uk.

// November 2005.

#define COPYRIGHT_NOTICE "Copyright (C) 2005 J G Wolff"

#define PROGRAM "SP62"

#define VERSION "3.1"

/*
**
** ABSTRACT:	A program for the formation of multiple alignments in the
**				SP framework.
**
** ENVIRONMENT: The module requires the header file "SP62_head.h" 
**				to be on the `include' path.  It also makes use of 
**				routines defined in "SP62_lib.cpp"
**
** AUTHOR:		Gerry Wolff
**
** MODIFIED BY:
**
** 3.0 JGW 11/8/05 -	From SP62a_comp.cpp, v 1.2. A switch is 
**						added to determine whether or not
**						alignments can be combined to make bigger
**						alignments.
** 3.1 JGW 10/11/05 -	A parameter is added to determine whether two
**						or more New patterns are to be processed as
**						one batch or as a succession of batches.
**						[2005-11-25] sequence::write_alignment_vertical()
**						has been completely re-written. The date and
**						time in the output file is given in ISO 8601
**						format
*/

#include "SP62_head.h"

#define MARK_UNIFY_BEST FALSE // Switches on or off print routines which
	// mark the start and end of each call to unify_best.
#define PRINT_CD_VALUES FALSE
#define MAX_DR_PATTS_ALIGNMENT 10

char *in_filename ;
char *parameters_filename ;
char *out_filename ;
char *latex_filename ;
FILE *input_file ;
FILE *parameters_file ;
FILE *output_file ;
FILE *latex_file ;

group *corpus ;
group *new_patterns ; // New patterns supplied at the start of processing
group *old_patterns ; // Old patterns supplied at the start of processing.
	// In 'learning' mode, there may not be any patterns in old_patterns.
group *original_symbols_in_corpus ;
group *symbols_in_old_after_learning ;
group *current_alignments_from_parsing ; // This will hold newly-formed 
	// alignments from the current cycle.
group *cumulative_alignments_from_parsing ;
group *driving_alignments ;
group *created_code_patterns ;
group *symbols_in_new ; // A list of the types of symbols (symbols with
	// a given name) found in new_patterns.
group *set_of_combinations ;
group *best_combinations ;
sequence *current_new_pattern ;
hit_node *hit_root ; /* A pointer to the root of the hit structure */
hit_node *hn_master ; /* Used in add_hit(). */
int original_alphabet_size ; // The size of the alphabet of symbols 
	// used in the original corpus.
int current_alphabet_size ; // The size of the alphabet after learning
	// has been done, including the creation of new code symbols.
int group_ID_number = 1 ;
int sequence_ID_number = 1 ;
int hit_node_ID_number = 1 ;
int start_int_pos ;
int stop_int_pos ;
int number_of_cnp_symbols ;
int window ;
int combination_ID_number = 1 ;
int number_of_New_patterns ;

/*** ARRAYS, CONSTANTS AND INDICES FOR STORING SYMBOL COUNTS FOR
EACH SYMBOL IN NEW AND FOR STORING THE HIT STRUCTURE ***/

/*** STRUCTS, ARRAYS, CONSTANTS AND INDICES ***/

struct pnc_entry
{
	sequence *al1 ; // The alignment from which the pnc comes.
	sequence *pnc ; // A 2-row alignment which represents the pnc.
	int next_set ; // An index pointer to the next set of pncs.
	int next_pnc ; // An index pointer to the next pnc in
				// given set.
} ; // Used in probabilities_of_inferences

hit_node **leaf_array ; /* leaf_array[] is used for recording the
	leaf nodes of the hit structure. */

int *sort_array ; /* sort_array[] is used for sorting the
	entries in leaf_array[]. */

int fe_sort ; /* The first empty entry in sort_array[]. */

sequence **alignments_array ;

struct sub_alignment_entry *sub_alignments_array ; // Dynamic array
	// used in combine_sub_alignments.

int fe_sub_alignments_array ; // Used in combine_sub_alignments.

sequence *original_patterns_array[MEDIUM_SCRATCH_ARRAY_SIZE] ;

int half_hit_structure_rows ;

struct counts_record
{
	sequence *pattern1 ;
	int frequency, max_cioa, count_in_one_alignment ;
} ;

struct counts_record *counts_array ;
int size_of_old_patterns ;

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compress the patterns in New in terms of patterns in Old (if any).
**
** CALLING SEQUENCE:
**
**	void compress_new_into_old()
**
** FORMAL ARGUMENTS:
**
**	Return value:			void
**
** IMPLICIT INPUTS:
**
**	The corpus
**
** IMPLICIT OUTPUTS:
**
**	Alignments and new patterns
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void compress_new_into_old()
{
	sequence *best_alignment = NIL, *al1 ;
	int counter ;

	current_number_of_alignments = 0 ;
	new_pattern_number = 0 ;

	list_for(current_new_pattern, sequence, new_patterns)
	{
		new_pattern_number++ ;

		if (true) // audit_trail
		{
			fprintf(output_file, "NEXT PATTERN FROM NEW:\n\n") ;
			current_new_pattern->print_ID() ;
			fprintf(output_file, " ") ;
			current_new_pattern->write_tree_object(PRINT_NO_FREQUENCIES) ;
		}

		number_of_cnp_symbols =
			current_new_pattern->count_number_of_symbols() ;
		start_int_pos = 0 ;
		
		window = 1 ; // Counts the number of windows into the
			// current_new_pattern ;

		stop_int_pos = window_size ; // stop_int_pos is the limiting
			// position of symbols in the current_new_pattern for
			// the current window.

		while (start_int_pos < number_of_cnp_symbols)
		{
			cycle = 1 ; // This counter counts the number of
				// cycles within the 'recognise()' function
				// for this current_new_pattern and this window
				// within that pattern.

			if (audit_trail)
			{
				fprintf(output_file, "CURRENT ALIGNMENTS FROM PARSING:\n\n") ;
				current_alignments_from_parsing->write_IDs() ;
			}

			// Now try to 'recognise' the New pattern.

			while (recognise()) ;

			set_of_combinations->delete_children() ;

			if (cumulative_alignments_from_parsing->is_empty())
			{
				if (audit_trail)
				{
					fprintf(output_file, 
						"BEST ALIGNMENT FOUND SINCE START: None\n\n") ;		
					fprintf(output_file, "END OF ") ;
					print_pattern_window_cycle() ;
					fprintf(output_file, "\n\n") ;
				}

				best_alignment = NIL ;

				goto L1 ; 
			}
					
			cumulative_alignments_from_parsing->sort_by_compression_difference() ;

			best_alignment = (sequence *)
				cumulative_alignments_from_parsing->get_first_child() ;

			if (audit_trail)
			{
				fprintf(output_file, "%s%d%s",
					"BEST ALIGNMENTS FOUND UP TO ",
					best_alignment_print_limit,
					" (") ;
				print_pattern_window_cycle() ;
				fprintf(output_file, ")\n\n") ;

				counter = 0 ;
				list_for(al1, sequence, cumulative_alignments_from_parsing)
				{
					counter++ ;
					al1->write_out_fully(NIL, NIL, NIL,
						write_section_chars_length, NULL_VALUE, false) ;
					if (counter >= best_alignment_print_limit) break ;
				}

				fprintf(output_file, "END OF (") ;
				print_pattern_window_cycle() ;
				fprintf(output_file, ")\n\n") ;
			}

			L1: window++ ;
			start_int_pos = stop_int_pos ;
			stop_int_pos += window_size ;
		} // The next window should be taken automatically.

		cumulative_alignments_from_parsing->sort_by_compression_difference() ;

		if (only_one_new_pattern_in_each_alignment == false
			&& number_of_New_patterns > 1)
		{
			fprintf(output_file, "%s%d%s%d%s%d%s",
				"IN THIS ONE BATCH OF ",
				number_of_New_patterns,
				" NEW PATTERNS,\nTHE BEST ALIGNMENTS FOUND UP TO ",
				best_alignment_print_limit,
				" FOR THE NEW PATTERN ID",
				current_new_pattern->get_ID(),
				" ARE:\n\n") ;
		}
		else
		{
			fprintf(output_file, "%s%d%s%d%s",
				"THE BEST ALIGNMENTS FOUND UP TO ",
				best_alignment_print_limit,
				" FOR THE NEW PATTERN ID",
				current_new_pattern->get_ID(),
				" ARE:\n\n") ;
		}
				
		counter = 0 ;
		list_for(al1, sequence, cumulative_alignments_from_parsing)
		{
			if (match_all_new_symbols)
			{
				if (al1->all_new_symbols_matched(current_new_pattern) == false)
					continue ;
			}

			if (only_one_new_pattern_in_each_alignment)
			{
				if (al1->number_of_new_patterns_in_alignment() > 1)
					continue ;
			}

			fprintf(output_file, "ALIGNMENT:\n\n") ;
			counter++ ;
			al1->write_out_fully(NIL, NIL, NIL,
				write_section_chars_length, NULL_VALUE, false) ;
			if (counter >= best_alignment_print_limit) break ;
		}

		// If only_one_new_pattern_in_each_alignment is false, then the 
		// alignments in cumulative_alignments_from_parsing are
		// retained and carried forward to the processing of the
		// next New pattern (and subsequent New patterns). If
		// only_one_new_pattern_in_each_alignment is true, then the alignments
		// in cumulative_alignments_from_parsing are deleted and
		// processing starts afresh with the next New pattern.

		if (only_one_new_pattern_in_each_alignment == false) continue ;

		cumulative_alignments_from_parsing->delete_children() ;

	} // Loop: taking successive patterns from New. Get next 
		// current_new_pattern from New.

	fprintf(output_file, "END OF PROCESSING\n\n") ;

	/* Calculate probabilities of inferences from the best
	alignment and all other alignments which encode
	the same symbols from New. */
	
	if (probabilities)
	{
		if (best_alignment == NIL)
		{
			if (audit_trail)
			{
				fprintf(output_file,
					"NIL best_alignment - no probabilities calculated.\n\n") ;
			}
			return ;
		}
		
		probabilities_of_inferences(best_alignment) ;
	}

} // compress_new_into_old

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Assigns status and types to symbols in corpus, compiles an alphabet
**	of symbols in the corpus and prints out its size, compiles a list
**	of classes, finds frequencies of symbols, computes corresponding
**	information costs and assigns these values to symbols in the
**	corpus.
**
** CALLING SEQUENCE:
**
**	void process_symbols()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	The corpus, original_alphabet_size
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void process_symbols()
{
	if (!identification_symbols_marked)
		assign_status_to_symbols() ;
	else
	{
		// Make sure that brackets at level 0 have
		// BOUNDARY_MARKER status.

		sequence *pattern1 ;

		list_for(pattern1, sequence, old_patterns)
			pattern1->assign_brackets_status() ;
	}

	assign_type_to_symbols() ;

	original_symbols_in_corpus = 
		corpus->compile_alphabet(&original_alphabet_size) ; 
		// This function needs to follow the previous one to ensure 
		// that type and status information is on each symbol in 
		// the alphabet.

	if (audit_trail)
	{
		fprintf(output_file, "%s%d%s",
			"Original alphabet size = ",
			original_alphabet_size,
			"\n\n") ;
	}

	find_symbol_frequencies(original_symbols_in_corpus, 
		true, true) ; /* Find the frequency of
		each symbol type by scanning over the patterns in the corpus
		and adding the frequency of the pattern to the frequency of the
		symbol type for each occurrence of each symbol in the
		pattern in which it occurs. */
	
	find_symbol_type_information_costs(original_symbols_in_corpus) ; 
		/* From the frequencies of symbol types this
		function computes an information cost for each symbol type 
		in the corpus. Calculates average_symbol_type_cost. 
		In addition, the function assigns the frequency of its type 
		to each symbol in the corpus and the associated cost, 
		calculated by the Shannon-Fano-Elias method. */
		
} // process_symbols

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	If XML-style brackets are used, check that they are used in a legal
**	manner. This means that the name of each LEFT_XML_BRACKET must start with '<'
**	and end with '>' and there must not be any other bracket symbols in between; that
**	the name of each RIGHT_XML_BRACKET must start with '</' and end with '>'and there
**	must be no instances of '<', '/' or '>' in between; that, within each pattern,
**	every left bracket must be paired with a right bracket; and that between every pair of brackets,
**	there must either be no other brackets or, if there are, they must be balanced
**	pairs.
**
** CALLING SEQUENCE:
**
**	bool check_xml_brackets()
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if brackets are valid as described above, false otherwise.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

bool check_xml_brackets()
{
	if (recognise_xml_brackets == false) return(true) ;

	sequence *pattern1 ;

	list_for(pattern1, sequence, new_patterns)
		if (pattern1->check_xml_brackets() == false) return(false) ;

	list_for(pattern1, sequence, old_patterns)
		if (pattern1->check_xml_brackets() == false) return(false) ;

	if (audit_trail)
	{
		fprintf(output_file, "XML-style brackets in input are OK.\n\n") ;
	}

	return(true) ;
} // check_xml_brackets

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Parsing of New in terms of Old with setting up and winding down.
**
** CALLING SEQUENCE:
**
**	void set_up_and_parsing()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void set_up_and_parsing()
{
	// Preliminaries.

	// Set up arrays and clear them.

	sort_array = new int[hit_structure_rows] ;
	leaf_array = new hit_node*[hit_structure_rows] ;
	half_list = new hit_node*[half_hit_structure_rows] ;
	alignments_array = new sequence*[hit_structure_rows] ;

	int index, row1, col1 ;

	for (index = 0; index < hit_structure_rows; index++)
		sort_array[index] = NULL_VALUE ;

	for (index = 0; index < hit_structure_rows; index++)
		leaf_array[index] = NIL ;

	for (index = 0; index < half_hit_structure_rows;index++)
		half_list[index] = NIL ;

	for (index = 0; index < hit_structure_rows; index++)
		alignments_array[index] = NIL ;

	// Clear hit_node_array[].
	
	for (index = 0; index < HIT_NODE_ARRAY_SIZE; index++)
	{
		hit_node_array[index].new_hit_node = NIL ;
		hit_node_array[index].hn_parent = NIL ;
	}
	
	// Clear symbol_selection_array[][].
	
	for (row1 = 0; row1 < MAX_NEW_ARRAY_ROWS; row1++)
	{
		for (col1 = 0; col1 < MCOLS; col1++)
			symbol_selection_array[row1][col1] = NIL ;
	}

	// Set up a sequence containing all the symbols in all the New patterns
	// (for use in make_selection()).

	symbol *symbol1 ;
	sequence *pattern1 ;
	length_of_concatenated_new_patterns = 0 ;
	list_for(pattern1, sequence, new_patterns)
	{
		list_for(symbol1, symbol, pattern1)
		{
			concatenated_new_patterns->add_child(symbol1) ;
			length_of_concatenated_new_patterns++ ;
		}
	}

	// Make counts.

	number_of_symbols_in_new =
		new_patterns->count_number_of_symbols() ; // Count the
			// number of symbols in New.

	number_of_symbols_in_old =
		old_patterns->count_number_of_symbols() ; // Count the
			// number of symbols in Old.

	// Find the numbers of patterns in New and Old.
	
	sequence *sequence1 ;

	number_of_patterns_in_new = 0 ;

	list_for(sequence1, sequence, new_patterns)
		number_of_patterns_in_new++ ;
	
	number_of_patterns_in_old = 0 ;

	list_for(sequence1, sequence, old_patterns)
		number_of_patterns_in_old++ ;
	
	// Print out the corpus.
	
	if (true) // audit_trail is alternative test.
	{
		fprintf(output_file,
			"THE WHOLE CORPUS (INCL. FREQUENCIES GREATER THAN ONE)\n\n") ;
		
		if (new_patterns->is_empty())
			fprintf(output_file, "NEW PATTERNS: None\n\n") ;
		else 
		{
			fprintf(output_file, "NEW PATTERNS:\n\n") ;
			list_for(sequence1, sequence, new_patterns)
			{
				sequence1->print_ID() ;
				fprintf(output_file, " ") ;
				sequence1->write_tree_object(PRINT_ALL_FREQUENCIES) ;
			}
		}

		if (old_patterns->is_empty())
			fprintf(output_file, "OLD PATTERNS: None\n\n") ;
		else 
		{
			fprintf(output_file, "OLD PATTERNS:\n\n") ;
			list_for(sequence1, sequence, old_patterns)
			{
				sequence1->print_ID() ;
				fprintf(output_file, " ") ;
				sequence1->write_tree_object(PRINT_ALL_FREQUENCIES) ;
			}
		}
		fflush(output_file) ;
	}

	process_symbols() ;

	// Compute the size of the corpus in bits.

	double size_of_corpus_at_start = corpus->total_of_symbol_sizes() ;

	if (audit_trail)
	{
		fprintf(output_file, "%s%1.2g%s",
			"Size of corpus at start: ",
			size_of_corpus_at_start,
			"\n\n") ;
	}
	
	// Now find EC for patterns in Old.
		
	list_for(sequence1, sequence, old_patterns)
		sequence1->compute_costs() ;
		
	// Notice that the following block must come after the function
	// assign_status_to_symbols() (called in process_symbols(), above)
	// because the former depends on the original order of the symbols.
		
	compute_scaling_factors() ; /* Computes the factors to be applied
		to reduce the effective info cost of a hit symbol, depending
		on the gap between the given hit and the previous hit. */

	// Compile an index of left and right brackets and associated ID-symbols
	// if any brackets are used in old_patterns.

	if (recognise_brackets)
	{
		check_brackets(new_patterns) ;
		check_brackets(old_patterns) ;
	}
		
	check_xml_brackets() ;

	if (fe_key_array == 0 && use_abbreviations != false) // fe_key_array
		// is used to detect whether a key to the abbreviations 
		// of symbols has been supplied or not.
	{
		if (use_abbreviations == true)
			make_letter_abbreviations() ;
		
		// if (use_abbreviations == DIGITS) 
		//	make_digit_abbreviations() ; Development of the
		// use of digit abbreviations has been halted.
	}

	// if (true) // audit_trail.
	//	write_new_and_old() ;
		
	if (audit_trail) write_new_and_old() ;
		
	/* Compress the corpus. */

	compress_new_into_old() ;

	sequence *alignment ;
	while(alignment = (sequence *)current_alignments_from_parsing->
			extract_first_child())
		delete alignment ;

	/* Now write out statistics */
	
	if (audit_trail)
	{
		fprintf(output_file, "STATISTICS\n\n") ;
		fprintf(output_file, "%s%d%s%d%s%d%s%d%s",
			"Original size of alphabet = ", original_alphabet_size,
			"\n\nCurrent size of alphabet = ", current_alphabet_size,
			"\n\nNumber of original patterns in Old = ",
			number_of_patterns_in_old,
			"\n\nFinal number of alignments = ",
			current_number_of_alignments,
			"\n\n") ;
	}

	delete sort_array ;
	delete leaf_array ;
	delete half_list ;
	delete alignments_array ;

} // set_up_and_parsing

/*********************************************************************/

/* The main part of SP */

/*
** FUNCTIONAL DESCRIPTION:
**
**	Entry point to the SP system.
**
** CALLING SEQUENCE:
**
**	void SP_main()	  
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	'output_file', 'latex_file'
**
** SIDE EFFECTS:
**
**	If the session logfile cannot be successfully opened, the function
**	exits with an error code.
**	As described in `show_menu', the order of options displayed in
**	that function is implied here.	If the order in `show_menu' were
**	to change then the `switch (selection)' case labels would have to
**	be adjusted accordingly. Selecting menu option 1 (edit) has no effect. 
*/

void SP_main()
{
	int i ;
	char yr[5] ;
	
	if ((input_file = fopen(in_filename, "r")) == NIL) 
		abort_run("Fatal: Unable to open input_file\n") ;
	if ((parameters_file = fopen(parameters_filename, "r")) == NIL)
		abort_run("Unable to open parameters file.\n");
	if ((output_file = fopen(out_filename, "w")) == NIL)
		abort_run("Unable to open write file.\n");
	if ((latex_file = fopen(latex_filename, "w")) == NIL) 
		abort_run("Fatal: Unable to open latex_file\n") ;

	if (true) // audit_trail
	{
		year(yr) ;
		fprintf(output_file, "%s%s%s%s%s%s%s%s%s%s",
			"SP62 version ", 
			VERSION,
			". Copyright (C) ",
			yr,
			" J G Wolff.\n",
			"This program is released to the public domain without\n",
			"any restrictions. You can redistribute it and/or modify\n",
			"it as you wish. Please retain this notice on all\n",
			"versions and acknowledge J G Wolff as the originator\n",
			"of the program in all publications.\n\n") ;

		fprintf(output_file, "%s%s%s%s%s",
			"RESULTS FROM ", PROGRAM, ", VERSION ", VERSION, "\n\n") ;
		
		time_now() ;
		
		// fprintf(output_file, "%s%s%s",
		//	"INPUT FILE: ", in_filename, "\n\n") ;
	}

	// Read parameters.

	read_parameters() ;
	
	half_hit_structure_rows = hit_structure_rows / 2 ;

	/* if (probabilities == true)
	{
		fprintf(output_file,
			"PROBABILITIES parameter changed from ON to OFF.\n\n") ;
		probabilities = false ;
	} */

	// Print out parameters.

	char *str_value ;
	
	fprintf(output_file, "PARAMETERS:\n\n") ;

	str_value = convert(NULL_VALUE, probabilities) ;
	fprintf(output_file, "%s%s%s", 
		"PROBABILITIES = ", str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, show_al_structure) ;
	fprintf(output_file, "%s%s%s", 
		"SHOW_AL_STRUCTURE = ", str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, show_hit_structure) ;
	fprintf(output_file, "%s%s%s", 
		"SHOW_HIT_STRUCTURE = ", str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, show_unselected_alignments) ;
	fprintf(output_file, "%s%s%s", 
		"SHOW_UNSELECTED_ALIGNMENTS = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, verbose) ;
	fprintf(output_file, "%s%s%s", 
		"VERBOSE = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, use_abbreviations) ;
	fprintf(output_file, "%s%s%s", 
		"USE_ABBREVIATIONS = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, show_all_parsing_alignments) ;
	fprintf(output_file, "%s%s%s", 
		"SHOW_ALL_PARSING_ALIGNMENTS = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, id_c_symbol_constraint) ;
	fprintf(output_file, "%s%s%s",
		"ID_C_SYMBOL_CONSTRAINT = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, recognise_brackets) ;
	fprintf(output_file, "%s%s%s",
		"RECOGNISE_BRACKETS = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, recognise_xml_brackets) ;
	fprintf(output_file, "%s%s%s",
		"RECOGNISE_XML_BRACKETS = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, brackets_id_symbol_constraint) ;
	fprintf(output_file, "%s%s%s",
		"BRACKETS_ID_SYMBOL_CONSTRAINT = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, match_all_data_symbols) ;
	fprintf(output_file, "%s%s%s",
		"MATCH_ALL_DATA_SYMBOLS = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE,
		match_all_data_symbols_in_composite_alignments) ;
	fprintf(output_file, "%s%s%s",
		"MATCH_ALL_DATA_SYMBOLS_IN_COMPOSITE_ALIGNMENTS = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE,
		match_all_data_symbols_in_hierarchical_alignments) ;
	fprintf(output_file, "%s%s%s",
		"MATCH_ALL_DATA_SYMBOLS_IN_HIERARCHICAL_ALIGNMENTS = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, match_all_new_symbols) ;
	fprintf(output_file, "%s%s%s",
		"MATCH_ALL_NEW_SYMBOLS = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE, combine_alignments) ;
	fprintf(output_file, "%s%s%s",
		"COMBINE_ALIGNMENTS = ",
		str_value, "\n\n") ;
	str_value = convert(NULL_VALUE,
		only_one_new_pattern_in_each_alignment) ;
	fprintf(output_file, "%s%s%s",
		"ONLY_ONE_NEW_PATTERN_IN_EACH_ALIGNMENT = ",
		str_value, "\n\n") ;
	str_value = convert(alignment_format, true) ;
	fprintf(output_file, "%s%s%s", 
		"ALIGNMENT_FORMAT = ",
		str_value, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"HIT_STRUCTURE_ROWS = ", hit_structure_rows, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"FAIL_SCORE = ", fail_score, "\n\n") ;
	fprintf(output_file, "%s%1.2f%s", 
		"COST_FACTOR = ", cost_factor, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"KEEP_ROWS = ", keep_rows, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"DRIVING_KEEP_ROWS = ",
		driving_keep_rows, "\n\n") ;
	fprintf(output_file, "%s%d%s",
		"BEST_ALIGNMENT_PRINT_LIMIT = ",
		best_alignment_print_limit, "\n\n") ;
	fprintf(output_file, "%s%s%s", 
		"FIGURE_ID = ",
		figure_ID, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"MAX_NEW_GAP = ", 
		max_new_gap, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"MAX_OLD_GAP = ", 
		max_old_gap, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"MAX_UNSUPPORTED_CYCLES = ",
		max_unsupported_cycles, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"MAX_ALIGNMENTS_IN_ONE_CYCLE = ",
		max_alignments_in_one_cycle, "\n\n") ;
	str_value = convert(orientation, true) ;
	fprintf(output_file, "%s%s%s", 
		"ORIENTATION = ",
		str_value, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"WRITE_AL_CHARS_LENGTH = ",
		write_al_chars_length, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"WRITE_SECTION_CHARS_LENGTH = ",
		write_section_chars_length, "\n\n") ;
	fprintf(output_file, "%s%1.2f%s", 
		"MINIMUM_FONT_SIZE = ",
		minimum_font_height, "\n\n") ;
	fprintf(output_file, "%s%1.2f%s", 
		"MAXIMUM_FONT_SIZE = ",
		maximum_font_height, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"WINDOW_SIZE = ",
		window_size, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"COMBINATION_LIMIT = ", 
		combination_limit, "\n\n") ;
	str_value = convert(NULL_VALUE, audit_trail) ;
	fprintf(output_file, "%s%s%s", 
		"AUDIT_TRAIL = ", 
		str_value, "\n\n") ;

	/* Load SP structure */
	
	load() ;
	
	/* The first object in the corpus is taken to be 'New' */
	
	corpus->initialise() ;
	new_patterns = (group *)corpus->get_next_child() ;
	if (new_patterns == NIL) abort_run("main(): invalid New object") ;
	
	old_patterns = (group *)corpus->get_next_child() ;
	if (old_patterns == NIL) abort_run("main(): invalid Old object") ;

	// Count number of New patterns.

	number_of_New_patterns =
		new_patterns->count_number_of_children() ;

	// Internal constants not set in input file.
	
	fprintf(output_file, "%s%d%s", 
		"MAX_ALIGNMENT_LENGTH = ",
		max_alignment_length, "\n\n") ;
	
	fflush(output_file) ;
	
	// Make global objects and arrays.
		
	hn_master = new hit_node ;
	hit_root = new hit_node ;
	current_alignments_from_parsing = new group ;
	cumulative_alignments_from_parsing = new group ;
	driving_alignments = new group ;
	created_code_patterns = new group ;
	symbols_in_new = new_patterns->symbol_set() ;
	concatenated_new_patterns = new sequence ;
	set_of_combinations = new group ;
	best_combinations = new group ;
	used_driving_patterns = new group ;
	hit_sequences = new group ;
	new_hit_sequences = new group ;

	write_al_chars = new char*[MAX_ALIGNMENT_DEPTH] ;
	write_al_vacant_slots = new char*[MAX_ALIGNMENT_DEPTH] ;

	for (i = 0; i < MAX_ALIGNMENT_DEPTH; i++)
	{
		write_al_chars[i] = new char[write_al_chars_length] ;
		write_al_vacant_slots[i] = new char[write_al_chars_length] ;
	}

	set_up_and_parsing() ;
		
	exit_routine(0) ;
} /* main */