// SP62: a program for the formation of multiple alignments in the ICMAUS framework.

// This program is released to the public domain without any restrictions.
// You can redistribute it and/or modify it as you wish.

// In any publication that refers to any version of this program, please
// acknowledge J G Wolff as the originator of the program. 

// A description of this program may be found in "Syntax, parsing and
// production of natural language in a framework of information compression
// by multiple alignment, unification and search", J G Wolff, Journal of Universal
// Computer Science 6(8), 781-829, 2000. See also
// www.cognitionresearch.org.uk/sp.htm.

// Please retain this notice on all versions.

// Dr J G Wolff, www.CognitionResearch.org.uk. Email: jgw@cognitionresearch.org.uk.

// November 2005.

/*
** TITLE:		SP62_lib.cpp
**
** VERSION: 	SP62, v 3.1
**
** ABSTRACT:	General purpose methods and functions for the SP system.
**
** ENVIRONMENT: The file "SP62_head.h" must be on the `include' 
**				path at compile time.
**
** AUTHOR:		Gerry Wolff
**
** MODIFIED BY:
**
** 3.0 JGW 11/8/05 -	From SP62a_lib.cpp, v 1.2.
** 3.1 JGW 10/11/05 -	See SP62_comp.cpp.
*/

#include "SP62_head.h"

/* PARAMETERS. */

/* Optional, ON or OFF. */

bool probabilities ; /* Calculations of probabilities at the end
	of each program run. */
bool show_al_structure ; // If true, the structure of each column of each
	// alignment is printed out, otherwise they are not printed.
bool show_hit_structure ; // If true, the hit structure is printed,
	// otherwise it is not printed.
bool show_unselected_alignments ; /* If ON, all the alignments which are
	formed are shown before selection (as well as showing the
	alignments which are selected. */
bool verbose ; // If ON, lots of messages are printed. Otherwise,
	// minimal messages are printed. AUDIT_TRAIL must be ON.
bool use_abbreviations ; /* When ON, alignments are printed
	using abbreviated names of symbols in alphabetic form (any symbol which
	starts with a digit is left as it is). When OFF, no abbreviations
	are used. */

bool show_all_parsing_alignments ;  /* If ON, all alignments produced by
	by compressing New against Old are shown. If OFF, only the best
	ones for any given pattern from New are shown. */

bool brackets_id_symbol_constraint ; /* This parameter is only relevant if
	RECOGNISE_BRACKETS is ON. When ON, any hit between two LEFT_BRACKETs
	must be immediately followed by a hit between two ID-symbols. Here,
	'immediately' means that the driving left bracket and the driving
	ID-symbol are next to each other and likewise for the target symbols.*/

bool match_all_data_symbols ; /* When ON, the program excludes all alignments in which
	the C-symbols of any Old pattern are not fully matched. When OFF, the
	program may deliver alignments where the C-symbols in one or more
	Old patterns are not fully matched. */

bool match_all_data_symbols_in_composite_alignments ; /* When ON, the program
	excludes all composite alignments in which the DATA_SYMBOLs of any
	Old pattern are not fully matched. When OFF, the program may deliver 
	composite alignments where the DATA_SYMBOLs in one or more
    Old patterns are not fully matched. */

bool match_all_data_symbols_in_hierarchical_alignments ; /* When ON, the program
	excludes all 'hierarchical' alignments in which the DATA_SYMBOLs of any
	Old pattern are not fully matched. When OFF, the program may deliver 
	hierarchical alignments where the DATA_SYMBOLs in one or more
    Old patterns are not fully matched. */

bool match_all_new_symbols ; /* When ON, the program excludes all alignments in which
	the New symbols are not fully matched. When OFF, the
	program may deliver alignments where the New are not fully matched. */

bool combine_alignments ; /* When ON, the program attempts to combine whole alignments
	into larger alignments on each cycle of recognise(). Otherwise it does not
	try to combine alignments. This parameter should only be ON if
	ONLY_ONE_NEW_PATTERN_IN_EACH_ALIGNMENT is also ON because it is not
	meaningful to combine alignments where there may be two or more
	New patterns in each alignment. If COMBINE_ALIGNMENTS is ON and
	ONLY_ONE_NEW_PATTERN_IN_EACH_ALIGNMENT is OFF, then
	COMBINE_ALIGNMENTS will be set to OFF. */

bool only_one_new_pattern_in_each_alignment ; /* When ON, there may be only one New
	pattern in any one alignment. When OFF, there may be one, two or more New patterns
	in any one alignment. */

/* Other Parameters. */

char alignment_format ; /* The format of alignments ('V' for vertical
	and 'H' for horizontal) */

// char alignment_order ; /* The order of columns in alignments ('F' for forward
//	and 'R' for reversed) */

int hit_structure_rows ; /* The number of rows in the
	sort_array[] and leaf_array[]. */

int key_array_size ;

int fail_score ; /* Any compression score which is equal to this or
	smaller counts as a fail. */

double cost_factor ; /* The multiplier to calculate the bit_cost of
	each data symbol from the minimum cost needed to discriminate it
	from other symbols. */

int keep_rows ; /* Controls the number of rows in the
	symbol_selection_array[][] which are used to determine which
	sequences are *not* purged at the end of each cycle. */

int driving_keep_rows ; /* Of the patterns selected by
						keep_rows and patterns_to_keep, the best
						driving_keep_rows are selected to be driving
						patterns for the next cycle. */

bool recognise_xml_brackets ; /* If ON, the program recognises brackets of the
	style <start> and </end>. If OFF, there is no recognition of these
	types of brackets. recognise_brackets and recognise_xml_brackets
	should not both be ON. */

int best_alignment_print_limit ; /* This limits the number of best alignments that are
	printed out at the end of the run. */

bool id_c_symbol_constraint ; /* If this is ON, ID-symbols can only ever be matched
	with C-symbols and vice versa. If the parameter is OFF, the constraint is
	not applied. This parameter is normally OFF. */

bool recognise_brackets ; /* If ON, the program recognises '<' as LEFT_BRACKET and
	'>' as RIGHT_BRACKET and applies a constraint tries to ensure that only
	valid matches of left and right brackets are recorded. If OFF, '<' and '>'
	brackets are not recognised and the constraint is not applied. */

char figure_ID[20] ; /* Records the id for the data, for printing
	out in the latex_file. */

int max_new_gap ; /* Sets the maximum permitted gap between hit symbols
	in any New pattern. */

int max_old_gap ; /* Sets the maximum permitted gap between hit symbols
	in any original old pattern (not an alignment formed during
	current processing). */

int max_unsupported_cycles ; /* For any one New pattern, this parameter
	sets the maximum number of cycles which build alignments without
	any gain in maximum compression score compared with the best
	cycle for the current New pattern. */

int max_alignments_in_one_cycle ; /* This can be used to limit the number
	of alignments formed in one cycle. It is applied as alignments are
	formed, in order of the approximate compression scores calculated for
	the hit structure. If this limit is applied, there is a possibility
	of cutting short the alignments before the best one has been made. */

int orientation ; /* Orientation of the output may be specified as
	LANDSCAPE or PORTRAIT. */

int write_al_chars_length ; // The maximum size of any alignment when written
	// out. Any excess is truncated with ellipses.

int write_section_chars_length ; // The maximum size of a section of
	// any alignment when written out.

double minimum_font_height ; /* The smallest font to be used in printing
	alignments. */

double maximum_font_height ; /* The largest font to be used in printing
	alignments. */

int window_size ; // The maximumn size of each window into each New pattern.

int combination_limit ; /* The number of combinations to be selected
	on each phase of combine_alignments(). */

bool audit_trail ; /* If true, an audit trail of the programs intermediate
	results is sent to output_file. Otherwise, only the final results
	are output. WARNING: The audit trail for each New pattern can be
	very large so this parameter should only be switched on when
	there is one New pattern and only one New pattern. */

bool identification_symbols_marked ; // This is set by examining the symbols in the
	// corpus. If any of them have a '!' prefix, then IDENTIFICATION
	// symbols are set in the input by the use of this mark
	// and CONTENTS symbols are those without it.

/* End of Parameters. */

char *syntax_buffer ;
char *syn_buf_ptr ;

symbol *driving_left_check_bracket, *target_left_check_bracket ;
int new_pattern_number ;
int code_number = 0 ;
int cycle ; /* Counts the recognition cycles within one New pattern. */
int number_of_alignments_purged ;
int current_number_of_alignments ;
int new_hits ; /* Used in update_hit_structure() and in add_hit(). */
int driving_span ;
int target_span ;
int span ;
int number_of_al_constrs ;
int best_cycle_this_cpfn ; // For a given New pattern, this is the cycle
	// in which the best score for the pattern was obtained.
int number_of_symbols_in_new, number_of_symbols_in_old ;
int number_of_patterns_in_new, number_of_patterns_in_old ;
int row_0_last_int_pos ; // For row 0, entry in row_0_array[] corresponding
    // to patt_last_el_pos in row_array[].

const double log_10_2 = log10(2) ;

clock_t time1 ;
clock_t time2 ; 
double old_symbol_bit_cost, adjusted_symbol_cost ;

/*** ARRAYS ETC USED IN WRITE_ALIGNMENT ***/

double font_heights[FONT_SET_SIZE] =
{7, 8, 9, 10, 11, 12, 14, 17.28, 20.74, 24.88} ;

struct key_entry *key_array ; // Stores the key to the abbreviations of
// symbols in the input file.

int fe_key_array ;

struct row_entry
{
	sequence *row_pattern ;
	list_element *patt_last_el_pos ; // This is used to record the
		// last (ie, most recent) el_pos for row_pattern where the
		// corresponding symbol has been written into
		// write_al_chars[][].
	int fe_col_in_write_al_chars ; // Keeps track of the first empty
		// column ***for this row*** in write_al_chars[][].
	bool pattern_is_finished ; // true when all the symbols in
		// the row_pattern have been written, false otherwise.
		// For row 0, it is true when all the symbols in the last row_pattern
		// have been written. Otherwise it is false.
} row_array[MAX_ALIGNMENT_DEPTH] ; // Used to keep track of
	// symbols and positions for each row in
	// fill_in_non_hit_symbols(), except row 0.

struct row_0_entry
{
	symbol *row_0_symbol ;
	sequence *row_0_pattern ;
	bool is_a_hit ;
} row_0_array[VERY_LARGE_SCRATCH_ARRAY_SIZE] ; // Stores a copy of row 0, including
	// non-hit symbols.

int fe_row_0_array, current_row_0_array_position ;

int max_alignment_length = 100 ; // Determines the maximum length of
	// an alignment, measured in symbols. (Currently not
	// read in but it is used).

int write_alignment_char_rows = 
	MAX_ALIGNMENT_DEPTH * 2 ; /* The maximum
	depth of an sequence that can be written (characters). */

char **write_al_chars ;
char **write_al_vacant_slots ;

char complete_symbol[MEDIUM_SCRATCH_ARRAY_SIZE],
	new_abbreviation[MEDIUM_SCRATCH_ARRAY_SIZE] ; // Used in
	// make_letter_abbreviations.

bool al1_rows[MEDIUM_SCRATCH_ARRAY_SIZE] ;

double scaling_factors[SCALING_FACTORS_SIZE] ; /* Stores the scaling
	factors for ECs calculated for various gaps between hits. */

struct hit_entry
{
	hit_node *h_node ;
	symbol *newly_created_col ;
} hit_entries[MEDIUM_SCRATCH_ARRAY_SIZE] ; // Used in unify_best.

int fe_hit_entries ;

sequence *symbol_selection_array[MAX_NEW_ARRAY_ROWS][MCOLS] ; /* Each	
	symbol corresponding to a symbol in New receives zero or more pointers
	to sequences which contains that symbol. At the end of each cycle, the
	array is used for selection of target sequences for the following cycle. */

sequence *concatenated_new_patterns ; // Used by make_selection().
int length_of_concatenated_new_patterns ; // Used by make_selection().

hit_node **half_list ; // Used to store the worst 50% of the leaf nodes in the
	// leaf_array[].

int fe_half_list ;

struct hn_entry hit_node_array[HIT_NODE_ARRAY_SIZE] ;

int fe_hit_node_array ;

struct best_alignment_entry
{
	sequence *al1 ;
	bool is_contained_in_another_alignment ;
	double rel_P ;
	bool done ;
} best_alignments[MEDIUM_SCRATCH_ARRAY_SIZE] ; /* Stores a set of
	alignments, including best_alignment, which code the
	same symbols from New as best_alignment. It also stores
	the abs_P and rel_P values calculated for each of
	these alignments. Used in probabilities_of_inferences(). */

int fe_best_alignments ;

group *brackets_index ; // Used to index left and right brackets
	// and associated ID-symbols.

group *used_driving_patterns ; // Used in match_driving_pattern().

group *hit_sequences ;  // Used in 
	// compare_patterns_with_brackets()
group *new_hit_sequences ;

// Used in compare_patterns_with_brackets() and
// compare_patterns_with_xml_brackets()

symbol *dp_array[VERY_LARGE_SCRATCH_ARRAY_SIZE],
	*tp_array[VERY_LARGE_SCRATCH_ARRAY_SIZE] ;

/*** Arrays used in sequence::write_alignment_vertical() ***/

symbol *unmatched_new_symbols[MEDIUM_SCRATCH_ARRAY_SIZE] ;
display_row *available_display_rows[MEDIUM_SCRATCH_ARRAY_SIZE] ;
int name_length_array[MEDIUM_SCRATCH_ARRAY_SIZE] ;
char d_chars_array[LARGE_SCRATCH_ARRAY_SIZE][LARGE_SCRATCH_ARRAY_SIZE] ;
int d_col_first_chars[MEDIUM_SCRATCH_ARRAY_SIZE] ;
char column_numbers[MEDIUM_SCRATCH_ARRAY_SIZE] ;
char col_number[SMALL_SCRATCH_ARRAY_SIZE] ;

/*** Mirror fields for hn_master ***/

int master_driving_int_pos ;
sequence *master_driving_pattern ;
symbol *master_driving_symbol ;
hit_node *master_parent ;
int master_target_int_pos ;
sequence *master_target_pattern ;
symbol *master_target_symbol ;

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Finds zero or more 'good' combinations of alignments in any given
**	cycle of the alignment-building process and adds them to
**	parsing_alignments.
**
**	The alignments that are combined are 'full' alignments in the sense
**	that all their Old DATA_SYMBOLs must be matched but the New
**	symbols are not necessarily all matched (if they were, then no
**	combination could be made as described next).
**
**	Alignments can be combined if, for any two alignments, none of the New
**	symbols between the first New symbol in alignment A and the last New
**	symbol in alignment A falls between the first New symbol in alignment
**	B and the last New symbol in alignment B - and vice versa. However,
**	there may be one or more unmatched New symbols between the last New
**	symbol of the left alignment and the first New symbol of the
**	right alignment.
**
**	Because the alignments in combine_al_array[] are in descending
**	order of compression difference, the simplifying assumption is
**	made that good combinations of alignments can be found by
**	adding alignments from successive entries in the array rather
**	than considering all possible combinations of alignments. Because
**	combinations are formed by adding alignments in this way, it 
**	is assumed that combinations containing n + 1 alignments are
**	always better than combinations containing n alignments.
**
** CALLING SEQUENCE:
**
**	void make_combinations(sequence *pattern1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	pattern1:		The pattern being parsed.
**
** IMPLICIT INPUTS:
**
**	alignments in parsing_alignments.
**
** IMPLICIT OUTPUTS:
**
**	Alignments added to parsing_alignments.
**
** SIDE EFFECTS:
**
**	NONE
*/

void make_combinations(sequence *pattern1)
{
	struct combine_al_record
	{
		sequence *alignment ;
		int new_int_pos_first, new_int_pos_last ;
	} *combine_al_array ;

	int number_of_combinable_alignments ;
	group *current_set_of_combinations = new group ;

	sequence *al1 ;

	if (audit_trail)
	{
		fprintf(output_file, "Start of make_combinations()\n(") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, ")\n\n") ;
	}

	current_alignments_from_parsing->sort_by_compression_difference() ;

	// Compile information about each full (B) alignment that is needed
	// for the creation of combinations.

	number_of_combinable_alignments = 0 ;
	list_for(al1, sequence, current_alignments_from_parsing)
	{
		if (match_all_data_symbols_in_composite_alignments)
		{
			if (al1->get_degree_of_matching() != FULL_B) continue ;
		}
		number_of_combinable_alignments++ ;
	}

	int counter, new_int_pos, new_int_pos_first, new_int_pos_last ;
	symbol *symbol1, *symbol2 ;

	combine_al_array =
		new struct combine_al_record[number_of_combinable_alignments] ;

	counter = 0 ;
	list_for(al1, sequence, current_alignments_from_parsing)
	{
		if (match_all_data_symbols_in_composite_alignments)
		{
			if (al1->get_degree_of_matching() != FULL_B) continue ;
		}
		// if (al1->get_row_pattern(0) != pattern1) continue ;
		new_int_pos_first = NULL_VALUE ;
		combine_al_array[counter].alignment = al1 ;
		list_for(symbol1, symbol, al1)
		{
			symbol2 = symbol1->get_row_symbol(0) ;
			if (symbol2 == NIL) continue ;
			new_int_pos = symbol1->get_row_orig_int_pos(0) ;
			if (new_int_pos_first == NULL_VALUE)
				new_int_pos_first = new_int_pos ;
			new_int_pos_last = new_int_pos ;
		}
		combine_al_array[counter].new_int_pos_first = new_int_pos_first ;
		combine_al_array[counter].new_int_pos_last = new_int_pos_last ;
		counter++ ;
	}

	// Now combine alignments, in descending order of CD score and growing
	// each combination as large as possible. Any alignment that does not
	// fit into the current series is used to start a new series.

	int i, j, length_pattern1 = pattern1->get_number_of_children() ;
	combination *comb1, *comb2 ;
	bool second_series_started ;
	for (i = 0; i < number_of_combinable_alignments; i++)
	{
		new_int_pos_first = combine_al_array[i].new_int_pos_first ;
		new_int_pos_last = combine_al_array[i].new_int_pos_last ;
		al1 = combine_al_array[i].alignment ;
		comb1 = new combination(length_pattern1) ;
		comb1->add_sub_alignment(al1,
			new_int_pos_first, new_int_pos_last, pattern1) ;
		second_series_started = false ;
		for (j = 0; j < number_of_combinable_alignments; j++)
		{
			if (j == i) continue ;
			new_int_pos_first = combine_al_array[j].new_int_pos_first ;
			new_int_pos_last = combine_al_array[j].new_int_pos_last ;
			al1 = combine_al_array[j].alignment ;
			if (comb1->can_accept(new_int_pos_first, new_int_pos_last))
				comb1->add_sub_alignment(al1, new_int_pos_first, 
							new_int_pos_last, pattern1) ;
		}

		if (comb1->number_of_alignments() > 1)
		{
			// Check to make sure that the alignments in comb1 are not
			// simply a subset of any combinations in
			// current_set_of_combinations or set_of_combinations
			// (from previous cycles). If it is, delete comb1,
			// otherwise add it to current_set_of_combinations.

			list_for(comb2, combination, current_set_of_combinations)
			{
				if (comb1->is_subset_of(comb2))
				{
					#if DIAGNOSTIC5

					if (audit_trail)
					{
						fprintf(output_file, "%s%d%s%d%s",
							"Combination C_ID",
							comb1->get_C_ID(),
							" is a subset of, or the same as, combination C_ID",
							comb2->get_C_ID(),
							" and is deleted.\n\n") ;
					}

					#endif

					delete comb1 ;
					goto L1 ;
				}
			}

			list_for(comb2, combination, set_of_combinations)
			{
				if (comb1->is_subset_of(comb2))
				{
					#if DIAGNOSTIC5

					if (audit_trail)
					{
						fprintf(output_file, "%s%d%s%d%s",
							"Combination C_ID",
							comb1->get_C_ID(),
							" is a subset of, or the same as, combination C_ID",
							comb2->get_C_ID(),
							" and is deleted.\n\n") ;
					}

					#endif

					delete comb1 ;
					goto L1 ;
				}
			}

			current_set_of_combinations->add_child(comb1) ;
		}
		else
		{
			#if DIAGNOSTIC5

			if (audit_trail)
			{
				fprintf(output_file, "%s%d%s",
					"Combination C_ID",
					comb1->get_C_ID(),
					" contains only one alignment and is deleted.\n\n") ;
			}

			#endif

			delete comb1 ;
		}
		L1: ;
	}

	// Find the best combinations up to combination_limit, print them
	// out and make them into composite alignments. At the same time,
	// add the combination to set_of_combinations.

	double best_score, temp_score ;
	combination *best_comb ;
	int total_number_of_combinations =
		current_set_of_combinations->count_number_of_children() ;

	if (audit_trail)
	{
		fprintf(output_file, "%s%d%s",
			"Total number of combinations = ",
			total_number_of_combinations,
			"\n\n") ;
	}

	sequence *composite_alignment ;
	counter = 0 ;
	while (counter < combination_limit)
	{
		best_score = NULL_VALUE ;
		list_for(comb1, combination, current_set_of_combinations)
		{
			temp_score = comb1->get_combination_score() ;
			if (temp_score > best_score)
			{
				best_comb = comb1 ;
				best_score = temp_score ;
			}
		}

		if (best_score > NULL_VALUE)
		{
			current_set_of_combinations->extract_child(best_comb) ;
			best_comb->print_combination() ;
			composite_alignment =
				best_comb->make_composite_alignment(pattern1) ;
			composite_alignment->find_degree_of_matching(pattern1) ;

			// Check for anomalies in the ordering of symbols in
			// current_new_pattern within the alignment.

			if (composite_alignment->check_ordering_of_new() == false)
			{
				if (verbose)
				{
					if (audit_trail)
					{
						fprintf(output_file, 
							"Anomalous ordering of symbols from New in composite alignment ") ;
						composite_alignment->print_ID() ;
						fprintf(output_file, ". Alignment is discarded.\n\n") ;
					}
				}
				delete composite_alignment ;
				continue ;
			}
	
			// CHECK AND MARK ANY BRACKETS IN THE NEW ALIGNMENT

			if (recognise_brackets)
				composite_alignment->check_brackets_in_one_pattern() ;

			// Mark get_keep as DRIVING

			composite_alignment->set_keep(DRIVING) ;

			current_alignments_from_parsing->add_child(composite_alignment) ;
			set_of_combinations->add_child(best_comb) ;
		}
		else break ;
		counter++ ;
	}

	if (audit_trail)
	{
		fprintf(output_file, "End of combining alignments.\n(") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, ")\n\n") ;
	}

	if (audit_trail)
		fprintf(output_file, "End of make_combinations()\n\n") ;

	// Delete arrays and temporary objects.

	current_set_of_combinations->delete_children() ;
	delete current_set_of_combinations ;
	delete[] combine_al_array ;

} // make_combinations

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Deletes a set of hit sequences.
**
** CALLING SEQUENCE:
**
**	void delete_hit_sequences(group *set_of_sequences)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	set_of_sequences:	The set to be deleted (hit_sequences or
**						new_hit_sequences.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Deletes the set of hit sequences
**
** SIDE EFFECTS:
**
**	NONE
*/

void delete_hit_sequences(group *set_of_sequences)
{
	hit_pair *first_current, *next1, *next2 ;

	// First step though the initial hit pairs of each hit sequence
	// and delete everything *except* the initial hit pair.

	list_for(first_current, hit_pair, set_of_sequences)
	{
		next1 = first_current->get_next() ;
		while (next1 != NIL)
		{
			next2 = next1->get_next() ;
			delete next1 ;
			next1 = next2 ;
		}
	}

	// Now delete the initial hit pairs and the set_of_sequences group.

	set_of_sequences->delete_children() ;
	delete set_of_sequences ;

} // delete_hit_sequences

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Creates an alignment from a list of basic alignments.
**
** CALLING SEQUENCE:
**
**	sequence *combination::make_composite_alignment(sequence *cnp)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	cnp:			The 'current' New pattern
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	An encoding of the compposite alignment is added to 
**	current_code_patterns.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

sequence *combination::make_composite_alignment(sequence *cnp)
{
	sequence *sub_alignment,
		*composite_alignment = new sequence(COMPOSITE_ALIGNMENT) ;
	int overall_depth = 0, ba_depth, new_row1, new_row2, old_row ;
	symbol *sub_alignment_symbol, *newly_made_column,
		*symbol1 ;
	alignment_element *al_el1, *al_el2 ;
	char *symbol_name ;

	composite_alignment->set_composite_alignment(true) ;

	// Calculate depth of new composite alignment as
	// D = SUM (d_i - 1) + 1, summing over the set of basic alignments
	// and with d representing the depth of each individual alignment.
	// The reasoning is that row 0 is the same for all alignments so
	// we don't need to include it in the depth of each alignment, but
	// we need to add in row 0.

	list_for(sub_alignment, sequence, sub_alignment_list)
	{
		ba_depth = sub_alignment->get_sequence_depth() - 1 ;
		overall_depth += ba_depth ;
	}
	
	overall_depth += 1 ; // Add 1 for the New pattern in row 0.
	
	composite_alignment->set_sequence_depth(overall_depth) ;

	// Now copy the basic alignments into a new composite alignment
	// with each symbol having a depth equal to overall_depth and
	// with each row of each basic alignment given a unique row
	// in the composite alignment.

	symbol *first_symbol, *last_symbol ;

	new_row1 = 1 ; // Row 0 is the same in the original symbol and
		// the new symbol. So the first row in the new symbol
		// for purposes of copying is row 1.

	list_for(sub_alignment, sequence, sub_alignment_list)
	{
		ba_depth = sub_alignment->get_sequence_depth() ;

		first_symbol = (symbol *)sub_alignment->get_first_child() ;
		last_symbol = (symbol *)sub_alignment->get_last_child() ;

		list_for(sub_alignment_symbol, symbol, sub_alignment)
		{
			new_row2 = new_row1 ;
			
			symbol_name = sub_alignment_symbol->get_name() ;
			
			newly_made_column =
				new symbol(symbol_name, overall_depth, NULL_VALUE) ;
			newly_made_column->set_bit_cost(sub_alignment_symbol->
												get_bit_cost()) ;
			newly_made_column->set_is_a_hit(sub_alignment_symbol->
												is_a_hit()) ;
			newly_made_column->set_status(sub_alignment_symbol->
												get_status()) ;
			newly_made_column->set_type(sub_alignment_symbol->
												get_type()) ;
			
			// Now set values for row 0.
			
			al_el1 = sub_alignment_symbol->get_al_el(0) ;
			al_el2 = newly_made_column->get_al_el(0) ;
			al_el2->set_el_obj(al_el1->get_el_obj()) ;
			al_el2->set_original_pattern(al_el1->
				get_original_pattern()) ;
			
			// Set values for other rows.
			
			for (old_row = 1; old_row < ba_depth; old_row++)
			{
				if (new_row2 >= overall_depth)
					abort_run("Anomaly in value of new_row \
						in make_composite_alignment()") ;
				al_el1 = sub_alignment_symbol->get_al_el(old_row) ;
				al_el2 = newly_made_column->get_al_el(new_row2) ;
				al_el2->set_el_obj(al_el1->get_el_obj()) ;
				al_el2->set_original_pattern(al_el1->
					get_original_pattern()) ;
				new_row2++ ;
			}
			
			// Set values for same_column_above,
			// same_column_below and symbol_is_a_hit.
			
			newly_made_column->set_symbol_matches() ;

			composite_alignment->add_child(newly_made_column) ;
		} // End of one basic alignment
		new_row1 = new_row2 ;
	} // End of sub_alignment_list

	composite_alignment->mark_parent_and_int_positions_non_recursive() ;

	// Fill in values for original_pattern in every row of every
	// column. First, find the first column that has a non-NIL
	// value for original_pattern in this row. Skip over row 0.
	
	sequence *original_pattern ;
	for (new_row1 = 1; new_row1 < overall_depth; new_row1++)
	{
		// Find the original_pattern for this row.
		
		for (symbol1 = (symbol *)composite_alignment->get_first_child();
			symbol1 != NIL; symbol1 = (symbol *)
				composite_alignment->get_next_child())
		{
			original_pattern = symbol1->get_row_pattern(new_row1) ;
			if (original_pattern != NIL) break ;
		}
		
		// Now fill it in right across the alignment.
		
		for (symbol1 = (symbol *)composite_alignment->get_first_child();
			symbol1 != NIL; symbol1 = (symbol *)
				composite_alignment->get_next_child())
		{
			al_el1 = symbol1->get_al_el(new_row1) ;
			al_el1->set_original_pattern(original_pattern) ;
		}
	}

	// Calculate values for NSC, EC, CD, CR and abs_P for
	// the alignment.
	
	composite_alignment->make_code(false) ;

	composite_alignment->set_symbol_matches() ;

	if (show_al_structure)
		composite_alignment->show_al_structure(NIL) ;

	// Print new composite alignment as alignment and as
	// a flat pattern, with and without details.

	if (audit_trail)
	{
		composite_alignment->write_out_fully("COMPOSITE ALIGNMENT", NIL, NIL,
			write_section_chars_length, NULL_VALUE, true) ;
	}

	return(composite_alignment) ;

} // combination::make_composite_alignment

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Tests whether the list of basic alignments in 'this' is a subset
**	of the list of basic alignments in test_comb.
**
** CALLING SEQUENCE:
**
**		bool combination::is_subset_of(combination *test_comb)
**
** FORMAL ARGUMENTS:
**
**		Return value:	true if the basic alignments in 'this' are a subset
**						of the basic alignments in test_comb, false otherwise.
**
**		test_comb:		The combination against which 'this' is to be tested.
**
** IMPLICIT INPUTS:
**
**		NONE
**
** IMPLICIT OUTPUTS:
**
**		NONE
**
** SIDE EFFECTS:
**
**		NONE
**
*/

bool combination::is_subset_of(combination *test_comb)
{
	group *test_sub_alignment_list =
		test_comb->get_sub_alignment_list() ;
	sequence *ba_this, *ba_test ;

	// 'this' combination must not be the same as test_comb.

	if (this == test_comb)
		abort_run("Invalid comparison in combination::is_subset_of().") ;

	list_for(ba_this, sequence, sub_alignment_list)
	{
		list_for(ba_test, sequence, test_sub_alignment_list)
		{
			if (ba_this == ba_test) break ;
		}
		if (ba_test == NIL) return(false) ; // An instance of ba_this
					// has been found which is not present on
					// test_sub_alignment_list. Therefore,
					// sub_alignment_list is *not* a subset of
					// test_sub_alignment_list.
	}
	return(true) ; // Every instance of ba_this is found on
				// test_sub_alignment_list. Therefore,
				// sub_alignment_list *is* a subset of
				// test_sub_alignment_list.
} // combination::is_subset_of

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Prints the score and the list of basic alignments in a combination.
**
** CALLING SEQUENCE:
**
**	void combination::print_combination()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Output to file
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void combination::print_combination()
{
	sequence *sub_alignment ;

	if (audit_trail)
	{
		fprintf(output_file, "%s%d%s%1.2f%s",
			"Combination C_ID",
			C_ID,
			" (score ",
			combination_score,
			") sub-alignments:\n\n") ;
	}
	
	bool first_in_list = true ;
	list_for(sub_alignment, sequence, sub_alignment_list)
	{
		if (first_in_list)
		{
			if (audit_trail)
			{
				fprintf(output_file, "%s%d",
					"ID",
					sub_alignment->get_ID()) ;
			}
			first_in_list = false ;
		}
		else
		{
			if (audit_trail)
			{
				fprintf(output_file, "%s%d",
					", ID",
					sub_alignment->get_ID()) ;
			}
		}
	}

	if (audit_trail)
	{
		fprintf(output_file, "\n\n") ;
	}
} // combination::print_sub_alignments

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Copy constructor for class combination.
**
** CALLING SEQUENCE:
**
**	combination::combination(combination &comb1) : base_object(comb1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	Creates a copy of the input combination
**
**	comb1:			The combination to be copied.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

combination::combination(combination &comb1) : base_object(comb1)
{
	sequence *sub_alignment ;
	int i ;

	C_ID = combination_ID_number++ ; 
	
	combination_score = comb1.combination_score ;
	
	len_fit_seq = comb1.len_fit_seq ;

	if (len_fit_seq >= MEDIUM_SCRATCH_ARRAY_SIZE)
		abort_run("len_fit_seq too long in combination") ;
	for (i = 0; i < len_fit_seq; i++)
		fitting_sequence[i] = comb1.fitting_sequence[i] ;
	
	group *temp_list = comb1.sub_alignment_list ;
	sub_alignment_list = new group ;
	for (sub_alignment = (sequence *)temp_list->get_first_child();
		sub_alignment != NIL; sub_alignment = (sequence *)
			temp_list->get_next_child())
	{
		sub_alignment_list->add_child(sub_alignment) ;
	}
} // combination::combination(combination &comb1)

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	This method in class combination tests whether or not a given basic
**	alignment 'fits' into the combination. A basic combination fits if
**	all the symbols from New that it covers have not already been marked
**	as true in the fitting _sequence of 'this'.
**
** CALLING SEQUENCE:
**
**	bool combination::can_accept(int pos_new_first, 
**		int pos_new_last)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if sub_alignment fits into
**					fitting _sequence, otherwise false.
**
**	pos_new_first:	The integer position of the first hit
**					symbol from current_new_pattern in bal1.
**	pos_new_last:	The integer position of the last hit
**					symbol from current_new_pattern in bal1.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool combination::can_accept(int pos_new_first, 
	int pos_new_last)
{
	for (int i = pos_new_first; i <= pos_new_last; i++)
		if (fitting_sequence[i] == true) return(false) ;
		
	return(true) ;
} // combination::can_accept

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Adds a 'basic' alignment to a combination.
**
** CALLING SEQUENCE:
**
**	void combination::add_sub_alignment(sequence *bal1,
**		int bal1_pos_new_first, int bal1_pos_new_last,
**		sequence *cnp)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	bal1:				The sub_alignment to be added to 'this'.
**	bal1_pos_new_first:	The integer position of the first hit
**						symbol from current_new_pattern in bal1.
**	bal1_pos_new_last:	The integer position of the last hit
**						symbol from current_new_pattern in bal1.
**	cnp:				The New pattern being processed.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Addition to the structure of 'this'. The 'coverage' of the
**	current pattern from New by the newly-added basic alignment
**	is marked in fitting_sequence.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void combination::add_sub_alignment(sequence *bal1,
	int bal1_pos_new_first, int bal1_pos_new_last,
	sequence *cnp)
{
	sequence *bal2 ;
	symbol *symbol1, *symbol_new ;
	int bal2_pos_new_first ;
	
	// Add sub_alignment to sub_alignment_list in the
	// position in the list that corresponds to the order
	// of the alignments relative to the current pattern from New.
	// This is done to facilitate the building of composite alignments
	// in make_composite_alignment.

	list_for(bal2, sequence, sub_alignment_list)	
	{
		// Find the first symbol in bal2 that contains a symbol
		// from current_new_pattern and find the position of
		// that symbol in current_new_pattern.

		list_for(symbol1, symbol, bal2)					
		{
			symbol_new = symbol1->get_row_symbol(0) ;
			if (symbol_new != NIL) break ;
		}
					
		if (symbol1 == NIL) abort_run("Error in add_sub_alignment()") ;
					
		bal2_pos_new_first = symbol1->get_row_orig_int_pos(0) ;
					
		// Compare this with the corresponding integer position
		// of the first hit symbol from current_new_pattern
		// in sub_alignment.
					
		if (bal1_pos_new_first < bal2_pos_new_first) break ;
	}
				
	if (bal2 == NIL) sub_alignment_list->add_child(bal1) ;
	else sub_alignment_list->precede(bal1, bal2) ;
				
	combination_score += bal1->get_compression_difference() ;
				
	// Now mark the entries in 'fitting_sequence' that are
	// 'covered' by the newly-added basic alignment.
				
	for (int i = bal1_pos_new_first; i <= bal1_pos_new_last; i++)
		fitting_sequence[i] = true ;

	#if DIAGNOSTIC5

	if (audit_trail)
	{
		fprintf(output_file, "%s%d%s",
			"Combination C_ID",
			C_ID,
			" is derived from alignments: ") ;

		sub_alignment_list->write_IDs(NULL_VALUE, cnp) ;
	}

	#endif

} // combination::add_sub_alignment

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Counts the number of New patterns in 'this' alignment.
**
** CALLING SEQUENCE:
**
**	int sequence::number_of_new_patterns_in_alignment()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

int sequence::number_of_new_patterns_in_alignment()
{
	symbol *symbol1 ;
	sequence *new_pattern1, *new_pattern2 = NIL ;
	int counter = 0 ;

	list_for(symbol1, symbol, this)
	{
		new_pattern1 = symbol1->get_row_pattern(0) ;
		if (new_pattern1 == NIL) continue ;
		if (new_pattern1 != new_pattern2)
			counter++ ;
		new_pattern2 = new_pattern1 ;
	}

	return(counter) ;

} // sequence::number_of_new_patterns_in_alignment

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks a newly-constructed alignment to make sure that symbols from 
**	current_new_pattern are in their correct order.
**
** CALLING SEQUENCE:
**
**	bool sequence::check_ordering_of_new()
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if the symbols from current_new_pattern
**					are in their correct order, false otherwise.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool sequence::check_ordering_of_new()
{
	int new_symbol_int_pos1 = NULL_VALUE, new_symbol_int_pos2 ;
	symbol *symbol1 ;
	sequence *pattern1 ;

	list_for(symbol1, symbol, this)
	{
		pattern1 = symbol1->get_row_pattern(0) ;
		if (pattern1 != current_new_pattern) continue ;
		new_symbol_int_pos2 = symbol1->get_row_orig_int_pos(0) ;
		if (new_symbol_int_pos2 != NULL_VALUE)
		{
			if (new_symbol_int_pos1 != NULL_VALUE)
			{
				if (new_symbol_int_pos2 <= new_symbol_int_pos1)
					return(false) ;
			}
			new_symbol_int_pos1 = new_symbol_int_pos2 ;
		}
	}

	return(true) ;
} // sequence::check_ordering_of_new

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks that gaps between hit symbols in a newly-formed alignment
**	conform to relevant parameters.
**
** CALLING SEQUENCE:
**
**	bool sequence::al_gaps_OK()
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if the gaps are within limits,
**					false otherwise.
**
**	al1:			The alignment to be tested.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool sequence::al_gaps_OK()
{
	symbol *col1, *new_symbol, *old_symbol ;
	int previous_new_hit_int_pos = NULL_VALUE, this_new_hit_int_pos,
		this_old_hit_int_pos, row ;

	if (sequence_depth >= MEDIUM_SCRATCH_ARRAY_SIZE)
		abort_run("Alignment too deep in al_gaps_OK()") ;

	int previous_old_hit_int_pos[MEDIUM_SCRATCH_ARRAY_SIZE] ;

	for (row = 1; row < sequence_depth; row++)
		previous_old_hit_int_pos[row] = NULL_VALUE ;
	
	// Test gaps between hit symbols in the New pattern and
	// test gaps between hit symbols that are also C-symbols
	// in each of the Old patterns.

	list_for(col1, symbol, this)
	{
		new_symbol = col1->get_row_symbol(0) ;
		if (new_symbol == NIL) continue ; // We are only interested
			// in hits between a New symbol and an Old symbol. If
			// there is a non-NIL New symbol then there must be
			// such a hit in this column.

		this_new_hit_int_pos = col1->get_row_orig_int_pos(0) ;
		
		if (previous_new_hit_int_pos == NULL_VALUE)
			goto L1 ;
		
		if ((this_new_hit_int_pos - 
			previous_new_hit_int_pos - 1) > max_new_gap)
				return(false) ;
		
		L1: previous_new_hit_int_pos = this_new_hit_int_pos ;

		// Test gaps between hit symbols (which are also C-symbols)
		// in Old patterns.

		for (row = 1; row < sequence_depth; row++)
		{
			old_symbol = col1->get_row_symbol(row) ;
			if (old_symbol == NIL) goto L2a ;
			if (old_symbol->get_status() != CONTENTS) goto L2a ;

			this_old_hit_int_pos = col1->get_row_orig_int_pos(row) ;
			
			if (previous_old_hit_int_pos[row] == NULL_VALUE)
				goto L2 ;
			
			if ((this_old_hit_int_pos - 
					previous_old_hit_int_pos[row] - 1) > max_old_gap)
						return(false) ;
			
			L2: previous_old_hit_int_pos[row] = this_old_hit_int_pos ;
			L2a: ;
		}
	}

	return(true) ;
} // sequence::al_gaps_OK

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Tests whether or not all the symbols from cnp
**	are matched in the alignment.
**
** CALLING SEQUENCE:
**
**	bool sequence::all_new_symbols_matched(sequence *cnp)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if all the symbols in cnp
**					are matched, false otherwise.
**
**	cnp:			The New pattern being processed.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool sequence::all_new_symbols_matched(sequence *cnp)
{
	// Check that this is an alignment and not a flat pattern.

	if (sequence_depth < 2)
		abort_run("Invalid alignment in sequence::all_new_symbols_matched()") ;

	symbol *col1, *symbol_new1, *symbol_new2 ;
	sequence *pattern_new = current_new_pattern ;

	this->initialise() ;
	list_for(symbol_new1, symbol, pattern_new)
	{
		while (col1 = (symbol *)this->get_next_child())
		{
			symbol_new2 = col1->get_row_symbol(0) ;
			if (symbol_new2 != NIL) break ;
		}

		if (col1 == NIL) // The alignment has been exhausted without
			// finding all the symbols in cnp.
				return(false) ;

		if (col1->is_a_hit() == false) return(false) ;
		if (symbol_new1 != symbol_new2) return(false) ;
	}

	return(true) ;
} // sequence::all_new_symbols_matched

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks whether or not an alignment is FULL_A, FULL_B, FULL_C or PARTIAL
**	as described in SP62a_head.h.
**
** CALLING SEQUENCE:
**
**	int sequence::find_degree_of_matching(sequence *cnp)
**
** FORMAL ARGUMENTS:
**
**	Return value:		FULL_A, FULL_B, FULL_C or PARTIAL.
**
**	cnp:				The New pattern being processed.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

int sequence::find_degree_of_matching(sequence *cnp)
{
	symbol *col1, *symbol1 ;
	int i, return_value, row ;
	bool all_new_symbols_are_matched,
		all_old_data_symbols_matched = true ;

	// Check to see whether all the New symbols are matched.

	all_new_symbols_are_matched = this->all_new_symbols_matched(cnp) ;

	// Check whether all the DATA_SYMBOLs of all the Old
	// patterns are matched.

	list_for(col1, symbol, this)
	{
		if (col1->is_a_hit()) continue ;

		// Find the single symbol in an Old pattern in this column.
		// There should not be any New symbol because, in general,
		// unmatched New symbols are not entered into alignments.

		symbol1 = col1->find_unmatched_symbol() ;

		/* Diagnostic.

		row = col1->get_row_by_symbol(symbol1) ;
		sequence *pattern1 = col1->get_row_pattern(row) ; */
			
		if (symbol1 == NIL)
			abort_run("Empty column in sequence::find_degree_of_matching().") ;

		if (symbol1->get_type() == DATA_SYMBOL)
		{
			all_old_data_symbols_matched = false ;

			/* Diagnostic.

			pattern1->write_with_details() ;
			fprintf(output_file, "%s%d%s%s%s%d%s%d%s",
				"In alignment ID",
				ID, 
				", unmatched symbol is ",
				symbol1->get_name(),
				" in pattern ID",
				pattern1->get_ID(),
				" in row ",
				row,
				"\n\n") ; */

			break ;
		}
	}

	// If all the New symbols are not matched, check to see whether
	// any of the unmatched symbols fall within the scope of
	// the DATA_SYMBOLs of any one Old pattern.

	int sequence_length = count_number_of_children(),
		index1, new_int_pos1, new_int_pos2 ;
	symbol **col_array, *new_symbol, *old_symbol ;
	bool unmatched_new_within_scope = false ;

	if (all_new_symbols_are_matched == false)
	{
		// Set up the sequence as an array.

		col_array = new symbol*[sequence_length] ;
		i = 0 ;
		list_for(col1, symbol, this)
		{
			col_array[i] = col1 ;
			i++ ;
		}

		for (row = 1; row < sequence_depth - 1; row++)
		{
			new_int_pos1 = new_int_pos2 = NULL_VALUE ;
			for (index1 = 0; index1 < sequence_length; index1++)
			{
				col1 = col_array[index1] ;
				old_symbol = col1->get_row_symbol(row) ;
				if (old_symbol == NIL) continue ;
				if (old_symbol->get_type() != DATA_SYMBOL)
					continue ;
				new_symbol = col1->get_row_symbol(0) ;
				if (new_symbol == NIL) continue ;
				new_int_pos1 = col1->get_row_orig_int_pos(0) ;
				if (new_int_pos2 != NULL_VALUE)
				{
					if (new_int_pos1 - new_int_pos2 > 1)
					{
						unmatched_new_within_scope = true ;
						goto L2 ;
					}
				}
				new_int_pos2 = new_int_pos1 ;
			}
		}

		L2: delete[] col_array ;
	}

	if (all_new_symbols_are_matched && all_old_data_symbols_matched)
	{
		set_degree_of_matching(FULL_A) ;
		return_value = FULL_A ;
	}
	else if (all_old_data_symbols_matched
		&& unmatched_new_within_scope == false)
	{
		set_degree_of_matching(FULL_B) ;
		return_value = FULL_B ;
	}
	else if (all_old_data_symbols_matched
		&& unmatched_new_within_scope == true)
	{
		set_degree_of_matching(FULL_C) ;
		return_value = FULL_C ;
	}
	else
	{
		set_degree_of_matching(PARTIAL) ;
		return_value = PARTIAL ;
	}

	return(return_value) ;

} // sequence::find_degree_of_matching

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Assigns status to bracket symbols in a pattern (simple brackets
**	or XML_style brackets.
**
** CALLING SEQUENCE:
**
**	void sequence::assign_brackets_status()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	status of bracket symbols.
**
** SIDE EFFECTS:
**
**	NONE
*/

void sequence::assign_brackets_status()
{
	symbol *symbol1 ;
	int level = -1 ;
	char *symbol_name ;

	list_for(symbol1, symbol, this)
	{
		if (recognise_brackets)
		{
			if (strcmp(symbol1->get_name(), "<") == 0)
			{
				level++ ;
				if (level == 0)
					symbol1->set_status(BOUNDARY_MARKER) ;
				else symbol1->set_status(CONTENTS) ;
			}
			else if (strcmp(symbol1->get_name(), ">") == 0)
			{
				if (level == 0)
					symbol1->set_status(BOUNDARY_MARKER) ;
				else symbol1->set_status(CONTENTS) ;
				level-- ;
			}
		}
		else if (recognise_xml_brackets)
		{
			symbol_name = symbol1->get_name() ;
			if (*symbol_name == '<'
				&& *(symbol_name + 1) != '/')
			{
				level++ ;
				if (level == 0)
					symbol1->set_status(BOUNDARY_MARKER) ;
				else symbol1->set_status(CONTENTS) ;
			}
			else if (*symbol_name == '<'
				&& *(symbol_name + 1) == '/')
			{
				if (level == 0)
					symbol1->set_status(BOUNDARY_MARKER) ;
				else symbol1->set_status(CONTENTS) ;
				level-- ;
			}
		}
	}
} // sequence::assign_brackets_status

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks the brackets in one pattern.
**
** CALLING SEQUENCE:
**
**	bool sequence::check_brackets_in_one_pattern()
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if brackets are valid, false otherwise.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

bool sequence::check_brackets_in_one_pattern()
{
	int symbol_type, depth = -1, temp_int_pos ;
	symbol *symbol1, *left_bracket ;
	struct symbol_and_position
	{
		symbol *bracket ;
		int bracket_int_pos ;
	} brackets_array[SMALL_SCRATCH_ARRAY_SIZE] ;
	list_element *el_pos1 = NIL ;

	list_for_el_pos(symbol1, symbol, this, el_pos1)
	{
		symbol_type = symbol1->get_type() ;

		if (symbol_type == LEFT_BRACKET)
		{
			plus_one(&depth, SMALL_SCRATCH_ARRAY_SIZE,
				"Overflow of brackets_array[] in check_brackets().") ;
			brackets_array[depth].bracket = symbol1 ;
			brackets_array[depth].bracket_int_pos =
				el_pos1->get_position() ;
		}
		else if (symbol_type == RIGHT_BRACKET)
		{
			left_bracket = brackets_array[depth].bracket ;
			left_bracket->set_partner_int_pos(el_pos1->get_position()) ;
			temp_int_pos = brackets_array[depth].bracket_int_pos ;
			symbol1->set_partner_int_pos(temp_int_pos) ;
			depth-- ;
		}
	}

	if (depth == -1) return(true) ;
	else return(false) ;
} // sequence::check_brackets_in_one_pattern

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	If brackets are used on old_patterns, check that they are used in a legal
**	manner. This means that every left bracket must be paired with a right
**	bracket at the corresponding depth of nesting.
**
**	This function also records the partner bracket for any given left or
**	right bracket.
**
** CALLING SEQUENCE:
**
**	void check_brackets(group *set_of_patterns)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	set_of_pattens:		The set of patterns to be checked.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void check_brackets(group *set_of_patterns)
{
	sequence *pattern1 ;

	list_for(pattern1, sequence, set_of_patterns)
	{
		if (pattern1->check_brackets_in_one_pattern() == false)
		{
			if (audit_trail)
			{
				fprintf(output_file, "%s%d%s",
					"Invalid brackets in pattern ID",
					pattern1->get_ID(),
					"\n\n") ;
			}
			abort_run("Program terminated.") ;
		}
	}
} // check_brackets

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compares two strings of characters for their relative alphabetic
**	position, regardless of the case of the letters.
**
** CALLING SEQUENCE:
**
**	int compare_strings(char *string1, char *string2)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

int compare_strings(char *string1, char *string2)
{
	int i ;
	char c1, c2 ;

	i = 0 ;
	c1 = string1[i] ;
	if (isupper(c1))
		c1 = tolower(c1) ;
	c2 = string2[i] ;
	if (isupper(c2))
		c2 = tolower(c2) ;
	while (c1 != '\0' && c2 != '\0')
	{
		if (c1 > c2) return(1) ;
		else if (c1 == c2) goto L1 ;
		else return (-1) ;

		L1: i++ ;
		c1 = string1[i] ;
		if (isupper(c1))
			c1 = tolower(c1) ;
		c2 = string2[i] ;
		if (isupper(c2))
			c2 = tolower(c2) ;
	}

	if (c1 == '\0' && c2 == '\0') return(0) ;
	else if (c1 != '\0' && c2 == '\0') return(1) ;
	else return(-1) ;
} // compare_strings

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Increments and index by 1, checks that it does not exceed a limit
**	and aborts the program, printing out a message if the limit is
**	exceeded.
**
** CALLING SEQUENCE:
**
**	void increment_index(int *index, int limit, char *message)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	index:			A pointer to the index to be incremented.
**	limit:			The limit to be checked.
**	message:		The message to be printed out.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void increment_index(int *index, int limit, char *message)
{
	(*index)++ ;
	if ((*index) >= limit)
		abort_run(message) ;

} // increment_index

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Prints the number of the pattern from New and, within that
**	pattern, the window and the cycle.
**
** CALLING SEQUENCE:
**
**	void print_pattern_window_cycle()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void print_pattern_window_cycle()
{
	fprintf(output_file, "%s%d%s%d%s%d",
		"PATTERN ",
		new_pattern_number,
		", WINDOW ",
		window,
		", CYCLE ",
		cycle) ;
	fflush(output_file) ;
} // print_pattern_window_cycle

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Makes a new column for insertion in al_constr
**	(in unify_best).
**
** CALLING SEQUENCE:
**
**	symbol *make_column(int al_constr_depth,
**			sequence *driving_pattern, symbol *driving_symbol, 
**			int driving_int_pos, sequence *target_pattern,
**			symbol *target_symbol, int target_int_pos, 
**			int row_one_of_t_patt_in_al_constr, int hn_ID,
**			bool hit_this_cycle, bool rows_0_to_be_merged)
**
** FORMAL ARGUMENTS:
**
**	Return value:			A pointer to the newly-created column. Value
**							is NIL if no column is created.
**
**	al_constr_depth:		The depth of the new alignment.
**	driving_pattern:		The driving pattern.
**	driving_symbol: 		The driving symbol containing information to be
**							copied into the new column.
**	driving_int_pos:		The int position of the driving_symbol in
**							the driving pattern.
**	target_pattern: 		The target pattern.
**	target_symbol:			The target symbol containing information to be
**							copied into the new column.
**	target_int_pos: 		The int position of the target_symbol in
**							the target pattern.
**	row_one_of_t_patt_in_al_constr: The row of the new column into which the
**							first row of the target symbol should be copied.
**	hn_ID:					The ID of the hit node (corresponding to this
**							new column. NULL_VALUE if there is no hit node.
**	hit_this_cycle: 		true if the column records a hit on this cycle,
**							false otherwise.
**	rows_0_to_be_merged:	Records whether or not row 0 of the driving
**							pattern and row 0 of the target pattern are to
**							be merged in row 0 of the alignment.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

symbol *make_column(int al_constr_depth,
	sequence *driving_pattern, symbol *driving_symbol, 
	int driving_int_pos, sequence *target_pattern,
	symbol *target_symbol, int target_int_pos, 
	int row_one_of_t_patt_in_al_constr, int hn_ID,
	bool hit_this_cycle, bool rows_0_to_be_merged)
{
	symbol *new_al_col1, *symbol1, *driving_new_symbol ;
	alignment_element *al_el1, *al_el2 ;
	int driving_pattern_depth, target_pattern_depth,
		target_patt_al_constr_last_row, i, j ;
	
	if (driving_symbol != NIL) symbol1 = driving_symbol ;
	else if (target_symbol != NIL) symbol1 = target_symbol ;
	else abort_run("driving_symbol and target_symbol are both NIL in \
		make_column") ;
	
	new_al_col1 = new symbol(symbol1->get_name(), 
		al_constr_depth, hn_ID) ;
	
	if (hit_this_cycle) new_al_col1->set_is_a_hit(true) ;
	
	new_al_col1->set_bit_cost(symbol1->get_bit_cost()) ;
	new_al_col1->set_type(symbol1->get_type()) ;

	if (driving_symbol != NIL)
	{
		driving_pattern_depth = driving_pattern->get_sequence_depth() ;
		if (driving_pattern_depth == 1) // Driving pattern is New
		{
			al_el1 = new_al_col1->get_al_el(0) ;
			al_el1->set_el_obj(driving_symbol) ;
			al_el1->set_original_pattern(driving_pattern) ;
		}
		else for (i = 0; i < driving_pattern_depth; i++)
		{
			al_el1 = new_al_col1->get_al_el(i) ;
			al_el2 = driving_symbol->get_al_el(i) ;
			*al_el1 = *al_el2 ;
		}
		if (driving_symbol->is_a_hit())
			new_al_col1->set_is_a_hit(true) ;
	}
	
	if (target_symbol != NIL)
	{
		target_pattern_depth = target_pattern->get_sequence_depth() ;
		if (target_pattern_depth == 1) // Target pattern is one of
			// original patterns in Old
		{
			al_el1 = new_al_col1->
				get_al_el(row_one_of_t_patt_in_al_constr) ;
			driving_new_symbol = (symbol *)al_el1->get_el_obj() ;
			if (driving_new_symbol != NIL &&
				driving_new_symbol != target_symbol)
				abort_run("Anomaly in make_column \
				(target_pattern_depth == 1)") ;
			al_el1->set_el_obj(target_symbol) ;
			al_el1->set_original_pattern(target_pattern) ;
		}
		else if (rows_0_to_be_merged == true)
		{
			// We know that the depth of the target pattern is > 1.
			
			target_patt_al_constr_last_row = 
				row_one_of_t_patt_in_al_constr +
				target_pattern_depth - 2 ;
			
			// Install the first row of the target symbol in
			// row 0 of the new alignment, provided that,
			// if there is a symbol there already, it is
			// the same as the one to be installed.
			
			al_el1 = new_al_col1->get_al_el(0) ;
			driving_new_symbol = (symbol *)al_el1->get_el_obj() ;
			al_el2 = target_symbol->get_al_el(0) ;
			
			if (driving_new_symbol == NIL) *al_el1 = *al_el2 ;
			
			// Now install the remaining rows of the target
			// symbol.
			
			j = 1 ; // j will index the rows of the
				// target alignment after the first.
			for (i = row_one_of_t_patt_in_al_constr;
				i <= target_patt_al_constr_last_row; i++)
			{
				al_el1 = new_al_col1->get_al_el(i) ;
				al_el2 = target_symbol->get_al_el(j++) ;
				*al_el1 = *al_el2 ;
			}
		}
		else
		{
			// We know that rows_0_to_be_merged == false.
			
			target_patt_al_constr_last_row = 
				row_one_of_t_patt_in_al_constr +
				target_pattern_depth - 1 ;
			
			j = 0 ; // j will index the rows of the
				// target alignment after the first.
			for (i = row_one_of_t_patt_in_al_constr;
			i <= target_patt_al_constr_last_row; i++)
			{
				al_el1 = new_al_col1->get_al_el(i) ;
				al_el2 = target_symbol->get_al_el(j++) ;
				*al_el1 = *al_el2 ;
			}
		}		
		if (target_symbol->is_a_hit())
			new_al_col1->set_is_a_hit(true) ;
	}
	
	return(new_al_col1) ;
} // make_column

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Creates a character string representing the current year as 4 digits
**	plus string terminating character.
**
** CALLING SEQUENCE:
**
**	void year(char *yr)
**
** FORMAL ARGUMENTS:
**
**	Return value:	Character string.
**
**	yr				An array of 5 characters.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void year(char *yr)
{
	struct tm *t_now ;
	time_t s_now ;

	tzset() ;
	time(&s_now) ;
	t_now = localtime(&s_now) ;
	strftime(yr, 5,
		"%Y",
		t_now);
	yr[5] = '\0' ;
} // year

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Converts integer or boolean values to character strings.
**
** CALLING SEQUENCE:
**
**	char *convert(int v, bool tf)
**
** FORMAL ARGUMENTS:
**
**	Return value:	A character string.
**
**	v				An integer value to be converted. NULL_VALUE means 
**					that the boolean should be converted.
**	tf				A boolean value to be converted to "ON" or "OFF".
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

char *convert(int v, bool tf)
{
	if (v == NULL_VALUE)
	{
		if (tf == true) return("ON") ;
		else return("OFF") ;
	}
	else
	{
		if (v == PORTRAIT) return("PORTRAIT") ;
		if (v == LANDSCAPE) return("LANDSCAPE") ;
		if (v == 'H') return("HORIZONTAL") ;
		if (v == 'V') return("VERTICAL") ;
	}

	return("INVALID CONVERSION") ;
} // convert

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Prints the current time.
**
** CALLING SEQUENCE:
**
**	void time_now()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	From the computer clock.
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void time_now()
{
	struct tm *tm_now ;
	time_t sec_now ;
	char str[80] ;
	
	tzset() ;
	time(&sec_now) ;
	tm_now = localtime(&sec_now) ;
	strftime(str, 80,
		"Date and time, GMT (ISO 8601): 20%y-%m-%d %I:%M (%A)",
		tm_now);
	fprintf(output_file, "%s\n\n", str);
} // time_now

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Prepares the ground for unify_best by copying the nodes of
**	a hit sequence into hit_entries[]. This means that it is easy
**	to step through the nodes in left-to-right order - which is
**	more intuitive than always using them in reverse order.
**
** CALLING SEQUENCE:
**
**	void prepare_hit_sequence(hit_node *leaf_node)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	leaf_node:	The leaf node of the hit sequence in the hit structure.
**
** IMPLICIT INPUTS:
**
**	The hit structure
**
** IMPLICIT OUTPUTS:
**
**	hit_entries[]
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void prepare_hit_sequence(hit_node *leaf_node)
{
	hit_node *h_node, *left_h_node ;
	int node_position ;
	
	// First count the number of nodes in the hit sequence
	// so that a check can be made that it does not exceed the
	// size of hit_entries[] and to find the correct place
	// in hit_entries[] to begin loading the nodes.
	
	node_position = -1 ;
	for (h_node = leaf_node; h_node != hit_root;
	h_node = h_node->get_hn_parent())
		node_position++ ;
	
	fe_hit_entries = node_position + 1 ;
	if (fe_hit_entries >= MEDIUM_SCRATCH_ARRAY_SIZE)
		abort_run("Chain of hit_nodes is too long in unify_best") ;
	
	for (left_h_node = leaf_node; left_h_node != hit_root;
		left_h_node = left_h_node->get_hn_parent())
			hit_entries[node_position--].h_node = left_h_node ;

} // prepare_hit_sequence

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks for mismatches in a hit sequence (in unify_best).
**
** CALLING SEQUENCE:
**
**	bool mismatch_found()
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if there are mismatches, false otherwise.
**
** IMPLICIT INPUTS:
**
**	hit_entries[] and fe_hit_entries
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool mismatch_found(hit_node *leaf_node)
{
	hit_node *h_node, *left_h_node, *right_h_node ;
	sequence *driving_pattern, *target_pattern ;
	symbol *last_driving_symbol, *last_target_symbol ;
	
	// Unless the driving pattern is the same as New, a 'mis-match'
	// occurs if, at any point in the alignment between one hit (for the
	// the current cycle) and the next hit (for the current cycle), there 
	// are one or more non-hit symbols from the driving pattern AND
	// one or more non-hit symbols from the target pattern. If there
	// is any mis-match in an alignment, it is not valid and should
	// not be formed.
	
	h_node = leaf_node ;
	driving_pattern = h_node->get_driving_pattern() ;
	if (driving_pattern == current_new_pattern) return(false) ;
	
	// The driving_pattern is not the same as New.
	
	// Check for mis-matches *after* the last hit (for the current cycle).
	
	target_pattern = h_node->get_target_pattern() ;
	last_driving_symbol = (symbol *)
		driving_pattern->get_last_child() ;
	last_target_symbol = (symbol *)
		target_pattern->get_last_child() ;
	if (last_driving_symbol != h_node->get_driving_symbol()
		&& last_target_symbol != h_node->get_target_symbol())
	{
		if (verbose)
		{
			if (audit_trail)
			{
				fprintf(output_file, "%s%d%s%d%s%d%s",
					"Mismatch (#",
					leaf_node->get_ID(),
					"):\n     driving_pattern ID",
					driving_pattern->get_ID(),
					"\n     target_pattern ID",
					target_pattern->get_ID(),
					" after last hit.\n\n") ;
			}
		}
		return(true) ;
	}
	
	// Check for mis-matches between the first and last hits.
	
	left_h_node = NIL ;
	for (int i = 0; i < fe_hit_entries; i++)
	{
		right_h_node = hit_entries[i].h_node ;
		if (left_h_node == NIL)
		{
			left_h_node = right_h_node ;
			continue ;
		}
		if (right_h_node->get_driving_int_pos() -
			left_h_node->get_driving_int_pos() > 1
			&& right_h_node->get_target_int_pos() -
			left_h_node->get_target_int_pos() > 1)
		{
			if (verbose)
			{
				if (audit_trail)
				{
					fprintf(output_file, "%s%d%s%d%s%s%s%d%s%s%s",
						"Mismatch (#",
						leaf_node->get_ID(),
						"):\n     driving_pattern ID",
						driving_pattern->get_ID(),
						"\n     after driving_symbol ",
						(left_h_node->get_driving_symbol())->get_name(),
						"\n     target_pattern ID",
						target_pattern->get_ID(),
						"\n     after target_symbol ",
						(left_h_node->get_target_symbol())->get_name(),
						"\n\n") ;
				}
			}
			return(true) ;
		}
		left_h_node = right_h_node ;
	}
	
	// Now check for mis-matches before the first hit.
	
	h_node = hit_entries[0].h_node ;
	if (h_node->get_driving_int_pos() > 0
		&& h_node->get_target_int_pos() > 0)
	{
		if (verbose)
		{
			if (audit_trail)
			{
				fprintf(output_file, "%s%d%s%d%s%d%s",
					"Mismatch (#",
					leaf_node->get_ID(),
					"):\n     driving_pattern ID",
					driving_pattern->get_ID(),
					"\n     target_pattern ID",
					target_pattern->get_ID(),
					" before first hit.\n\n") ;
			}
		}
		return(true) ;
	}
	
	// No mismatch has been detected.
	
	return(false) ; 		
} // mismatch_found

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Forms an alignment from a hit sequence. This function is called
**	from unify_best.
**
** CALLING SEQUENCE:
**
**	sequence *make_one_alignment(hit_node *leaf_node)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The alignment which is formed. NIL if, for some reason,
**					the hit sequence proves to be invalid.
**
**	leaf_node:		The leaf node of the hit sequence from which the new
**					alignment is to be formed.
**
** IMPLICIT INPUTS:
**
**	hit_entries[], fe_hit_entries
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

sequence *make_one_alignment(hit_node *leaf_node)
{
	sequence *driving_pattern, *target_pattern, *temp_patt, *al_constr ;
	symbol *newly_created_col, *driving_symbol, *target_symbol,
		*hit_target_symbol, *newly_created_hit_col, *symbol1,
		*symbol2 ;
	int driving_pattern_depth, target_pattern_depth,
		al_constr_depth, hit_node_index,
		row_one_of_t_patt_in_al_constr,
		driving_int_pos, target_int_pos, i, j, row1 ;
	hit_node *h_node ;
	alignment_element *al_el1, *al_el2 ;
	bool rows_0_to_be_merged ;
	
	// 1 CALCULATE THE DEPTH OF THE NEW ALIGNMENT AS FOLLOWS:
	// * If the driving pattern and the target pattern both have
	//   a depth of 1, depth of new alignment A = 2.
	// * If the driving pattern has a depth D = 1 and the target
	//   pattern has a depth T > 1, then A = T. This is because
	//   the driving pattern will be the second or subsequent of
	//	 the patterns in New and it will occupy the same row 0
	//   as the target pattern.
	// * If D > 1 and T > 1, then A = D + T - 1. This is because
	//   the driving alignment and the target alignment will 
	//   share the same row 0.
	// * If D > 1 and T = 1, then A = D + T. This is because the 
	//   target pattern will be one of the original patterns in Old
	//   and it will not share a row 0 with the driving pattern.
		
	driving_pattern = leaf_node->get_driving_pattern() ;
	driving_pattern_depth = driving_pattern->get_sequence_depth() ;
	target_pattern = leaf_node->get_target_pattern() ;
	target_pattern_depth = target_pattern->get_sequence_depth() ;

	if (driving_pattern_depth == 1 && target_pattern_depth == 1)
	{
		al_constr_depth = 2 ;
		rows_0_to_be_merged = false ;
	}
	else if (driving_pattern_depth == 1 && target_pattern_depth > 1)
	{
		al_constr_depth = target_pattern_depth ;
		rows_0_to_be_merged = true ;
	}
	else if (driving_pattern_depth > 1 && target_pattern_depth > 1)
	{
		al_constr_depth = driving_pattern_depth + target_pattern_depth - 1 ;
		rows_0_to_be_merged = true ;
	}
	else if (driving_pattern_depth > 1 && target_pattern_depth == 1)
	{
		al_constr_depth = driving_pattern_depth + target_pattern_depth ;
		rows_0_to_be_merged = false ;
	}
	else abort_run("Anomalous values for driving_pattern_depth and \
		target_pattern_depth in make_one_alignment()") ;

	if (al_constr_depth >= MAX_ALIGNMENT_DEPTH)
		abort_run("Too many rows in alignment in unify_best") ;
	
	row_one_of_t_patt_in_al_constr = driving_pattern_depth ;
	
	// 2 CREATE AN OBJECT FOR A NEW ALIGNMENT. 
	
	al_constr = new sequence ;
	
	// 3 IF THE DRIVING PATTERN IS NOT THE SAME AS NEW, COPY IN
	// ALL THE COLUMNS FOR THE DRIVING PATTERN BUT
	// USING COLUMNS WITH A LENGTH OF AL_CONSTR_DEPTH. IN hit_entries[]
	// KEEP A RECORD OF WHICH OF THE NEWLY-CREATED COLUMNS REPRESENTS
	// A HIT.
	
	// IF THE DRIVING PATTERN IS THE SAME AS NEW,
	// DO THE SAME BUT ONLY FOR THOSE SYMBOLS IN THE DRIVING_PATTERN
	// THAT FORM HITS WITH THE TARGET PATTERN.
	
	hit_node_index = 0 ;
	if (driving_pattern == current_new_pattern)
	{
		// The driving pattern is the same as New. Only make
		// columns for those symbols in the driving pattern which
		// form hits with the target pattern.
		
		for (hit_node_index = 0; hit_node_index < fe_hit_entries; hit_node_index++)
		{
			h_node = hit_entries[hit_node_index].h_node ;
			driving_symbol = h_node->get_driving_symbol() ;
			driving_pattern = h_node->get_driving_pattern() ;
			driving_int_pos = driving_pattern->
				get_int_pos_by_child(driving_symbol) ;
			target_symbol = h_node->get_target_symbol() ;
			target_pattern = h_node->get_target_pattern() ;
			target_int_pos = target_pattern->
				get_int_pos_by_child(target_symbol) ;
			newly_created_col = make_column(al_constr_depth,
				driving_pattern, driving_symbol, 
				driving_int_pos, target_pattern, target_symbol, 
				target_int_pos, row_one_of_t_patt_in_al_constr, 
				h_node->get_ID(), true, rows_0_to_be_merged) ;
			
			al_constr->add_child(newly_created_col) ;

			hit_entries[hit_node_index].newly_created_col = 
				newly_created_col ;
		}
	}
	else
	{
		// The driving pattern is not the same as New.

		driving_pattern->initialise() ;
		while (driving_symbol = (symbol *)
			driving_pattern->get_next_child())
		{
			driving_int_pos = driving_pattern->
				get_int_pos_by_child(driving_symbol) ;
			
			// Note: the last field below (for whether or not
			// this column contains a hit) is set to false
			// as a default. The value is set prooerly a few
			// lines down.
			
			newly_created_col = make_column(al_constr_depth,
				driving_pattern, driving_symbol, 
				driving_int_pos, target_pattern, NIL, 
				NULL_VALUE, row_one_of_t_patt_in_al_constr, 
				NULL_VALUE, false, false) ;
			
			al_constr->add_child(newly_created_col) ;
			
			// Now check to see whether this driving symbol is
			// part of a hit. If it is, make a note in hit_entries
			// to show the correspondence between this hit
			// and newly_created_col.
			
			if (hit_node_index >= fe_hit_entries) continue ;
			h_node = hit_entries[hit_node_index].h_node ;
			if (driving_symbol == h_node->get_driving_symbol())
			{
				hit_entries[hit_node_index].newly_created_col = 
					newly_created_col ;
				newly_created_col->
					set_h_node_ID(h_node->get_ID()) ;
				newly_created_col->
					set_is_a_hit(true) ;
				hit_node_index++ ;
			}
		}
	}
	
	// 4 STEP THROUGH THE HIT SEQUENCE FILLING IN THE DETAILS OF
	// THE TARGET SYMBOL IN EACH COLUMN REPRESENTING A HIT.
	
	for (hit_node_index = 0; hit_node_index < fe_hit_entries; hit_node_index++)
	{
		h_node = hit_entries[hit_node_index].h_node ;
		newly_created_col = hit_entries[hit_node_index].newly_created_col ;
		
		target_symbol = h_node->get_target_symbol() ;
		
		if (target_symbol->get_sequence_depth() == 1)
		{
			target_pattern = h_node->get_target_pattern() ;
			target_int_pos = h_node->get_target_int_pos() ;
			al_el2 = newly_created_col->
				get_al_el(row_one_of_t_patt_in_al_constr) ;
			al_el2->set_el_obj(target_symbol) ;
			al_el2->set_original_pattern(target_pattern) ;
		}
		else if (rows_0_to_be_merged)
		{
			// The depth of the target pattern is > 1.
			
			// Copy the alignment elements of the target
			// symbol into the lower alignment elements
			// in newly_created_col.
			
			// Row 0 of the target symbol is treated separately.
			// If row 0 contains a symbol (from New)
			// and if there is not already a symbol (from New)
			// in al_constr, then copy the alignment_element
			// across into row 0 of the alignment.
			
			al_el1 = target_symbol->get_al_el(0) ;
			al_el2 = newly_created_col->get_al_el(0) ;
			if (al_el1->get_el_obj() != NIL)
			{
				if (al_el2->get_el_obj() == NIL)
				{
					*al_el2 = *al_el1 ;
				}
			}
			
			j = row_one_of_t_patt_in_al_constr ;
			for (i = 1; i < target_pattern_depth; i++)
			{
				al_el1 = target_symbol->get_al_el(i) ;
				al_el2 = newly_created_col->get_al_el(j) ;
				*al_el2 = *al_el1 ;
				j++ ;
			}
		}
		else
		{
			// Depth of the target_pattern is > 1 and
			// rows_0_to_be_merged == false.
			
			// Copy the alignment elements of the target
			// symbol into the lower alignment elements
			// in newly_created_col.
			
			j = row_one_of_t_patt_in_al_constr ;
			for (i = 0; i < target_pattern_depth; i++)
			{
				al_el1 = target_symbol->get_al_el(i) ;
				al_el2 = newly_created_col->get_al_el(j) ;
				*al_el2 = *al_el1 ;
				j++ ;
			}
		}
	}
	
	// 5 STEP THROUGH THE TARGET PATTERN AND THE TABLE OF HIT_ENTRIES[],
	// INSERTING NEW COLUMNS IN AL_CONSTR CORRESPONDING TO SYMBOLS
	// IN THE TARGET PATTERN THAT HAVE NOT FORMED HITS
	// WITH THE DRIVING PATTERN.
	
	target_symbol = (symbol *)target_pattern->get_first_child() ;
	for (hit_node_index = 0; hit_node_index < fe_hit_entries;
		hit_node_index++)
	{
		h_node = hit_entries[hit_node_index].h_node ;
		newly_created_hit_col = hit_entries[hit_node_index].newly_created_col ;
		hit_target_symbol = h_node->get_target_symbol() ;
		while (target_symbol != hit_target_symbol)
		{
			newly_created_col = make_column(al_constr_depth,
				driving_pattern, NIL, 
				NULL_VALUE, target_pattern, target_symbol, 
				target_pattern->get_int_pos_by_child(target_symbol), 
				row_one_of_t_patt_in_al_constr,
				NULL_VALUE, false, rows_0_to_be_merged) ;
			al_constr->precede(newly_created_col,
				newly_created_hit_col) ;
			target_symbol = (symbol *)
				target_pattern->get_next_child() ;
		}
		
		// Step past the target symbol which corresponds to the
		// current hit_target_symbol and get the next target
		// symbol.
		
		target_symbol = (symbol *)target_pattern->get_next_child() ;
	}
	
	// 6 DO TAIL OF THE TARGET PATTERN (IF ANY).
	
	while (target_symbol != NIL)
	{
		newly_created_col = make_column(al_constr_depth,
			driving_pattern, NIL, NULL_VALUE, 
			target_pattern, target_symbol, target_pattern->
			get_int_pos_by_child(target_symbol), 
			row_one_of_t_patt_in_al_constr, NULL_VALUE,
			false, rows_0_to_be_merged) ;
		al_constr->add_child(newly_created_col) ;
		target_symbol = (symbol *)target_pattern->get_next_child() ;
	}

	// Check for anomalies in the ordering of symbols in
	// current_new_pattern within the alignment.

	if (al_constr->check_ordering_of_new() == false)
	{
		delete al_constr ;
		if (verbose)
		{
			if (audit_trail)
			{
				fprintf(output_file, 
					"Anomalous ordering of symbols from New in hit sequence ") ;
				leaf_node->print_ID() ;
				fprintf(output_file, ". No new alignment is formed.\n\n") ;
			}
		}
		return(finish_u_b(NIL)) ;
	}
	
	al_constr->set_sequence_depth(al_constr_depth) ;
	al_constr->set_leaf_node_ID(leaf_node->get_ID()) ;
	
	// Check that gaps between hits conform with relevant parameters.
	
	if (al_constr->al_gaps_OK() == false)
	{
		if (verbose)
		{
			if (audit_trail)
			{
				fprintf(output_file, "Gaps too large in alignment ") ;
				al_constr->print_ID() ;
				fprintf(output_file, ". Alignment is discarded.\n\n") ;
			}
		}
		delete al_constr ;
		return(finish_u_b(NIL)) ;
	}
	
	// 7 MAKE SURE THAT EVERY CELL OF EACH ROW OF THE ALIGNMENT AFTER ROW 0 CONTAINS
	// A REFERENCE TO THE PATTERN(S) FOR THAT ROW. Row 0 may contain references
	// to more than one pattern and contains NIL entries for alignment symbols
	// that contain no symbol in row 0 - so this row is skipped over
	
	for (row1 = 1; row1 < al_constr_depth; row1++)
	{
		// First find the first reference to the pattern for
		// this row.

		list_for(symbol1, symbol, al_constr)		
		{
			temp_patt = symbol1->get_row_pattern(row1) ;
			if (temp_patt != NIL) break ;
		}
		
		// Now put the value of temp_patt in every cell for this row.
		
		list_for(symbol1, symbol, al_constr)		
		{
			al_el1 = symbol1->get_al_el(row1) ;
			al_el1->set_original_pattern(temp_patt) ;
		}
	}

	// 8 MARK POSITIONS OF SYMBOLS IN INTEGERS AND SHOW PARENT
	// OF EACH SYMBOL. MARK HIT COLUMNS AND NON-HIT COLUMNS.
	
	al_constr->mark_parent_and_int_positions_non_recursive() ;
	list_for(symbol1, symbol, al_constr)
		symbol1->check_is_a_hit() ;

	// 9 MARK STATUS OF SYMBOLS

	list_for(symbol1, symbol, al_constr)
	{
		if (symbol1->is_a_hit())
			symbol1->set_status(CONTENTS) ;
		else
		{
			// Find the status of the single symbol in this
			// column and assign this status to the column.

			symbol2 = symbol1->get_first_symbol_in_column() ;
			symbol1->set_status(symbol2->get_status()) ;
		}
	}

	// Set status of bracket symbols separately.

	al_constr->assign_brackets_status() ;

	// 10 Fill in values for built_from_driving and built_from_target and, each column, values
	// for same_column_above and same_column_below.

	al_constr->set_built_from_driving(driving_pattern) ;
	al_constr->set_built_from_target(target_pattern) ;

	symbol *col1 ;
	list_for(col1, symbol, al_constr)
		col1->set_symbol_matches() ;

	return(al_constr) ;
} // make_one_alignment

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Finishes off unify_best.
**
** CALLING SEQUENCE:
**
**	sequence *finish_u_b(sequence *al_constr)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The alignment (if any) formed by unify_best. The
**					value is NIL if no new alignment is formed.
**
**	al_constr:		The alignment (if any) formed by unify_best. The
**					value is NIL if no new alignment is formed.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

sequence *finish_u_b(sequence *al_constr)
{
#if MARK_UNIFY_BEST
	
	if (audit_trail)
	{
		fprintf(output_file, "FINISHED unify_best (") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, ") LEAF NODE ID = ") ;
		leaf_node->print_ID() ;
		fprintf(output_file, "\n\n") ;
		fflush(output_file) ;
	}
	
#endif
	return(al_constr) ;
} // finish_u_b

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	From the best hit sequence in the hit_structure, this function
**	forms a new alignment. 
**
** CALLING SEQUENCE:
**
**	sequence *unify_best(int leaf_index)
**
** FORMAL ARGUMENTS:
**
**	Return value:					The new sequence, NIL if none 
**									is formed.
**
**	leaf_index:						The position of the last node of 
**									the current hit sequence in the 
**									leaf_array[].
**
** IMPLICIT INPUTS:
**
**	The hit_structure
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

sequence *unify_best(int leaf_index)
{
	sequence *matching_sequence, *al_constr ;
	hit_node *leaf_node = leaf_array[leaf_index] ;
	
	if (leaf_node == NIL)
		abort_run("NIL value for leaf_node in unify_best") ;
	
#if MARK_UNIFY_BEST
	
	if (audit_trail)
	{
		fprintf(output_file, "STARTING unify_best (") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, ") LEAF NODE ID = ") ;
		leaf_node->print_ID() ;
		fprintf(output_file, "\n\n") ;
		fflush(output_file) ;
	}
	
#endif
	
	// 1 LOAD NODES OF THE HIT SEQUENCE INTO hit_entries[] SO THAT
	// THEY CAN BE ACCESSED EASILY IN LEFT-TO-RIGHT ORDER (WHICH
	// IS MORE INTUITIVE).
	
	prepare_hit_sequence(leaf_node) ;

	// 2 CHECK FOR MISMATCHES (WHEN DRIVING PATTERN != NEW)
	
	if (mismatch_found(leaf_node)) return(finish_u_b(NIL)) ;
	
	// 3 FORM AN ALIGNMENT
	
	al_constr = make_one_alignment(leaf_node) ;
	if (al_constr == NIL) return(finish_u_b(al_constr)) ;
	
	// 4 CLASSIFY ALIGNMENT BY DEGREE OF MATCHING
	// AND RECOMPUTE SCORES OF THE NEW ALIGNMENT

	al_constr->find_degree_of_matching(current_new_pattern) ;

	int degree_of_matching = al_constr->get_degree_of_matching() ;

	if (match_all_data_symbols
		|| match_all_data_symbols_in_hierarchical_alignments)
	{
		if (degree_of_matching == PARTIAL)
		{
			if (audit_trail && verbose)
			{
				fprintf(output_file, "%s%d%s",
					"Alignment ID",
					al_constr->get_ID(),
					" is only partially-matched and is deleted.\n\n") ;
			}
			delete al_constr ;
			return(NIL) ;
		}
	}

	al_constr->make_code(false) ;
	
	leaf_node->set_new_symbols_cost(al_constr->get_new_symbols_cost()) ;
	leaf_node->set_encoding_cost(al_constr->get_encoding_cost()) ;
	leaf_node->set_compression_ratio(al_constr->get_compression_ratio()) ;
	leaf_node->set_compression_difference(al_constr->
		get_compression_difference()) ;
	
	// 5 CHECK FOR LOW SCORING ALIGNMENT
	
	if (al_constr->get_compression_difference() <= fail_score)
	{
		if (verbose)
		{
			if (audit_trail)
			{
				fprintf(output_file, "Alignment ") ;
				al_constr->print_ID() ;
				fprintf(output_file, " for node ") ;
				leaf_node->print_ID() ;
				fprintf(output_file, " has a fail score and is deleted.\n\n") ;
				fflush(output_file) ;
			}
		}

		leaf_node->set_compression_difference(fail_score) ;
		
		delete al_constr ;
		
		al_constr = NIL ;
		return(finish_u_b(al_constr)) ;
	}
	
	if (show_al_structure)
		al_constr->show_al_structure(leaf_node) ;
	
	// 6 CHECK WHETHER NEW ALIGNMENT IS A DUPLICATE OF ANY
	// PREVIOUSLY-FORMED ALIGNMENT
	
	// Match the new sequence against sequences formed earlier in
	// this cycle and previous cycles. If there is a match, delete
	// new sequence and return NIL.
	
	if (matching_sequence = 
		al_constr->matches_earlier_alignment())
	{
		if (verbose)
		{
			if (audit_trail)
			{
				fprintf(output_file, "Alignment ") ;
				al_constr->print_ID() ;
				fprintf(output_file, " for node ") ;
				leaf_node->print_ID() ;
				fprintf(output_file, " matches sequence ") ;
				matching_sequence->print_ID(),
				fprintf(output_file, " and is discarded.\n\n") ;
				fflush(output_file) ;
			}
		}

		leaf_node->set_compression_difference(fail_score) ;		
		delete al_constr ;	
		al_constr = NIL ;
		return(finish_u_b(al_constr)) ;
	}
	
	// 8 DELIVER NEW ALIGNMENT

	alignments_array[leaf_index] = al_constr ;

	// 9 CHECK AND MARK ANY BRACKETS IN THE NEW ALIGNMENT

	if (recognise_brackets)
		al_constr->check_brackets_in_one_pattern() ;

	return(finish_u_b(al_constr)) ;
	
} /* unify_best */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Inserts an sequence into group_of_patterns in so that
**	sequences are in order of their CR.
**
** CALLING SEQUENCE:
**
**	void insert_sequence_in_order_of_CR(sequence *al1,
**		group *group_of_patterns)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	al1:				The sequence to be inserted.
**	group_of_patterns:	The set of sequences.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	group_of_patterns with addition of al1.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void insert_sequence_in_order_of_CR(sequence *al1,
	group *group_of_patterns)
{
	sequence *existing_sequence ;
	double al_CD = al1->get_compression_difference() ;
	
	group_of_patterns->initialise() ;
	while (existing_sequence =
		(sequence *)group_of_patterns->get_next_child())
	{
		if (al_CD < existing_sequence->get_compression_difference())
			continue ;
		
		group_of_patterns->precede(al1, existing_sequence) ;
		
		return ;	
	}
	
	/* If the program gets this far, the al1 has a larger
	CR than any of the existing sequences so it should be added
	at the end of the set. */
	
	group_of_patterns->add_child(al1) ;
	
} /* insert_sequence_in_order_of_CR */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Removes from current_alignments_from_parsing
**	any sequences which are not marked as KEEP. At the same time, clear the 
**	symbol_selection_array[][] ready for the next cycle.
**
** CALLING SEQUENCE:
**
**	void purge_al_constrs()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	alignments_array[]
**
** IMPLICIT OUTPUTS:
**
**	Alignments and codes removed from alignments_array[]
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void purge_al_constrs()
{
	sequence *sequence1, *sequence2 ;
	int row1, col1 ;
	
	if (audit_trail)
	{
		fprintf(output_file, "STARTING purge_al_constrs ") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, "\n\n") ;
		fflush(output_file) ;
	}
	
	// Clear symbol_selection_array[][] to avoid dangling pointers.
	
	for (row1 = 0; row1 < MAX_NEW_ARRAY_ROWS; row1++)
	{
		for (col1 = 0; col1 < MCOLS; col1++)
			symbol_selection_array[row1][col1] = NIL ;
	}
	
	/* Delete the sequences which are not marked as KEEP
	and retain the ones that are. At the same time, reset
	the value of keep ready for the next cycle. Each time
	an alignment deleted, it is necessary to check that
	it is not referenced by 'built_from_driving' or
	'built_from_target'. Relevant variable are set to NIL. */
	
	number_of_alignments_purged = 0 ;
	bool non_keep_alignment_found = true ;
	list_element *el_pos1 ;
	while (non_keep_alignment_found)
	{
		non_keep_alignment_found = false ;
		list_for(sequence1, sequence, current_alignments_from_parsing)
		{
			/* Check to see if sequence1 is marked as KEEP
			or DRIVING. */
		
			if (sequence1->get_keep() == KEEP
				|| sequence1->get_keep() ==	DRIVING)
					continue ;

			if (audit_trail)
			{
				fprintf(output_file, "Alignment ") ;
				sequence1->print_ID() ;
				fprintf(output_file, " is purged\n\n") ;
			}

			// Any variables 'built_from' that point to
			// sequence1 must be set to NIL.

			el_pos1 = NIL ;
			list_for_el_pos(sequence2, sequence, current_alignments_from_parsing,
				el_pos1)
			{
				if (sequence2->get_built_from_driving() == sequence1)
					sequence2->set_built_from_driving(NIL) ;
				if (sequence2->get_built_from_target() == sequence1)
					sequence2->set_built_from_target(NIL) ;
			}

			current_alignments_from_parsing->delete_child(sequence1) ;
			number_of_alignments_purged++ ;
			non_keep_alignment_found = true ;
			break ;
		}
	}
	
	current_number_of_alignments =
		current_alignments_from_parsing->count_number_of_children() ;
	
	if (audit_trail)
	{
		fprintf(output_file, "%s%d%s%d%s",
			"Number of alignments purged = ",
			number_of_alignments_purged,
			"\n\nNumber of alignments retained = ",
			current_number_of_alignments, "\n\n") ;
		
		fprintf(output_file, "FINISHED purge_al_constrs (") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, ")\n\n") ;
		fflush(output_file) ;
	}
	
} /* purge_al_constrs */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Selects a sub-set of the alignments in current_alignments_from_parsing,
**	including alignments formed in the current cycle and those from
**	previous cycles.
**
**	This function uses the symbol_selection_array[][] to ensure that
**	as many sections of New as possible are represented in the
**	selection.
**
**	The function computes values for the global variable 
**	current_number_of_alignments.
**
**	For each selected alignment, the function computes a code
**	corresponding to the alignment (provided it contains coded 
**	patterns).
**
**
** CALLING SEQUENCE:
**
**	void make_selection()
**
** FORMAL ARGUMENTS:
**
**	Return value:			void
**
** IMPLICIT INPUTS:
**
**	old_patterns
**
** IMPLICIT OUTPUTS:
**
**	Values for the global variable current_number_of_alignments.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void make_selection()
{
	symbol *B_symbol1, *symbol1, *symbol2 ;
	alignment_element *al_el1 ;
	int row1, row2, col,
		number_of_driving_alignments = 0,
		max_entries_in_symbol, col_n, symbol_counter, leaf_index ;
	sequence *sequence1 ;

	if (audit_trail)
	{
		fprintf(output_file, "Start of make_selection()\n\n") ;
	}
	
	// Transfer alignments to current_alignments_from_parsing.

	int hit_node_row = 0, counter = 0 ;
	sequence *al1 ;
	hit_node *h_node1 ;
	

	while (h_node1 = get_leaf_nodes_in_order(&hit_node_row, &al1, 
		&leaf_index))
	{
		if (al1 == NIL) continue ;
		alignments_array[leaf_index] = NIL ;
		current_alignments_from_parsing->add_child(al1) ;
		counter++ ;
	}

	if (audit_trail)
	{
		fprintf(output_file, "%s%d%s%d%s",
			"The number of alignments added to current_alignments_from_parsing = ",
			counter,
			"\n\nThe total number of alignments in current_alignments_from_parsing = ",
			current_alignments_from_parsing->count_number_of_children(),
			"\n\n") ;
	}

	// Sort current_alignments_from_parsing by compression score.

	current_alignments_from_parsing->sort_by_compression_difference() ;

	// Now make entries in symbol_selection_array[]. At the same time,
	// reset values for KEEP.

	list_for(sequence1, sequence, current_alignments_from_parsing)
	{
		/* Look for symbols from New in the sequence which form
		hits with symbols in Old. Make entries in the corresponding
		symbols of symbol_selection_array[][]. */

		sequence1->set_keep(NULL_VALUE) ;
		
		list_for(B_symbol1, symbol, sequence1)
		{
			al_el1 = B_symbol1->get_al_el(0) ;
			symbol1 = (symbol *)al_el1->get_el_obj() ;
			if (symbol1 == NIL) continue ;
			if (al_el1->get_same_column_below() == NULL_VALUE) 
				continue ;
			
			// al_el1 is in a symbol/column which contains a hit symbol
			// in row 0. Find the position of this symbol in
			// concatenated_new_patterns.

			col_n = 0 ;
			list_for(symbol2, symbol, concatenated_new_patterns)
			{
				if (symbol1 == symbol2) break ;
				col_n++ ;
			}
			
			/* Check whether there are any empty slots for
			this symbol position in symbol_selection_array[][].
			If there are, insert a pointer to sequence1
			and quit. If not, proceed to next column in the alignment. */
			
			for (row2 = 0; row2 < keep_rows; row2++)
			{
				if (symbol_selection_array[row2][col_n] == NIL)
				{
					symbol_selection_array[row2][col_n] = sequence1 ;
					break ;
				}
			}
		}
	}
	
	/* Now traverse the symbol_selection_array[][] marking
	alignments to be kept and, amongst those alignments, the alignments
	to be used as driving patterns on the next cycle. In each list,
	no sequence is included more than once. */
	
	// Alignments to keep:

	for (row1 = 0; row1 < keep_rows; row1++)
	{
		for (col = 0; col < length_of_concatenated_new_patterns; col++)
		{
			sequence1 = symbol_selection_array[row1][col] ;
			if (sequence1 == NIL) continue ;

			// Mark sequence1 to be kept.

			sequence1->set_keep(KEEP) ;
		}
	}
	
	// Driving alignments:
	
	for (row1 = 0; row1 < driving_keep_rows; row1++)
	{
		for (col = 0; col < length_of_concatenated_new_patterns; col++)
		{
			sequence1 = symbol_selection_array[row1][col] ;
			if (sequence1 == NIL) continue ;

			// Mark sequence1 as DRIVING.

			sequence1->set_keep(DRIVING) ;
		}
	}

	number_of_driving_alignments = 0 ;
	list_for(sequence1, sequence, current_alignments_from_parsing)
	{
		if (sequence1->get_keep() == DRIVING)
			number_of_driving_alignments++ ;
	}
	
	/* Find the maximum number of entries in any symbol of the
	symbol_selection_array[][] and print it out. */
	
	max_entries_in_symbol = 0 ;
	for (col = 0; col < length_of_concatenated_new_patterns; col++)
	{
		symbol_counter = 0 ;
		for (row1 = 0; row1 < keep_rows; row1++)
		{
			if (symbol_selection_array[row1][col] == NIL) break ;
			symbol_counter++ ;
		}
		if (symbol_counter > max_entries_in_symbol)
			max_entries_in_symbol = symbol_counter ;
	}
	
	if (audit_trail)
	{
		fprintf(output_file, "FOR ") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, "%s%s%d%s",
			",\nTHE MAXIMUM NUMBER OF ENTRIES IN A COLUMN OF ",
			"symbol_selection_array[][] IS ",
			max_entries_in_symbol,
			"\n\n") ;
	}
	
	// Remove from current_alignments_from_parsing any alignment
	// which is not marked as KEEP or DRIVING. At the same time, clear the
	// symbol_selection_array[][] ready for the next cycle.
	
	purge_al_constrs() ;

	if (show_all_parsing_alignments)
	{
		// Print out current_alignments_from_parsing.

		if (audit_trail)
		{
			fprintf(output_file,
				"ALIGNMENTS FROM PARSING AFTER ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, "\n\n") ;

			list_for(al1, sequence, current_alignments_from_parsing)
			{
				al1->write_out_fully("SELECTED ALIGNMENT", NIL, NIL,
					write_section_chars_length, NULL_VALUE, true) ;
			}
		}
	}
	
	// Print ref numbers, CR and CD values for alignments that
	// are marked as KEEP and those that are marked as DRIVING.

	if (audit_trail)
	{
		fprintf(output_file, "%s%d%s",
			"LIST OF ", current_number_of_alignments,
			" ALIGNMENTS AFTER ") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, "\n\n") ;

		counter = 0 ;
		list_for(sequence1, sequence, current_alignments_from_parsing)
		{
			counter++ ;
			if (counter > 5)
			{
				fprintf(output_file, "\n") ;
				counter = 1 ;
			}
		
			sequence1->print_ID() ;
			// fprintf(output_file, "%s%1.2f%s%1.2f",
			//	" CR = ",
			//	sequence1->get_compression_ratio(),
			//	", CD = ",
			//	sequence1->get_compression_difference()) ;
			if (current_alignments_from_parsing->this_is_last_child(sequence1))
				fprintf(output_file, "\n\n") ;
			else fprintf(output_file, ", ") ;
		}

		fprintf(output_file, "%s%d%s",
			"LIST OF ", number_of_driving_alignments,
			" DRIVING ALIGNMENTS AFTER ") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, "\n\n") ;

		int print_counter = 0 ;
		int al_counter = 0 ;
		list_for(sequence1, sequence, current_alignments_from_parsing)
		{
			if (sequence1->get_keep() != DRIVING) continue ;
			print_counter++ ;
			if (print_counter > 5)
			{
				fprintf(output_file, "\n") ;
				print_counter = 1 ;
			}

			al_counter++ ;
			sequence1->print_ID() ;
			// fprintf(output_file, "%s%1.2f%s%1.2f",
			//	" CR = ",
			//	sequence1->get_compression_ratio(),
			//	", CD = ",
			//	sequence1->get_compression_difference()) ;
			if (al_counter == number_of_driving_alignments)
				fprintf(output_file, "\n\n") ;
			else fprintf(output_file, ", ") ;
		}

		fflush(output_file) ;
		
		fprintf(output_file, "End of make_selection()\n\n") ;
	}

} /* make_selection */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Creates Shannon-Fano-Elias codes according to the method in
**	Cover and Thomas, pp 101-103 but modified to remove rounding
**	of values and the addition of 1. The output is put in sfe_array[].
**
**	The S-F-E method gives a code with this size:
**
**		l(x) = ceil(log_2(1 / p(x)) + 1.
**
**	This function does not construct an explicit code. It simply
**	calculates:
**
**		l(x) = log_2(1 / p(x)).
**
** CALLING SEQUENCE:
**
**	void modified_sizes_of_sfe_codes(group *object_list)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	object_list:	The list of objects for which sizes of codes are to
**					be calculated.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Codewords and lengths for codewords in sfe_array[].
**
** SIDE EFFECTS:
**
**	NONE
*/

void modified_sizes_of_sfe_codes(group *object_list)
{
	double total_frequency = 0 ;
	base_object *obj1 ;
	int i, fe_sfe_array, size_of_array = 
		object_list->get_number_of_children() ;
	struct sfe_entry
	{
		base_object *obj ;
		double prob_x ;
		double code_size ;
	} *sfe_array ; 

	sfe_array = new sfe_entry[size_of_array] ;
	
	/* Clear sfe_array[]. */
	
	for (i = 0; i < size_of_array; i++)
	{
		
		sfe_array[i].obj = NIL ;
		sfe_array[i].prob_x = 0 ;
		sfe_array[i].code_size = 0 ;
	}
	
	/* Enter objects in sfe_array[] and find total frequency for objects. */
	
	i = 0 ;
	object_list->initialise() ;
	while (obj1 = object_list->get_next_child())
	{
		sfe_array[i].obj = obj1 ;
		plus_one(&i, VERY_LARGE_SCRATCH_ARRAY_SIZE,
			"sfe_array[] is too small") ;
		total_frequency += obj1->get_frequency() ;
	}
	
	fe_sfe_array = i ;
	
	/* Calculate the probability of each object and enter values
	in sfe_array[]. */

	int obj_frequency ;
	for (i = 0; i < fe_sfe_array; i++)
	{
		obj_frequency = sfe_array[i].obj->get_frequency() ;
		if (obj_frequency == 0) sfe_array[i].prob_x = 0 ;
		else sfe_array[i].prob_x = obj_frequency / total_frequency ;
	}
	
	/* Calculate values for code_size. Each one is the size of the
	code for the corresponding object. */

	double prob_x ;
	for (i = 0; i < fe_sfe_array; i++)
	{
		prob_x = sfe_array[i].prob_x ;
		if (prob_x == 0) sfe_array[i].code_size = 0 ;
		else sfe_array[i].code_size = log_2(1 / prob_x) ;
	}

	/* Apply the sizes to the objects in object_list. */
	
	for (i = 0; i < fe_sfe_array; i++)
		sfe_array[i].obj->set_encoding_cost(sfe_array[i].code_size) ;

	delete[] sfe_array ;
	
} /* modified_sizes_of_sfe_codes */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Tests whether the symbols up to and including 'end_symbol' in
**	'pattern' render that pattern unique within Old.
**
** CALLING SEQUENCE:
**
**	bool is_unique(sequence *test_pattern, symbol *end_symbol, 
**		sequence **matching_pattern)
**
** FORMAL ARGUMENTS:
**
**	Return value:		true if the combination of symbols up to and
**						including 'end_symbol' render the
**						test_pattern unique within old_patterns,
**						false otherwise.
**
**	test_pattern:		A pattern within old_patterns.
**	end_symbol: 		A symbol in test_pattern which marks the end
**						of the set of symbols to be tested.
**	matching_pattern:	A pointer to a variable which is NIL if 
**						test_pattern is unique or takes the value
**						of a pattern where the leading symbols match
**						the test_pattern.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

bool is_unique(sequence *test_pattern, symbol *end_symbol, 
				   sequence **matching_pattern)
{
	list_element *el_pos1 = NIL, *el_pos2, *el_pos3 ;
	sequence *pattern ;
	symbol *symbol1, *symbol2 ;

/*
	while (pattern = (sequence *)
		new_patterns->get_next_child_by_el_pos(&el_pos1))
	{
		if (pattern == test_pattern) continue ;
		el_pos2 = el_pos3 = NIL ;
		while (symbol1 = (symbol *)
			test_pattern->get_next_child_by_el_pos(&el_pos2))
		{
			symbol2 = (symbol *)
				pattern->get_next_child_by_el_pos(&el_pos3) ;
			if (symbol2 == NIL)
			{
				*matching_pattern = pattern ;
				return(false) ;
			}
			if (!symbol1->name_matches(symbol2)) break ;
			if (symbol1 == end_symbol)
			{
				*matching_pattern = pattern ;
				return(false) ;
			}
		}
	}
*/

	el_pos1 = NIL ;

	while (pattern = (sequence *)
		old_patterns->get_next_child_by_el_pos(&el_pos1))
	{
		if (pattern == test_pattern) continue ;
		el_pos2 = el_pos3 = NIL ;
		while (symbol1 = (symbol *)
			test_pattern->get_next_child_by_el_pos(&el_pos2))
		{
			symbol2 = (symbol *)
				pattern->get_next_child_by_el_pos(&el_pos3) ;
			if (symbol2 == NIL)
			{
				*matching_pattern = pattern ;
				return(false) ;
			}
			if (!symbol1->name_matches(symbol2)) break ;
			if (symbol1 == end_symbol)
			{
				*matching_pattern = pattern ;
				return(false) ;
			}
		}
	}
	*matching_pattern = NIL ;
	return(true) ;
} /* is_unique */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	A function for calculating the probability of obtaining at least one
**	hit when a symbol is matched against all the symbols in a random string
**	of symbols of length D. Details of the calculation are given in
**	"Architecture of Chance", pp 64-65. Each probability is translated
**	into an equivalent measure of information content which is itself
**	converted into a factor for scaling the existing value of the bit_cost
**	of the symbol involved in the matching.
**
**	This function is derived from
**	.../dev/random_counts/prob_at_least_one_hit.c.
**
** CALLING SEQUENCE:
**
**	void compute_scaling_factors()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	original_alphabet_size (the size of the alphabet of symbols)
**
** IMPLICIT OUTPUTS:
**
**	Values in scaling_factors[]
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void compute_scaling_factors()
{
	double	A = original_alphabet_size, /* Size of the alphabet */
		p, /* Probability of one hit. */
		q, /* Probability of one non-hit. */
		n_a, /* The number of instances of a particular assortment
		   of hits and non-hits. */
		D_temp, h_temp, nh_temp, D_i,
		p_a, /* The probability of one particular assortment of
		   hits and non-hits. */
		PROB, /* The probability of one *or more* hits in D. */
		p_temp, q_temp,
		info_not_rounded, /* The amount of information (in bits) 
							 conveyed by the occurrence of a hit with
							 probability PROB (not rounded). */
		max_info, /* The largest info value (which is
				always the value when the number of matches
				is 1. */
		scaling ; /* This is info / maximum value of info (the
				value in the first row). The scaling value can
				be multiplied by the bit_cost for a symbol to 
				obtain a value which is scaled down to reflect 
				the reduced value of a hit which occurs after 
				one or more gaps. When the gap is
				zero, the scaling value is always 1. */
	int D, // Length of data.
		h, // h is the number of hits.
		nh, // nh is the number of non-hits.
		h_i, nh_i, // h_i and hn_i are indices for the
			// respective values.
		index ;
	
	p = 1 / A ;
	q = (A - 1) / A ;
	
	for (D = 1; D <= SCALING_FACTORS_SIZE; D++)
	{
		PROB = 0 ;
		for (h = 1; h <= D; h++)
		{
			nh = D - h ;
			
			/* Calculate the probability of one assortment
			of h hits and nh non-hits. */
			
			p_temp = 1 ;
			for (h_i = 1; h_i <= h; h_i++)
				p_temp *= p ;
			
			q_temp = 1 ;
			for (nh_i = 1; nh_i <= nh; nh_i++)
				q_temp *= q ;
			
			p_a = p_temp * q_temp ;
			
			/* Now calculate the number of assortments
			with h hits and nh non-hits using the formula
			n_a = (D! / (h! * nh!)). */
			
			D_temp = D ;
			D_i = D - 1 ;
			while (D_i > 1)
			{
				D_temp *= D_i ;
				D_i -= 1 ;
			}
			
			h_temp = h ;
			if (h_temp == 0) h_temp = 1 ;
			else
			{
				h_i = h - 1 ;
				while (h_i > 1)
				{
					h_temp *= h_i ;
					h_i -= 1 ;
				}
			}
			
			nh_temp = nh ;
			if (nh_temp == 0) nh_temp = 1 ;
			else
			{
				nh_i = nh - 1 ;
				while (nh_i > 1)
				{
					nh_temp *= nh_i ;
					nh_i -= 1 ;
				}
			}
			
			if (h_temp < 0 || nh_temp < 0)
			{
				printf("%s%s%d%s%s%d%s",
					"Overflow on h_temp or nh_temp\n\n",
					"h_temp = ", h_temp, "\n\n",
					"nh_temp = ", nh_temp, "\n\n") ;
				/* fflush(output_file) ; */
				exit_routine(1) ;
			}
			
			n_a = D_temp / (h_temp * nh_temp) ;
			
			PROB += (p_a * n_a) ;
		}
		
		/* Calculate the amount of information (in bits) conveyed
		by the occurrence of a hit with probability PROB
		(modified Shannon-Fano-Elias method). */
		
		info_not_rounded = log_2(1 / PROB) ;
		
		if (D == 1) max_info = info_not_rounded ;
		
		scaling = info_not_rounded / max_info ;
		
		index = D - 1 ;
		
		// The following test is necessary to weed out negative
		// numbers which can occur with extreme values, apparently
		// due to overflow of one or more variables.
		
		if (scaling > 4.8e-15) scaling_factors[index] = scaling ;
		else scaling_factors[index] = 0 ;
	}
	
	/* Print out values in scaling_factors[]. */

	if (audit_trail)
	{
		fprintf(output_file, "SCALING FACTORS:\n\n") ;
		
		for (index = 0; index < SCALING_FACTORS_SIZE; index++)
			fprintf(output_file, "%d%s%g%s",
				index, " ", scaling_factors[index], "\n") ;
		fprintf(output_file, "\n") ;
	}
} /* compute_scaling_factors */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	A function that prints out, in sorted order,
**	the compression scores of all the hit sequences in the
**	hit structure.
**
** CALLING SEQUENCE:
**
**	void print_hit_scores()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void print_hit_scores()
{
	int row, leaf_index ;
	sequence *al1 ;
	hit_node *h_node ;
	
	fprintf(output_file, "COMPRESSION SCORES OF ALIGNMENTS:\n\n") ;
	
	row = 0 ;
	while (h_node = get_leaf_nodes_in_order(&row, &al1, &leaf_index))
		fprintf(output_file, "%s%1.2f%s%1.2f%s",
		"CR = ",
		h_node->get_compression_ratio(),
		", CD = ",
		h_node->get_compression_difference(),
		"\n") ;
	fprintf(output_file, "\n") ;
	fflush(output_file) ;
} /* print_hit_scores */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Sets up a new (empty) structure for storing hit nodes.
**
** CALLING SEQUENCE:
**
**	void set_up_hit_structure()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void set_up_hit_structure()
{
	int i ;
	
	hit_root = new hit_node ; // It is assumed that any pre-existing
		// tree of hit nodes was deleted towards the end of recognise().

	for (i = 0; i < hit_structure_rows; i++)
	{
		leaf_array[i] = NIL ;
		alignments_array[i] = NIL ;
		sort_array[i] = i ;
	}
	fe_sort = 0 ;
} /* set_up_hit_structure */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Called from write_hit_structure().
**
** CALLING SEQUENCE:
**
**	void shs(hit_node *h_node, int indentation)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	h_node:			The top node of the structure.
**	indendation:	The amount of indentation.
**
** IMPLICIT INPUTS:
**
**	The hit structure
**
** IMPLICIT OUTPUTS:
**
**	Output
**
** SIDE EFFECTS:
**
**	NONE
*/

void shs(hit_node *h_node, int indentation)
{
	hit_node *child ;
	int i ;
	
	for (i = 0; i < indentation; i++) fprintf(output_file, "  ") ;
	
	if (h_node != hit_root)
	{
		h_node->print_ID() ;
		fprintf(output_file, ": ") ;
		(h_node->get_driving_pattern())->print_ID() ;
		fprintf(output_file, "%s%s%s%d%s",
			", ",
			(h_node->get_driving_symbol())->get_name(),
			", d_pos ",
			h_node->get_driving_int_pos(),
			"; ") ;
		(h_node->get_target_pattern())->print_ID() ;
		fprintf(output_file, "%s%s%s%d%s%1.2f%s%1.2f%s%1.2f%s",
			", ",
			(h_node->get_target_symbol())->get_name(),
			", t_pos ",
			h_node->get_target_int_pos(),
			" (NSC = ",
			h_node->get_new_symbols_cost(),
			" EC = ",
			h_node->get_encoding_cost(),
			", CD = ",
			h_node->get_compression_difference(),
			")") ;
	}
	else fprintf(output_file, "hit_root") ;
	fprintf(output_file, "\n") ;

	list_for(child, hit_node, h_node)
		shs(child, indentation + 1) ;
} /* shs */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Prints out hit structure (for diagnosis).
**
** CALLING SEQUENCE:
**
**	void write_hit_structure(hit_node *h_node, int indentation) ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	h_node:			The top node of the structure.
**	indendation:	The amount of indentation.
**
** IMPLICIT INPUTS:
**
**	The hit structure
**
** IMPLICIT OUTPUTS:
**
**	Output
**
** SIDE EFFECTS:
**
**	NONE
*/

void write_hit_structure(hit_node *h_node, int indentation)
{
	fprintf(output_file, "HIT STRUCTURE (") ;
	print_pattern_window_cycle() ;
	fprintf(output_file, ")\n\n") ;
	shs(h_node, indentation) ;
	fprintf(output_file, "\n") ;
	fflush(output_file) ;
} /* write_hit_structure */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	The frequencies of symbol types are used to compute the
**	information cost of each symbol type using a simplified version
**	of the Shannon-Fano-Elias method.
**	 
**	Calculates average_symbol_type_cost.
**
**	In addition, the function assigns the frequency of its type 
**	to each symbol in the corpus and the associated bit_cost.
**
** CALLING SEQUENCE:
**
**	void find_symbol_type_information_costs(group *symbol_types)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	symbol_types:	A set of symbol types, each one with 
**					an associated frequency.
**
** IMPLICIT INPUTS:
**
**	The corpus
**
** IMPLICIT OUTPUTS:
**
**	bit_cost value on each symbol in the corpus.
**
** SIDE EFFECTS:
**
**	NONE
*/

void find_symbol_type_information_costs(group *symbol_types)
{
	symbol *symbol1 ;
	double total_min_symbol_type_cost = 0,
		total_actual_symbol_type_cost = 0 ;
	int alphabet_size = symbol_types->get_number_of_children() ;

	// Calculate sizes of symbols in symbol_types using a modified
	// version of the S-F-E method.

	double temp_bit_cost ;
	modified_sizes_of_sfe_codes(symbol_types) ;

	// Multiply the sizes of the DATA_SYMBOLs by the cost_factor.

	list_for(symbol1, symbol, symbol_types)
	{
		if (symbol1->get_type() == DATA_SYMBOL)
		{
			temp_bit_cost = symbol1->get_bit_cost() ;
			symbol1->set_bit_cost(temp_bit_cost * cost_factor) ;
		}
	}

	/* Print results. */
	
	if (true) // audit_trail
	{
		fprintf(output_file, "%s%s",
			"FREQUENCIES OF SYMBOL TYPES\n",
			"AND INFORMATION COSTS (IN BITS)\n\n") ;
		write_lines(output_file, "", START) ;
		list_for(symbol1, symbol, symbol_types)
		{
			fprintf(output_file, "%s%s",
				symbol1->get_name(),
				" type = ") ;
			if (symbol1->get_type() == LEFT_BRACKET)
				fprintf(output_file, "LB") ;
			if (symbol1->get_type() == RIGHT_BRACKET)
				fprintf(output_file, "RB") ;
			if (symbol1->get_type() == CODE_SYMBOL)
				fprintf(output_file, "CDS") ;
			if (symbol1->get_type() == DATA_SYMBOL)
				fprintf(output_file, "DS") ;
			fprintf(output_file, "%s%d%s%1.2f%s",
				", fr = ", symbol1->get_frequency(),
				", bit_cost = ", symbol1->get_bit_cost(),
				"\n") ;
			total_min_symbol_type_cost += symbol1->get_bit_cost() ;
		}
		
		fprintf(output_file, "\n") ;
	}
	
	double average_min_symbol_type_cost = 
		total_min_symbol_type_cost / alphabet_size ;
	double average_actual_symbol_type_cost =
		total_actual_symbol_type_cost / alphabet_size ;
	
	if (audit_trail)
	{
		fprintf(output_file, "%s%1.2f%s",
			"Average of bit_costs for symbol types = ",
			average_min_symbol_type_cost, "\n\n") ;
	}

	// Now assign the values for frequency and bit_cost
	// to the symbol instances in the corpus.
	
	assign_symbol_frequencies_and_costs(symbol_types) ; 

} /* find_symbol_type_information_costs */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Gets a new empty entry in leaf_array[]. If the number of entries
**	is exhausted, it returns false, otherwise true.
**
** CALLING SEQUENCE:
**
**	int get_new_leaf_entry() ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	The leaf_index value for the next available empty
**					cell in leaf_array. If none is available, NULL_VALUE
**					is returned.
**
** IMPLICIT INPUTS:
**
**	fe_sort
**
** IMPLICIT OUTPUTS:
**
**	new value for fe_sort
**
** SIDE EFFECTS:
**
**	NONE
*/

int get_new_leaf_entry()
{
	int leaf_index ;
	
	if (fe_sort >= hit_structure_rows) return(NULL_VALUE) ;
	leaf_index = sort_array[fe_sort++] ;
	return(leaf_index) ;
} /* get_new_leaf_entry */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compiles an unordered list of the leaf nodes with the smallest
**	values for CD.
**
** CALLING SEQUENCE:
**
**	void find_worst_leaf_nodes()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	The hit structure.
**
** IMPLICIT OUTPUTS:
**
**	An unordered list of the worst leaf nodes in half_list[].
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void find_worst_leaf_nodes()
{
	int half_fe_sort = fe_sort / 2, row1, row2, best_worst_hl_index ;
	double best_worst_CD, temp_CD, time_started = clock(),
		time_finished, time_difference_secs, time_per_item_secs ;
	hit_node *h_node1 ;
	
	fe_half_list = 0 ;
	
	if (half_fe_sort < 1) abort_run("Too few entries in half_list[]") ;
	if (half_fe_sort > half_hit_structure_rows)
		abort_run("Anomalous value for half_fe_sort in find_mid") ;
	
	// Start by filling up half_list with the first half of the
	// nodes in leaf_array[]. At the same time, find the
	// best leaf node in the half list.
	
#if PRINT_CD_VALUES
	if (audit_trail)
	{
		fprintf(output_file, "CD values for leaf nodes:\n\n") ;
	}
#endif
	
	best_worst_CD = fail_score ;
	for (row1 = 0; row1 < half_fe_sort; row1++)
	{
		h_node1 = leaf_array[sort_array[row1]] ;
		if (h_node1 == NIL)
		abort_run("NIL value for h_node1 \
			in find_worst_leaf_nodes") ;
		half_list[fe_half_list] = h_node1 ;
		temp_CD = h_node1->get_compression_difference() ;
		
#if PRINT_CD_VALUES
		if (audit_trail)
		{
			fprintf(output_file, "%1.2f", temp_CD) ;
			print_counter++ ;
			if (print_counter < 10) fprintf(output_file, ", ") ;
			else
			{
				fprintf(output_file, ",\n") ;
				print_counter = 0 ;
			}
		}
#endif
		
		if (temp_CD > best_worst_CD)
		{
			best_worst_CD = temp_CD ;
			best_worst_hl_index = fe_half_list ;
		}
		fe_half_list++ ;
	}
	
	// Now continue with the second half of sort array[] comparing CD
	// values in the second half with CD values in half_list[]. If
	// a CD value is worse than the best value in half_list, it replaces
	// that value.
	
	for (row1 = row1; row1 < fe_sort; row1++)
	{
		h_node1 = leaf_array[sort_array[row1]] ;
		if (h_node1 == NIL)
		abort_run("NIL value for h_node1 \
			in find_worst_leaf_nodes") ;
		temp_CD = h_node1->get_compression_difference() ;
		
#if PRINT_CD_VALUES
		if (audit_trail)
		{
			fprintf(output_file, "%1.2f", temp_CD) ;
			print_counter++ ;
			if (row1 + 1 >= fe_sort) fprintf(output_file, ".\n\n") ;
			else if (print_counter < 10) fprintf(output_file, ", ") ;
			else
			{
				fprintf(output_file, ",\n") ;
				print_counter = 0 ;
			}
		}
#endif
		
		if (temp_CD >= best_worst_CD) continue ;
		
		// temp_CD is worse than best_worst_CD so the corresponding
		// hit node replaces the hit node for the
		// best_worst_CD in half_list[]. Then it is necesssary to
		// find a new value for best_worst_CD and its corresponding
		// value for best_worst_hl_index.
		
		half_list[best_worst_hl_index] = h_node1 ;
		
		best_worst_CD = fail_score ;
		for (row2 = 0; row2 < half_fe_sort; row2++)
		{
			temp_CD = half_list[row2]->
				get_compression_difference() ;
			if (temp_CD > best_worst_CD)
			{
				best_worst_CD = temp_CD ;
				best_worst_hl_index = row2 ;
			}
		}
	}
	
#if PRINT_CD_VALUES
	
	if (audit_trail)
	{
		fprintf(output_file, "%s%f%s",
			"Value of best_worst_CD = ",
			best_worst_CD,
			"\n\n") ;
		
		fprintf(output_file, "Values in half_list[]:\n\n") ;
		
		print_counter = 0 ;
		for (row1 = 0; row1 < half_fe_sort; row1++)
		{
			h_node1 = half_list[row1] ;
			temp_CD = h_node1->get_compression_difference() ;
			fprintf(output_file, "%1.2f", temp_CD) ;
			print_counter++ ;
			if (row1 + 1 >= half_fe_sort) fprintf(output_file, ".\n\n") ;
			else if (print_counter < 10) fprintf(output_file, ", ") ;
			else
			{
				fprintf(output_file, ",\n") ;
				print_counter = 0 ;
			}
		}
	}
#endif
	
	time_finished = clock() ;
	time_difference_secs = (time_finished - time_started) / CLOCKS_PER_SEC ;
	time_per_item_secs = time_difference_secs / (double)fe_sort ;
	
	if (audit_trail)
	{
		fprintf(output_file,
			"Finding worst leaf nodes started and finished.\n(") ;
		
		print_pattern_window_cycle() ;
		
		fprintf(output_file, "%s%d%s%g%s%g%s",
			")\n\nNumber of items = ",
			fe_sort,
			",\nSifting time = ",
			time_difference_secs,
			" seconds,\nSifting time per item = ",
			time_per_item_secs,
			" seconds\n\n") ;
	}
} // find_worst_leaf_nodes

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Purges the hit structure. There is an assumption that the
**	leaf nodes have been sorted.
**
** CALLING SEQUENCE:
**
**	void purge_hit_structure() ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	The hit structure
**
** IMPLICIT OUTPUTS:
**
**	The purged hit structure
**
** SIDE EFFECTS:
**
**	NONE
*/

void purge_hit_structure()
{
	hit_node *h_node, *parent_h_node, **leaf_entry ;
	int row, fe_sort_at_start = fe_sort ;
	double start_end_fe_sort ;
	
	if (audit_trail)
	{
		fprintf(output_file, "PURGING OF HIT_STRUCTURE (") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, "%s%d%s",
			")\n\nValue of fe_sort at start = ",
			fe_sort,
			"\n\n") ;
	}
	
	find_worst_leaf_nodes() ; // Compiles an unordered list of the
	// worst leaf nodes. Then these nodes can be purged.
	
	/* half_list[] contains the bottom 50% of the leaf nodes. Delete
	each of these hit nodes, pruning each branch
	up to the first node that has more than one child. */
	
	for (row = 0; row < fe_half_list; row++)
	{
		h_node = half_list[row] ;
		if (h_node == NIL) 
		abort_run("NIL value for h_node in \
		purge_hit_structure") ;
		
		leaf_entry = h_node->get_leaf_entry() ; // This is the cell
			// in leaf_array[] which contains a pointer to
			// a pointer to a leaf node.
		(*leaf_entry) = NIL ;
		while (h_node != hit_root)
		{
			parent_h_node = h_node->get_hn_parent() ;
			parent_h_node->extract_child(h_node) ;
			delete h_node ;
			
			/* Test whether the parent_h_node has any other
			children. If so, stop extracting nodes
			up this branch of the tree. */
			
			if (parent_h_node->get_first_child() != NIL) break ;
			h_node = parent_h_node ;
		}
	}
	
	// Now compute new values for fe_sort so that the NIL values in
	// leaf_array[] are not referenced by any cells in fe_sort[] between
	// 0 and fe_sort.
	
	fe_sort = 0 ;
	for (row = 0; row < hit_structure_rows; row++)
		if (leaf_array[row] != NIL) sort_array[fe_sort++] = row ;
		
		int i = fe_sort ;
		for (row = 0; row < hit_structure_rows; row++)
			if (leaf_array[row] == NIL) sort_array[i++] = row ;
			
		start_end_fe_sort = (double)fe_sort / (double)fe_sort_at_start ;

		if (audit_trail)
		{
			fprintf(output_file, "%s%d%s%1.2f%s",
				"Value of fe_sort at end = ",
				fe_sort,
				"\n\nCurrent value of fe_sort is ",
				start_end_fe_sort,
				" of value before purging\n\n") ;
				
			fprintf(output_file, "PURGE IS COMPLETED (") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, ")\n\n") ;
		}
			
} /* purge_hit_structure */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Estimates compression scores for hit sequences added to the
**	hit structure.
**
**	The adjustment for gaps between hits used in earlier versions
**	has been dropped.
**
** CALLING SEQUENCE:
**
**	void estimate_compression_scores(double *NSC, 
**		double *EC, hit_node *parent_h_node)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	NSC:			A pointer to a double variable in the calling function.
**	EC:				A pointer to a double variable in the calling function.
**	parent_h_node:	The parent hit node.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void estimate_compression_scores(double *NSC, 
	double *EC, hit_node *parent_h_node)
{
	// The adjustment for gaps between hits used in earlier versions
	// has been dropped.

	if (cycle == 1)
	{
		/* Driving pattern is one of the New patterns. */

		if (parent_h_node == hit_root)
		{
			*NSC = master_driving_symbol->get_bit_cost() +
				master_target_pattern->get_new_symbols_cost() ;
			*EC = master_target_pattern->get_encoding_cost() ;
		}
		else
		{
			*NSC = parent_h_node->get_new_symbols_cost() +
				master_driving_symbol->get_bit_cost() ;
			*EC = parent_h_node->get_encoding_cost() ;
		}
	}
	else
	{
		// Driving pattern is an alignment from a previous cycle.

		if (parent_h_node == hit_root)
		{
			*NSC = master_driving_pattern->get_new_symbols_cost() +
				master_target_pattern->get_new_symbols_cost() ;
			*EC = master_driving_pattern->get_encoding_cost() +
				master_target_pattern->get_encoding_cost() ;
		}
		else
		{
			*NSC = parent_h_node->get_new_symbols_cost() ;
			*EC = parent_h_node->get_encoding_cost() ;
		}

		// Make an adjustment to EC if the current hit involves at
		// least one IDENTIFICATION symbol.

		if (master_driving_symbol->get_status() == IDENTIFICATION
			|| master_target_symbol->get_status() == IDENTIFICATION)
				*EC -= master_target_symbol->get_bit_cost() ;
	}

} // estimate_compression_scores

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	This function makes a hit_node. It is called from add_hit().
**
** CALLING SEQUENCE:
**
**	void make_hit_node(hit_node *parent_h_node)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	parent_h_node:	The parent of the hit_node to be created.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	A new hit_node attached to a chain of 'pending' hit nodes.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void make_hit_node(hit_node *parent_h_node)
{
	hit_node *copy_h_node = new hit_node(*hn_master) ;
	double NSC, EC ;
	
	estimate_compression_scores(&NSC, &EC, parent_h_node) ;

	copy_h_node->set_new_symbols_cost(NSC) ;
	copy_h_node->set_encoding_cost(EC) ;
	copy_h_node->set_compression_ratio(EC / NSC) ;
	copy_h_node->set_compression_difference(NSC - EC) ;
	
	/* Attach the new copy to the hit_node_array[]. */
	
	hit_node_array[fe_hit_node_array].new_hit_node = copy_h_node ;
	hit_node_array[fe_hit_node_array].hn_parent = parent_h_node ;
	plus_one(&fe_hit_node_array, HIT_NODE_ARRAY_SIZE, 
		"hit_node_array[] is too small.") ;
	copy_h_node->set_hn_parent(parent_h_node) ;
	
	new_hits++ ;
	
} /* make_hit_node */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Adds one or more records of a hit to the hit structure. 
**	The function traverses the tree of hit_nodes recursively. 
**	New nodes representing the hit are created wherever they are needed. 
**	They are not added to the hit structure immediately but are 
**	attached to a chain of hit nodes and then (in 
**	update_hit_structure()) added	to the hit structure if there 
**	is enough space to accommodate them. If there is not enough space,
**	the hit structure is purged and add_hit() is called again.
**
** CALLING SEQUENCE:
**
**	bool add_hit(hit_node *parent_h_node)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if a hit is added, false otherwise.
**
**	parent_h_node:	The top node of the (sub) tree.
**
** IMPLICIT INPUTS:
**
**	Global structures, including hn_master.
**
** IMPLICIT OUTPUTS:
**
**	Global structures
**
** SIDE EFFECTS:
**
**	NONE
*/

bool add_hit(hit_node *parent_h_node)
{
	hit_node *child ;
	bool hit_added = false ;

	list_for(child, hit_node, parent_h_node)
	{
		// compression_difference <= fail_score marks the child as
		// 'dead'. Nothing must be added below this node.
		
		if (child->get_compression_difference() <= fail_score)
			continue ;
		
		if (master_driving_pattern !=
			child->get_driving_pattern()) continue ;
		
		if (master_driving_int_pos <=
			child->get_driving_int_pos())
			continue ;
		
		if (master_target_pattern !=
			child->get_target_pattern()) continue ;
		
		if (master_target_int_pos <=
			child->get_target_int_pos())
			continue ;
		
		hit_added = add_hit(child) ;
	}
	
	if (hit_added) return(true) ;
	
	// There is no child at this level which match all the criteria
	// for going deeper in the tree of hit nodes. So the hit can be
	// installed as a new leaf node child of parent_h_node,
	// provided the gap between driving symbols in the new node and
	// its parent are less than max_driving_gap and likewise for
	// the target symbols.
	
	make_hit_node(parent_h_node) ;
	
	return(true) ;
} /* add_hit */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Updates the data structures which record hits.
**
** CALLING SEQUENCE:
**
**	void update_hit_structure()
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
** IMPLICIT INPUTS:
**
**	The hit structure
**
** IMPLICIT OUTPUTS:
**
**	The hit structure
**
** SIDE EFFECTS:
**
**	NONE
*/

void update_hit_structure()
{
	hit_node *parent_h_node, *node,
		**leaf_entry, *new_h_node ;
	int i, leaf_index ;
	
	new_hits = 0 ;
	fe_hit_node_array = 0 ;

	add_hit(hit_root) ;
	
	/* Check whether there is enough space in leaf_array[]. If not,
	purge the hit structure, clear the chain of hit nodes
	and try again. */
	
	if (fe_sort + new_hits >= hit_structure_rows)
	{
		purge_hit_structure() ;
		
		/* Now delete the new_hit_nodes in the hit_node_array[]. */
		
		for (i = 0; i < fe_hit_node_array; i++)
		{
			delete hit_node_array[i].new_hit_node ;
			hit_node_array[i].hn_parent = NIL ;
		}
		
		new_hits = 0 ;
		fe_hit_node_array = 0 ;
		
		add_hit(hit_root) ;
		
		if (fe_sort + new_hits >= hit_structure_rows)
			abort_run("Too little space in leaf_array[]") ;
	}
	
	/* Now put the copies of hn_master in their proper places in
	the hit_structure. */
	
	for (i = 0; i < fe_hit_node_array; i++)
	{
		new_h_node = hit_node_array[i].new_hit_node ;
		parent_h_node = hit_node_array[i].hn_parent ;
		
		/* Check whether the parent_h_node is a leaf node. If it is,
		use the parent_h_node's entry in leaf_array[]. If not,
		check whether the new_h_node is NIL. If it is, take a
		new entry in leaf_array[]. If the parent_h_node is
		the hit_root then, although at first it is a leaf node,
		its entry in leaf_array[] is not available so a new
		entry must be taken. */
		
		if (parent_h_node == hit_root)
		{
			leaf_index = get_new_leaf_entry() ;
			if (leaf_index == NULL_VALUE)
				abort_run("No leaf_array[] entries left") ;
			leaf_array[leaf_index] = new_h_node ;
			new_h_node->set_leaf_entry(&leaf_array[leaf_index]) ;
		}
		else
		{
			parent_h_node->initialise() ;
			node = (hit_node *)parent_h_node->get_next_child() ;
			if (node == NIL)
			{
				// parent_h_node is itself a leaf node.
				
				leaf_entry = parent_h_node->get_leaf_entry() ;
				(*leaf_entry) = new_h_node ;
				new_h_node->set_leaf_entry(leaf_entry) ;
				parent_h_node->set_leaf_entry(NIL) ;
			}
			else
			{
				// parent_h_node is not a leaf node but
				// the new_hit_node will be a new leaf node.
				
				leaf_index = get_new_leaf_entry() ;
				if (leaf_index == NULL_VALUE)
				abort_run("No leaf_array[] entries \
				left") ;
				leaf_array[leaf_index] = new_h_node ;
				new_h_node->
					set_leaf_entry(&leaf_array[leaf_index]);
			}
		}
		parent_h_node->add_child(new_h_node) ;
	}
	
	/* Now clear the hit_node_array[] to avoid problems when
	hit nodes are deleted. */
	
	for (i = 0; i < HIT_NODE_ARRAY_SIZE; i++)
	{
		hit_node_array[i].new_hit_node = NIL ;
		hit_node_array[i].hn_parent = NIL ;
	}
} /* update_hit_structure */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Tests whether any one symbol from Old is being matched
**	with itself. If it is, this is an illegal hit which is ignored.
**	If two matching columns contain the same symbol from New, this
**	is accepted because there is still a bona fide match between symbols
**	in Old. When the new alignment is formed, the two appearances
**	of one symbol in New can be merged.
**
**	If a New pattern is matched with itself, any one instance of a
**	symbol cannot be matched with itself.
**
** CALLING SEQUENCE:
**
**	bool anomalous_hit(symbol *driving_symbol, int driving_pattern_depth,
**		symbol *target_symbol, int target_pattern_depth)
**
** FORMAL ARGUMENTS:
**
**	Return value:			true if the hit anomalous. false otherwise.
**
**	driving_symbol:			The driving symbol.
**	driving_pattern_depth:	The depth of the driving pattern.
**	target_symbol:			The target symbol.
**	target_pattern_depth:	The depth of the target pattern.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

bool anomalous_hit(symbol *driving_symbol, int driving_pattern_depth,
					   symbol *target_symbol, int target_pattern_depth)
{
	symbol *d_symbol, *t_symbol ;
	int driving_row, target_row ;
	
	if (driving_symbol == target_symbol) return(true) ;
	
	if (driving_pattern_depth == 1 && target_pattern_depth == 1)
		// Since the two symbols have already been compared
		// and given a negative result, we know that
		// the two symbols are not the same.
		return(false) ;
	
	if (driving_pattern_depth == 1)
	{
		// We know that the target_pattern_depth must be > 1.
		
		for (target_row = 0; target_row < target_pattern_depth; 
		target_row++)
		{
			t_symbol = target_symbol->get_row_symbol(target_row) ;
			if (t_symbol == NIL) continue ;
			if (driving_symbol == t_symbol) return(true) ;
		}
		return(false) ;
	}
	
	if (target_pattern_depth == 1)
	{
		// We know that the driving_pattern_depth must be > 1.
		
		for (driving_row = 0; driving_row < driving_pattern_depth; 
		driving_row++)
		{
			d_symbol = driving_symbol->get_row_symbol(driving_row) ;
			if (d_symbol == NIL) continue ;
			if (target_symbol == d_symbol) return(true) ;
		}
		return(false) ;
	}
	
	// driving_pattern_depth and target_pattern_depth must both be > 1.
	
	for (driving_row = 0; driving_row < driving_pattern_depth; 
	driving_row++)
	{
		d_symbol = driving_symbol->get_row_symbol(driving_row) ;
		if (d_symbol == NIL) continue ;
		for (target_row = 0; target_row < target_pattern_depth; 
		target_row++)
		{
			t_symbol = target_symbol->get_row_symbol(target_row) ;
			if (t_symbol == NIL) continue ;
			if (d_symbol == t_symbol) return(true) ;
		}
	}
	return(false) ;
} /* anomalous_hit */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compares two symbols and makes various tests to ensure that any
**	hit is a valid hit.
**
** CALLING SEQUENCE:
**
**	bool match_symbols(sequence *driving_pattern, symbol *driving_symbol, 
**		sequence *target_pattern, symbol *target_symbol)
**
** FORMAL ARGUMENTS:
**
**	Return value:		true if all tests are passed, false otherwise.
**
**	driving_pattern:	The driving pattern.
**	driving_symbol: 	The driving symbol.
**	target_pattern: 	The target pattern.
**	target_symbol:		The target symbol.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Additions to the hit structure
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool match_symbols(sequence *driving_pattern, symbol *driving_symbol, 
	sequence *target_pattern, symbol *target_symbol)
{
	if (!(driving_symbol->name_matches(target_symbol)))
		return(false) ;
	
	int driving_pattern_depth = driving_pattern->get_sequence_depth(),
		target_pattern_depth = target_pattern->get_sequence_depth() ;
	
	if (anomalous_hit(driving_symbol, driving_pattern_depth, target_symbol, 
		target_pattern_depth)) 
		return(false) ;
	
	return(true) ;
	
} // match_symbols

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compares a driving symbol with a target pattern.
**
** CALLING SEQUENCE:
**
**	void compare_symbol_with_pattern(symbol *driving_symbol,
**		list_element *ds_el_pos, sequence *driving_pattern,
**		sequence *target_pattern)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	driving_symbol: 	The driving symbol.
**	ds_el_pos:			The el_pos position of the driving symbol
**						in the driving pattern.
**	driving_pattern:	The driving pattern
**	target_pattern: 	The target pattern
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void compare_symbol_with_pattern(symbol *driving_symbol,
	list_element *ds_el_pos, sequence *driving_pattern,
	sequence *target_pattern)
{
	list_element *ts_el_pos = NIL ;
	symbol *target_symbol ;
	int target_symbol_int_pos, driving_symbol_int_pos = ds_el_pos->get_position() ;
	
	while (target_symbol = (symbol *)target_pattern->
		get_next_child_by_el_pos(&ts_el_pos))
	{
		// When the driving pattern is the same as the
		// target pattern, the test which follows
		// prevents the comparison of symbol A with symbol B
		// and then, later, the comparison of symbol B
		// with symbol A.
		
		target_symbol_int_pos = ts_el_pos->get_position() ;
		
		if (driving_pattern == target_pattern)
			if (target_symbol_int_pos >= driving_symbol_int_pos)
				break ;

		// Constraint on the matching of IDENTIFICATION symbols against
		// CONTENTS symbols and vice versa.

		if (id_c_symbol_constraint)
		{
			if (driving_symbol->get_status() == target_symbol->get_status())
				continue ;
		}

		if (match_symbols(driving_pattern, driving_symbol,
			target_pattern, target_symbol))
		{
			hn_master->set_target_symbol(target_symbol) ;
			master_target_symbol = target_symbol ;
			hn_master->set_target_int_pos(target_symbol_int_pos) ;
			master_target_int_pos = target_symbol_int_pos ;
				
			/* Copies of hn_master which are created during 
			add_hit() are attached to an array of hit
			nodes for later attachement to the hit 
			structure, with purging of the hit structure,
			if necessary. */
				
			update_hit_structure() ;
		}
	}
} // compare_symbol_with_pattern

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compares a driving pattern/alignment with a target pattern or
**	alignment.
**
**	If any pattern contains brackets, then constraints are applied: 1 For each
**	left bracket there must be a corresponding right
**	bracket and the two must be 'balanced', meaning that any brackets 
**	lying between them in the same pattern must also be paired.
**	2 A balanced pair in the driving pattern must match a balanced pair 
**	in the target pattern and vice versa.
**
** CALLING SEQUENCE:
**
**	void compare_patterns(sequence *driving_pattern,
**		sequence *target_pattern)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void.
**
**	driving_pattern:	The pattern which 'drives' the
**						search for matching patterns.
**	target_pattern:		The target pattern.
**
** IMPLICIT INPUTS:
**
**	old_patterns
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void compare_patterns(sequence *driving_pattern,
	sequence *target_pattern)
{
	symbol *driving_symbol ;
	int driving_symbol_int_pos ;
	list_element *dr_el_pos ;

	if (driving_pattern == current_new_pattern)
	{
		if (start_int_pos == 0)
			dr_el_pos = NIL ;
		else dr_el_pos = 
			driving_pattern->get_el_pos_by_int_pos(start_int_pos - 1) ;
	}
	else dr_el_pos = NIL ;

	list_for_el_pos(driving_symbol, symbol, driving_pattern, dr_el_pos)
	{
		driving_symbol_int_pos = dr_el_pos->get_position() ;

		if (driving_pattern == current_new_pattern)
		{
			if (driving_symbol_int_pos >= stop_int_pos)
				break ;
		}

		hn_master->set_driving_symbol(driving_symbol) ;
		master_driving_symbol = driving_symbol ;
		hn_master->set_driving_int_pos(driving_symbol_int_pos) ;
		master_driving_int_pos = driving_symbol_int_pos ;
		
		compare_symbol_with_pattern(driving_symbol, dr_el_pos,
			driving_pattern, target_pattern) ;
	}
} // compare_patterns

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	A diagnostic function that displays the contents of hit_sequences[]
**
** CALLING SEQUENCE:
**
**	void show_hit_sequences()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void show_hit_sequences()
{
	int dr_int_pos, t_int_pos ;
	hit_pair *hp1 ;

	fprintf(output_file, "Contents of hit_sequences[]:\n\n") ;

	list_for (hp1, hit_pair, hit_sequences)
	{
		while (hp1 != NIL)
		{
			dr_int_pos = hp1->get_dr_int_pos() ;
			t_int_pos = hp1->get_t_int_pos() ;
			fprintf(output_file, "%s%s%s%s%d%s%d%s",
				(dp_array[dr_int_pos])->get_name(),
				",",
				(tp_array[t_int_pos])->get_name(),
				" (",
				dr_int_pos,
				",",
				t_int_pos,
				") ") ;

				hp1 = hp1->get_next() ;
		}
		fprintf(output_file, "\n\n") ;
	}

	fflush(output_file) ;

} // show_hit_sequences

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For a given left bracket hit in hit_sequences[], this function adds
**	the corresponding right bracket hit.
**
** CALLING SEQUENCE:
**
**	void add_right_bracket(sequence *driving_pattern,
**		sequence *target_pattern, hit_pair *hp1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	hp1:			The hit pair that records the left bracket hit.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void add_right_bracket(sequence *driving_pattern,
	sequence *target_pattern, hit_pair *hp1)
{
	hit_pair *hp2 ;
	symbol *driving_symbol, *target_symbol, *temp_dr_symbol,
		*temp_t_symbol ;
	int dr_lb_int_pos = hp1->get_dr_int_pos(),
		t_lb_int_pos = hp1->get_t_int_pos(),
		dr_rb_int_pos, t_rb_int_pos ;

	driving_symbol = hp1->get_driving_symbol() ;
	dr_rb_int_pos = driving_symbol->get_partner_int_pos() ;

	target_symbol = hp1->get_target_symbol() ;
	t_rb_int_pos = target_symbol->get_partner_int_pos() ;

	temp_dr_symbol = (symbol *)
		driving_pattern->get_child_by_int_pos(dr_rb_int_pos) ;		
	temp_t_symbol = (symbol *)
		target_pattern->get_child_by_int_pos(t_rb_int_pos) ;
	hp2 = new hit_pair(temp_dr_symbol, dr_rb_int_pos, target_symbol,
		t_rb_int_pos) ;
	hp1->set_next(hp2) ;

} // add_right_bracket

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For any left brackets in a leading sequence of hits in a new
**	hit sequence, fill in the corresponding right brackets.
**
** CALLING SEQUENCE:
**
**	void fill_in_right_brackets(sequence *driving_pattern,
**		sequence *target_pattern, hit_pair *first_of_new_sequence,
**		hit_pair *hp_marker)
**
** FORMAL ARGUMENTS:
**
**	Return value:			void
**
**	driving_pattern:		The driving pattern.
**	target_pattern:			The target pattern.
**	first_of_new_sequence:	The first hit_pair ofthe current
**							hit sequence.
**	hp_marker:				The hit pair that marks the end of the
**							newly-created hit sequence, before any right
**							brackets are added.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	New hit pairs in the current hit sequence.
**
** SIDE EFFECTS:
**
**	NONE
*/

void fill_in_right_brackets(sequence *driving_pattern,
	sequence *target_pattern, hit_pair *first_of_new_sequence,
	hit_pair *hp_marker)
{
	hit_pair *hp1, *hp2, *hp3, *hp4, *hp5 ;
	bool hp_marker_not_passed = true ;
	symbol *driving_symbol, *target_symbol ;
	int dr_rb_int_pos, t_rb_int_pos ;
	bool pair_inserted ;

	hp1 = first_of_new_sequence ;

	while (hp_marker_not_passed)
	{
		if (hp1->is_left_bracket() == false) goto L1 ;

		// hp1 represents a match between two left brackets.

		driving_symbol = hp1->get_driving_symbol() ;
		dr_rb_int_pos = driving_symbol->get_partner_int_pos() ;
		driving_symbol = (symbol *)
			driving_pattern->get_child_by_int_pos(dr_rb_int_pos) ;
		target_symbol = hp1->get_target_symbol() ;
		t_rb_int_pos = target_symbol->get_partner_int_pos() ;
		target_symbol = (symbol *)
			target_pattern->get_child_by_int_pos(t_rb_int_pos) ;

		// Check that this pair of right brackets is not already
		// in the leading sequence of hit pairs.

		hp5 = first_of_new_sequence ;
		while (hp5 != hp_marker) // There is no need to check the
								// hp_marker itself because this would
								// never be a pair of right brackets.
		{
			if (hp5->get_dr_int_pos() == dr_rb_int_pos
				&& hp5->get_t_int_pos() == t_rb_int_pos)
					goto L1 ; // The pair of right brackets already
								// exists in the leading sequence.
			hp5 = hp5->get_next() ;
		}

		// Now place the hit pair in the right position in the hit
		// sequence, if possible.

		hp3 = hp_marker ;
		hp4 = hp3->get_next() ;
		pair_inserted = false ;
		while (hp4 != NIL)
		{
			if (dr_rb_int_pos > hp3->get_dr_int_pos()
				&& t_rb_int_pos > hp3->get_t_int_pos()
					&& dr_rb_int_pos < hp4->get_dr_int_pos()
						&& t_rb_int_pos < hp4->get_t_int_pos())
			{
				// Insert the hit pair.

				hp2 = new hit_pair(driving_symbol, dr_rb_int_pos,
					target_symbol, t_rb_int_pos) ;
				hp3->set_next(hp2) ;
				hp2->set_next(hp4) ;
				pair_inserted = true ;
				break ;
			}

			hp3 = hp4 ;
			hp4 = hp3->get_next() ;
		}

		if (pair_inserted == false)
		{
			if (dr_rb_int_pos > hp3->get_dr_int_pos()
				&& t_rb_int_pos > hp3->get_t_int_pos())
			{
				hp2 = new hit_pair(driving_symbol, dr_rb_int_pos,
					target_symbol, t_rb_int_pos) ;
				hp3->set_next(hp2) ;
			}
			else
			{
				// The sequence is not legal and must be deleted.

				hp1 = first_of_new_sequence->get_next() ;
				while (hp1 != NIL)
				{
					hp2 = hp1->get_next() ;
					delete hp1 ;
					hp1 = hp2 ;
				}

				new_hit_sequences->delete_child(first_of_new_sequence) ;

				return ;
			}
		}

		L1: if (hp1 == hp_marker)
			hp_marker_not_passed = false ;
		hp1 = hp1->get_next() ;
	}

} // fill_in_right_brackets

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Removes duplicate hit sequences by marking them.
**
** CALLING SEQUENCE:
**
**	void remove_duplicate_hit_sequences()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Removal of hit sequences from hit_sequences[]
**
** SIDE EFFECTS:
**
**	NONE
*/

void remove_duplicate_hit_sequences()
{
	hit_pair *hp1, *hp2, *first_hp2, *first_hp1, *temp_hp ;

	list_for(first_hp2, hit_pair, hit_sequences)
	{
		// Compare the ith sequence with each of the jth sequences
		// where j ranges from 0 up to i - 1.

		hp2 = first_hp2 ;

		list_for(first_hp1, hit_pair, hit_sequences)
		{
			if (first_hp1 == first_hp2) break ;

			// Compare the two sequences.

			hp1 = first_hp1 ;

			while (hp1 != NIL)
			{
				if (hp2 == NIL) goto L1 ; // The two sequences are
					// not the same so try another pairing.

				if (hp1->get_dr_int_pos() != hp2->get_dr_int_pos()
					|| hp1->get_t_int_pos() != hp2->get_t_int_pos())
						goto L1 ; // The two sequences are not
							// the same so try another pairing.
				

				hp1 = hp1->get_next() ;
				hp2 = hp2->get_next() ;
			}

			if (hp2 == NIL) // This means that the two sequences are
				// the same length and have the same hit pairs. Since
				// they are identical, the later of them is deleted.
				// There is no need to continue comparing
				// it with earlier sequences.
			{
				hp2 = first_hp2->get_next() ;
				while (hp2 != NIL)
				{
					temp_hp = hp2->get_next() ;
					delete hp2 ;
					hp2 = temp_hp ;
				}

				hit_sequences->delete_child(first_hp2) ;

				break ;
			}

			L1: hp2 = first_hp2 ;
		}
	}
} // remove_duplicate_hit_sequences

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Records a hit between two symbols found in
**	compare_patterns_with_brackets() in one or more hit sequences
**	attached to the hit_sequences[] array.
**
** CALLING SEQUENCE:
**
**	void add_hit_pair(sequence *driving_pattern, symbol *driving_symbol,
**		int dr_int_pos, sequence *target_pattern,
**		symbol *target_symbol, int t_int_pos,
**		bool is_left_bracket)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	driving_pattern:	The driving pattern.
**	driving_symbol:		The driving symbol.
**	dr_int_pos:			The position of the driving hit symbol.
**	target_pattern:		The target pattern.
**	target_symbol:		The target symbol.
**	t_int_pos:			The position of the target hit symbol.
**	is_left_bracket:	true if the match is between left brackets,
**						false otherwise.
*
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	One or more entries in hit sequences attached to hit_sequences[].
**
** SIDE EFFECTS:
**
**	NONE
*/

void add_hit_pair(sequence *driving_pattern, symbol *driving_symbol,
	int dr_int_pos, sequence *target_pattern,
	symbol *target_symbol, int t_int_pos,
	bool is_left_bracket)
{
	hit_pair *hp1, *hp1a, *hp2, *hp3, *hp4, *hp5, *hp5_previous,
		*first_of_new_sequence ;
	bool new_record_made = false ;

	// Step through the hit sequences recorded in hit_sequences[].

	list_for (hp1, hit_pair, hit_sequences)
	{
		// For each hit sequence, look to see if there is a place
		// in the sequence where the new hit can be added. If there
		// is, add a new hit_pair at the appropriate point to record
		// the hit. If there is not, create a new hit sequence with or
		// without leading hits from the current hit sequence.
		// Then pass on to later hit sequences.

		hp1a = hp1 ;

		while (hp1a != NIL)
		{
			hp2 = hp1a->get_next() ;

			if (dr_int_pos > hp1a->get_dr_int_pos()
				&& t_int_pos > hp1a->get_t_int_pos())
			{
				if (hp2 == NIL) // End of hit sequence has been
					// reached.
				{
					// Insert hit at
					// the end of the hit sequence.

					// Make a new record for the hit.

					hp3 = new hit_pair(driving_symbol, dr_int_pos,
						target_symbol, t_int_pos) ;
					hp1a->set_next(hp3) ;

					if (is_left_bracket)
						add_right_bracket(driving_pattern,
							target_pattern, hp3) ;

					new_record_made = true ;
					goto L1 ;
				}
				else if (dr_int_pos > hp2->get_dr_int_pos()
					&& t_int_pos > hp2->get_t_int_pos())
				{
					hp1a = hp2 ; // Carry on down the chain of hits.
					continue ;
				}
				else if (dr_int_pos < hp2->get_dr_int_pos()
					&& t_int_pos < hp2->get_t_int_pos())
				{
					// Insert new hit pair within the current
					// sequence.

					// Make a new record for the hit.

					hp3 = new hit_pair(driving_symbol, dr_int_pos,
						target_symbol, t_int_pos) ;
					hp3->set_next(hp2) ;
					hp1a->set_next(hp3) ;
				}
				else
				{
					// Start a new sequence, copy the leading
					// entries of the current sequence into it
					// and add a new entry for the current hit
					// at the end.

					hp4 = hp1 ;
					hp5 = new hit_pair(*hp4) ;
					new_hit_sequences->add_child(hp5) ;
					first_of_new_sequence = hp5 ;
					hp4 = hp4->get_next() ;
					while (hp4 != hp2)
					{
						hp5_previous = hp5 ;
						hp5 = new hit_pair(*hp4) ;
						hp5_previous->set_next(hp5) ;
						hp4 = hp4->get_next() ;	
					}

					// Make a new record for the hit.

					hp3 = new hit_pair(driving_symbol, dr_int_pos,
						target_symbol, t_int_pos) ;
					hp5->set_next(hp3) ;

					// For any left brackets in the leading sequence,
					// fill in the corresponding right brackets.

					fill_in_right_brackets(driving_pattern,
						target_pattern, first_of_new_sequence, hp3) ;

					new_record_made = true ;
					goto L1 ;
				}
			}
			else hp1a = hp2 ; // Go round the loop again.
		}
		L1: ;
	}

	if (new_record_made)
	{
		hit_sequences->merge_children(new_hit_sequences) ;

		goto L2 ; // At least one new record of the
		// hit has been inserted in a hit sequence.
	}

	// Since no new record has been inserted in a hit sequence, a new
	// hit sequence is created with the current hit as its first
	// entry.

	hp3 = new hit_pair(driving_symbol, dr_int_pos,
		target_symbol, t_int_pos) ;
	hit_sequences->add_child(hp3) ;

	if (is_left_bracket)
		add_right_bracket(driving_pattern, target_pattern, hp3) ;

	// Check for duplicate hit sequences and delete duplicates.

	L2: ;

	// Remove duplicates from hit_sequences[].

	remove_duplicate_hit_sequences() ;

	// Check that new_hit_sequences is empty.

	if (new_hit_sequences->is_empty() == false)
		abort_run("new_hit_sequences is not empty.") ;

} // add_hit_pair

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	This function, called from add_sequence_to_hit_structure(),
**	adds one hit to the hit structure.
**
** CALLING SEQUENCE:
**
**	hit_node *add_one_hit(hit_node *parent_h_node)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The new parent for the next hit.
**
**	parent_h_node:	The parent hit node for the current hit.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

hit_node *add_one_hit(hit_node *parent_h_node)
{
	hit_node *child ;

	list_for(child, hit_node, parent_h_node)
	{
		if (child->get_driving_pattern() != master_driving_pattern)
			continue ;
		
		if (child->get_target_pattern() != master_target_pattern)
			continue ;

		if (child->get_driving_int_pos() != master_driving_int_pos)
			continue ;

		if (child->get_target_int_pos() != master_target_int_pos)
			continue ;

		// A copy of the hit to be added is already present in the
		// hit structure. In this case, all we need to do is return
		// a pointer to child so that it can serve as the parent
		// of the next hit in the sequence.

		return(child) ;
	}

	// No copy of the current hit has been found at this level in
	// the hit structure tree. So we can create a new entry and
	// insert it in the tree.

	hit_node *copy_h_node = new hit_node(*hn_master) ;

	double NSC, EC ;

	estimate_compression_scores(&NSC, &EC, parent_h_node) ;
	
	copy_h_node->set_new_symbols_cost(NSC) ;
	copy_h_node->set_encoding_cost(EC) ;
	copy_h_node->set_compression_ratio(EC / NSC) ;
	copy_h_node->set_compression_difference(NSC - EC) ;
	
	parent_h_node->add_child(copy_h_node) ;
	copy_h_node->set_hn_parent(parent_h_node) ;

	return(copy_h_node) ;

} // add_one_hit

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	This function, called from compare_patterns_with_brackets(), adds
**	a sequence of hits to the hit structure.
**
** CALLING SEQUENCE:
**
**	void add_sequence_to_hit_structure(hit_pair *hp1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	hp1:			The first hit_pair of the sequence
**					that is to be deleted.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void add_sequence_to_hit_structure(hit_pair *hp1)
{
	hit_node *parent_h_node, *new_h_node = hit_root ;
	int dr_int_pos, t_int_pos ;
	symbol *driving_symbol, *target_symbol ;

	while (hp1 != NIL)
	{
		parent_h_node = new_h_node ; // This starts with hit_root
			// as parent_h_node.
		dr_int_pos = hp1->get_dr_int_pos() ;
		driving_symbol = dp_array[dr_int_pos] ;
		t_int_pos = hp1->get_t_int_pos() ;
		target_symbol = tp_array[t_int_pos] ;

		hn_master->set_driving_symbol(driving_symbol) ;
		master_driving_symbol = driving_symbol ;

		hn_master->set_driving_int_pos(dr_int_pos) ;
		master_driving_int_pos = dr_int_pos ;

		hn_master->set_target_symbol(target_symbol) ;
		master_target_symbol = target_symbol ;

		hn_master->set_target_int_pos(t_int_pos) ;
		master_target_int_pos = t_int_pos ;

		new_h_node = add_one_hit(parent_h_node) ;

		hp1 = hp1->get_next() ;
	}

	// The last new_hit_node is a new leaf node. Make an entry
	// for it in leaf_array[] and a reverse pointer to that
	// entry in new_h_node.
	
	int leaf_index = get_new_leaf_entry() ;
	if (leaf_index == NULL_VALUE)
		abort_run("No leaf_array[] entries left in \
			add_sequence_to_hit_structure()") ;
	leaf_array[leaf_index] = new_h_node ;
	new_h_node->set_leaf_entry(&leaf_array[leaf_index]);

} // add_sequence_to_hit_structure

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks a hit sequence to see whether any hit between two LEFT_BRACKETs
**	is immediately followed by a hit between two ID-symbols. Here,
**	'immediately' means that the driving left bracket and the driving
**	ID-symbol are next to each other and likewise for the target symbols.
**
** CALLING SEQUENCE:
**
**	bool brackets_id_symbol_check_OK(hit_pair *hp1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if the constraint is valid, false otherwise.
**
**	hp1:			The first hit pair in the hit sequence.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

bool brackets_id_symbol_check_OK(hit_pair *hp1)
{
	hit_pair *hp2 ;
	int pos_difference ;

	while (hp1 != NIL)
	{
		hp2 = hp1->get_next() ;

		if ((dp_array[hp1->get_dr_int_pos()])->get_type() != LEFT_BRACKET)
		{
			hp1 = hp2 ;
			continue ;
		}

		// hp1 represents a hit between two left_brackets.

		if (hp2 == NIL) return(false) ;

		pos_difference = hp2->get_dr_int_pos() - hp1->get_dr_int_pos() ;
		if (pos_difference != 1) return(false) ;

		pos_difference = hp2->get_t_int_pos() - hp1->get_t_int_pos() ;
		if (pos_difference != 1) return(false) ;

		hp1 = hp2 ;
	}

	return(true) ;

} // brackets_id_symbol_check_OK

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Matches one pattern against another when both patterns contain brackets.
**	This means that if two left brackets have been matched then their
**	their corresponding right brackets must be matched.
**
** CALLING SEQUENCE:
**
**	void compare_patterns_with_brackets(sequence *driving_pattern,
**		sequence *target_pattern)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	driving_pattern:	The driving pattern.
**	target_pattern:		The target pattern.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void compare_patterns_with_brackets(sequence *driving_pattern, 
	sequence *target_pattern)
{
	symbol *driving_symbol, *target_symbol ;
	int dr_int_pos, t_int_pos, fe_dp, fe_tp ;
	bool is_left_bracket ;

	// Put driving_pattern and target_pattern into arrays to
	// facilitate the finding of positions.

	dr_int_pos = 0 ;
	list_for(driving_symbol, symbol, driving_pattern)
	{
		dp_array[dr_int_pos] = driving_symbol ;
		plus_one(&dr_int_pos, VERY_LARGE_SCRATCH_ARRAY_SIZE,
			"Overflow of dp_array[] in \
				compare_patterns_with_xml_brackets()") ;
	}

	fe_dp = dr_int_pos ;

	t_int_pos = 0 ;
	list_for(target_symbol, symbol, target_pattern)
	{
		tp_array[t_int_pos] = target_symbol ;
		plus_one(&t_int_pos, VERY_LARGE_SCRATCH_ARRAY_SIZE,
			"Overflow of tp_array[] in \
				compare_patterns_with_xml_brackets()") ;
	}

	fe_tp = t_int_pos ;

	// Now look for hit sequences, with the constraint that, for every
	// matching pair of left brackets, the corresponding pair of right
	// brackets must be recorded as a hit and no other pair of right
	// brackets. This is achieved by using the 'partner' field in
	// the class 'symbol' and by not recording hits between right brackets
	// directly.

	int symbol_type ;

	for(dr_int_pos = 0; dr_int_pos < fe_dp; dr_int_pos++)
	{
		driving_symbol = dp_array[dr_int_pos] ;
		if (driving_symbol->get_type() == RIGHT_BRACKET) continue ;
		for(t_int_pos = 0; t_int_pos < fe_tp; t_int_pos++)
		{
			target_symbol = tp_array[t_int_pos] ;
			symbol_type = target_symbol->get_type() ;
			if (symbol_type == RIGHT_BRACKET) continue ;

			if (match_symbols(driving_pattern, driving_symbol,
				target_pattern, target_symbol) == false) continue ;

			if (symbol_type == LEFT_BRACKET)
			{
				// Regardless of the value of ID_C_SYMBOL_CONSTRAINT,
				// the rule is enforced that two left bracket symbols
				// can only form a successful match if one has the
				// status BOUNDARY_MARKER and the other has the
				// status CONTENTS, or vice versa. If this rule is
				// applied to left brackets, it is automatically 
				// applied to the corresponding right brackets.

				if (target_symbol->get_status() == BOUNDARY_MARKER)
				{
					if (driving_symbol->get_status() != CONTENTS)
						continue ;
				}
				else if (target_symbol->get_status() == CONTENTS)
				{
					if (driving_symbol->get_status() != BOUNDARY_MARKER)
						continue ;
				}

				is_left_bracket = true ;
			}
			else is_left_bracket = false ;

			add_hit_pair(driving_pattern, driving_symbol, dr_int_pos,
				target_pattern, target_symbol, t_int_pos, is_left_bracket) ;
		}
	}

	// Diagnostic

	// show_hit_sequences() ;

	// Copy the hit sequences in hit_sequences[] into the main
	// hit structure and clear the hit_sequences[] array
	// at the same time.

	hn_master->set_driving_pattern(driving_pattern) ;
	master_driving_pattern = driving_pattern ;
	hn_master->set_target_pattern(target_pattern) ;
	master_target_pattern = target_pattern ;

	hit_pair *hp1, *hp2 ;
	list_for(hp1, hit_pair, hit_sequences)
	{
		// Check that there is enough space in the hit structure and,
		// if not, purge it. The figure of '3' in the next line is
		// represents an outside value for the number of new leaf nodes
		// that will be created by one sequence. The true value in this
		// case is only 1.

		if (fe_sort + 3 >= hit_structure_rows)
			purge_hit_structure() ;

		if (brackets_id_symbol_constraint)
		{
			if (brackets_id_symbol_check_OK(hp1))
				add_sequence_to_hit_structure(hp1) ;
		}
		else add_sequence_to_hit_structure(hp1) ;

		// Delete hit sequence (excluding first hit_pair).

		hp1 = hp1->get_next() ;
		while (hp1 != NIL)
		{
			hp2 = hp1->get_next() ;
			delete hp1 ;
			hp1 = hp2 ;
		}
	}

	hit_sequences->delete_children() ;

} // compare_patterns_with_brackets

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Matches one pattern against another when both patterns contain XML-style brackets:
**	For the time being, it is assumed that any one type of XML-style bracket can only
**	occur once in any one pattern. In other words, recursive structures within one
**	pattern are ruled out.
**
** CALLING SEQUENCE:
**
**	void compare_patterns_with_xml_brackets(sequence *driving_pattern,
**		sequence *target_pattern)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	driving_pattern:	The driving pattern.
**	target_pattern:		The target pattern.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void compare_patterns_with_xml_brackets(sequence *driving_pattern,
	sequence *target_pattern)
{
	// First phase: find matches between brackets.

	symbol *driving_symbol, *target_symbol ;
	int symbol_type, i, j, j_start, fe_dp, fe_tp ;
	sequence *hit_sequence = new sequence ;
	hit_node *h_node1, *h_node2, *h_node3, *h_node4 ;

	// Put driving_pattern and target_pattern into arrays to
	// facilitate the finding of positions.

	i = 0 ;
	list_for(driving_symbol, symbol, driving_pattern)
	{
		dp_array[i] = driving_symbol ;
		plus_one(&i, VERY_LARGE_SCRATCH_ARRAY_SIZE,
			"Overflow of dp_array[] in \
				compare_patterns_with_xml_brackets()") ;
	}

	fe_dp = i ;

	j = 0 ;
	list_for(target_symbol, symbol, target_pattern)
	{
		tp_array[j] = target_symbol ;
		plus_one(&j, VERY_LARGE_SCRATCH_ARRAY_SIZE,
			"Overflow of tp_array[] in \
				compare_patterns_with_xml_brackets()") ;
	}

	fe_tp = j ;

	j_start = 0 ; // Used to record start position in target_pattern.
	for(i = 0; i < fe_dp; i++)
	{
		driving_symbol = dp_array[i] ;
		symbol_type = driving_symbol->get_type() ;
		if (symbol_type != LEFT_XML_BRACKET
			&& symbol_type != RIGHT_XML_BRACKET)
				continue ;
		for(j = j_start; j < fe_tp; j++)
		{
			target_symbol = tp_array[j] ;
			symbol_type = target_symbol->get_type() ;
			if (symbol_type != LEFT_XML_BRACKET
				&& symbol_type != RIGHT_XML_BRACKET)
					continue ;
			if (match_symbols(driving_pattern, driving_symbol,
				target_pattern, target_symbol))
			{
				// There is a bone fide hit.

				h_node1 = new hit_node ;
				h_node1->set_driving_pattern(driving_pattern) ;
				h_node1->set_driving_symbol(driving_symbol) ;
				h_node1->set_driving_int_pos(i) ;
				h_node1->set_target_pattern(target_pattern) ;
				h_node1->set_target_symbol(target_symbol) ;
				h_node1->set_target_int_pos(j) ;
				hit_sequence->add_child(h_node1) ;

				j_start = j + 1 ;
				break ;
			}
		}
	}

	// Diagnostic: print hit_sequence

	// list_for(h_node1, hit_node, hit_sequence)
	//	h_node1->write_hit_node() ;

	// Second phase: find matches (if any) between non-bracket
	// symbols that appear between symbol alignments found in
	// the first phase.

	list_element *el_pos1 = NIL, *el_pos2 = NIL ;
	int int_pos_d1, int_pos_d2, int_pos_t1, int_pos_t2 ;
	list_for_el_pos(h_node1, hit_node, hit_sequence, el_pos2)
	{
		if (el_pos1 == NIL)
		{
			el_pos1 = el_pos2 ;
			continue ;
		}
		h_node1 = (hit_node *)el_pos1->get_el_obj() ;
		h_node2 = (hit_node *)el_pos2->get_el_obj() ;
		int_pos_d1 = h_node1->get_driving_int_pos() ; 
		int_pos_d2 = h_node2->get_driving_int_pos() ; 
		int_pos_t1 = h_node1->get_target_int_pos() ; 
		int_pos_t2 = h_node2->get_target_int_pos() ;

		// Now look for hits between non-hit symbols between
		// the hits recorded in h_node1 and h_node2.

		j_start = int_pos_t1 + 1 ;
		h_node3 = h_node1 ;
		for (i = int_pos_d1 + 1; i < int_pos_d2; i++)
		{
			driving_symbol = dp_array[i] ;
			for (j = j_start; j < int_pos_t2; j++)
			{
				target_symbol = tp_array[j] ;
				if (match_symbols(driving_pattern, driving_symbol,
					target_pattern, target_symbol))
				{
					// There is a bone fide hit.

					h_node4 = new hit_node ;
					h_node4->set_driving_pattern(driving_pattern) ;
					h_node4->set_driving_symbol(driving_symbol) ;
					h_node4->set_driving_int_pos(i) ;
					h_node4->set_target_pattern(target_pattern) ;
					h_node4->set_target_symbol(target_symbol) ;
					h_node4->set_target_int_pos(j) ;
					hit_sequence->follow(h_node3, h_node4) ;
					h_node3 = h_node4 ; // Next hit will follow
						// the present h_node4.

					j_start = j + 1 ;
					break ;
				}
			}
		}
		el_pos1 = el_pos2 ;
	}

	// Now put the hit sequence into the main hit structure, working
	// from left to right. Delete the hit nodes after they have
	// been used.

	while (h_node1 = (hit_node *)hit_sequence->extract_first_child())
	{
		*hn_master = *h_node1 ;
		master_driving_pattern = h_node1->get_driving_pattern() ;
		master_driving_symbol = h_node1->get_driving_symbol() ;
		master_driving_int_pos = h_node1->get_driving_int_pos() ;
		master_target_pattern = h_node1->get_target_pattern() ;
		master_target_symbol = h_node1->get_target_symbol() ;
		master_target_int_pos = h_node1->get_target_int_pos() ;
		update_hit_structure() ;
		delete h_node1 ;
	}

	delete hit_sequence ;

} // compare_patterns_with_xml_brackets

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Matches a driving pattern with a target pattern.
**
** CALLING SEQUENCE:
**
**	void match_target_pattern(sequence *driving_pattern,
**		sequence *target_pattern)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	driving_pattern:	The driving pattern.
**	target_pattern:		The target pattern.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void match_target_pattern(sequence *driving_pattern,
	sequence *target_pattern)
{
	if (recognise_brackets)
	{
		if (driving_pattern->has_brackets()
			&& target_pattern->has_brackets()) // It is only necessary to
				// check the first symbol in the pattern because
				// a pattern with internal brackets but no external
				// brackets is not 'legal' (as determined by check_brackets().
				compare_patterns_with_brackets(driving_pattern, target_pattern) ;
		else compare_patterns(driving_pattern, target_pattern) ;
	}
	else if (recognise_xml_brackets)
	{
		if (driving_pattern->has_xml_brackets()
			&& target_pattern->has_xml_brackets()) // It is only necessary to
				// check the first symbol in the pattern because
				// a pattern with internal brackets but no external
				// brackets is not 'legal'.
				compare_patterns_with_xml_brackets(driving_pattern,
					target_pattern) ;
		else compare_patterns(driving_pattern, target_pattern) ;
	}
	else compare_patterns(driving_pattern, target_pattern) ;
} // match_target_pattern

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Matches one driving pattern with target patterns
**
** CALLING SEQUENCE:
**
**	void match_driving_pattern(sequence *driving_pattern)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	driving_pattern:	The driving pattern
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void match_driving_pattern(sequence *driving_pattern)
{
	sequence *target_pattern ;
	list_element *el_pos1 ;

	if (audit_trail)
	{
		fprintf(output_file, "NEXT DRIVING PATTERN (") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, "%s%d%s",
			") ID",
			driving_pattern->get_ID(),
			"\n\n") ;
	}
			
	if (driving_pattern->get_number_of_children() >= MCOLS)
		abort_run("MCOLS is too small.") ;

	hn_master->set_driving_pattern(driving_pattern) ;
	master_driving_pattern = driving_pattern ;

	// Take target patterns from old_patterns.

	el_pos1 = NIL ;	
	list_for_el_pos(target_pattern, sequence, old_patterns, el_pos1)
	{
		if (target_pattern->get_number_of_children() >= MCOLS)
			abort_run("MCOLS is too small.") ;

		hn_master->set_target_pattern(target_pattern) ;
		master_target_pattern = target_pattern ;

		match_target_pattern(driving_pattern, target_pattern) ;		
	}

	// Take target patterns from cumulative_alignments_from_parsing.

	el_pos1 = NIL ;
	list_for_el_pos(target_pattern, sequence,
		cumulative_alignments_from_parsing, el_pos1)
	{
		// Don't use the target pattern if it has already been
		// used as a driving pattern.

		if (used_driving_patterns->contains(target_pattern)) continue ;

		if (target_pattern->get_number_of_children() >= MCOLS)
			abort_run("MCOLS is too small.") ;

		hn_master->set_target_pattern(target_pattern) ;
		master_target_pattern = target_pattern ;			

		match_target_pattern(driving_pattern, target_pattern) ;		
	}

} // match_driving_pattern

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Finds hit sequences between driving patterns and patterns in
**	Old and in alignments already formed.
**
** CALLING SEQUENCE:
**
**	bool find_best_matches()
**
** FORMAL ARGUMENTS:
**
**	Return value:			true if one or more hit sequences are 
**							formed. false otherwise.
**
** IMPLICIT INPUTS:
**
**	old_patterns
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

bool find_best_matches()
{
	sequence *driving_pattern ;
	
	if (audit_trail)
	{
		fprintf(output_file, "DRIVING PATTERNS (") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, ")\n\n") ;
	}

	if (cycle == 1)
	{
		driving_pattern = current_new_pattern ;
		match_driving_pattern(driving_pattern) ;
	}
	else
	{
		list_for(driving_pattern, sequence, driving_alignments)
		{
			match_driving_pattern(driving_pattern) ;
			used_driving_patterns->add_child(driving_pattern) ;
		}
		used_driving_patterns->release_children() ;
	}
	
	if (hit_root->get_first_element() == NIL)
	{
		if (audit_trail)
		{
			fprintf(output_file, "FOR ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file,
				", NO HIT SEQUENCES HAVE BEEN FOUND.\n\n") ;
		}
		return(false) ;
	}
	
	if (show_hit_structure)
		write_hit_structure(hit_root, 0) ;
	
	sort_leaf_array(0, fe_sort - 1) ;
	
	/* Quit if no new hit sequences have been found. */
	
	if (fe_sort < 1)
	{
		if (audit_trail)
		{
			fprintf(output_file, "FOR ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file,
				", NO HIT SEQUENCES HAVE BEEN FOUND.\n\n") ;
		}
		return(false) ;
	}
	
	if (show_unselected_alignments)
	{
		if (audit_trail)
		{
			fprintf(output_file,
				"SET OF ALIGNMENTS FORMED, BEFORE SELECTION, IN ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, "\n\n") ;
			fflush(output_file) ;
		}
	}
	
	return(true) ;
	
} /* find_best_matches */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes the best alignment to latex_file in a form ready
**	for printing using latex.
**
** CALLING sequence:
**
**	void output_alignment(sequence *sequence1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	sequence1:	The sequence to be output.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void output_alignment(sequence *sequence1)
{
	int al_length_in_characters = 0, al_lines, i ;
	double target_font_height_in_points,
		actual_font_height_in_points, actual_font_width_in_points,
		actual_font_width_in_mm,
		baselineskip_in_points, base_line_skip_in_mm,
		printing_height, printing_width,
		al_actual_depth_in_mm,
		font_heights[FONT_SET_SIZE] =
	{7, 8, 9, 10, 11, 12, 14, 17.28, 20.74, 24.88} ;

	// double actual_font_height_in_mm ;
	
	/* Note re fontsizes: there are 72.27 'points' to an inch which
	is 2.8452756 points to 1 mm (assuming 1 inch = 25.4 mm). This
	is defined in sp.h as PRINTING_POINTS_PER_MM.
	
	Here are values to go in \fontsize{17.28pt}{20.74pt} in Latex files:
	  
		\fontsize{07.00pt}{08.40pt}
		\fontsize{08.00pt}{09.60pt}
		\fontsize{09.00pt}{10.80pt}
		\fontsize{10.00pt}{12.00pt}
		\fontsize{12.00pt}{14.40pt}
		\fontsize{14.00pt}{16.80pt}
		\fontsize{17.28pt}{20.74pt}
		\fontsize{20.74pt}{24.89pt}
		\fontsize{24.88pt}{29.86pt}
		
		The second figure is 1.2 times the first figure.
		  
		A 'standard' (12 pt?) character is 2.125 mm wide and 7.6 mm high.
			
		An A4 sheet of paper has dimensions as follows:
			  
		Width: 210 mm. With margins of 25 mm, this gives a
		printing space of 160 mm (6.25 inch).
				
		Height: 297 mm. With margins of 25 mm, this gives a
		printing space of 247 mm (9.6 inch).
				  
		For the two orientations, and assuming the letter size above,
		these figures mean:
					
		Portrait: 75.29 letters / line, 32.5 lines.
					  
		Landscape: 116.24 letters / line, 21 lines.
						
		Font families: \fontfamily{family}, families include:
		cmr for Computer Modern Roman, cmss for Computer Modern Sans Serif,
		cmtt for Computer Modern Typewriter.
		
		Standard fontsizes in LaTeX are 7, 8, 9, 10, 11, 12, 14, 17.28, 20.74,
		and 24.88. The baselineskip_in_points (distance from the
		bottom of one line to the bottom of the next) is normally 1.2 times
		the point size. */
	
	if (orientation == LANDSCAPE)
	{
		printing_height = PRINTING_SPACE_SHORT_SIDE ;
		printing_width = PRINTING_SPACE_LONG_SIDE ;
	}
	else
	{
		printing_height = PRINTING_SPACE_LONG_SIDE ;
		printing_width = PRINTING_SPACE_SHORT_SIDE ;
	}
	
	/* Find the length of the sequence in characters. 1 is
	added to the length of each symbol except the last to represent
	the space between symbols. */
	
	al_length_in_characters = sequence1->compute_character_length() ;
	al_lines = (sequence1->get_sequence_depth() * 2) - 1 ; // The number
	// of lines in the printed alignment is twice the number
	// of rows (to allow for an extra line between each row)
	// less 1 because the extra line is not needed for the
	// last row.
	
	// Calculate a 'target' font size. This may then be adjusted in
	// the light of available font sizes and maximum_font_height
	// and minimum_font_height.
	
	target_font_height_in_points = (printing_width /
		al_length_in_characters)
		* PRINTING_POINTS_PER_MM
		* CHARACTER_HEIGHT_WIDTH_RATIO ;
	
	// Find available font size which is >= minimum_font_height
	// and <= maximum_font_height and is nearest to the target_font size.
	// Start by looking for the available font size which is nearest
	// to the target font size.
	
	double difference, smallest_difference = HIGH_VALUE ;
	int index_best ;
	
	for (i = 0; i < FONT_SET_SIZE; i++)
	{
		if (target_font_height_in_points < font_heights[i])
			difference = font_heights[i] -
			target_font_height_in_points ;
		else difference = target_font_height_in_points -
			font_heights[i] ;
		if (difference < smallest_difference)
		{
			smallest_difference = difference ;
			index_best = i ;
		}
	}
	
	actual_font_height_in_points = font_heights[index_best] ;
	
	if (actual_font_height_in_points < minimum_font_height)
		actual_font_height_in_points = minimum_font_height ;
	
	if (actual_font_height_in_points > maximum_font_height)
		actual_font_height_in_points = maximum_font_height ;
	
	actual_font_width_in_points = actual_font_height_in_points /
		CHARACTER_HEIGHT_WIDTH_RATIO ;
	
	// Calculated the distance between the bottom of one line and
	// the bottom of the next line below.
	
	baselineskip_in_points = actual_font_height_in_points * 1.2 ;
	
	// Convert measurements in points to measurements in mm.
	
	// actual_font_height_in_mm = actual_font_height_in_points /
	// PRINTING_POINTS_PER_MM ;
	actual_font_width_in_mm = actual_font_width_in_points /
		PRINTING_POINTS_PER_MM ;
	base_line_skip_in_mm = baselineskip_in_points / PRINTING_POINTS_PER_MM ;
	
	fprintf(output_file, "%s%1.2f%s",
		"actual_font_height_in_points = ",
		actual_font_height_in_points,
		" pt\n\n") ;
	
	// Calculate how many character columns of an alignment can
	// be fitted across the printing_width without, at this stage,
	// allowing for the row number printed at the beginning and end
	// of each row. This adjustment will be made in write_alignment.
	// Most sections will be shorter than the calculated number of
	// columns because alignments will only be split at blank columns.
	
	int write_section_chars_length_latex = (int)
		ceil(printing_width / actual_font_width_in_mm) ;
	
	// Calculate the number of sections of the alignment can
	// be fitted into the printing_height.
	
	al_actual_depth_in_mm = base_line_skip_in_mm * (al_lines + 2) ;
	// al_lines is increased by 2 to allow for the 2 blank
	// lines between one section and the next or between
	// the last section and the "Figure xxx" label. No attempt
	// has been made to allow for this bottom line because
	// there are likely to be few occasions when there is
	// not a spare line to accommodate it.
	
	int sections_per_page = (int)floor(printing_height /
		al_actual_depth_in_mm) ;
	
	fprintf(latex_file, "%s%s%s%s%s%s",
		"\\documentclass{article}\n",
		"\\textheight 254mm\n",
		"\\textwidth 177.8mm\n",
		"\\topmargin -20mm\n",
		"\\oddsidemargin 0mm\n",
		"\\evensidemargin 0mm\n") ;
	
	
	if (orientation == LANDSCAPE)
		fprintf(latex_file, "\\usepackage{lscape}\n") ;
	
	fprintf(latex_file, "\\begin{document}\n") ;
	
	if (orientation == LANDSCAPE)
		fprintf(latex_file, "\\begin{landscape}\n") ;
	
	fprintf(latex_file, "%s%1.2f%s%1.2f%s",
		"\\fontsize{",
		actual_font_height_in_points,
		"pt}{",
		baselineskip_in_points,
		"pt}\n") ;
	
	fprintf(latex_file, "%s%s",
		"\\pagestyle{empty}\n",
		"\\begin{verbatim}\n") ;

	sequence1->write_alignment_horizontal(latex_file, 
		write_section_chars_length_latex,
		sections_per_page) ;
	
	fprintf(latex_file, "%s%s%s",
		"\nFigure ",
		figure_ID,
		"\n") ;
	
	fprintf(latex_file, "\\end{verbatim}\n") ;
	
	if (orientation == LANDSCAPE)
		fprintf(latex_file, "\\end{landscape}\n") ;
	
	fprintf(latex_file, "\\end{document}\n") ;
	
	fflush(latex_file) ;
} /* output_alignment */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compares two double variables.
**
** CALLING SEQUENCE:
**
**	bool is_better_than(double A, double B)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if A is better than B, false otherwise
**
**	A, B:			The two variables to be compared.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool is_better_than(double A, double B)
{
	if (A < B) return(true) ;
	else return(false) ;
} // is_better_than

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Calculates the logarithm to the base 2 according to the formula
**	given in Knuth's "The Art of Computer Programming", vol 2,
**	page 23.
**
** CALLING SEQUENCE:
**
**	double log_2(double x)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The required log to the base 2 of x.
**
**	x:				The number whose logarithm is to be calculated.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

double log_2(double x)
{
	return(log10(x) / log_10_2) ;
} // log_2

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Looks up a symbol name in key_array[] to find the corresponding
**	abbreviation. If no abbreviation can be found, the original
**	string is returned. Any symbol beginning with '#' is
**	returned directly.
**
** CALLING SEQUENCE:
**
**	char *find_symbol_abbreviation(char *symbol_name_1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	A character string for the abbreviation.
**
**	symbol_name_1:	The name whose abbreviation is to be found.
**
** IMPLICIT INPUTS:
**
**	key_array[]
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

char *find_symbol_abbreviation(char *symbol_name_1)
{
	int key_index ;
	char *char_string ;
	
	for (key_index = 0; key_index < fe_key_array; key_index++)
	{
		char_string = key_array[key_index].complete_symbol ;
		if (strcmp(symbol_name_1, char_string) == 0)
				return(key_array[key_index].abbreviated_symbol) ;
	}
	
	// If this point is reached, then no match was found for symbol_name_2.
	// In this case, the function returns symbol_name_1.
	
	return(symbol_name_1) ;
} // find_symbol_abbreviation

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  In write_alignment, fills in the non-hit symbols before a hit
**  symbol.
**
** CALLING SEQUENCE:
**
**  bool fill_in_non_hit_symbols(int s_row1, symbol *hit_symbol,
**      int *fe_col)
**
** FORMAL ARGUMENTS:
**
**  Return value:   true if the write_al_chars_length has been breached.
**  false otherwise.
**
**  s_row1:     The relevant row in row_array[].
**  hit_symbol: The hit symbol which terminates the sequence of
**              symbols. If hit_symbol == NIL then all the trailing
**              symbols for each row are added to the alignment.
**  fe_col:     The first empty column in the given row of
**              write_al_chars[][] when the last non-hit symbol
**              before the hit_symbol has been filled in.
**
** IMPLICIT INPUTS:
**
**  arrays for write_alignment.
**
** IMPLICIT OUTPUTS:
**
**  arrays for write_alignment.
**
** SIDE EFFECTS:
**
**  NONE
**
*/

bool fill_in_non_hit_symbols(int s_row1, symbol *hit_symbol, int *fe_col)
{
    symbol *symbol1 ;
    bool char_column_not_OK ;
    sequence *pattern1 = row_array[s_row1].row_pattern ;
    int end_col, c_col1, c_col2, slot_width, c_row1 = s_row1 * 2 ;
    char *c, *symbol_name ;
    
    pattern1->set_current_el_pos(row_array[s_row1].patt_last_el_pos) ;
    c_col1 = row_array[s_row1].fe_col_in_write_al_chars ;
    while (symbol1 = (symbol *)pattern1->get_next_child())
    {
        if (symbol1 == hit_symbol) break ;
        
        if (use_abbreviations == true)
            symbol_name = find_symbol_abbreviation(symbol1->
            get_name()) ;
        else symbol_name = symbol1->get_name() ;
        
        // Find the next available space in this row of
        // write_al_chars[][] to take the symbol name plus
        // space for one space character.
        
        slot_width = strlen(symbol_name) + 1 ; // slot_width
        // is the enough space for the symbol name
        // plus 1 character for the space following the symbol.
        
        char_column_not_OK = true ;
        while (char_column_not_OK)
        {
            char_column_not_OK = false ;
            end_col = c_col1 + slot_width ;
            for (c_col2 = c_col1; c_col2 < end_col; c_col2++)
            {
                if (write_al_vacant_slots[s_row1][c_col2] ==
                    'x')
                {
                    char_column_not_OK = true ;
                    c_col1 = c_col2 + 1;
                    break ;
                }
            }
        }
        
        // Check that we are still within the bounds
        // of write_al_chars[][].
        
        if (c_col1 + slot_width >= write_al_chars_length)
        {
            // The writing out of this row of the alignment
            // will be truncated at this point, with "..." to
            // show that there is more to be written.
            
            *fe_col = c_col1 ;
            return(true) ;
        }
        else
        {
            // Fill in the symbol name.
            
            for (c = symbol_name; *c != '\0'; c++)
                write_al_chars[c_row1][c_col1++] = *c ;
        }
        
        // Add 1 to c_col to allow for the space character following
        // the symbol.
        
        c_col1++ ;
    }
    
    // Test to see whether the end of the pattern has been reached.
    
    if (symbol1 == NIL) row_array[s_row1].pattern_is_finished = true ;
    
    // Update the value of patt_last_el_pos for this row. At this
    // point, there is no need to update the value of
    // fe_col_in_write_al_chars because this will be updated
    // in write_alignment after the hit symbol has been written.
    
    row_array[s_row1].patt_last_el_pos = pattern1->get_current_el_pos() ;
    
    *fe_col = c_col1 ;
    return(false) ;
    
} // fill_in_non_hit_symbols

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Partitions sort_array[]. (Used by sort_leaf_array()).
**
** CALLING SEQUENCE:
**
**	void partition(int low, int high, int *mid)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	low:		The 'low' position in sort_array[].
**	high:		The 'high' position in sort_array[].
**	mid:		A pointer to the 'mid' position in sort_array[]
**				(in the calling function).
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void partition(int low, int high, int *mid)
{
	int left_ptr, right_ptr, temp_num ;
	double pivot_score ;
	hit_node *h_node ;
	
	left_ptr = low ;
	right_ptr = high ;
	while (left_ptr < right_ptr)
	{
	/* Search for the rightmost element greater than
	h_node->get_compression_difference() for the 'low' position
	in leaf_array[] (which is the 'pivot'), starting from
	the right. */
		
		h_node = leaf_array[sort_array[low]] ;
		pivot_score = h_node->get_compression_difference() ;
		h_node = leaf_array[sort_array[right_ptr]] ;
		while (h_node->get_compression_difference() < pivot_score)
			h_node = leaf_array[sort_array[--right_ptr]] ;
		
		/* Seach for the leftmost element greater than the
		pivot score, starting from the left. */
		
		h_node = leaf_array[sort_array[left_ptr]] ;
		while (left_ptr < right_ptr
			&& (h_node->get_compression_difference() >= 
			pivot_score))
			h_node = leaf_array[sort_array[++left_ptr]] ;
		
		/* If the pointers have not crossed, switch the values. */
		
		if (left_ptr < right_ptr)
		{
			temp_num = sort_array[left_ptr] ;
			sort_array[left_ptr] = sort_array[right_ptr] ;
			sort_array[right_ptr] = temp_num ;
		}
	}
	*mid = right_ptr ;
	temp_num = sort_array[*mid] ;
	sort_array[*mid] = sort_array[low] ;
	sort_array[low] = temp_num ;
} /* partition */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Sorts rows of leaf_array[] by compression_ratio in leaf_node.
**	This version uses the Quicksort method.
**
** CALLING SEQUENCE:
**
**	void sort_leaf_array(int low, int high) ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	low:			The 'low' position in sort_array[].
**	high:			The 'high' position in sort_array[].
**
** IMPLICIT INPUTS:
**
**	sort_array[] and leaf_array[]
**
** IMPLICIT OUTPUTS:
**
**	sort_array[] and leaf_array[]
**
** SIDE EFFECTS:
**
**	NONE
*/

void sort_leaf_array(int low, int high)
{
	int mid ;
	
	if (low < high)
	{
		partition(low, high, &mid) ; /* Partition sort_array[] */
		sort_leaf_array(low, mid - 1) ; /* Sort the lower portion */
		sort_leaf_array(mid + 1, high) ; /* Sort the upper portion */
	}
} /* sort_leaf_array */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Makes new alignments.
**
** CALLING SEQUENCE:
**
**	bool make_alignments()
**
** FORMAL ARGUMENTS:
**
**	Return value:		true if one or more new sequences 
**						are formed, false otherwise.
**
** IMPLICIT INPUTS:
**
**	Global data structures
**
** IMPLICIT OUTPUTS:
**
**	New unifications
**
** SIDE EFFECTS:
**
**	NONE
*/

bool make_alignments()
{
	sequence *al1, *al2 ;
	hit_node *node1 ;
	int hit_node_row, leaf_index ;
	bool al_constr_formed = false ;
	
	if (audit_trail)
	{
		fprintf(output_file, "STARTING make_alignments (") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, ")\n\n") ;
		fflush(output_file) ;
	}
	
	number_of_al_constrs = 0 ;
	
	hit_node_row = 0 ;
	while (node1 = get_leaf_nodes_in_order(&hit_node_row, &al1, 
		&leaf_index))
	{
		if (node1->get_compression_difference() <= fail_score)
			continue ;
		
			/* Make a new sequence and, if it is not NIL and
			does not match a sequence formed on this cycle and all
			earlier cycles, add it to alignments_array[]. */
		
		al2 = unify_best(leaf_index) ;          // that is formed.

		if (al2 == NIL) continue ;
		
		if (show_unselected_alignments)
		{
			if (audit_trail)
			{
				fprintf(output_file, "NEW ALIGNMENT\n\n") ;
				al2->print_ID() ;
				fprintf(output_file, " : ") ;
				(node1->get_driving_pattern())->print_ID() ;
				fprintf(output_file, " : ") ;
				(node1->get_target_pattern())->print_ID() ;
				fprintf(output_file, " : ") ;
				node1->print_ID() ;
				fprintf(output_file, 
					"%s%1.2f%s%1.2f%s%1.2f%s%1.2f%s%1.12g",
					" NSC = ", al2->get_new_symbols_cost(),
					", EC = ", al2->get_encoding_cost(),
					", CR = ", al2->get_compression_ratio(),
					", CD = ", al2->get_compression_difference(),
					",\nAbsolute P = ", al2->get_abs_P()) ;

					if (al2->get_degree_of_matching() == FULL_A)
						fprintf(output_file, ", FULL_A") ;
					else if (al2->get_degree_of_matching() == FULL_B)
						fprintf(output_file, ", FULL_B") ;
					else if (al2->get_degree_of_matching() == FULL_C)
						fprintf(output_file, ", FULL_C") ;
					else if (al2->get_degree_of_matching() == PARTIAL)
						fprintf(output_file, ", PARTIAL") ;
					else fprintf(output_file, "NULL_VALUE") ;

					fprintf(output_file, "\n\n") ;
				
				al2->write_alignment(output_file,  
					write_section_chars_length, NULL_VALUE, 
					alignment_format) ;
				
				fflush(output_file) ;
			}
		}
		
		al_constr_formed = true ;
		
		number_of_al_constrs++ ;
		
		if (number_of_al_constrs > max_alignments_in_one_cycle)
		{
			if (audit_trail)
			{
				fprintf(output_file, "%s%s%d%s",
					"Formation of alignments cut short by ",
					"MAX_ALIGNMENTS_IN_ONE_CYCLE (",
					max_alignments_in_one_cycle,
					")\n\n") ;
			}
			break ;
		}
	}
	
	sort_leaf_array(0, fe_sort - 1) ;
	
	if (audit_trail)
	{
		fprintf(output_file, "FINISHED make_alignments (") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, ")\n\n") ;
		fflush(output_file) ;
	}
	
	if (al_constr_formed == false)
	{
		if (audit_trail)
		{
			fprintf(output_file,"FOR ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file,
				"\nNO NEW ALIGNMENTS HAVE BEEN FOUND.\n\n") ;
		}
	}
	
	return(al_constr_formed) ;
	
} /* make_alignments */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
** Checks whether the maximum number of 'unsuported' cycles for a given pattern
** from New has been reached.
**
** CALLING SEQUENCE:
**
**	bool max_unsupported_cycles_reached()
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if, in this New pattern, there have been more
**					than max_unsupported_cycles since the
**					best CD value for the pattern was reached.
**					false otherwise.
**
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool max_unsupported_cycles_reached()
{
	if (cycle - best_cycle_this_cpfn >= max_unsupported_cycles)
	{
		// We have reached the set limit on the number of
		// cycles which may be performed without any gain
		// in compression score.
		
		if (audit_trail)
		{
			fprintf(output_file, "FOR ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, "%s%d%s",
				", THE MAXIMUM NUMBER OF UNSUPPORTED CYCLES (",
				max_unsupported_cycles,
				") HAS BEEN REACHED\n\n") ;
		}
		
		return(true) ;
	}
	else return(false) ;
} // max_unsupported_cycles_reached

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compresses a sequence from New by repeated applications of
**	find_best_matches.
**
** CALLING SEQUENCE:
**
**	bool recognise()
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if unifiable matchings have been found,
**					false otherwise.
**
** IMPLICIT INPUTS:
**
**	old_patterns
**
** IMPLICIT OUTPUTS:
**
**	old_patterns	
**
** SIDE EFFECTS:
**
**	NONE
*/

bool recognise()
{
	bool return_value = true ;
	sequence *al1 ;
	bool partial_al_found, full_A_al_found,
		full_B_al_found ;
	int type_of_alignment ;

	if (audit_trail)
	{
		fprintf(output_file, "START OF ") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, ":\n\n") ;
	}
	
	set_up_hit_structure() ;

	if (audit_trail)
	{
		if (window == 1 && cycle == 1)
		{
			fprintf(output_file, "DRIVING PATTERN FROM NEW (") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, ")\n\n") ;
			current_new_pattern->print_ID() ;
			fprintf(output_file, " ") ;
			current_new_pattern->write_tree_object(PRINT_NO_FREQUENCIES) ;
		}
	}
		
	if (!find_best_matches())
	{
		return_value = false ;
		goto L1 ;
	}

	if (!make_alignments())
	{
		return_value = false ;
		goto L1 ;
	}
	
	make_selection() ;
	
	// Determine which of FULL_A, FULL_B or PARTIAL alignments
	// have been found.

	partial_al_found = false ;
	full_A_al_found = false ;
	full_B_al_found = false ;	

	list_for(al1, sequence, current_alignments_from_parsing)
	{
		type_of_alignment = al1->get_degree_of_matching() ;
		if (type_of_alignment == FULL_A) full_A_al_found = true ;
		if (type_of_alignment == FULL_B) full_B_al_found = true ;
		if (type_of_alignment == PARTIAL) partial_al_found = true ;
	}

	current_alignments_from_parsing->sort_by_compression_difference() ;

	if (combine_alignments)
	{
		if (match_all_data_symbols_in_composite_alignments)
		{
			if (full_B_al_found)
			{
				make_combinations(current_new_pattern) ;
			}
			else if (audit_trail)
			{
				fprintf(output_file, "%s%s%s",
					"make_combinations() not run because\n",
					"match_all_data_symbols_in_composite_alignments == true\n",
					"but full_B_al_found == false.\n\n") ;
			}
		}
		else make_combinations(current_new_pattern) ;
	}

	current_alignments_from_parsing->sort_by_ID() ;

	// Print ID numbers of alignments.

	if (audit_trail)
	{
		fprintf(output_file, "%s%s%s",
			"In the following lists, alignments that are new on this cycle\n",
			"are marked with an asterisk ('*').\n\n",
			"SELECTED ALIGNMENTS FROM ") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, "\n\n") ;

		current_alignments_from_parsing->write_IDs() ;

		fflush(output_file) ;

		if (full_A_al_found)
		{
			fprintf(output_file, "SELECTED FULL ALIGNMENTS (A) FROM ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, "\n\n") ;

			current_alignments_from_parsing->write_IDs() ;
		}
		else
		{
			fprintf(output_file, "NO FULL_A ALIGNMENTS FROM ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, "\n\n") ;
		}

		if (full_B_al_found)
		{
			fprintf(output_file, "SELECTED FULL ALIGNMENTS (B) FROM ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, "\n\n") ;

			current_alignments_from_parsing->write_IDs() ;
		}
		else
		{
			fprintf(output_file, "NO FULL_B ALIGNMENTS FROM ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, "\n\n") ;
		}

		if (partial_al_found)
		{
			fprintf(output_file, "SELECTED PARTIAL ALIGNMENTS FROM ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, "\n\n") ;
			current_alignments_from_parsing->write_IDs() ;
		}
		else
		{
			fprintf(output_file, "NO PARTIAL ALIGNMENTS FROM ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, "\n\n") ;
		}
	}

	if (max_unsupported_cycles_reached()) return_value = false ;
	
	L1: delete hit_root ; // This deletes the whole tree of hit
		// nodes, recursively.

	driving_alignments->release_children() ; // Release the driving
		// alignments from this cycle.

	while (al1 = (sequence *)
			current_alignments_from_parsing->extract_first_child())
	{
		cumulative_alignments_from_parsing->add_child(al1) ;
		if (al1->get_keep() == DRIVING)
			driving_alignments->add_child(al1) ; // Prepare driving
				// alignments for next cycle.
	}

	if (audit_trail)
	{
		fprintf(output_file, "END OF ") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, "\n\n") ;
	}
	
	cycle++ ;
	
	return(return_value) ;	
} /* recognise */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For each of the sequences in best_alignments[], this function
**	compiles a list of the other best sequences, if any, which
**	contain the given sequence.
**
** CALLING SEQUENCE:
**
**	void find_containing_sequences()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	best_alignments[]
**
** IMPLICIT OUTPUTS:
**
**	For each best_sequence[], a list of containing sequences which
**	may be NIL.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void find_containing_sequences()
{
	int i, j ;
	
	for (i = 0; i < fe_best_alignments; i++)
	{
		for (j = 0; j < fe_best_alignments; j++)
		{
			if (i == j) continue ;
			if ((best_alignments[i].al1)->contained_in(best_alignments[j].al1))
			{
				best_alignments[i].is_contained_in_another_alignment = true ;
				if (audit_trail)
				{
					fprintf(output_file, "Alignment ") ;
					best_alignments[i].al1->print_ID() ;
					fprintf(output_file, " is contained in ") ;
					best_alignments[j].al1->print_ID() ;
					fprintf(output_file, "\n\n") ;
				}
			}
		}
	}
} // find_containing_sequences

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Fetches symbols by row order and symbol order from an sequence,
**	excluding any symbols in the first row (which represents New)
**	which are not aligned with any symbol in Old. Each symbol type
**	is returned only once, ie second and subsequent instances of a symbol
**	type are skipped.
**
** CALLING SEQUENCE:
**
**	symbol *get_next_symbol_from_sequence(sequence *al1, int status)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The next symbol type in the sequence.
**
**	al1:			The sequence from which symbol types are fetched.
**	status: 		If START, the sequence is searched from the
**					beginning, otherwise it continues from where it
**					was previously.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

symbol *get_next_symbol_from_sequence(sequence *al1, int status)
{
	static symbol *B_symbol1 ;
	static alignment_element *al_el1 ;
	static symbol *symbol_array[MEDIUM_SCRATCH_ARRAY_SIZE] ;
	symbol *symbol1 ;
	int i, row1, sequence_depth = al1->get_sequence_depth() ;
	
	if (al1 == NIL) abort_run("Invalid value for al1 in \
		get_next_symbol_from_sequence") ;
	
	if (status == START)
	{
		// Initialise symbol_array[] (which is used for checking for 
		// duplicates in symbols returned).
		
		for (i = 0; i < MEDIUM_SCRATCH_ARRAY_SIZE; i++)
			symbol_array[i] = NIL ;
		return(NIL) ;
	}
	
	al1->initialise() ;
	while (B_symbol1 = (symbol *)al1->get_next_child())
	{
		for (row1 = 0; row1 < sequence_depth; row1++)
		{
			al_el1 = B_symbol1->get_al_el(row1) ;
			symbol1 = (symbol *)al_el1->get_el_obj() ;
			if (symbol1 == NIL) continue ;
			
			// Check whether al_el1 is at the top of a symbol.
			// If it is below the top of a symbol,
			// skip to next symbol.
			
			if (al_el1->get_same_column_above() != NULL_VALUE) 
				break ;
			
			// Check whether al_el1 is in the first row
			// (corresponding to New) and is also not part part
			// of any hit. If so, skip.
			
			if (row1 == 0 && al_el1->get_same_column_below() == 
				NULL_VALUE)
				break ;
			
			// Now check whether a matching symbol has been
			// retrieved from this sequence before. If it has,
			// skip to next symbol. Otherwise add symbol1 to 
			// symbol_array[] and return it as the next symbol.
			
			for (i = 0; i < MEDIUM_SCRATCH_ARRAY_SIZE; i++)
			{
				if (symbol_array[i] == NIL)
				{
					symbol_array[i] = symbol1 ;
					return(symbol1) ;
				}
				
				if (strcmp(symbol1->get_name(),
					symbol_array[i]->get_name()) == 0)
					
					// There is no need to
					// continue down this symbol
					// because only symbols of
					// the same type will be found.
					
					goto L1 ;
			}
			
			// If the program reaches this point, it means that
			// symbol_array[] is too small.
			
			abort_run("symbol_array[] is too small in \
				get_next_symbol_from_sequence") ;
			
		}
		L1: ;
	}
	
	return(NIL) ;
} // get_next_symbol_from_sequence

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks whether an alignment contains one or more redundant rows.
**
**	A row is redundant if it is in Old and if every symbol in it
**	forms a hit with a symbol in one other row in Old (see sp61_od, %49).
**	No attempt is made to check for rows which may be redundant
**	because they form hits across two or more rows in Old.
**
**	If one or more rows are redundant, the alignment is edited to remove
**	the redundant rows and the function returns true. Otherwise it
**	returns false.
**
** CALLING SEQUENCE:
**
**	bool check_for_redundant_rows(sequence *al1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if al1 has been edited. false otherwise.
**
**	al1:			The alignment to be checked.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool check_for_redundant_rows(sequence *al1)
{
	// Check each row in the alignment against every other row.
	// A row is redundant if:
	// * Every symbol in the row is a hit symbol.
	// * There is at least one other row which forms a hit with
	//	every symbol of the given row.
	//
	// The second condition means that the other row must be at least
	// as long as the given row. If they are equal in size, then one
	// of them must be chosen arbitrarily as the redundant row.
	
	symbol *al_col1, *al_col2, *symbol1, 
		*new_alignment_array[VERY_LARGE_SCRATCH_ARRAY_SIZE] ;
	alignment_element *al_el1 ;
	int row1, row2, al1_depth = al1->get_sequence_depth(),
		al1_length = al1->get_number_of_children(),
		this_row_pattern_length, that_row_pattern_length,
		this_row_int_pos, temp_row_int_pos ;
	bool redundant_rows_array[MEDIUM_SCRATCH_ARRAY_SIZE], // Used to
		// record whether a row is redundant and is to
		// be deleted (true) or is to be preserved (false).
		that_row_matches_this_row[MEDIUM_SCRATCH_ARRAY_SIZE]; // Used
	// to record when 'that' row matches 'this' row.
	sequence *this_row_pattern, *that_row_pattern, *pattern1 ;
	list_element *this_el_pos_start, *pos1 ;
	
	if (al1_depth >= MEDIUM_SCRATCH_ARRAY_SIZE)
	abort_run("redundant_rows_array[] is too small in \
	check_for_redundant_rows") ;
	if (al1_length >= VERY_LARGE_SCRATCH_ARRAY_SIZE)
	abort_run("new_alignment_array[] is too short in \
		check_for_redundant_row") ;
	
	redundant_rows_array[0] = false ; // Make sure that the first
		// row (New) is preserved in all circumstances.
	
	
	for (row1 = 1; row1 < al1_depth; row1++)
	{
		redundant_rows_array[row1] = false ; // By default, each row
		// is to be preserved, not edited out.
		this_row_pattern = al1->get_row_pattern(row1) ;
		this_row_pattern_length = 
			this_row_pattern->get_number_of_children() ;
		
		// Search for the first non-NIL entry in row1.
		
		this_el_pos_start = NIL ;
		while (al_col1 = (symbol *)
			al1->get_next_child_by_el_pos(&this_el_pos_start))
		{
			al_el1 = al_col1->get_al_el(row1) ;
			symbol1 = (symbol *)al_el1->get_el_obj() ;
			if (symbol1 != NIL) break ;
		}
		
		// Check whether the first hit symbol for this row is
		// the first symbol in the pattern for this row. If it
		// is not, then this row cannot be contained in any other
		// row.

		this_row_int_pos =
			this_row_pattern->get_int_pos_by_child(symbol1) ;
		if (this_row_int_pos != 0) continue ;
		
		// Now step along 'this' row, comparing it with every other
		// row except row 0. If there are any unmatched symbols in
		// 'this' row, break out of the loop because 'this' row
		// cannot be completely matched to any other row.
		// Whenever a mis-match appears between 'that'
		// row and 'this' row, a false value is entered into
		// that_row_matches_this_row[].
		
		// Initialise that_row_matches_this_row[].
		// Check the lengths of the patterns in each
		// row against the length of the pattern for 'this' row.
		// If a row contains a pattern that is shorter than the 
		// pattern in 'this' row, then 'this' row cannot be 
		// completely matched against it.
		
		for (row2 = 1; row2 < al1_depth; row2++)
		{
			if (row2 == row1)
			{
				that_row_matches_this_row[row2] = false ;
				continue ;
			}
			
			// Compare lengths of patterns.
			
			that_row_pattern = al1->get_row_pattern(row2) ;
			that_row_pattern_length = 
				that_row_pattern->get_number_of_children() ;
			if (that_row_pattern_length < this_row_pattern_length)
			{
				that_row_matches_this_row[row2] = false ;
				continue ;
			}
			
			// Check whether there is a symbol in al_col1 in
			// this row.
			
			al_el1 = al_col1->get_al_el(row2) ;
			if (al_el1->get_el_obj() == NIL)
				that_row_matches_this_row[row2] = false ;
			else that_row_matches_this_row[row2] = true ;
		}
		
		// At this point, we know that this_row_int_pos == 0.
		
		pos1 = this_el_pos_start ;
		while (al_col1 = (symbol *)
			al1->get_next_child_by_el_pos(&pos1))
		{
			al_el1 = al_col1->get_al_el(row1) ;
			
			// Skip over any column which does not contain
			// a symbol for row1.

			symbol1 = (symbol *)al_el1->get_el_obj() ;
			if (symbol1 == NIL) continue ;
			
			// Check whether there are any unmatched symbols
			// in 'this' row between the current column and
			// the last one. If there is, 'this' row cannot
			// be completely contained in any other.

			pattern1 = al1->get_row_pattern(row1) ;
			temp_row_int_pos = pattern1->get_int_pos_by_child(symbol1) ;
			if (temp_row_int_pos - this_row_int_pos > 1) break ;
			
			// 'this' row contains a symbol and there are no
			//	unmatched symbols preceding the current symbol
			// in this row.
			
			this_row_int_pos = temp_row_int_pos ;
			
			// Now examine all the other rows for this column.
			// If there is no symbol or an unmatched symbol
			// in any row then it cannot be a matching row
			// for 'this' row. Also, there is no point checking
			// rows which are already known to be false.
			
			for (row2 = 1; row2 < al1_depth; row2++)
			{
				if (that_row_matches_this_row[row2] == false)
					continue ;
				
				if (row2 == row1)
				{
					that_row_matches_this_row[row2] =
						false ;
					continue ;
				}
				
				// Check whether there is a symbol in
				// al_col1 in this row.
				
				al_el1 = al_col1->get_al_el(row2) ;
				if (al_el1->get_el_obj() == NIL)
					that_row_matches_this_row[row2] =
					false ;
				else that_row_matches_this_row[row2] = true ;
			}
		}
		
		// Now examine that_row_matches_this_row[] to see if
		// there are any rows that are true. If there are, this
		// means that row1 is redundant and this can be entered
		// into redundant_rows_array[].
		
		// If Old contains duplicate
		// patterns (which is possible but unlikely), then there
		// may be two rows which match each other completely and
		// are the same length. It is assumed here that Old does
		// not contain duplicate patterns.
		
		for (row2 = 1; row2 < al1_depth; row2++)
			if (that_row_matches_this_row[row2] == true) break ;
			
			if (row2 < al1_depth)
			{
				// This means that there is at least one row
				// which matches row1 completely.
				
				redundant_rows_array[row1] = true ;
			}
	}
	
	// Now check to see whether any rows need to be edited out and 
	// do it where necessary.
	
	int new_al_depth = al1_depth ;
	for (row1 = 0; row1 < al1_depth; row1++)
		if (redundant_rows_array[row1] == true) new_al_depth-- ;
		
		if (new_al_depth == al1_depth) return(false) ;
		
		// al1 is now to be edited. Do this by replacing each column by
		// a new column. The 'shell' of al1 is preserved to avoid upsetting
		// any pointers to it which may exist in the global data structure.
		
		alignment_element *al_el2 ;
		int col1 = -1 ;
		al1->initialise() ;
		while (al_col1 = (symbol *)al1->get_next_child())
		{
			al_col2 = new symbol(al_col1->get_name(), new_al_depth,
				NULL_VALUE) ;
			plus_one(&col1, al1_length, "col1 is too large in \
				check_for_redundant_rows") ;
			new_alignment_array[col1] = al_col2 ;
			row2 = -1 ;
			for (row1 = 0; row1 < al1_depth; row1++)
			{
				if (redundant_rows_array[row1] == true) continue ;
				// This row is to be omitted from the
				// edited version of al1.
					
				// This row is to be preserved.
					
				plus_one(&row2, new_al_depth,
					"row2 is too large in check_for_redundant_rows") ;
					
				al_el1 = al_col1->get_al_el(row1) ;
				al_el2 = al_col2->get_al_el(row2) ;
				*al_el2 = *al_el1 ;
			}
		}
		
		// Delete all the columns in al1.
		
		while (al_col1 = (symbol *)al1->get_first_child())
		{
			al1->extract_child(al_col1) ;
			delete al_col1 ;
		}
		
		// Now put in the new columns currently stored in
		// new_alignment_array[].
		
		for (col1 = 0; col1 < al1_length; col1++)
			al1->add_child(new_alignment_array[col1]) ;
		
		// Re-set sequence_depth in al1.
		
		al1->set_sequence_depth(new_al_depth) ;
		
		// Re-set values in each column of al1.
		
		al1->initialise() ;
		int symbol_count ;
		while (al_col1 = (symbol *)al1->get_next_child())
		{
			al_col1->set_sequence_depth(new_al_depth) ;
			
			// Does this column contain at least one hit?
			
			symbol_count = 0 ;
			for (row1 = 0; row1 < new_al_depth; row1++)
			{
				al_el1 = al_col1->get_al_el(row1) ;
				if (al_el1->get_el_obj() != NIL)
					symbol_count++ ;
			}
			
			if (symbol_count > 1) al_col1->set_is_a_hit(true) ;
			else al_col1->set_is_a_hit(false) ;
			
			// For each columnm, put the correct values in for 
			// same_column_above and same_column_below.
			
			for (row1 = 0; row1 < new_al_depth; row1++)
			{
				al_el1 = al_col1->get_al_el(row1) ;
				if (al_el1->get_el_obj() == NIL) continue ;
				for (row2 = row1 + 1; row2 < new_al_depth; row2++)
				{
					al_el2 = al_col1->get_al_el(row2) ;
					if (al_el2->get_el_obj() != NIL) break ;
				}
				if (row2 >= new_al_depth) break ;
				al_el1->set_same_column_below(row2) ;
				al_el2->set_same_column_above(row1) ;
			}
		}
				
		return(true) ;
		
} // check_for_redundant_rows

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	If there is a best sequence, this function calculates the
**	probabilities of inferences from the best sequence and all
**	other sequences which encode the same symbols from New.
**
**	The function calculates 'Absolute P' which is the absolute probability
**	of each sequence (derived directly from its 'old symbols cost')
**	and 'Relative P' which is the probability of each sequence relative
**	to other sequences which encode the same symbols from New.
**
**	The function also calculates the probabilities of patterns and
**	symbols within the sequences as described in sp_ideas6, %37.
**	For each pattern found anywhere amongst the set of sequences 
**	which encode the same symbols from New, the function adds up
**	the values of Small p for the sequences in which it is found,
**	giving a value of finding at least one instance of the pattern
**	in an alignment. Likewise for individual symbols.
**
** CALLING SEQUENCE:
**
**	void probabilities_of_inferences(sequence *best_alignment)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	best_alignment: The best sequence found at the end of compression.
**
** IMPLICIT INPUTS:
**
**	Global variables
**
** IMPLICIT OUTPUTS:
**
**	Prints all sequences which code the same symbols from New as
**	the best sequence and calculates probabilities of the
**	associated inferences.
**
** SIDE EFFECTS:
**
**	NONE
*/

void probabilities_of_inferences(sequence *best_alignment)
{
	symbol *B_symbol1 ;
	alignment_element *al_el1 ;
	sequence *pattern ;
	symbol *new_hit_symbol, *symbol1, *symbol2, *hit_symbol ;
	group *new_hits = new group() ; /* This is
		the head of a list of symbols from New
		which have formed hits in the best_alignment. */
	sequence *al1, *al2 ;
	struct best_alignment_entry *sequence_entry ;
	struct patterns_entry
	{
		sequence *pattern ;
		double probability ;
		bool in_current_sequence, done ;
	} patterns_in_sequences[MEDIUM_SCRATCH_ARRAY_SIZE] ;
	
	struct symbols_entry
	{
		symbol *symbol1 ;
		double probability ;
		bool in_current_sequence, done ;
	} symbols_in_sequences[MEDIUM_SCRATCH_ARRAY_SIZE] ;
	
	double total_abs_P ;
	int i, j, k, fe_p_in_als, fe_s_in_als, row1,
		sequence_depth ;
	bool sequence_found, symbol_found ;
	
	if (best_alignment == NIL)
	{
		if (audit_trail)
		{
			fprintf(output_file,
				"NO GOOD ALIGNMENTS OR INFERENCES FOUND\n\n") ;
		}
		return ;
	}
	
	fe_best_alignments = 0 ;
	
	best_alignments[fe_best_alignments].al1 = best_alignment ;
	best_alignments[fe_best_alignments].is_contained_in_another_alignment = false ;
	fe_best_alignments++ ;
	
	/* Compile symbols from New which are hit_symbols. */

	best_alignment->initialise() ;
	while (B_symbol1 = (symbol *)best_alignment->get_next_child())
	{
		al_el1 = B_symbol1->get_al_el(0) ;
		if (al_el1->get_same_column_below() == NULL_VALUE) continue ;
		
		hit_symbol = (symbol *)al_el1->get_el_obj() ;
		new_hits->add_child(hit_symbol) ;
	}
	
	// Now look for sequences (other than best_alignment), which
	// have the same hit_symbols in New, neither more nor less.
	// For each one found, check for redundant rows and remove them
	// (see sp61_od, %49). For each alignment which is edited in this
	// way, check to see that it is not identical to one already in
	// the reference set of alignments. If it is, do not add it to the
	// set.
	
	bool alignment_is_OK ;
	sequence *modified_alignments[MEDIUM_SCRATCH_ARRAY_SIZE] ;
	int fe_modified_alignments = 0, modified_alignments_index ;

	list_for(al1, sequence, cumulative_alignments_from_parsing)
	{
		if (al1 == best_alignment) continue ;
		
		new_hits->initialise() ;
		new_hit_symbol = (symbol *)new_hits->get_next_child() ;
		al1->initialise() ;
		while (B_symbol1 = (symbol *)al1->get_next_child())
		{
			al_el1 = B_symbol1->get_al_el(0) ;
			if (al_el1->get_el_obj() == NIL) continue ;
			if (al_el1->get_same_column_below() == NULL_VALUE) 
				continue ;
			
			hit_symbol = (symbol *)al_el1->get_el_obj() ;
			if (hit_symbol != new_hit_symbol)
				/* There is a mis-match. */
				goto L1 ;
			
			new_hit_symbol = (symbol *)new_hits->get_next_child() ;
		}
		
		if (new_hit_symbol != NIL) continue ;
		
		// All the symbols in new_hits have been matched
		// and none others - which means that this
		// sequence is OK.
		
		// Check to see whether the alignment contains
		// any redundant rows. If it has, remove it or them, then place
		// the modified alignment in modified_alignments[] to
		// be checked later to see whether it duplicates any other
		// alignment in current_alignments_from_parsing.
		
		if (check_for_redundant_rows(al1))
		{
			if (audit_trail)
			{
				fprintf(output_file, "%s%d%s",
					"Alignment %",
					al1->get_ID(),
					" has been edited.\n\n") ;
				al1->write_alignment(output_file, 
					write_section_chars_length, NULL_VALUE, 
					alignment_format) ;
			}
			modified_alignments[fe_modified_alignments] = al1 ;
			plus_one(&fe_modified_alignments,
				MEDIUM_SCRATCH_ARRAY_SIZE,
				"modified_alignments[] is too small") ;
		}
		else
		{
			best_alignments[fe_best_alignments].al1 = al1 ;
			best_alignments[fe_best_alignments].is_contained_in_another_alignment =
				false ;
			plus_one(&fe_best_alignments,
				MEDIUM_SCRATCH_ARRAY_SIZE,
				"best_alignments[] too small.") ;
		}
		
		L1: ;
	}
	
	for (modified_alignments_index = 0; modified_alignments_index <
		fe_modified_alignments; modified_alignments_index++)
	{
		al1 = modified_alignments[modified_alignments_index] ;
		alignment_is_OK = true ;
		list_for(al2, sequence, current_alignments_from_parsing)
		{
			if (al2 == al1) continue ;
			if (al2->alignment_matches(al1))
			{
				if (audit_trail)
				{
					if (verbose)
					{
						fprintf(output_file, "Alignment ") ;
						al1->print_ID() ;
						fprintf(output_file, " matches sequence ") ;
						al2->print_ID(),
						fprintf(output_file, " and is discarded.\n\n") ;
						fflush(output_file) ;
					}
				}
				
				current_alignments_from_parsing->delete_child(al1) ;
				alignment_is_OK = false ;
				break ;
			}
		}
		
		if (alignment_is_OK)
		{
			if (audit_trail)
			{
				fprintf(output_file, "%s%d%s",
					"After editing, alignment %",
					al1->get_ID(),
					" does not match any earlier alignment\n\n") ;
				al1->make_code(false) ;
				fprintf(output_file, "%s%d%s",
					"New scores for alignment %",
					al1->get_ID(),
					" are:\n\n") ;
				fprintf(output_file,
					"%s%1.2f%s%1.2f%s%1.2f%s%1.2f%s%1.12g",
					"NSC = ", al1->get_new_symbols_cost(),
					", EC = ", al1->get_encoding_cost(),
					", CR = ", al1->get_compression_ratio(),
					", CD = ", al1->get_compression_difference(),
					",\nAbsolute P = ", al1->get_abs_P()) ;

				if (al1->get_degree_of_matching() == FULL_A)
					fprintf(output_file, ", FULL_A") ;
				else if (al1->get_degree_of_matching() == FULL_B)
					fprintf(output_file, ", FULL_B") ;
				else if (al1->get_degree_of_matching() == FULL_C)
					fprintf(output_file, ", FULL_C") ;
				else if (al1->get_degree_of_matching() == PARTIAL)
					fprintf(output_file, ", PARTIAL") ;
				else fprintf(output_file, "NULL_VALUE") ;

				fprintf(output_file, "\n\n") ;
			}
			
			best_alignments[fe_best_alignments].al1 = al1 ;
			best_alignments[fe_best_alignments].is_contained_in_another_alignment =
				false ;
			plus_one(&fe_best_alignments,
				MEDIUM_SCRATCH_ARRAY_SIZE,
				"best_alignments[] too small.") ;
		}
	}

	// For each sequence in best_alignments[], compile a list of the
	// other best sequences which contain it, if any.
	
	find_containing_sequences() ;

	// Calculate total_abs_P.
	
	total_abs_P = 0 ;
	for (i = 0; i < fe_best_alignments; i++)
	{
		if (best_alignments[i].is_contained_in_another_alignment) continue ;
		al1 = best_alignments[i].al1 ;
		total_abs_P += al1->get_abs_P() ;
	}
	
	/* Calculate rel_P for each of the alignments in best_alignments[].
	Then write out the alignments and probabilities in order of
	relative probability. */
	
	if (audit_trail)
	{
		fprintf(output_file, "%s%s%1.12g%s",
			"BEST ALIGNMENTS AND PROBABILITIES\n\n",
			"Total absolute P = ",
			total_abs_P, "\n\n") ;
	}

	for (i = 0; i < fe_best_alignments; i++)
	{
		if (best_alignments[i].is_contained_in_another_alignment)
			continue ;
		sequence_entry = &best_alignments[i] ;
		al1 = sequence_entry->al1 ;
		sequence_entry->rel_P = al1->get_abs_P() / total_abs_P ;
		sequence_entry->done = false ; // To be used in printing.
	}
	
	double biggest_rel_P, temp_rel_P ;
	int index_of_biggest ;
	
	while (true)
	{
		biggest_rel_P = NULL_VALUE ;
		index_of_biggest = NULL_VALUE ;
		for (i = 0; i < fe_best_alignments; i++)
		{
			if (best_alignments[i].is_contained_in_another_alignment)
				continue ;
			sequence_entry = &best_alignments[i] ;
			if (sequence_entry->done == true) continue ;
			temp_rel_P = sequence_entry->rel_P ;
			if (temp_rel_P > biggest_rel_P)
			{
				biggest_rel_P = temp_rel_P ;
				index_of_biggest = i ;
			}
		}
		
		if (index_of_biggest == NULL_VALUE) break ;
		sequence_entry = &best_alignments[index_of_biggest] ;
		sequence_entry->done = true ;
		al1 = sequence_entry->al1 ;
		
		if (audit_trail)
		{
			fprintf(output_file, "ALIGNMENT:\n\n") ;
			sequence_entry->al1->print_ID() ;
			fprintf(output_file,	
				"%s%1.2f%s%1.2f%s%1.2f%s%1.2f%s%1.12g%s%1.12g%s",
				", NSC = ", 
				al1->get_new_symbols_cost(),
				", EC = ", 
				al1->get_encoding_cost(),
				", CR = ", 
				al1->get_compression_ratio(),
				", CD = ", 
				al1->get_compression_difference(),
				",\nAbsolute P = ", al1->get_abs_P(),
				", Relative P = ", sequence_entry->rel_P,
				"\n\n") ;
			
			al1->write_alignment(output_file, 
				write_section_chars_length, NULL_VALUE, 
				alignment_format) ;
		}
	}
	
	/* Compile a set of patterns each of which appears at least
	once in one of best_alignments[]. At the same time, calculate
	the probability of each pattern as the sum of the rel_P
	probabilities of the sequences in which it appears. */
	
	for (i = 0; i < MEDIUM_SCRATCH_ARRAY_SIZE; i++)
	{
		patterns_in_sequences[i].pattern = NIL ;
		patterns_in_sequences[i].probability = 0 ;
		patterns_in_sequences[i].in_current_sequence = false ;
		patterns_in_sequences[i].done = false ;
	}
	
	fe_p_in_als = 0 ;
	
	for (i = 0; i < fe_best_alignments; i++)
	{
		if (best_alignments[i].is_contained_in_another_alignment) continue ;
		al1 = best_alignments[i].al1 ;
		
		/* Initialise values of in_current_sequence in
		patterns_in_sequences[]. */
		
		for (j = 0; j < fe_p_in_als; j++)
			patterns_in_sequences[j].in_current_sequence =
			false ;
		
		B_symbol1 = (symbol *)al1->get_first_child() ;
		al1->set_current_el_pos(al1->get_first_el_pos()) ;
		sequence_depth = al1->get_sequence_depth() ;
		
		/* Exclude the first row of the sequence, which
		corresponds to the current_new_pattern. */
		
		for (row1 = 1; row1 < sequence_depth; row1++)
		{
			pattern = B_symbol1->get_row_pattern(row1) ;
			
			/* Check patterns_in_sequences[] to see if there
			is already an entry for pattern. If there is,
			use that entry, otherwise make a new entry. */
			
			sequence_found = false ;
			for (k = 0; k < fe_p_in_als; k++)
			{
				if (patterns_in_sequences[k].pattern ==
					pattern)
				{
					sequence_found = true ;
					break ;
				}
			}
			
			if (!sequence_found)
			{
				patterns_in_sequences[fe_p_in_als].pattern =
					pattern ;
				
				k = fe_p_in_als ;
				plus_one(&fe_p_in_als, 
					MEDIUM_SCRATCH_ARRAY_SIZE,
					"patterns_in_sequences[] \
					is too small") ;
			}
			
			/* Now add to the probability value for this
			pattern if it has not already been done in
			this sequence. */
			
			if (patterns_in_sequences[k].in_current_sequence ==
				true)
				continue ;
			
			patterns_in_sequences[k].in_current_sequence = true ;
			patterns_in_sequences[k].probability +=
				best_alignments[i].rel_P ;
		}
	}
	
	if (audit_trail)
	{
		fprintf(output_file, "PROBABILITIES OF PATTERNS:\n\n") ;
	}
	
	double biggest_probability, temp_probability ;
	
	while (true)
	{
		biggest_probability = NULL_VALUE ;
		index_of_biggest = NULL_VALUE ;
		for (i = 0; i < fe_p_in_als; i++)
		{
			if (patterns_in_sequences[i].done == true) continue ;
			temp_probability =
				patterns_in_sequences[i].probability ;
			if (temp_probability > biggest_probability)
			{
				biggest_probability = temp_probability ;
				index_of_biggest = i ;
			}
		}
		
		if (index_of_biggest == NULL_VALUE) break ;
		patterns_in_sequences[index_of_biggest].done = true ;
		
		pattern = patterns_in_sequences[index_of_biggest].pattern ;
		if (audit_trail)
		{
			pattern->print_ID() ;
			fprintf(output_file, "%s%1.12g%s",
				", p = ",
				patterns_in_sequences[index_of_biggest].probability,
				" ") ;
			
			pattern->write_tree_object(PRINT_NO_FREQUENCIES) ;
		}
	}
	
	/* Compile a set of symbols each of which appears at least
	once in one of best_alignments[]. At the same time, calculate
	the probability of each symbol as the sum of the rel_P
	probabilities of the sequences in which it appears. */
	
	for (i = 0; i < MEDIUM_SCRATCH_ARRAY_SIZE; i++)
	{
		symbols_in_sequences[i].symbol1 = NIL ;
		symbols_in_sequences[i].probability = 0 ;
		symbols_in_sequences[i].in_current_sequence = false ;
		symbols_in_sequences[i].done = false ;
	}
	
	fe_s_in_als = 0 ;
	
	for (i = 0; i < fe_best_alignments; i++)
	{
		if (best_alignments[i].is_contained_in_another_alignment) continue ;
		al1 = best_alignments[i].al1 ;
		
		/* Initialise values of in_current_sequence in
		symbols_in_sequences[]. */
		
		for (j = 0; j < fe_s_in_als; j++)
			symbols_in_sequences[j].in_current_sequence =
			false ;
		
		get_next_symbol_from_sequence(al1, START) ;
		while (symbol2 = get_next_symbol_from_sequence(al1, CONTINUE))
		{
			/* Check symbols_in_sequences[] to see if there
			is already an entry for symbol2. If there is,
			use that entry, otherwise make a new entry. */
			
			symbol_found = false ;
			for (k = 0; k < fe_s_in_als; k++)
			{
				symbol1 = symbols_in_sequences[k].symbol1 ;
				if (strcmp(symbol1->get_name(),
					symbol2->get_name()) == 0)
				{
					symbol_found = true ;
					break ;
				}
			}
			
			if (!symbol_found)
			{
				symbols_in_sequences[fe_s_in_als].symbol1 =
					symbol2 ;
				
				k = fe_s_in_als ;
				plus_one(&fe_s_in_als,
					MEDIUM_SCRATCH_ARRAY_SIZE,
					"symbols_in_alingments[] \
					is too small") ;
			}
			
			/* Now add to the probability value for this
			symbol if it has not already been done in
			this sequence. */
			
			if (symbols_in_sequences[k].in_current_sequence ==
				true)
				continue ;
			
			symbols_in_sequences[k].in_current_sequence = true ;
			symbols_in_sequences[k].probability +=
				best_alignments[i].rel_P ;
		}
	}
	
	if (audit_trail)
	{
		fprintf(output_file, "PROBABILITIES OF SYMBOLS:\n\n") ;
	}
	
	while (true)
	{
		biggest_probability = NULL_VALUE ;
		index_of_biggest = NULL_VALUE ;
		for (i = 0; i < fe_s_in_als; i++)
		{
			if (symbols_in_sequences[i].done == true) continue ;
			temp_probability =
				symbols_in_sequences[i].probability ;
			if (temp_probability > biggest_probability)
			{
				biggest_probability = temp_probability ;
				index_of_biggest = i ;
			}
		}
		
		if (index_of_biggest == NULL_VALUE) break ;
		symbols_in_sequences[index_of_biggest].done = true ;
		
		symbol2 = symbols_in_sequences[index_of_biggest].symbol1 ;
		if (audit_trail)
		{
			fprintf(output_file, "%s%s%1.12g%s",
				symbol2->get_name(), ", p = ",
				symbols_in_sequences[index_of_biggest].probability,
				" ") ;
			fprintf(output_file, "\n") ;
		}
	}
	
	if (audit_trail)
	{
		fprintf(output_file, "\n") ;
	}
	
	new_hits->release_children() ;
	delete new_hits ;
} /* probabilities_of_inferences */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  Using existing markings of New or Old symbols with IDENTIFICATION
**  status, this function compiles a list of symbol types that have
**  that status. It then applies the list to patterns in Old and New,
**  marking all symbols that appear on the list as being type
**  CODE_SYMBOL and all other symbols as type DATA_SYMBOL.
**
**  This function also marks types LEFT_BRACKET, RIGHT_BRACKET, LEFT_XML_BRACKET
**  and RIGHT_XML_BRACKET and, with respect to the latter two types, it marks
**  symbols that are partners.
**
** CALLING SEQUENCE:
**
**  void assign_type_to_symbols()
**
** FORMAL ARGUMENTS:
**
**  Return value:   void
**
** IMPLICIT INPUTS:
**
**  Patterns in Old.
**
** IMPLICIT OUTPUTS:
**
**  Marking of symbols in Old and New with type CODE_SYMBOL or DATA_SYMBOL.
**
** SIDE EFFECTS:
**
**  NONE
**
*/

void assign_type_to_symbols()
{
    group *code_type_symbols = new group ;
    symbol *symbol1, *symbol2, *symbol3 ;
    sequence *pattern1 ;
    char c, c_successor, *symbol_name ;

    // Compile a list of symbol types which are of CODE_SYMBOL type.
    // It is assumed that these symbol types are the types of the
    // symbols in New or Old that have IDENTIFICATION status or they
    // are symbols that have names that start with '%' or '#'. The
    // symbols '<' and '>' are, slightly redundantly, given the type
    // LEFT_BRACKET and RIGHT_BRACKET respectively, provided that
    // recognise_brackets is true. Symbols like <name> and </name>
    // are given the type LEFT_XML_BRACKET and RIGHT_XML_BRACKET
    // provided that recognise_xml_brackets is true.

    list_for(pattern1, sequence, old_patterns)
    {
        list_for(symbol1, symbol, pattern1)
        {
            symbol_name = symbol1->get_name() ;
            c = *symbol_name ;
            if (symbol1->get_status() != IDENTIFICATION
                && c != '%' && c != '#') continue ;

            if (recognise_brackets)
            {
                if (strcmp(symbol_name, "<") == 0
                    || strcmp(symbol_name, ">") == 0) continue ;
            }

            if (recognise_xml_brackets)
            {
                if (c == '<') continue ;
            }

            // Check whether the type of symbol1 is already registered
            // in code_type_symbols. If not, add a copy of symbol1
            // to the list. If it is, continue.

            list_for(symbol2, symbol, code_type_symbols)
                if (symbol2->name_matches(symbol1)) goto L1 ;

            // No matching symbol type has been found. Add a copy
            // of symbol1 to code_type_symbols.

            symbol3 = symbol1->shallow_copy() ;
            code_type_symbols->add_child(symbol3) ;

            L1: ;
        }
    }

	// It is assumed that all New symbols are DATA_SYMBOLs so
	// there is no need to process the New pattern(s).

/*
    list_for(pattern1, sequence, new_patterns)
    {
        list_for(symbol1, symbol, pattern1)
        {
            symbol_name = symbol1->get_name() ;
            c = *symbol_name ;

            if (symbol1->get_status() == IDENTIFICATION) continue ;

            if (recognise_brackets)
            {
                if (strcmp(symbol_name, "<") == 0
                    || strcmp(symbol_name, ">") == 0) continue ;
            }

            if (recognise_xml_brackets)
            {
                if (c == '<') continue ;
            }

            // Check whether the type of symbol1 is already registered
            // in code_type_symbols. If not, add a copy of symbol1
            // to the list. If it is, continue.

            list_for(symbol2, symbol, code_type_symbols)
                if (symbol2->name_matches(symbol1)) goto L2 ;

            // No matching symbol type has been found. Add a copy
            // of symbol1 to code_type_symbols.

            symbol3 = symbol1->shallow_copy() ;
            code_type_symbols->add_child(symbol3) ;

            L2: ;
        }
    }
*/

    // Now traverse New and Old, marking symbols as type CODE_SYMBOL if
    // a copy appears in code_type_symbols, and marking them as type
    // DATA_SYMBOL if a copy does not appear on the list. The symbols
    // '<' and '>' are marked as LEFT_BRACKET and RIGHT_BRACKET
    // respectively, provided that 'recognise_brackets' is true.
    // Symbols whose names have the form <...> and </...> have the
    // type LEFT_XML_BRACKET and RIGHT_XML_BRACKET, provided that
    // 'recognise_xml_brackets' is true.

    list_for(pattern1, sequence, new_patterns)
    {
        list_for(symbol1, symbol, pattern1)
        {
            symbol_name = symbol1->get_name() ;

            if (recognise_brackets)
            {
                if (strcmp(symbol_name, "<") == 0)
                {
                    symbol1->set_type(LEFT_BRACKET) ;
                    continue ;
                }

                if (strcmp(symbol_name, ">") == 0)
                {
                    symbol1->set_type(RIGHT_BRACKET) ;
                    continue ;
                }
            }

            if (recognise_xml_brackets)
            {
                c = *symbol_name ;
                c_successor = *(symbol_name + 1) ;
                if (c == '<')
                {
                    if (c_successor == '/')
                    {
                        symbol1->set_type(RIGHT_XML_BRACKET) ;
                        continue ;
                    }
                    else
                    {
                        symbol1->set_type(LEFT_XML_BRACKET) ;
                        continue ;
                    }
                }
            }

            // Check whether a copy of symbol1 appears
            // in code_type_symbols. If it does, mark symbol1 as
            // the type CODE_SYMBOL, otherwise mark it as type
            // DATA_SYMBOL.

            if (code_type_symbols->contains_copy_of(symbol1))
                symbol1->set_type(CODE_SYMBOL) ;
            else symbol1->set_type(DATA_SYMBOL) ;
        }
    }

    list_for(pattern1, sequence, old_patterns)
    {
        list_for(symbol1, symbol, pattern1)
        {
            symbol_name = symbol1->get_name() ;

            if (recognise_brackets)
            {
                if (strcmp(symbol_name, "<") == 0)
                {
                    symbol1->set_type(LEFT_BRACKET) ;
                    continue ;
                }

                if (strcmp(symbol_name, ">") == 0)
                {
                    symbol1->set_type(RIGHT_BRACKET) ;
                    continue ;
                }
            }

            if (recognise_xml_brackets)
            {
                c = *symbol_name ;
                c_successor = *(symbol_name + 1) ;
                if (c == '<')
                {
                    if (c_successor == '/')
                    {
                        symbol1->set_type(RIGHT_XML_BRACKET) ;
                        continue ;
                    }
                    else
                    {
                        symbol1->set_type(LEFT_XML_BRACKET) ;
                        continue ;
                    }
                }
            }

            // Check whether a copy of symbol1 appears
            // in code_type_symbols. If it does, mark symbol1 as
            // the type CODE_SYMBOL, otherwise mark it as type
            // DATA_SYMBOL.

            if (code_type_symbols->contains_copy_of(symbol1))
                symbol1->set_type(CODE_SYMBOL) ;
            else symbol1->set_type(DATA_SYMBOL) ;
        }
    }

    // list_for(pattern1, sequence, new_patterns)
    //  pattern1->find_xml_bracket_pairs() ;

    // list_for(pattern1, sequence, old_patterns)
    //  pattern1->find_xml_bracket_pairs() ;

    // Delete code_type_symbols and all its children.

    delete code_type_symbols ;

} // assign_type_to_symbols

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For each pattern in Old, identifies IDENTIFICATION
**	symbols and CONTENTS symbols. Finds the coding cost of each
**	pattern in Old as described in sp70_od, %3 - total of the
**	bit_costs for the 'coding' symbols for each pattern.
**
**	The IDENTIFICATION symbols for any pattern are: the one
**	or more symbols at the start of the pattern which are needed to
**	identify the pattern uniquely amongst the patterns in the corpus plus
**	the termination symbol for the pattern (if any).
**
**	The status of symbols in New are either the default status of
**	CONTENTS, or they are explicitly marked in the input, or they
**	acquired their status in a previous iteration of 
**	set_up_and_parsing().
**
** CALLING SEQUENCE:
**
**	void assign_status_to_symbols()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void assign_status_to_symbols()
{
	sequence *pattern1, *matching_pattern ;
	symbol *symbol1, *symbol2 ;
	bool end_of_leading_code_sequence_found ;
	char *symbol_name1, *symbol_name2 ;
	
	list_for(pattern1, sequence, old_patterns)
	{
		end_of_leading_code_sequence_found = false ;
		list_for(symbol1, symbol, pattern1)
		{
			symbol1->set_status(IDENTIFICATION) ;
			if (is_unique(pattern1, symbol1, &matching_pattern))
			{
				end_of_leading_code_sequence_found = true ;
				break ;
			}
		}
		
		if (!end_of_leading_code_sequence_found)
		{
			if (audit_trail)
			{
				fprintf(output_file, "FIRST PATTERN: ") ;
				pattern1->print_ID() ;
				fprintf(output_file, "\n") ;
				fprintf(output_file, "SECOND PATTERN: ") ;
				matching_pattern->print_ID() ;
				fprintf(output_file, "\n\n") ;
			}
			abort_run("Non-unique pattern found in old_patterns") ;
		}
		
		// Now mark the remaining symbols in pattern1 as CONTENTS,
		// except the last symbol if its contents character string
		// begins with '#'. If the last symbol is '>', set its status
		// as IDENTIFICATION if the first symbol is '<'.
		
		while (symbol1 = (symbol *)pattern1->get_next_child())
		{
			if (pattern1->this_is_last_child(symbol1))
			{
				symbol_name1 = symbol1->get_name() ;
				if (*symbol_name1 == '#')
				{
					symbol1->set_status(IDENTIFICATION) ;
					break ;
				}

				if (strcmp(symbol_name1, ">") == 0)
				{
					symbol2 = (symbol *)pattern1->get_first_child() ;
					symbol_name2 = symbol2->get_name() ;
					if (strcmp(symbol_name2, "<") == 0)
					{
						symbol1->set_status(IDENTIFICATION) ;
						break ;
					} 
				}
			}
			symbol1->set_status(CONTENTS) ;
		}
	}
} // assign_status_to_symbols

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Using the frequencies and costs of symbol types recorded in 
**	set_of_symbols, this function assigns these values to
**	instances of symbols in the corpus. The costs of DATA_SYMBOLs
**	are multiplied by the cost_factor to ensure that they are treated
**	as if they were relatively large chunks of information.
**
** CALLING SEQUENCE:
**
**	void assign_symbol_frequencies_and_costs(group *set_of_symbols)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	set_of_symbols:		The alphabet of symbol types (each with values
**						for frequency and bit_cost) that
**						will be used in this function. This is derived
**						from the set of New and Old patterns.
**
** IMPLICIT INPUTS:
**
**	original_symbols_in_corpus
**
** IMPLICIT OUTPUTS:
**
**	frequency values of symbols in old_patterns
**
** SIDE EFFECTS:
**
**	NONE
*/

void assign_symbol_frequencies_and_costs(group *set_of_symbols)
{
	sequence *pattern ;
	symbol *symbol_type, *symbol1 ;
	
	/* Run through all symbols in the corpus and, for each one,
	look for the matching symbol type in set_of_symbols. When
	a match is found, give the symbol in the corpus the frequency
	of the matching symbol type in set_of_symbols and the costs of
	that symbol type. */

	list_for(pattern, sequence, new_patterns)
	{
		for (symbol1 = (symbol *)pattern->get_first_child();
		symbol1 != NIL; symbol1 = (symbol *)pattern->
			get_next_child() )
		{
			list_for(symbol_type, symbol, set_of_symbols)
			{
				if (strcmp(symbol_type->get_name(),
					symbol1->get_name()) != 0)
					continue ;
				symbol1->set_frequency(symbol_type->
					get_frequency()) ;

				symbol1->set_bit_cost(symbol_type->
					get_bit_cost()) ;
				break ;
			}
		}
	}

	list_for(pattern, sequence, old_patterns)
	{
		list_for(symbol1, symbol, pattern)
		{
			list_for(symbol_type, symbol, set_of_symbols)
			{
				if (strcmp(symbol_type->get_name(),
					symbol1->get_name()) != NIL)
					continue ;
				symbol1->set_frequency(symbol_type->
					get_frequency()) ;
				symbol1->set_bit_cost(symbol_type->
					get_bit_cost()) ;
				break ;
			}
		}
	}
	
} /* assign_symbol_frequencies_and_costs */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	This function finds the frequencies of symbol types by adding
**	the frequency of the pattern in which it appears to the
**	symbol type for each appearance of a symbol type in the corpus.
**
** CALLING SEQUENCE:
**
**	void find_symbol_frequencies(group *symbol_set, bool do_new, 
**		bool do_old)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	symbol_set:		The alphabet of symbols whose frequencies are to
**					to be set.
**	do_new:			Include new_patterns in the count.
**	do_old:			Include old_patterns in the count.
**
** IMPLICIT INPUTS:
**
**	NONE 
**
** IMPLICIT OUTPUTS:
**
**	frequencies of symbols recorded in symbol_set.
**
** SIDE EFFECTS:
**
**	NONE
*/

void find_symbol_frequencies(group *symbol_set, bool do_new, bool do_old)
{
	symbol *symbol1 ;
	
	/* Initialise frequencies of symbols in symbol_set. */
	
	symbol_set->initialise() ;
	while (symbol1 = (symbol *)symbol_set->get_next_child())
		symbol1->set_frequency(0) ;

	if (do_new) new_patterns->compile_frequencies(symbol_set) ;

	if (do_old) old_patterns->compile_frequencies(symbol_set) ;

} /* find_symbol_frequencies */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Write out New and Old patterns for reference.
**
** CALLING SEQUENCE:
**
**	void write_new_and_old()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	Data structures
**
** IMPLICIT OUTPUTS:
**
**	Output to output_file.
**
** SIDE EFFECTS:
**
**	NONE
*/

void write_new_and_old()
{
	/* Identify alignments in New and Old and write them out for reference,
	with the coding costs for patterns in Old. */
	
	fprintf(output_file, "NEW PATTERNS WITH DETAILS:\n\n") ;
	new_patterns->initialise() ;

	if (new_patterns->is_empty()) fprintf(output_file, "NONE\n\n") ;
	else new_patterns->write_patterns_with_details() ;
	fprintf(output_file, "%s%d%s", 
		"Size of New (symbols): ", number_of_symbols_in_new, "\n") ;
	fprintf(output_file, "%s%d%s", 
		"Size of New (patterns): ", number_of_patterns_in_new, "\n\n") ;
	
	fprintf(output_file, "OLD PATTERNS WITH DETAILS:\n\n") ;
	old_patterns->initialise() ;
	if (old_patterns->is_empty()) fprintf(output_file, "NONE\n\n") ;
	else old_patterns->write_patterns_with_details() ;
	fprintf(output_file, "%s%d%s", 
		"Size of Old (symbols): ", number_of_symbols_in_old, "\n") ;
	fprintf(output_file, "%s%d%s", 
		"Size of Old (patterns): ", number_of_patterns_in_old, "\n\n") ;
	
	fflush(output_file) ;
	
} /* write_new_and_old */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For diagnostic purposes.
**
** CALLING SEQUENCE:
**
**	void alignment_element::show_al_structure()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Printed output
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void alignment_element::show_al_structure()
{
	symbol *symbol1 = NIL ;

	if (el_obj != NIL)
	{
		fprintf(output_file, "Symbol ") ;
		symbol1 = (symbol *)el_obj ;

		fprintf(output_file, "%s", symbol1->get_name()) ;
	}
	else fprintf(output_file, "No symbol") ;
	fprintf(output_file, "%s%d%s%d%s",
		", same_column_above ", same_column_above,
		", same_column_below ", same_column_below,
		", original_pattern ") ;
	if (original_pattern != NIL)
		original_pattern->print_ID() ;
	else fprintf(output_file, "NIL") ;

	int orig_patt_int_pos ;

	if (symbol1 != NIL && original_pattern != NIL)
	{
		orig_patt_int_pos =
			original_pattern->get_int_pos_by_child(symbol1) ;
		fprintf(output_file, "%s%d%s",
			",\norig_patt_int_pos ", orig_patt_int_pos, "\n") ;

	}
	else fprintf(output_file, "%s%s%s",
		",\norig_patt_int_pos ", "NULL", "\n") ;
} // alignment_element::show_al_structure

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For a row within a multi-row symbol, this method finds the position
**	of the symbol (if any) in a given row in the original pattern
**	corresponding to that row.
**
** CALLING SEQUENCE:
**
**	int symbol::get_row_orig_int_pos(int row)
**
** FORMAL ARGUMENTS:
**
**	Return value:	Returns the position of the symbol (if any) in
**					the pattern corresponding to the row.
**
**	row:			The target row.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

int symbol::get_row_orig_int_pos(int row)
{
	if (sequence_depth == 1 ||
	row > sequence_depth || row < 0)
		abort_run("Invalid value for row in get_row_symbol") ;
	symbol *s1 = get_row_symbol(row) ;
	sequence *p1 = get_row_pattern(row) ;
	if (s1 != NIL) return(p1->get_int_pos_by_child(s1)) ;
	else return(NULL_VALUE) ;
} // get_row_orig_int_pos

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For a column within an alignment, puts in the correct values in
**	for same_column_above and same_column_below. The method also
**	sets symbol_is_a_hit as false if the column contains one symbol
**	and sets it as true if the column contains more than one symbol.
**
** CALLING SEQUENCE:
**
**	void symbol::set_symbol_matches()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Values for same_column_above and same_column_below in
**	column1.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void symbol::set_symbol_matches()
{
	int i, j, column_depth = this->get_sequence_depth(),
		symbol_counter = 0 ;
	alignment_element *al_el1, *al_el2 ;
	
	for (i = 0; i < column_depth; i++)
	{
		al_el1 = this->get_al_el(i) ;
		if (al_el1->get_el_obj() == NIL) continue ;
		symbol_counter++ ;
		for (j = i + 1; j < column_depth; j++)
		{
			al_el2 = this->get_al_el(j) ;
			if (al_el2->get_el_obj() != NIL) break ;
		}
		if (j >= column_depth) break ;
		al_el1->set_same_column_below(j) ;
		al_el2->set_same_column_above(i) ;
	}

	if (symbol_counter == 1) symbol_is_a_hit = false ;
	else if (symbol_counter > 1) symbol_is_a_hit = true ;
	else abort_run("Invalid value for symbol_count \
		in symbol::set_symbol_matches()") ;
} // symbol::set_symbol_matches

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	When 'this' symbol is a LEFT_XML_BRACKET, this method tests whether another
**	symbol is a matching RIGHT_XML_BRACKET. 
**
** CALLING SEQUENCE:
**
**	bool symbol::xml_left_bracket_matches_right_bracket(symbol *rb)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

bool symbol::xml_left_bracket_matches_right_bracket(symbol *rb)
{
	if (type != LEFT_XML_BRACKET) return(false) ;
	if (rb->get_type() != RIGHT_XML_BRACKET) return(false) ;

	char *l_name = name, *r_name = rb->get_name() ;
	if (strlen(r_name) != strlen(l_name) + 1) return(false) ; // The right bracket
		// is one character longer than the left bracket because it contains '/'.

	int i = 1, j = 2 ;
	char c = *(l_name + i) ;
	while (c != '>')
	{
		if (*(r_name + j) != c) return(false) ;
		i++ ;
		j++ ;
		c = *(l_name + i) ;
	}

	return(true) ;
} // symbol::xml_left_bracket_matches_right_bracket

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Constructor for class symbol.
**
** CALLING SEQUENCE:
**
**	symbol::symbol(char *nm, int al_depth, int hn_ID) : base_object()
**
** FORMAL ARGUMENTS:
**
**	Return value:
**
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

symbol::symbol(char *nm, int al_depth, int hn_ID) : base_object()
{
	name = NIL ;
	if (nm == NIL) {type = NULL_VALUE ;}
	else set_name(nm) ;
	bit_cost = 0 ;
	status = NULL_VALUE ;
	type = NULL_VALUE ;
	sequence_depth = al_depth ;
	h_node_ID = hn_ID ;
	symbol_is_a_hit = false ;
	partner_int_pos = NULL_VALUE ;
	if (al_depth == NULL_VALUE || al_depth == 1) al_el_array = NIL ;
	else
	{
		al_el_array = new alignment_element[al_depth] ;
		for (int i = 0; i < al_depth; i++)
			al_el_array[i].set_el_obj(NIL) ;
	}
} // symbol::symbol

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Makes a copy of a symbol. If the symbol contains 2 or more rows,
**	it contains a new al_el_array[] but all the pointers within this
**	array are to the original structures pointed to from the symbol
**	which is being copied. Those structures are not themselves copied.
**
** CALLING SEQUENCE:
**
**	symbol::symbol(symbol &s) : base_object(s)
**
** FORMAL ARGUMENTS:
**
**	Return value:	NONE
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

symbol::symbol(symbol &s) : base_object(s)
{
	// The value of el_obj for the copy points to the start of
	// a new array of al_els.
	
	int row1 ;
	
	name = NIL ;
	name = new char[strlen(s.get_name()) + 1] ;
	strcpy(name, s.get_name()) ;
	bit_cost = s.get_bit_cost() ;
	type = s.get_type() ;
	status = s.get_status() ;
	sequence_depth = s.get_sequence_depth() ;
	h_node_ID = s.get_h_node_ID() ;
	symbol_is_a_hit = s.symbol_is_a_hit ;
	partner_int_pos = NULL_VALUE ; // When copies are made, we need
		// to find partners again.
	
	if (sequence_depth != NULL_VALUE && sequence_depth != 1)
	{
		// Patterns and symbols with a depth of 1
		// do not have any al_el_array[].
		
		al_el_array = new alignment_element[sequence_depth] ;
		
		for (row1 = 0; row1 < sequence_depth; row1++)
			*(al_el_array + row1) = *((s.al_el_array) + row1) ;
	}
	else al_el_array = NIL ;
} // symbol::symbol(&s)

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks down the column of an alignment to see whether there
**	are any symbols in the column that have the status CONTENTS. If
**	there is one or more such symbols, the method returns true,
**	otherwise it returns false.
**
** CALLING SEQUENCE:
**
**	bool symbol::contains_contents_symbol()
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if the column contains a CONTENTS symbol,
**					false otherwise.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool symbol::contains_contents_symbol()
{
	symbol *symbol1 ;
	int i ;

	for (i = 0; i < sequence_depth; i++)
	{
		symbol1 = get_row_symbol(i) ;
		if (symbol1 != NIL)
		{
			if (symbol1->get_status() == CONTENTS) return(true) ;
		}
	}

	return(false) ;
	
} // symbol::contains_contents_symbol

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Makes a copy of 'this' but without any information about rows. It
**	assumes a depth of 1.
**
** CALLING SEQUENCE:
**
**	symbol *symbol::shallow_copy()
**
** FORMAL ARGUMENTS:
**
**	Return value:
**
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

symbol *symbol::shallow_copy()
{
	symbol *copy_symbol = new symbol(this->get_name(), 1, NULL_VALUE) ;

	copy_symbol->set_bit_cost(this->get_bit_cost()) ;
	copy_symbol->set_type(this->get_type()) ;
	copy_symbol->set_status(this->get_status()) ;
	copy_symbol->set_sequence_depth(1) ;
	copy_symbol->set_h_node_ID(this->get_h_node_ID()) ;
	copy_symbol->set_is_a_hit(this->is_a_hit()) ;
	copy_symbol->set_al_el_array(NIL) ;

	return(copy_symbol) ;

} // symbol::shallow_copy

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For diagnostic purposes.
**
** CALLING SEQUENCE:
**
**	void symbol::show_al_structure()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Printout of the name of the column
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void symbol::show_al_structure()
{
	int i ;
	
	fprintf(output_file, "%s%d%s",
		"sequence_depth = ", sequence_depth,
		", hit node ID = ") ;

	if (h_node_ID == 0) fprintf(output_file, "NIL") ;
	else if (h_node_ID == NULL_VALUE)
		fprintf(output_file, "NULL_VALUE") ;
	else fprintf(output_file, "%d",
		h_node_ID) ;

	fprintf(output_file, ", is_a_hit = ") ;

	if (symbol_is_a_hit)
		fprintf(output_file, "true") ;
	else fprintf(output_file, "false") ;

	fprintf(output_file, "\n") ;
	for (i = 0; i < sequence_depth; i++)
		al_el_array[i].show_al_structure() ;
	fprintf(output_file, "\n") ;
} // symbol::show_al_structure

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Extracts the children of the parameter tree object and
**	adds them to 'this' tree object.
**
** CALLING SEQUENCE:
**
**	void tree_object::merge_children(tree_object *obj2)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	obj2:			The tree object whose children are to
**					be extracted and added to 'this' tree
**					object.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	'This' tree object receives new children and obj2 loses
**	them.
**
** SIDE EFFECTS:
**
**	NONE
*/

void tree_object::merge_children(tree_object *obj2)
{
	base_object *child1 ;

	while (child1 = obj2->extract_first_child())
		this->add_child(child1) ;

} // tree_object::merge_children

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Recursive copy constructor for a tree object.
**
** CALLING SEQUENCE:
**
**	tree_object::tree_object(tree_object &tr_obj) : base_object(tr_obj)
**
** FORMAL ARGUMENTS:
**
**	Return value:	NONE
**
**	tr_obj: 	The object to be copied.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

tree_object::tree_object(tree_object &tr_obj) : base_object(tr_obj)
{
	base_object *child, *new_child ;
	
	first_element = NIL ;
	current_el_pos = NIL ;
	number_of_children = tr_obj.number_of_children ;
	
	// Now make a 'deep' copy of the structure of sequence columns and
	// sequence elements.
	
	tr_obj.initialise() ;
	while (child = tr_obj.get_next_child())
	{
		// The following statement makes a copy of al_col2 including
		// copies of the constituent al_els.
		
		new_child = child->clone() ;
		
		this->add_child(new_child) ;
	}
	
	this->mark_parent_and_int_positions_non_recursive() ; // take care!
		// This alters the value of the 'parent' field.
	
} // tree_object::tree_object(tree_object &)

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Arranges the symbols in a sequence or a group in alpha-numeric order of
**	their name fields. The method is not valid (and will fail) if
**	the sequence or group is not composed entirely of symbols. 
**
** CALLING SEQUENCE:
**
**	void tree_object::sort_by_name()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void tree_object::sort_by_name()
{
	int i, string_comparison, fe_symbol_array ;
	symbol *symbol_array[VERY_LARGE_SCRATCH_ARRAY_SIZE], *symbol1, *symbol2 ;
	const char *s1, *s2 ;
	base_object *obj1 ;
	
	// Transfer children from 'this' to symbol_array[] and release
	// from 'this'.
	
	this->initialise() ;
	i = 0 ;
	while (obj1 = this->get_next_child())
	{
		if (obj1->is_symbol() == false)
			abort_run("Invalid object in sort_by_name()") ;
		symbol1 = (symbol *)obj1 ;
		symbol_array[i++] = symbol1 ;
		if (i >= VERY_LARGE_SCRATCH_ARRAY_SIZE)
			abort_run("Array overflow in sort_by_name") ;
	}
	fe_symbol_array = i ;
	this->release_children() ;
	
	// Now transfer them back to this, inserting them in
	// alpha-numeric order.
	
	for (i = 0; i < fe_symbol_array; i++)
	{
		symbol1 = symbol_array[i] ;
		if (this->is_empty())
		{
			this->add_child(symbol1) ;
			continue ;
		}
		
		// There is at least one symbol in this ;
		
		this->initialise() ;
		while (symbol2 = (symbol *)this->get_next_child())
		{
			s1 = symbol1->get_name() ;
			s2 = symbol2->get_name() ;
			string_comparison = strcmp(s1, s2) ;
			if (string_comparison < 0)
			{
				this->precede(symbol1, symbol2) ;
				goto L1 ;
			}
		}
		
		// If this point is reached, this means that symbol1 comes
		// after the last symbol in this.
		
		this->add_child(symbol1) ;
		
		L1: ;
	}
} // tree_object::sort_by_name

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Adds a new object to the end of the message-receiving
**	object. If the new object is NIL, nothing happens.
**
**	Notice that the position of current_el_pos
**	is not changed by adding a new child to an object.
**
** CALLING SEQUENCE:
**
**	void tree_object::add_child(base_object *child)
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
**	child:			The new base_object to be added.
**
** IMPLICIT INPUTS:
**
**	Functions:		new_element
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE. Notice, in particular, that the position of current_el_pos
**	is not changed by adding a new child to an object.
*/

void tree_object::add_child(base_object *child)
{
	list_element *temp1, *temp2, *next ;
	
	if (child == NIL) return ;
	temp1 = new list_element(child) ;
	temp2 = first_element ;
	if (temp2 == NIL)
	{
		first_element = temp1 ;
		this->mark_parent_and_int_positions_non_recursive() ; // take 
		// care! This alters the value of the 'parent' field.
		return ;
	}
	while (next = temp2->get_next()) temp2 = next;
	temp2->set_next(temp1) ;
	this->mark_parent_and_int_positions_non_recursive() ; // take 
	// care! This alters the value of the 'parent' field.
} /* tree_object::add_child */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	This is like add_child but checks that the child to be added
**	does not duplicate any child already added (meaning the identical
**	object, not merely one with a matching name). If the proposed
**	addition is the same as a child already added, the proposed addition
**	is not added.
**
** CALLING SEQUENCE:
**
**	void tree_object::add_child_without_duplicates(base_object *child)
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
**	child:			The new base_object to be added.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE. Notice, in particular, that the position of current_el_pos
**	is not changed by adding a new child to an object.
*/

void tree_object::add_child_without_duplicates(base_object *child)
{
	list_element *temp1, *temp2, *next ;
	base_object *existing_child ;
	
	if (child == NIL) return ;
	temp2 = first_element ;
	if (temp2 == NIL)
	{
		temp1 = new list_element(child) ;
		first_element = temp1 ;
		this->mark_parent_and_int_positions_non_recursive() ; // take 
		// care! This alters the value of the 'parent' field.
		return ;
	}
	else
	{
		existing_child = temp2->get_el_obj() ;
		if (existing_child == child) return ;
	}

	while (next = temp2->get_next())
	{
		existing_child = next->get_el_obj() ;
		if (existing_child == child) return ;
		temp2 = next ;
	}

	temp1 = new list_element(child) ;
	temp2->set_next(temp1) ;
	this->mark_parent_and_int_positions_non_recursive() ; // take 
		// care! This alters the value of the 'parent' field.
} /* tree_object::add_child_without_duplicates */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Adds a new object to the beginning of 'this' object.
**	If the new object is NIL, nothing happens.
**
** CALLING SEQUENCE:
**
**	void tree_object::add_child_at_start(base_object *child)
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
**	child:			The new object to be added.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void tree_object::add_child_at_start(base_object *child)
{
	list_element *temp1, *temp2 ;
	
	if (child == NIL) 
		abort_run("NIL child for add_child_at_start") ;
	temp1 = new list_element(child) ;
	temp2 = first_element ;
	first_element = temp1 ;
	temp1->set_next(temp2) ;
	this->mark_parent_and_int_positions_non_recursive() ; // take 
	// care! This alters the value of the 'parent' field.
} /* tree_object::add_child_at_start */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Extracts (unlinks) an specific object from a composite object.
**
** CALLING SEQUENCE:
**
**	void tree_object::extract_child(base_object *child) ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	void.
**
**	child:			Node to be extracted.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	The object is removed from the parent structure but is not deleted
**	(i.e. it's heap storage is still allocated).
*/

void tree_object::extract_child(base_object *child)
{
	list_element *position, *posx;
	base_object *temp;
	
	if (child == NIL) abort_run("Invalid parameter to extract_child") ;
	position = NIL ;
	temp = this->get_next_child_by_el_pos(&position) ;
	
	if (temp == child)
	{
		first_element = position->get_next();
		delete position ;
		this->mark_parent_and_int_positions_non_recursive() ; // take 
		// care! This alters the value of the 'parent' field.
		return ;
	}
	
	while (true)
	{
		posx = position;
		temp = this->get_next_child_by_el_pos(&position) ;
		if (temp == NIL)
			abort_run("Child not found in extract_child") ;
		if (temp == child) break;
	}
	
	posx->set_next(position->get_next()) ;
	delete position ;
	this->mark_parent_and_int_positions_non_recursive() ; // take 
	// care! This alters the value of the 'parent' field.
} /* tree_object::extract_child */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	'Gets' the children of an object in succession, ie it fetches
**	their addresses without disturbing their positions in the object.
**	If the value of *position is NIL, the function
**	gets the first child and sets *position to the position of that
**	child. Otherwise it gets the child following *position and 
**	updates *position. It returns NIL when there are no more children
**	to be fetched.
**
** CALLING SEQUENCE:
**
**	base_object *tree_object::get_next_child()
**
** FORMAL ARGUMENTS:
**
**	Return value:	The child being fetched. NIL if there are no
**					more children to be fetched.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

base_object *tree_object::get_next_child()
{
	if (current_el_pos == NIL)
	{
		/* This is the signal to get the first child of 'this',
		if any. */
		
		current_el_pos = this->first_element;
		if (current_el_pos == NIL) return(NIL) ;
		else return((current_el_pos)->get_el_obj()) ;
	} 
	else
	{
		/* In this case up-date current_el_pos
		and get the corresponding child. */
		
		current_el_pos = (current_el_pos)->get_next() ;
		if (current_el_pos == NIL)
			return(NIL) ;
		else return((current_el_pos)->get_el_obj()) ;
	}
} /* tree_object::get_next_child */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	'Gets' the children of an object in succession, ie it fetches
**	their addresses without disturbing their positions in the object.
**	If the value of *position is NIL, the function
**	gets the first child and sets *position to the position of that
**	child. Otherwise it gets the child following *position and 
**	updates *position. It returns NIL when there are no more children
**	to be fetched.
**
** CALLING SEQUENCE:
**
**	base_object *tree_object::get_next_child_by_el_pos(list_element 
**					**position)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The child being fetched. NIL if there are no
**					more children to be fetched.
**
**	position:		The address of the address of the element
**					of the child which was last returned.
**
** IMPLICIT INPUTS:
**
**	position
**
** IMPLICIT OUTPUTS:
**
**	The value of *position in the function which calls this
**	function is up-dated.
**
** SIDE EFFECTS:
**
**	NONE
*/

base_object *tree_object::get_next_child_by_el_pos(list_element **position)
{
	if (*position == NIL)
	{
		/* This is the signal to get the first child of 'this',
		if any. */
		
		*position = this->first_element;
		if (*position == NIL) return(NIL) ;
		else return((*position)->get_el_obj()) ;
	} 
	else
	{
		/* In this case up-date *position and get the corresponding
		child. */
		
		*position = (*position)->get_next() ;
		if (*position == NIL)
			return(NIL) ;
		else return((*position)->get_el_obj()) ;
	}
} /* tree_object::get_next_child_by_el_pos */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	'Gets' the list elements of an object in (reverse) succession,
**	ie it fetches their addresses without disturbing their positions
**	in the object. If the value of position is NIL, the function
**	gets the last el_pos. Otherwise it gets the el_pos preceding *position.
**	It returns NIL when there are no more list_elements to be fetched.
**
** CALLING SEQUENCE:
**
**	list_element *tree_object::get_preceding_el_pos(list_element *position)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The list_element being fetched. NIL if there are no
**					more list_elements to be fetched.
**
**	position:		The reference position.
**
** IMPLICIT INPUTS:
**
**	position
**
** IMPLICIT OUTPUTS:
**
**	NONE.
**
** SIDE EFFECTS:
**
**	NONE
*/

list_element *tree_object::get_preceding_el_pos(list_element *position)
{
	list_element *pos1, *pos2 ;
	
	if (first_element == NIL)
		return(NIL) ;
	
	if (position == NIL)
	{
		// This is the signal to get the last list_element of 'this',
		// if any.
		
		return(get_last_el_pos()) ;
	}
	
	// position is not NIL.
	
	pos1 = NIL ;
	pos2 = first_element ;
	while (pos2 != position)
	{
		pos1 = pos2 ;
		pos2 = pos2->get_next() ;
	}
	
	return(pos1) ;
	
} // tree_object::get_preceding_el_pos

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	'Gets' the child of an object preceding a given child. If the
**	given child is the first child, or if there is only one child,
**	the function returns NIL. If the given child is NIL,
**	the function returns an error.
**
** CALLING SEQUENCE:
**
**	base_object *tree_object::get_preceding_child(base_object *child)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The child before child. NIL if there is
**					only one child in 'this'.
**
**	child:			The given child whose predecessor is to be found.
**
** IMPLICIT INPUTS:
**
**	position
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

base_object *tree_object::get_preceding_child(base_object *child)
{
	base_object *preceding_child = NIL, *child1 ;
	list_element *pos1 ;
	
	if (child == NIL) abort_run("NIL parameter for get_preceding_child") ;
	
	pos1 = NIL ;
	while (child1 = this->get_next_child_by_el_pos(&pos1))
	{
		if (child1 == child) return(preceding_child) ;
		preceding_child = child1 ;
	}
	
	abort_run("child not found in get_preceding_child") ;
	
	return(NIL) ;
} /* tree_object::get_preceding_child */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	'Gets' the children of an object in (reverse) succession, ie it fetches
**	their addresses without disturbing their positions in the object.
**	If the value of position is NIL, the function
**	gets the last child and sets *position to the position of that
**	child. Otherwise it gets the child preceding *position and 
**	updates *position. It returns NIL when there are no more children
**	to be fetched.
**
** CALLING SEQUENCE:
**
**	base_object *tree_object::
**		get_preceding_child_by_el_pos(list_element **position)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The child being fetched. NIL if there are no
**					more children to be fetched.
**
**	position:		The address of the address of the element
**					of the child which was last returned.
**
** IMPLICIT INPUTS:
**
**	position
**
** IMPLICIT OUTPUTS:
**
**	The value of *position in the function which calls this
**	function is up-dated.
**
** SIDE EFFECTS:
**
**		NONE
*/

base_object *tree_object::get_preceding_child_by_el_pos(list_element **position)
{
	base_object *child ;
	list_element *pos1, *pos2 ;
	
	if (first_element == NIL)
	{
		*position = NIL ;
		return(NIL) ;
	}
	
	if (*position == NIL)
	{
		/* This is the signal to get the last child of 'this',
		if any. */
		
		*position = this->get_last_el_pos() ;
		child = this->get_child_by_el_pos(position) ;
		return(child) ;
	}
	
	/* *position is not NIL. Up-date *position and get the
	corresponding child. */
	
	pos1 = NIL ;
	pos2 = first_element ;
	while (pos2 != *position)
	{
		pos1 = pos2 ;
		pos2 = pos2->get_next() ;
	}
	
	*position = pos1 ;
	if (pos1 == NIL) return(NIL) ;
	else return(pos1->get_el_obj()) ;
	
} /* tree_object::get_preceding_child_by_el_pos */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Tests whether a child object is the last child of 'this' parent.
**
** CALLING SEQUENCE:
**
**	bool tree_object::this_is_last_child(base_object *child)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if child is the last child, false otherwise.
**
**	child:			The child to be tested.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool tree_object::this_is_last_child(base_object *child)
{
	if (child == NIL) abort_run("NIL child in this_is_last_child") ;
	list_element *el_pos1 = this->get_el_pos_by_child(child) ;
	if (el_pos1->get_next() == NIL) return(true) ;
	else return(false) ;
} // tree_object::this_is_last_child

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Recursively marks the 'parent' and 'position' fields of a tree_object.
**	This function also sets the value of 'number_of_children'.
**
** CALLING SEQUENCE:
**
**	void tree_object::mark_parent_and_int_positions_recursive()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void tree_object::mark_parent_and_int_positions_recursive()
{
	list_element *pos1 ;
	int position = 0 ;
	tree_object *obj ;
	
	pos1 = first_element ;
	if (pos1 == NIL) return ;
	while (pos1 != NIL)
	{
		pos1->set_parent(this) ;
		pos1->set_position(position++) ;
		obj = (tree_object *)pos1->get_el_obj() ;
		obj->mark_parent_and_int_positions_recursive() ;
		pos1 = pos1->get_next() ;
	}
	number_of_children = position ;
} /* tree_object::mark_parent_and_int_positions_recursive */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Marks the 'parent' and 'position' fields of a tree_object. This is the 
**	non-recursive version of 
**	tree_object::mark_parent_and_int_positions_recursive().
**
**	Notice that this method does not disturb the value of
**	current_el_pos.
**
**	But take care! This method alters the value of the 'parent' field.
**
** CALLING SEQUENCE:
**
**	void tree_object::mark_parent_and_int_positions_non_recursive()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void tree_object::mark_parent_and_int_positions_non_recursive()
{
	list_element *pos1 ;
	int position = 0 ;
	
	pos1 = first_element ;
	if (pos1 == NIL) return ;
	while (pos1 != NIL)
	{
		pos1->set_parent(this) ;
		pos1->set_position(position++) ;
		pos1 = pos1->get_next() ;
	}
	number_of_children = position ;
} /* tree_object::mark_parent_and_int_positions_non_recursive */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	'Gets' the child of an object by its ordinal position, ie it fetches
**	its address without disturbing its position in the object, using
**	ordinal position to determine which is accessed.
**
** CALLING SEQUENCE:
**
**	base_object *tree_object::get_child_by_int_pos(int driving_patt_int_pos)
**
** FORMAL ARGUMENTS:
**
**	Return value:			The child being fetched. NIL if there is no
**							child at the given position.
**
**	driving_patt_int_pos:	The ordinal position (first position is 0).
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	The value of current_el_pos in 'this' is set to the
**	value corresponding to the child which is returned.
**
** SIDE EFFECTS:
**
**	NONE
*/

base_object *tree_object::get_child_by_int_pos(int driving_patt_int_pos)
{
	base_object  *child ;
	int counter = 0 ;
	
	this->initialise() ;
	child = this->get_next_child() ;
	if (child == NIL) return(NIL) ;
	while (counter++ < driving_patt_int_pos)
	{
		child = this->get_next_child() ;
		if (child == NIL) return(NIL) ;
	}
	return(child) ;
} /* tree_object ::get_child_by_int_pos */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Return a pointer (and element) to the last object within 'this'
**	object. If there are no children, the function returns
**	NIL.
**
** CALLING SEQUENCE:
**
**	base_object *tree_object::get_last_child() ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	Pointer to the last object within 'this' object.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	None, but see `position' above.
**
** SIDE EFFECTS:
**
**	NONE
*/

base_object *tree_object::get_last_child()
{
	list_element *position = first_element ;
	if (position == NIL) return(NIL) ;
	while (position != NIL)
	{
		current_el_pos = position ;
		position = position->get_next() ;
	}
	return(current_el_pos->get_el_obj()) ;
} /* tree_object::get_last_child */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Adds a new child to 'this' node so that it immediately
**	precedes another specified child within the parent object.
**	This is the complementary function to `follow'.
**
** CALLING SEQUENCE:
**
**	void tree_object::precede(base_object *child1, base_object *child2)
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
**	child1: 		The new object to be inserted before child2.
**	child2: 		The child that child1 is to be inserted
**					in front of.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void tree_object::precede(base_object *child1, base_object *child2)
{
	list_element *tempx, *temp1, *temp2 ;
	
	if (child1 == NIL || child2 == NIL || child1 == child2)
		abort_run("Invalid parameter(s) for precede") ;
	temp2 = first_element ;
	if (temp2 == NIL) abort_run("child2 not found in precede") ;
	
	if (temp2->get_el_obj() == child2)
	{
		temp1 = new list_element(child1) ;
		temp1->set_next(temp2) ;
		first_element = temp1 ;
		
		// Revise position values for list_elements which follow
		// the inserted child1.
		
		this->mark_parent_and_int_positions_non_recursive() ; // take
		// care! This alters the value of the 'parent' field.
		return ;
	}
	
	// child2 is not the first object in the list.
	
	tempx = temp2;
	temp2 = temp2->get_next() ;
	while (temp2 != NIL) 
	{
		if (temp2->get_el_obj() == child2)
		{
			temp1 = new list_element(child1) ;
			tempx->set_next(temp1) ;
			temp1->set_next(temp2) ;
			
			// Revise position values for list_elements which
			// follow the inserted child1.
			
			this->mark_parent_and_int_positions_non_recursive() ;
			// take care! This alters the value of
			// the 'parent' field.
			return;
		}
		tempx = temp2;
		temp2 = temp2->get_next() ;
	}
	
	abort_run("child2 not found in precede") ;
} /* tree_object::precede */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Adds a new child to 'this' node so that it immediately
**	follows another specified child within the parent object.
**	This is the complementary function to `precede'.
**
** CALLING SEQUENCE:
**
**	void tree_object::follow(base_object *child1, base_object *child2)
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
**	child1: 		child within the parent object that child2 
**					is to follow
**	child2:			Object to be inserted after child1.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void tree_object::follow(base_object *child1, base_object *child2)
{
	list_element *temp1, *temp2 ;
	
	if (child1 == NIL || child2 == NIL || child1 == child2) 
		abort_run("Invalid parameter(s) for follow") ;;
	temp1 = first_element ;
	while (temp1 != NIL) 
	{
		if (temp1->get_el_obj() == child1)
		{
			temp2 = new list_element(child2) ;
			temp2->set_next(temp1->get_next()) ;
			temp1->set_next(temp2) ;
			
			// Revise position values for list_elements which
			// follow the inserted child.
			
			this->mark_parent_and_int_positions_non_recursive() ;
			// take care! This alters the value of
			// the 'parent' field.
			return;
		}
		temp1 = temp1->get_next() ;
	}
	
	abort_run("child1 not found in follow") ;
} /* tree_object::follow */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Returns a stored integer representing the ordinal position of a child
**	within its parent (starting at 0). 
**
** CALLING SEQUENCE:
**
**	int tree_object::get_int_pos_by_child(base_object *child) ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	Position of child.
**
**	child:			The child whose position is to be found.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

int tree_object::get_int_pos_by_child(base_object *child)
{
	list_element *pos1 ;
	
	if (child == NIL) abort_run("NIL child in get_int_pos_by_child") ;
	pos1 = first_element ;
	while ((pos1) != NIL)
	{
		if ((pos1)->get_el_obj() == child) break;
		pos1 = (pos1)->get_next() ;
	}
	if ((pos1) == NIL)
		abort_run("No child found in get_int_pos_by_child") ;
	return((pos1)->get_position()) ;
} /* tree_object::get_int_pos_by_child */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Returns a stored integer representing the ordinal position of a child
**	within its parent (starting at 0). By contrast, with 
**	tree_object::get_int_pos_by_child(), this method obtains the
**	position by counting. It does not set the value in the
**	list element corresponding to the child.
**
** CALLING SEQUENCE:
**
**	int tree_object::find_int_pos_by_child(base_object *child) ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	Position of child.
**
**	child:			The child whose position is to be found.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

int tree_object::find_int_pos_by_child(base_object *child)
{
	list_element *pos1 ;
	int counter = 0 ;
	
	if (child == NIL) abort_run("NIL child in find_int_pos_by_child") ;
	pos1 = first_element ;
	while ((pos1) != NIL)
	{
		if ((pos1)->get_el_obj() == child) break;
		pos1 = (pos1)->get_next() ;
		counter++ ;
	}
	if ((pos1) == NIL)
		abort_run("No child found in get_int_pos_by_child") ;
	return(counter) ;
} /* tree_object::find_int_pos_by_child */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Unlinks the top-level children of 'this' object.  The storage
**	for the children is left allocated but the elements between the
**	'this' and the children are deleted.
**
** CALLING SEQUENCE:
**
**	void tree_object::release_children()
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	The tree_object storage for the children is still allocated.
*/

void tree_object::release_children()
{
	base_object *child ;
	list_element *position ;
	
	position = NIL ;
	child = this->get_next_child_by_el_pos(&position) ;
	while (child != NIL)
	{
		first_element = position->get_next() ;
		delete position ;
		position = NIL ;
		child = this->get_next_child_by_el_pos(&position) ;
	}
	number_of_children = 0 ;
	current_el_pos = NIL ;
} /* tree_object::release_children */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Completes the deletion of a tree_object or hit_node. This function is
**	called from ~hit_node() or ~tree_object().
**
** CALLING SEQUENCE:
**
**	void tree_object::finish_deletion()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void tree_object::finish_deletion()
{
	list_element *position, *temp ;
	base_object *child ;
	
	if (first_element == NIL) return ;
	
	position = NIL ;
	child = this->get_next_child_by_el_pos(&position) ;
	while (child != NIL)
	{
		delete child ;
		temp = position ;
		child = this->get_next_child_by_el_pos(&position) ;
		delete temp ;
	}
} // tree_object::finish_deletion.

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Traverses 'this' pattern, finds pairs of LEFT and RIGHT XML-style brackets
**	and marks each pair appropriately. This method takes account of the depth of
**	bracketting in case there is a recursive pair within the pattern. 
**
** CALLING SEQUENCE:
**
**	void sequence::find_xml_bracket_pairs()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

/* void sequence::find_xml_bracket_pairs()
{
	symbol *symbol1, *symbol_left ;
	int level = -1 ;

	symbol *left_xml_bracket[SMALL_SCRATCH_ARRAY_SIZE] ;

	list_for(symbol1, symbol, this)
	{
		if (symbol1->get_type() != LEFT_XML_BRACKET
			&& symbol1->get_type() != RIGHT_XML_BRACKET) continue ;
		// Keep record of symbol1 and its 'level' in nested bracketting.

		if (symbol1->get_type() == LEFT_XML_BRACKET)
		{
			plus_one(&level, SMALL_SCRATCH_ARRAY_SIZE,
				"Overflow in sequence::find_xml_bracket_pairs().") ;
			left_xml_bracket[level] = symbol1 ;
		}

		if (symbol1->get_type() == RIGHT_XML_BRACKET)
		{
			symbol_left = left_xml_bracket[level] ;

			if (symbol_left->xml_left_bracket_matches_right_bracket(symbol1) == false)
			{
				fprintf(output_file, "%s%d%s",
					"XML bracketting anomaly in pattern ID",
					ID,
					"\n\n") ;
				abort_run("Run aborted.") ;
			}

			symbol1->set_partner_xml_bracket(symbol_left) ;
			symbol_left->set_partner_xml_bracket(symbol1) ;

			level-- ;
		}
	}
} // sequence::find_xml_bracket_pairs */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks that, within 'this' pattern, each LEFT_XML_BRACKET has a
**	corresponding RIGHT_XML_BRACKET and that between each such pair of
**	brackets, there are either no brackets or balanced pairs of brackets.
**
** CALLING SEQUENCE:
**
**	bool sequence::check_xml_brackets()
**
** FORMAL ARGUMENTS:
**
**	Return value:		true if all brackets are paired, false otherwise.
**
**	set_of_patterns:	new_patterns or old_patterns.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

bool sequence::check_xml_brackets()
{
	symbol *symbol1, *symbol_left ;
	char *symbol_name1, *symbol_name_left ;
	bool return_value = true ;
	int i, j, level = -1 ;

	symbol *left_xml_bracket[MEDIUM_SCRATCH_ARRAY_SIZE] ;

	// Check that, if the pattern contains any XML-style brackets,
	// the first and last symbols in the pattern
	// are of type LEFT_XML_BRACKET and RIGHT_XML_BRACKET, respectively.

	if (contains_xml_brackets())
	{
		symbol1 = (symbol *)get_first_child() ;
		if (symbol1->get_type() != LEFT_XML_BRACKET)
		{
			if (audit_trail)
			{
				fprintf(output_file, "Pattern ") ;
				write_tree_object(PRINT_NO_FREQUENCIES) ;
			}
			abort_run("\nshould have XML-style bracket at the beginning.") ;
		}

		symbol1 = (symbol *)get_last_child() ;
		if (symbol1->get_type() != RIGHT_XML_BRACKET)
		{
			if (audit_trail)
			{
				fprintf(output_file, "Pattern ") ;
				write_tree_object(PRINT_NO_FREQUENCIES) ;
			}
			abort_run("\nshould have XML-style bracket at the end.") ;
		}
	}

	list_for(symbol1, symbol, this)
	{
		if (symbol1->get_type() != LEFT_XML_BRACKET
			&& symbol1->get_type() != RIGHT_XML_BRACKET) continue ;

		// Check that the name has a legal form.

		symbol_name1 = symbol1->get_name() ;

		if (*symbol_name1 != '<')
		{
			fprintf(output_file, "%s%s",
				symbol_name1,
				" is invalid XML-style bracket.\n\n") ;
			return_value = false ;
		}

		// Check the inner characters.

		if (symbol1->get_type() == LEFT_XML_BRACKET)
			i = 1 ;
		else i = 2 ;

		j = i + 1 ;

		while (*(symbol_name1 + j) != '\0')
		{
			if (*(symbol_name1 + i) == '<' || *(symbol_name1 + i) == '>'
				|| *(symbol_name1 + i) == '/')
			{
				fprintf(output_file, "%s%s",
					symbol_name1,
					" is invalid XML-style bracket.\n\n") ;
				return_value = false ;
			}
			i++ ;
			j++ ;
		}

		// Check the last character in the name.

		if (*(symbol_name1 + i) != '>')
		{
			fprintf(output_file, "%s%s",
				symbol_name1,
				" is invalid XML-style bracket.\n\n") ;
			return_value = false ;
		}

		// Keep record of symbol1 and its 'level' in nested bracketting.

		if (symbol1->get_type() == LEFT_XML_BRACKET)
		{
			plus_one(&level, MEDIUM_SCRATCH_ARRAY_SIZE,
				"Overflow in sequence::check_xml_brackets().") ;
			left_xml_bracket[level] = symbol1 ;
		}

		if (symbol1->get_type() == RIGHT_XML_BRACKET)
		{
			if (level < 0)
			{
				fprintf(output_file, "%s%d%s",
					"Unbalanced XML-style brackets in pattern ID",
					ID,
					"\n\n") ;
					break ;
			}

			symbol_left = left_xml_bracket[level] ;

			// Check that the 'inner' parts of the two symbol names match.

			i = 1;
			j = 2 ;
			symbol_name1 = symbol1->get_name() ;
			symbol_name_left = symbol_left->get_name() ;

			while (*(symbol_name1 + j) != '>')
			{
				if (*(symbol_name_left + i) != *(symbol_name1 + j))
				{
					fprintf(output_file, "%s%s%s%s",
						symbol_name1,
						" does not match ",
						symbol_name_left,
						"\n\n") ;
					return_value = false ;
					break ;
				}

				i++ ;
				j++ ;
			}

			level-- ;
		}
	}

	if (level != -1)
	{
		fprintf(output_file, "%s%d%s",
			"Unbalanced XML-style brackets in pattern ID",
			ID,
			"\n\n") ;
		return_value = false ;
	}

	return(return_value) ;

} // sequence::check_xml_brackets

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Called from find_containing_sequences.
**
** CALLING SEQUENCE:
**
**	bool sequence::contained_in(sequence *test_sequence)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if 'this' is contained in test_sequence,
**					false otherwise.
**
**	test_sequence:	The sequence being tested to see whether it
**					contains the given sequence or not.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool sequence::contained_in(sequence *test_sequence)
{
	if (test_sequence == NIL) return(false) ;

	sequence *built_from_driving = test_sequence->get_built_from_driving(),
		*built_from_target = test_sequence->get_built_from_target() ;

	if (this == built_from_driving) return(true) ;
	if (this == built_from_target) return(true) ;
	if (this->contained_in(built_from_driving)) return(true) ;
	if (this->contained_in(built_from_target)) return(true) ;
	return(false) ;

} // sequence::contained_in

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For each column in 'this' alignment, the method finds the values
**	of frequency and bit_cost for the column (from any symbol within
**	the column) and assigns these values to the column itself.
**
** CALLING SEQUENCE:
**
**	void sequence::correct_column_values()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Corrected values for frequency and bit_cost for each column in 'this'.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void sequence::correct_column_values()
{
	symbol *col1, *symbol1 ;
	int i ;

	list_for(col1, symbol, this)
	{
		for (i = 1; i < sequence_depth; i++)
		{
			symbol1 = col1->get_row_symbol(i) ;
			if (symbol1 != NIL) break ;
		}

		col1->set_frequency(symbol1->get_frequency()) ;
		col1->set_bit_cost(symbol1->get_bit_cost()) ;
	}
} // sequence::correct_column_values

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Tests whether or not all the symbols from current_new_pattern
**	are matched in the alignment.
**
** CALLING SEQUENCE:
**
**	bool sequence::all_new_CONTENTS_symbols_matched()
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if all the CONTENTS symbols in current_new_pattern
**					are matched, false otherwise.
**
** IMPLICIT INPUTS:
**
**	current_new_pattern
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool sequence::all_new_CONTENTS_symbols_matched()
{
	// Check that this is an alignment and not a flat pattern.

	if (sequence_depth < 2)
		abort_run("Invalid alignment in sequence::all_new_ID_symbols_matched()") ;

	symbol *col1, *symbol_new1, *symbol_new2 ;
	sequence *pattern_new = get_row_pattern(0) ;

	this->initialise() ;
	list_for(symbol_new1, symbol, pattern_new)
	{
		if (symbol_new1->get_status() != CONTENTS) continue ;
		while (col1 = (symbol *)this->get_next_child())
		{
			symbol_new2 = col1->get_row_symbol(0) ;
			if (symbol_new2 != NIL) break ;
		}

		if (col1 == NIL) // The alignment has been exhausted without
			// finding all the symbols in current_new_pattern.
				return(false) ;

		if (col1->is_a_hit() == false) return(false) ;
		if (symbol_new1 != symbol_new2) return(false) ;
	}

	return(true) ;
} // sequence::all_new_ID_symbols_matched

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks whether or not a given alignment matches a sub-alignment 
**	of al1. As with sequence::alignment_matches(), there is no need
**	for the rows to be in the same order in the candidate alignment
**	and in the sub-alignment.
**
** CALLING SEQUENCE:
**
**	bool sequence::matches_sub_alignment(sequence *al1, int al1_depth,
**		int target_alignment_depth, int target_alignment_length,
**		list_element *el_pos1, list_element *el_pos2)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if there is a positive match, false otherwise.
**
**	al1:			The alignment of which the sub-alignment is a part.
**	al1_depth:		The number of rows in al1.
**	target_alignment_depth:		The number of rows in the 
**								sub-alignment, excluding the old_row.
**	target_alignment_length:	The number of columns in the 
**								sub-alignment.
**	el_pos1:		The position of the first column of the
**					sub-alignment.
**	el_pos2:		The position of the last column of the
**					sub-alignment.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Print out of result.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool sequence::matches_sub_alignment(sequence *al1, int al1_depth,
	int target_alignment_depth, int target_alignment_length,
	list_element *el_pos1, list_element *el_pos2)
{
	symbol *al_col1, *al_col2, *symbol1, *symbol2 ;
	alignment_element *al_el1, *al_el2 ;
	sequence *orig_patt1, *orig_patt2 ;
	int row1, row2 ;
	bool match_found ;
	list_element *el_pos3 ;
	
	/* Compare the depths and lengths of the sequences. */
	
	if (sequence_depth != target_alignment_depth) return(false) ;
	
	if (number_of_children != target_alignment_length) return(false) ;

	// Check for a match in the basic sequence of columns.

	el_pos3 = el_pos1 ;
	for (symbol1 = (symbol *)this->get_first_child();symbol1 != NIL ;
		symbol1 = (symbol *)this->get_next_child())
	{
		symbol2 = (symbol *)el_pos3->get_el_obj() ;
		if (symbol1->name_matches(symbol2) == false) return(false) ;
		if (el_pos3 == el_pos2)
		{
			symbol1 = (symbol *)this->get_next_child() ;
			break ;
		}
		el_pos3 = el_pos3->get_next() ;
		if (el_pos3 == NIL) return(false) ;
	}

	if (symbol1 != NIL || el_pos3 != el_pos2)
		return(false) ;
	
	/* Now check if actual sequences match. Two sequences can
	be the same even though the rows are in a different order.
	However, if they are to be recognised as the same, for every
	row in one sequence, there must be a row in the other sequence
	which matches it exactly. */
	
	/* Initialise al1_rows[] which is used to avoid unnecessary
	re-checking of rows in al1 which have already been found to
	match a row in *this. */
	
	for (row2 = 0; row2 < al1_depth; row2++) al1_rows[row2] = false ;
	
	for (row1 = 0; row1 < sequence_depth; row1++)
	{
		el_pos3 = el_pos1 ;
		match_found = false ;
		for (row2 = 0; row2 < al1_depth; row2++)
		{
			if (al1_rows[row2] == true) continue ;
			
			al_col1 = (symbol *)this->get_first_child() ;
			this->set_current_el_pos(this->get_first_el_pos()) ;
			al_el1 = al_col1->get_al_el(row1) ;
			orig_patt1 = al_el1->get_original_pattern() ;
			
			al_col2 = (symbol *)el_pos3->get_el_obj() ;
			al_el2 = al_col2->get_al_el(row2) ;
			orig_patt2 = al_el2->get_original_pattern() ;
			
			if (orig_patt2 != orig_patt1) continue ;
			
			symbol1 = (symbol *)al_el1->get_el_obj() ;
			symbol2 = (symbol *)al_el2->get_el_obj() ;
			if (symbol1 != symbol2) continue ;

			// In the following loop, we can avoid certain checks
			// because we know that 'this' and the sub-alignment
			// are the same length.
			
			while (al_col1 = (symbol *)
				this->get_next_child())
			{
				symbol1 = al_col1->get_row_symbol(row1) ;
				el_pos3 = el_pos3->get_next() ;
				al_col2 = (symbol *)el_pos3->get_el_obj() ;
				symbol2 = al_col2->get_row_symbol(row2) ;
				if (symbol1 != symbol2) goto L1 ;
			}
			match_found = true ;
			al1_rows[row2] = true ;
			break ;
			L1: ;
		}
		if (!match_found) return(false) ;
	}

	if (audit_trail)
	{
		fprintf(output_file, "%s%d%s%d%s",
			"Sub-alignment of ID",
			al1->get_ID(),
			" has been found: ID",
			this->get_ID(),
			"\n\n") ;
	}

	return(true) ;
} // sequence::matches_sub_alignment

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks whether a newly-formed alignment matches any one formed
**	previously, on this cycle or previous cycles.
**
** CALLING SEQUENCE:
**
**	sequence *sequence::matches_earlier_alignment()
**
** FORMAL ARGUMENTS:
**
**	Return value:	The matching sequence if a match is found,
**					NIL otherwise.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

sequence *sequence::matches_earlier_alignment()
{
	sequence *al2 ;
	hit_node *node1 ;
	int row, leaf_index ;
	
	/* The program searches for matches amongst the sequences
	formed so far in the current cycle. */
	
	row = 0 ;
	while (node1 = get_leaf_nodes_in_order(&row, &al2, &leaf_index))
	{
		if (node1->get_compression_difference() == fail_score)
			continue ;
		if (al2 == NIL) continue ;
		
		// Because each new sequence is always matched 
		// against sequences formed in the current cycle, it is
		// assumed that there are no other duplicates amongst
		// the new sequences so it is safe to break out of the
		// loop now.
		
		if (al2->alignment_matches(this)) return(al2) ;
	}
	
	// Test whether sequences formed in all previous cycles are
	// to be matched as well.
	
	// if (!CHECK_ALL_ALIGNMENTS) return(NIL) ;
	
	// Check whether the newly-formed sequence matches anything
	// produced in previous cycles.

	list_for(al2, sequence, current_alignments_from_parsing)
	{
		// Because each new sequence is always matched 
		// against sequences formed in the earlier cycles, it is 
		// assumed that there are no other duplicates amongst the 
		// earlier sequences so it is safe to break out of
		// the loop if a match is found now.
		
		if (al2->alignment_matches(this)) return(al2) ;
	}
		
	return(NIL) ;
} // matches_earlier_alignment

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes out a sequence as a flat pattern without details and 
**	a flat pattern with details. If the pattern is an alignment,
**	the rows of the alignment are written out. The method
**	also writes out any code derived from the alignment or sequence.
**
** CALLING SEQUENCE:
**
**	void sequence::write_out_fully(char *header, sequence *code,
**		hit_node *h_node, int write_section_chars_length, 
**		int sections_per_page, bool flat_pattern_and_details)
**
** FORMAL ARGUMENTS:
**
**	Return value:					void
**
**	header:							The heading.
**	code:							The code derived from the 
**									alignment. It may be NIL.
**	hit_node:						The hit node in the case of
**									alignments derived from the
**									hit structure. It may be NIL.
**	write_section_chars_length:		The number of character columns
**									that can be written before the
**									alignment must be broken into
**									a new section.
**	sections_per_page:				The number of sections of the
**									alignment that can be written
**									before a new page must be started
**									(shown with a Latex command for
**									use in latex_file).
**									If the value is NULL_VALUE, no
**									page break will be inserted.
**	flat_pattern_and_details:		If true, the alignment is printed
**									as a flat pattern and
**									also with its details. Otherwise,
**									these things are not printed.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Output to output_file
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void sequence::write_out_fully(char *header, sequence *code,
	hit_node *h_node, int write_section_chars_length, 
	int sections_per_page, bool flat_pattern_and_details)
{
	if (header != NIL)
	{
		fprintf(output_file, "%s%s",
			header,
			" (") ;
		print_pattern_window_cycle() ;
		fprintf(output_file, ")\n\n") ;
	}

	this->print_ID() ;
	
	if (built_from_driving == NIL)
		fprintf(output_file, " : NIL") ;
	else fprintf(output_file, "%s%d",
		" : ID",
		built_from_driving->get_ID()) ;
	if (built_from_target == NIL)
		fprintf(output_file, " : NIL") ;
	else fprintf(output_file, "%s%d",
		" : ID",
		built_from_target->get_ID()) ;

	if (h_node != NIL)
	{
		fprintf(output_file, " : ") ;
		h_node->print_ID() ;
	}

	fprintf(output_file,
		"%s%1.2f%s%1.2f%s%1.2f%s%1.2f%s%1.12g",
		", NSC = ", this->get_new_symbols_cost(),
		", EC = ", this->get_encoding_cost(),
		", CR = ", this->get_compression_ratio(),
		", CD = ", this->get_compression_difference(),
		",\nAbsolute P = ", this->get_abs_P()) ;

	if (degree_of_matching == FULL_A)
		fprintf(output_file, ", FULL_A") ;
	else if (degree_of_matching == FULL_B)
		fprintf(output_file, ", FULL_B") ;
	else if (degree_of_matching == FULL_C)
		fprintf(output_file, ", FULL_C") ;
	else if (degree_of_matching == PARTIAL)
		fprintf(output_file, ", PARTIAL") ;
	else fprintf(output_file, "NULL_VALUE") ;

	fprintf(output_file, "\n\n") ;
	
	if (sequence_depth > 1)
		this->write_alignment(output_file,
			write_section_chars_length, sections_per_page, 
			alignment_format) ; 

	if (flat_pattern_and_details)
	{
		if (sequence_depth > 1)
		{
			fprintf(output_file, "Sequence ") ;
			this->print_ID() ;
			fprintf(output_file, " as flat pattern:\n\n") ;
		}

		this->write_tree_object(PRINT_NO_FREQUENCIES) ;
		this->write_with_details() ;
	}

	if (code != NIL)
	{
		fprintf(output_file, "%s%d%s%d%s",
			"The code derived from sequence ID",
			this->get_ID(),
			" is: ID",
			code->get_ID(),
			" ") ;

		code->write_tree_object(PRINT_NO_FREQUENCIES) ;
	}

} // sequence::write_out_fully

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  In write_alignment, fills in the non-hit symbols before a hit
**  symbol in row 0.
**
** CALLING SEQUENCE:
**
**  bool fill_in_row_0_non_hit_symbols(symbol *hit_symbol, int *fe_col)
**
** FORMAL ARGUMENTS:
**
**  Return value:   true if the write_al_chars_length has been breached.
**                  false otherwise.
**
**  hit_symbol:     The hit symbol which terminates the sequence of
**                  symbols. If hit_symbol == NIL then all the trailing
**                  symbols for each row are added to the alignment.
**  fe_col:         The first empty column in the given row of
**                  write_al_chars[][] when the last non-hit symbol
**                  before the hit_symbol has been filled in.
**
** IMPLICIT INPUTS:
**
**  arrays for write_alignment.
**
** IMPLICIT OUTPUTS:
**
**  arrays for write_alignment.
**
** SIDE EFFECTS:
**
**  NONE
**
*/

bool fill_in_row_0_non_hit_symbols(symbol *hit_symbol, int *fe_col)
{
    symbol *symbol1 ;
    bool char_column_not_OK ;
    int end_col, c_col1, c_col2, slot_width, c_row1 = 0, row_0_index ;
    char *c, *symbol_name ;

    c_col1 = row_array[0].fe_col_in_write_al_chars ;
    for (row_0_index = row_0_last_int_pos; row_0_index < fe_row_0_array; row_0_index++)
    {
        symbol1 = row_0_array[row_0_index].row_0_symbol ;
        if (symbol1 == hit_symbol) break ;
        
        if (use_abbreviations == true)
            symbol_name = find_symbol_abbreviation(symbol1->get_name()) ;
        else symbol_name = symbol1->get_name() ;
        
        // Find the next available space in this row of
        // write_al_chars[][] to take the symbol name plus
        // space for one space character.
        
        slot_width = strlen(symbol_name) + 1 ; // slot_width
        // is the enough space for the symbol name
        // plus 1 character for the space following the symbol.
        
        char_column_not_OK = true ;
        while (char_column_not_OK)
        {
            char_column_not_OK = false ;
            end_col = c_col1 + slot_width ;
            for (c_col2 = c_col1; c_col2 < end_col; c_col2++)
            {
                if (write_al_vacant_slots[0][c_col2] == 'x')
                {
                    char_column_not_OK = true ;
                    c_col1 = c_col2 + 1;
                    break ;
                }
            }
        }
        
        // Check that we are still within the bounds
        // of write_al_chars[][].
        
        if (c_col1 + slot_width >= write_al_chars_length)
        {
            // The writing out of this row of the alignment
            // will be truncated at this point, with "..." to
            // show that there is more to be written.
            
            *fe_col = c_col1 ;
            return(true) ;
        }
        else
        {
            // Fill in the symbol name.
            
            for (c = symbol_name; *c != '\0'; c++)
                write_al_chars[c_row1][c_col1++] = *c ;
        }
        
        // Add 1 to c_col to allow for the space character following
        // the symbol.
        
        c_col1++ ;
    }
    
    // Test to see whether the end of the last pattern in row 0 has been reached.

    if (row_0_index >= fe_row_0_array) row_array[0].pattern_is_finished = true ;
    
    // Update the value of row_0_last_int_pos (for row 0). At this
    // point, there is no need to update the value of
    // fe_col_in_write_al_chars because this will be updated
    // in write_alignment after the hit symbol has been written.
    
    row_0_last_int_pos = row_0_index + 1 ; // Add 1 to take us on to the next
        // set of non-hit symbols (if any).
    
    *fe_col = c_col1 ;
    return(false) ;
    
} // fill_in_row_0_non_hit_symbols

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  Writes out an alignment of symbols.
**
** CALLING SEQUENCE:
**
**  void sequence::write_alignment_horizontal(FILE *output_file,
**      int write_section_chars_length, int sections_per_page)
**
** FORMAL ARGUMENTS:
**
**  Return value:           void
**
**  output_file:                    The output_file or the latex_file.
**  write_section_chars_length: The number of character columns
**                                  that can be written before the
**                                  alignment must be broken into
**                                  a new section.
**  sections_per_page:              The number of sections of the
**                                  alignment that can be written
**                                  before a new page must be started
**                                  (shown with a Latex command for
**                                  use in latex_file).
**                                  If the value is NULL_VALUE, no
**                                  page break will be inserted.
**
** IMPLICIT INPUTS:
**
**  NONE
**
** IMPLICIT OUTPUTS:
**
**  NONE.
**
** SIDE EFFECTS:
**
**  NONE
*/

void sequence::write_alignment_horizontal(FILE *output_file, 
    int write_section_chars_length, int sections_per_page)
{
    symbol *al_col1, *symbol1 ;
    sequence *pattern1 ;
    int alignment_depth = this->get_sequence_depth(),
        number_of_children = this->get_number_of_children() ;
    int s_row1, s_row2, c_row1, c_col1, end_col,
        symbol_length, c_col_for_link, al_char_depth, max_c_col,
        temp_c_col, column_width, s_row_lowest_symbol, al_c_depth ;
    char *c, *symbol_name ;
    alignment_element *al_el1 ;
    list_element *temp_el_pos ;
    bool highest_symbol_has_been_found, write_char_limit_reached ;
    
    al_c_depth = (alignment_depth * 2) - 1 ; // Calculate the
    // depth of the alignment in terms of written out characters.
    
    if (al_c_depth > write_alignment_char_rows)
        abort_run("write_al_chars[][] has too few rows.") ;
    
    if (number_of_children >= max_alignment_length)
        abort_run("max_alignment_length has been exceeded") ;

    // Copy all the patterns in row 0 with associated details. The aim is
    // to create something like a single pattern containing all the patterns
    // in the row, together with those patterns from New up to and including
    // the current_new_pattern that have not formed any hits in this alignment.
    // This will facilitate later processing.

    int row_0_index = 0 ;
    sequence *temp_row_pattern = NIL, *row_0_pattern ;
    symbol *al_symbol1, *row_0_symbol1, *row_0_symbol2 ;
    list_for(al_symbol1, symbol, this)
    {
        row_0_symbol1 = al_symbol1->get_row_symbol(0) ;
        row_0_pattern = al_symbol1->get_row_pattern(0) ;

        if (row_0_pattern != temp_row_pattern) // Transition from one pattern
            // to the next.
        {
            if (row_0_pattern != NIL)
            {
                // Check that there is a non-NIL symbol at this point.

                if (row_0_symbol1 == NIL)
                    abort_run("NIL symbol in sequence::write_alignment_horizontal()") ;

                // We have found a new pattern. Now copy all of it into the array.

                list_for(row_0_symbol2, symbol, row_0_pattern)
                {
                    row_0_array[row_0_index].row_0_symbol = row_0_symbol2 ;
                    row_0_array[row_0_index].row_0_pattern = row_0_pattern ;

                    plus_one(&row_0_index, VERY_LARGE_SCRATCH_ARRAY_SIZE,
                        "row_0_array[] too small in sequence::write_alignment_vertical()") ;
                }

                temp_row_pattern = row_0_pattern ;
            }
        }
    }

    fe_row_0_array = row_0_index ;

    // Identify the patterns in New up to and including the current_new_pattern
    // that have not formed hits in this alignment. Add these patterns to
    // row_0_array[].

    sequence *new_patt ;
    int temp_row_0_index = fe_row_0_array ; // Position in row_0_array[] where
        // non-hit patterns from New will be added.
    list_for(new_patt, sequence, new_patterns)
    {
        for (row_0_index = 0; row_0_index < fe_row_0_array; row_0_index++)
        {
            if (new_patt == row_0_array[row_0_index].row_0_pattern)
                // new_patt forms at least one hit in this alignment.
                goto L2 ;
        }

        // new_patt has not formed a hit in this alignment. Add it to row_0_array[].

        list_for(row_0_symbol2, symbol, new_patt)
        {
            row_0_array[temp_row_0_index].row_0_symbol = row_0_symbol2 ;
            row_0_array[temp_row_0_index].row_0_pattern = new_patt ;

            plus_one(&temp_row_0_index, VERY_LARGE_SCRATCH_ARRAY_SIZE,
                "row_0_array[] too small in sequence::write_alignment_vertical()") ;
        }

        L2: if (new_patt == current_new_pattern) break ;
    }

    fe_row_0_array = temp_row_0_index ;

    // Now traverse the array marking symbols that are hit symbols.

    row_0_index = 0 ;
    for (row_0_index = 0; row_0_index < fe_row_0_array; row_0_index++)
    {
        row_0_symbol1 = row_0_array[row_0_index].row_0_symbol ;

        // Look for this symbol in row 0 of 'this'.

        list_for(al_symbol1, symbol, this)
        {
            row_0_symbol2 = al_symbol1->get_row_symbol(0) ;
            if (row_0_symbol1 == row_0_symbol2) break ;
        }
    }

    /* Clear write_al_chars[][], row_array[] and write_al_vacant_slots[][].
    The first of these arrays will contain the alignment as the
    characters to be written out. The second one keeps track of
    how far through the alignment (from left to right), the
    process has got. The third one records (with 'x'),
    the character slots that may not contain any non-hit symbol. */
    
    al_col1 = (symbol *)this->get_first_child() ;   
    if (al_col1 == NIL) abort_run("Anomaly in write_alignment") ;
    
    for (s_row1 = 0; s_row1 < alignment_depth; s_row1++)
    {
        row_array[s_row1].row_pattern =
            al_col1->get_row_pattern(s_row1) ;

        row_array[s_row1].patt_last_el_pos = NIL ;
        row_array[s_row1].fe_col_in_write_al_chars = 0 ;
        row_array[s_row1].pattern_is_finished = false ;
        c_row1 = s_row1 * 2 ;
        for (c_col1 = 0; c_col1 < write_al_chars_length; c_col1++)
        {
            write_al_vacant_slots[s_row1][c_col1] = ' ' ;
            write_al_chars[c_row1][c_col1] = ' ' ;
            write_al_chars[c_row1 + 1][c_col1] = ' ' ;
        }
    }

	row_0_last_int_pos = 0 ;

    // Now fill in characters in write_al_chars[][].
    
    max_c_col = 0 ;
    list_for(al_col1, symbol, this)
    {
        highest_symbol_has_been_found = false ;
        
        // Find the lowest row that contains a symbol.
        
        s_row_lowest_symbol = NULL_VALUE ;
        
        for (s_row2 = alignment_depth - 1; s_row2 >= 0; s_row2--)
        {
            symbol1 = al_col1->get_row_symbol(s_row2) ;
            if (symbol1 != NIL)
            {
                s_row_lowest_symbol = s_row2 ;
                break ;
            }
        }
        
        if (s_row_lowest_symbol == NULL_VALUE)
            abort_run("Anomaly in value of s_row_lowest_symbol") ;
        
        // Find the length of the symbol
        // type for this column.
        
        if (use_abbreviations == true)
            symbol_name = find_symbol_abbreviation(al_col1->get_name()) ;
        else symbol_name = al_col1->get_name() ;
        
        symbol_length = strlen(symbol_name) ;
        column_width = symbol_length + 1 ;
        
        // For those rows which contain a hit symbol,
        // fill in the non-hit symbols which precede the hit
        // symbol.
        
        write_char_limit_reached = false ;
        for (s_row1 = 0; s_row1 < alignment_depth; s_row1++)
        {
            symbol1 = al_col1->get_row_symbol(s_row1) ;
            if (symbol1 == NIL) continue ;

            // We have a hit symbol in this row.
            
            // Fill in the non-hit symbols from after
            // the previous hit for this row (or from the
            // beginning of the pattern) up to but
            // not including the current hit_symbol.
            // If necessary, update the value of max_c_col.

			if (s_row1 == 0)
			{
			if (fill_in_row_0_non_hit_symbols(symbol1, &temp_c_col))
				write_char_limit_reached = true ;
			}
			else
 			{
				if (fill_in_non_hit_symbols(s_row1, symbol1, &temp_c_col))
						write_char_limit_reached = true ;
			}
             
            if (temp_c_col > max_c_col) max_c_col = temp_c_col ;
        }
        
        if (write_char_limit_reached) goto L1 ;
        
        // Allowing for the width of the 'hit' column, check that
        // we are still within the bounds of write_al_chars[][].
        
        if (max_c_col + column_width >= write_al_chars_length)
        {
            write_char_limit_reached = true ;
            goto L1 ;
        }
        
        // Blank off those sections of write_al_vacant_slots[][]
        // which correspond to a NIL symbol in this column.

		end_col = max_c_col + column_width ;
		for (s_row1 = 0; s_row1 < alignment_depth; s_row1++)
		{
			symbol1 = al_col1->get_row_symbol(s_row1) ;
			if (symbol1 == NIL)
			{
				for (c_col1 = max_c_col;
				c_col1 < end_col; c_col1++)
					write_al_vacant_slots[s_row1]
					[c_col1] = 'x' ;
			}
		}
        
        // Now, for each row which contains a hit symbol,
        // fill in the hit symbol. If the hit symbol is the
        // last symbol in that row, mark the row as finished.
        
        // Calculate the column position of the 'link' symbol ('|')
        // which connects the hit symbols.
        
        c_col_for_link = max_c_col + (column_width - (symbol_length / 2) - 2) ;
        
        for (s_row1 = 0; s_row1 <= s_row_lowest_symbol; s_row1++)
        {
            c_row1 = s_row1 * 2 ;
            al_el1 = al_col1->get_al_el(s_row1) ;
            symbol1 = (symbol *)al_el1->get_el_obj() ;
            pattern1 = al_el1->get_original_pattern() ;
            
            if (symbol1 == NIL)
            {
                if (!highest_symbol_has_been_found) continue ;
                
                // Fill in a linking symbol for this c_row1
                // and the one below it. We can be sure
                // the lowest symbol has not yet been reached
                // because, if it had, symbol1 would not be
                // equal to NIL.
                
                write_al_chars[c_row1][c_col_for_link] = '|' ;
                write_al_chars[c_row1+1][c_col_for_link] = '|' ;
                continue ;
            }
            else highest_symbol_has_been_found = true ;
            
            // symbol1 is not NIL. Now fill in the name
            // of the 'hit' symbol for this row. But first,
            // mark this row as finished if the hit symbol
            // is the last symbol in the row.

            if (s_row1 == 0)
            {
                if (row_0_last_int_pos >= fe_row_0_array)
					row_array[s_row1].pattern_is_finished = true ;
            }
            else
            {
                temp_el_pos = row_array[s_row1].patt_last_el_pos ;
                pattern1->set_current_el_pos(temp_el_pos) ;
                if (pattern1->this_is_last_child(symbol1))
                    row_array[s_row1].pattern_is_finished = true ;
            }
            
            c_col1 = max_c_col ;
            for (c = symbol_name; *c != '\0'; c++)
                write_al_chars[c_row1][c_col1++] = *c ;
            
            if (s_row1 < s_row_lowest_symbol)
                write_al_chars[c_row1+1][c_col_for_link] = '|' ;
            
            // Update the value for the first empty column for
            // this row stored in write_al_char_cols[].
            
            row_array[s_row1].fe_col_in_write_al_chars =
                c_col1 + 1 ; // Add 1 to allow for the
            // space character following
            // the symbol.
        }
        
        max_c_col += column_width ;
    }
    
    // Finish off any remaining non-hit symbols in each row of
    // the alignment.
    
    write_char_limit_reached = false ;
    for (s_row1 = 0; s_row1 < alignment_depth; s_row1++)
    {
        if (s_row1 == 0)
        {
			if (fill_in_row_0_non_hit_symbols(symbol1, &temp_c_col))
				write_char_limit_reached = true ;
        }
        else
        {
            if (fill_in_non_hit_symbols(s_row1, NIL, &temp_c_col))
                write_char_limit_reached = true ;
        }
        if (temp_c_col > max_c_col) max_c_col = temp_c_col ;
    }
    
    L1: ;
    
    if (write_char_limit_reached)
    {
        /* The writing out of the alignment will be
        truncated at this point. The rows which are
        not finished are marked with "...". */
        
        for (s_row1 = 0; s_row1 < alignment_depth; s_row1++)
        {
            if (row_array[s_row1].pattern_is_finished)
                continue ;
            c_col1 = max_c_col ;
            c_row1 = s_row1 * 2 ;
            write_al_chars[c_row1][c_col1++] = '.' ;
            write_al_chars[c_row1][c_col1++] = '.' ;
            write_al_chars[c_row1][c_col1++] = '.' ;
        }
        max_c_col += 3 ;
    }
    
    /* Now write out the characters in write_al_chars[][]. */
    
    // int sections_written = 0 ;
    int first_section_column = 0, last_section_column,
        al_char_length = max_c_col - 1,section_counter = 0 ;
    
    // Calculate the number of character rows in the alignment.
    
    al_char_depth = (alignment_depth * 2) - 1 ;
    
    // Adjust the value of write_section_chars_length in the
    // light of the row numbers to be written at the beginning and
    // end of each row. For each number, one or two columns is needed
    // for the number itself, plus one blank column between the
    // number and the alignment.
    
    int last_row = alignment_depth - 1, cols_for_number ;
    if (last_row < 10) cols_for_number = 2 ;
    else if (last_row >=10 && last_row < 20) cols_for_number = 3 ;
    else abort_run("Too many rows in alignment") ;
    
    write_section_chars_length -= (cols_for_number * 2) ;
    if (write_section_chars_length < 10)
        abort_run("write_section_chars_length is too short") ;
    
    // Check whether what remains of the alignment will fit across
    // the breadth of the page. If so, set the value of
    // last_section_column to correspond to the last character column
    // of the alignment. If not, find the last blank column
    // in the current section of write_al_chars[] and set the
    // value of last_section_column to point to the column before
    // that blank column.
    
    while (true)
    {
        last_section_column = first_section_column +
            write_section_chars_length - 1 ;
        if (last_section_column > al_char_length)
            last_section_column = al_char_length - 1 ;
        else
        {
            for (c_col1 = first_section_column +
                write_section_chars_length - 1;
            c_col1 >= first_section_column; c_col1--)
            {
                // See whether the column in
                // write_al_chars[] corresponding
                // to c_col1 is blank. If so, break.
                
                for (c_row1 = 0; c_row1 < al_char_depth;
                c_row1++)
                {
                    if (write_al_chars[c_row1][c_col1]
                        != ' ')
                        break ;
                }
                
                if (c_row1 >= al_char_depth) break ;
            }
            
            if (c_col1 <= first_section_column)
                abort_run("Anomaly in last_section_column") ;
            
            last_section_column = c_col1 - 1 ;
        }
        
        // Write out the current section. Write the row number
        // at the beginning and end of each row.
        
        bool number_required = false ;
        
        s_row1 = 0 ;
        
        for (c_row1 = 0; c_row1 < al_char_depth; c_row1++)
        {
            if (number_required == false) number_required = true ;
            else number_required = false ;
            
            if (number_required)
            {
                if (cols_for_number == 2)
                    fprintf(output_file, "%1d", s_row1) ;
                else fprintf(output_file, "%2d", s_row1) ;
                fprintf(output_file, " ") ;
            }
            else
            {
                if (cols_for_number == 2)
                    fprintf(output_file, "  ") ;
                else fprintf(output_file, "   ") ;
            }
            
            for (c_col1 = first_section_column; c_col1 <=
                last_section_column; c_col1++)
            {
                fprintf(output_file, "%c",
                    write_al_chars[c_row1][c_col1]) ;
            }
            
            if (number_required)
            {
                fprintf(output_file, " ") ;
                if (cols_for_number == 2)
                    fprintf(output_file, "%d%s", s_row1,
                    "\n") ;
                else fprintf(output_file, "%2d%s", s_row1,
                    "\n") ;
                
                s_row1++ ;
            }
            else fprintf(output_file, "\n") ;
        }
        
        first_section_column = last_section_column + 2 ;
        if (first_section_column >= al_char_length)
        {
            fprintf(output_file, "\n") ;
            break ;
        }
        
        if (sections_per_page != NULL_VALUE)
        {
            section_counter++ ;
            if (section_counter >= sections_per_page)
            {
                fprintf(output_file, "\\newpage\n") ;
                section_counter = 0 ;
            }
            else fprintf(output_file, "\n\n") ;
        }
        else fprintf(output_file, "\n\n") ;
    }

    fflush(output_file) ;
} // sequence::write_alignment_horizontal

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  Writes out an alignment in 'vertical' format: there is a column
**  for each row in the alignment and symbols are written horizontally.
**
** CALLING SEQUENCE:
**
**  void sequence::write_alignment_vertical(FILE *output_file)
**
** FORMAL ARGUMENTS:
**
**  Return value:   void
**
**  output_file:    The file into which the alignment is to be written.
**
** IMPLICIT INPUTS:
**
**  NONE
**
** IMPLICIT OUTPUTS:
**
**  NONE
**
** SIDE EFFECTS:
**
**  NONE
**
*/

void sequence::write_alignment_vertical(FILE *output_file)
{
	display_row *d_row1, *d_row2, *d_row3, *d_row4,
		*d_row_list, *d_row_previous, *d_row_first_available ;
	symbol *col1, *symbol1, *symbol2 ;
	int i, j, k, fe_unmatched_new_symbols, fe_available_display_rows ;
	sequence *row_pattern1, *row_pattern2 ;
	bool is_a_hit ;

	// Create a sequence of display rows corresponding to the columns
	// in the alignment.

	d_row2 = NIL ;
	list_for(col1, symbol, this)
	{
		is_a_hit = col1->is_a_hit() ;
		d_row1 = new display_row(sequence_depth, is_a_hit) ; // This 
			// will be a row in the display, which corresponds to a
			// column in the alignment or an unmatched New symbol.
		if (d_row2 == NIL) // d_row1 is the first row in the list.
			d_row_list = d_row1 ;

		for (i = 0; i < sequence_depth; i++) // Step through the rows
			// of the current column.
		{
			symbol1 = col1->get_row_symbol(i) ;
			d_row1->set_row_symbol(symbol1, i) ;

			row_pattern1 = col1->get_row_pattern(i) ;
			d_row1->set_row_pattern(row_pattern1, i) ;
		}

		if (d_row2 != NIL) d_row2->set_next(d_row1) ;
		d_row2 = d_row1 ;
	}

	// Now add unmatched New symbols to d_row_list. Where one or more
	// non-hit rows are available in the right positions, the unmatched
	// New symbols may be added to these rows, one per row. Where there
	// are no such rows, create a new row for each unmatched New symbol
	// and insert it in the correct position in the list.

	list_element *el_pos1 ;

	row_pattern2 = NIL ;
	d_row_previous = NIL ;
	for (d_row1 = d_row_list; d_row1 != NIL; d_row1 = d_row1->get_next())
	{
		if (d_row1->is_a_hit() == false) continue ;
		symbol1 = d_row1->get_row_symbol(0) ;
		if (symbol1 == NIL) continue ;

		// We have a New symbol that is part of a hit.

		row_pattern1 = d_row1->get_row_pattern(0) ;
		if (row_pattern1 == NIL)
			abort_run("Anomalous value for row_pattern1 in \
				sequence::write_alignment_vertical().") ;

		i = 0 ; // For indexing unmatched_new_symbols[].

		if (row_pattern1 != row_pattern2)
		{
			// We have started another New pattern. Before resetting
			// el_pos1 to NIL, we should list any trailing unmatched
			// New symbols in row_pattern2 if it is not NIL.

			if (row_pattern2 != NIL)
			{
				list_for_el_pos(symbol2, symbol, row_pattern2, el_pos1)
				{
					unmatched_new_symbols[i] = symbol2 ;
					plus_one(&i, MEDIUM_SCRATCH_ARRAY_SIZE,
						"Too many unmatched New symbols in \
							sequence::write_alignment_vertical().") ;
				}
			}

			el_pos1 = NIL ;
		}

		// Now make a list of unmatched symbols in row_pattern1
		// up to but not including symbol1.

		list_for_el_pos(symbol2, symbol, row_pattern1, el_pos1)
		{
			if (symbol2 == symbol1) break ;

			unmatched_new_symbols[i] = symbol2 ;
			plus_one(&i, MEDIUM_SCRATCH_ARRAY_SIZE,
				"Too many unmatched New symbols in \
					sequence::write_alignment_vertical().") ;
		}

		fe_unmatched_new_symbols = i ;

		// Make a list of the display rows that are available to
		// receive unmatched New symbols, one symbol per pattern.

		if (fe_unmatched_new_symbols > 0)
		{
			if (d_row_previous == NIL)
				d_row_first_available = d_row_list ;
			else d_row_first_available = d_row_previous->get_next() ;
			i = 0 ;
			for (d_row2 = d_row_first_available; d_row2 != NIL;
				d_row2 = d_row2->get_next())
			{
				if (d_row2 == d_row1) break ;
				if (d_row2->is_a_hit()) continue ;
				symbol2 = d_row2->get_row_symbol(0) ;
				if (symbol2 != NIL) continue ;

				// We have found a non-hit row with a vacant slot for
				// a New symbol.

				available_display_rows[i] = d_row2 ;
				plus_one(&i, MEDIUM_SCRATCH_ARRAY_SIZE,
					"Too many display rows in \
						sequence::write_alignment_vertical().") ;
			}

			fe_available_display_rows = i ;

			// Now assign unmatched New symbols to available display
			// rows.

			j = 0 ;
			for (i = 0; i < fe_unmatched_new_symbols; i++)
			{
				if (j >= fe_available_display_rows) break ;
				symbol2 = unmatched_new_symbols[i] ;
				d_row2 = available_display_rows[j++] ;
				d_row2->set_row_symbol(symbol2, 0) ;
			}

			// For any remaining unmatched New symbols, create new
			// display rows, one per symbol, and insert them in
			// the list of display rows just in front of d_row1.
			// First, find the display row immediately before d_row1.

			if (d_row1 == d_row_list) d_row3 = NIL ;
			else
			{
				for (d_row3 = d_row_list; d_row3 != NIL;
					d_row3 = d_row3->get_next())
						if (d_row3->get_next() == d_row1) break ;
			}

			for (i = i; i < fe_unmatched_new_symbols; i++)
			{
				symbol2 = unmatched_new_symbols[i] ;
				d_row4 = new display_row(sequence_depth, false) ;
				d_row4->set_row_symbol(symbol2, 0) ;
				d_row4->set_row_pattern(row_pattern1, 0) ;
				if (d_row3 == NIL) d_row_list = d_row4 ;
				else d_row3->set_next(d_row4) ;
				d_row3 = d_row4 ;
			}
			d_row3->set_next(d_row1) ;
		}

		row_pattern2 = row_pattern1 ;
		d_row_previous = d_row1 ;
	}

	// Now deal with any trailing unmatched New symbols in the last
	// New pattern that was processed.

	// First, list any trailing unmatched New symbols in that pattern.

	i = 0 ;
	list_for_el_pos(symbol2, symbol, row_pattern1, el_pos1)
	{
		unmatched_new_symbols[i] = symbol2 ;
		plus_one(&i, MEDIUM_SCRATCH_ARRAY_SIZE,
			"Too many unmatched New symbols in \
				sequence::write_alignment_vertical().") ;
	}

	fe_unmatched_new_symbols = i ;

	if (fe_unmatched_new_symbols > 0)
	{
		i = 0 ;
		for (d_row2 = d_row_previous->get_next(); d_row2 != NIL;
			d_row2 = d_row2->get_next())
		{
			if (d_row2->is_a_hit()) continue ;
			symbol2 = d_row2->get_row_symbol(0) ;
			if (symbol2 != NIL) continue ;

			// We have found a non-hit row with a vacant slot for
			// a New symbol.

			available_display_rows[i] = d_row2 ;
			plus_one(&i, MEDIUM_SCRATCH_ARRAY_SIZE,
				"Too many display rows in \
					sequence::write_alignment_vertical().") ;
		}

		fe_available_display_rows = i ;

		// Now assign unmatched New symbols to available display
		// rows.

		j = 0 ;
		for (i = 0; i < fe_unmatched_new_symbols; i++)
		{
			if (j >= fe_available_display_rows) break ;
			symbol2 = unmatched_new_symbols[i] ;
			d_row2 = available_display_rows[j++] ;
			d_row2->set_row_symbol(symbol2, 0) ;
		}

		// For any remaining unmatched New symbols, create new
		// display rows, one per symbol, and insert them in
		// the list of display rows at the end of the list.
		// First, find the end of the list.

		for (d_row3 = d_row_list; d_row3 != NIL;
			d_row3 = d_row3->get_next())
				if (d_row3->get_next() == NIL) break ;

		for (i = i; i < fe_unmatched_new_symbols; i++)
		{
			symbol2 = unmatched_new_symbols[i] ;
			d_row4 = new display_row(sequence_depth, false) ;
			d_row4->set_row_symbol(symbol2, 0) ;
			d_row4->set_row_pattern(row_pattern1, 0) ;
			d_row3->set_next(d_row4) ;
			d_row3 = d_row4 ;
		}
	}

	// For each display column, find the length of the longest symbol
	// name in the column. At the same time, find the number of display
	// rows.

	int name_length, number_of_d_rows = 0 ;

	if (sequence_depth >= MEDIUM_SCRATCH_ARRAY_SIZE)
		abort_run("sequence_depth too large in \
			sequence::write_alignment_vertical().") ;

	for (i = 0; i < sequence_depth; i++)
		name_length_array[i] = 0 ;

	for (d_row1 = d_row_list; d_row1 != NIL; d_row1 = d_row1->get_next())
	{
		number_of_d_rows++ ;
		for (i = 0; i < sequence_depth; i++)
		{
			symbol1 = d_row1->get_row_symbol(i) ;
			if (symbol1 == NIL) continue ;
			name_length = strlen(symbol1->get_name()) ;
			if (name_length > name_length_array[i])
				name_length_array[i] = name_length ;
		}
	}

	if (number_of_d_rows >= LARGE_SCRATCH_ARRAY_SIZE)
		abort_run("number_of_d_rows too large in \
			sequence::write_alignment_vertical().") ;

	// Now convert each display row of d_row_list into a row of characters
	// in the 2D array d_chars_array[][].

	int d_row_chars_length = 0, d_col_width, d_col_first_char_pos ;

	d_col_first_char_pos = 0 ;
	for (i = 0; i < sequence_depth; i++)
	{
		d_col_first_chars[i] = d_col_first_char_pos ;
		d_col_width = name_length_array[i] ;

		// There are 3 character spaces between one column and the next.
		// Add this value to the width of every column except the last.

		if (i < sequence_depth - 1) d_col_width += 3 ;
		d_row_chars_length += d_col_width ;
		d_col_first_char_pos += d_col_width ;
	}

	if (d_row_chars_length >= LARGE_SCRATCH_ARRAY_SIZE)
		abort_run("d_row_chars_length too large in \
			sequence::write_alignment_vertical().") ;

	// Fill 2D character array with space characters.

	for(i = 0; i < number_of_d_rows; i++)
	{
		for (j = 0; j < d_row_chars_length; j++)
			d_chars_array[i][j] = ' ' ;
	}

	// Now fill in non-space characters in d_chars_array[][].

	int d_c_row = 0, d_c_col, start_dash, end_dash ;
	char *symbol_name, c ;

	for (d_row1 = d_row_list; d_row1 != NIL; d_row1 = d_row1->get_next())
	{
		for (i = 0; i < sequence_depth; i++) // Step through the rows
			// of the current column.
		{
			symbol1 = d_row1->get_row_symbol(i) ;
			if (symbol1 == NIL) continue ;
			symbol_name = symbol1->get_name() ;

			// Copy characters of symbol_name into appropriate column
			// of d_chars_array[][] in d_c_row.

			d_c_col = d_col_first_chars[i] ;
			c = *symbol_name ;
			while (c != '\0')
			{
				d_chars_array[d_c_row][d_c_col] = c ;
				d_c_col++ ;
				c = *(++symbol_name) ;
			}
		}

		if (d_row1->is_a_hit() == false)
		{
			d_c_row++ ;
			continue ;
		}
		
		// This row represents a hit. Fill in dashes between
		// matching symbols.

		start_dash = end_dash = NULL_VALUE ;
		for (i = 0; i < d_row_chars_length; i++)
		{
			if (start_dash == NULL_VALUE)
			{
				// Look for the next position, if any, where
				// a non-space character is followed by a space
				// character.

				for (j = i; j < d_row_chars_length; j++)
				{
					if (d_chars_array[d_c_row][j] != ' '
						&& d_chars_array[d_c_row][j + 1] == ' ')
					{
						start_dash = j + 2 ; // There is one blank
							// character between the end of a symbol
							// and the start of the dashes.
						break ;
					}
				}
			}

			if (start_dash != NULL_VALUE)
			{
				// Look for the next position, if any, where
				// a space character is followed by a non-space
				// character.

				for (k = start_dash; k < d_row_chars_length - 1; k++)
				{
					if (d_chars_array[d_c_row][k] == ' '
						&& d_chars_array[d_c_row][k + 1] != ' ')
					{
						end_dash = k - 1 ; // There is one blank
							// character between the end of the dashes
							// and the start of the next symbol.
						break ;
					}
				}
			}

			if (start_dash != NULL_VALUE && end_dash != NULL_VALUE)
			{
				// Fill in the dashes.

				for (j = start_dash; j <= end_dash; j++)
					d_chars_array[d_c_row][j] = '-' ;

				i = end_dash ;
			}
			else break ;

			start_dash = end_dash = NULL_VALUE ;	
		}

		d_c_row++ ;
	}

	// Now write out d_chars_array[][] but first,
	// write out column numbers.

	for (i = 0; i < d_row_chars_length; i++)
		column_numbers[i] = ' ' ;

	for (i = 0; i < sequence_depth; i++)
	{
		sprintf(col_number, "%d", i) ;
		d_c_col = d_col_first_chars[i] ;
		c = *col_number ;
		k = 0 ;
		while (c != '\0')
		{
			column_numbers[d_c_col] = c ;
			d_c_col++ ;
			c = col_number[++k] ;
		}
	}

	for (d_c_col = 0; d_c_col < d_row_chars_length; d_c_col++)
	{
		fprintf(output_file, "%c",
			column_numbers[d_c_col]) ;
	}

	fprintf(output_file, "\n\n") ;

	for (d_c_row = 0; d_c_row < number_of_d_rows; d_c_row++)
	{
		for (d_c_col = 0; d_c_col < d_row_chars_length; d_c_col++)
		{
			fprintf(output_file, "%c",
				d_chars_array[d_c_row][d_c_col]) ;
		}

		fprintf(output_file, "\n") ;
	}

	// Now output column numbers again.

	fprintf(output_file, "\n") ;

	for (d_c_col = 0; d_c_col < d_row_chars_length; d_c_col++)
	{
		fprintf(output_file, "%c",
			column_numbers[d_c_col]) ;
	}

	fprintf(output_file, "\n\n") ;

	// Delete dynamically allocated data structures.

	for (d_row1 = d_row_list; d_row1 != NIL; d_row1 = d_row2)
	{
		d_row2 = d_row1->get_next() ;
		delete d_row1 ;
	}

} // sequence::write_alignment_vertical

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes out the symbols in a pattern with details of
**	the type, status and other information about symbols.
**
** CALLING SEQUENCE:
**
**	void sequence::write_with_details()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Output of the pattern with details of symbols
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void sequence::write_with_details()
{
	symbol *symbol1 ;
	int counter = 0, int_pos, tp ;
	char type[SMALL_SCRATCH_ARRAY_SIZE], 
		status[SMALL_SCRATCH_ARRAY_SIZE] ;

	list_for(symbol1, symbol, this)
	{
		counter++ ;
		if (counter >= 3)
		{
			fprintf(output_file, "\n") ;
			counter = 1 ;
		}

		tp = symbol1->get_type() ;

		if (tp == CODE_SYMBOL) strcpy(type, "CODE_SYMBOL") ;
		else strcpy(type, "DATA_SYMBOL") ;
			
		if (recognise_brackets)
		{
			if (tp == LEFT_BRACKET) strcpy(type, "LB") ;
			else if (tp == RIGHT_BRACKET) strcpy(type, "RB") ;
		}
		else if (recognise_xml_brackets)
		{
			if (tp == LEFT_XML_BRACKET) strcpy(type, "LXB") ;
			else if (tp == RIGHT_XML_BRACKET) strcpy(type, "RXB") ;
		}

		if (symbol1->get_status() == IDENTIFICATION)
			strcpy(status, "ID") ;
		else if (symbol1->get_status() == CONTENTS)
			strcpy(status, "CONTENTS") ;
		else strcpy(status, "BM") ;

				
		int_pos = this->get_int_pos_by_child(symbol1) ;
			
		fprintf(output_file, "%s%s%d%s%1.2f%s%s%s%s%s%d",
			symbol1->get_name(),
			" (", int_pos, ", ",
			symbol1->get_bit_cost(),
			", ", type,
			", ", status,
			", ", symbol1->get_partner_int_pos()) ;
				
		if (this->this_is_last_child(symbol1))
			fprintf(output_file, ").\n\n") ;
		else fprintf(output_file, "), ") ;
	}
} // sequence::write_with_details

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Tests whether the CONTENTS symbols in one sequence of symbols
**	match the CONTENTS symbols in another sequence.
**
** CALLING SEQUENCE:
**
**	bool sequence::contents_symbols_match(sequence *pattern1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	If the CONTENTS symbols match, true, otherwise false.
**
**	pattern1:		The sequence to be matched with 'this'.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool sequence::contents_symbols_match(sequence *pattern1)
{
	symbol *symbol1, *symbol2 ;
	
	symbol2 = (symbol *)this->get_first_child() ;
	for (symbol1 = (symbol *)pattern1->get_first_child();
	symbol1 != NIL; symbol1 = (symbol *)pattern1->get_next_child())
	{
		if (symbol1->get_status() != CONTENTS) continue ;
		if (symbol2 == NIL) return(false) ;
		
		while (symbol2->get_status() != CONTENTS)
		{
			symbol2 = (symbol *)this->get_next_child() ;
			if (symbol2 == NIL) return(false) ;
		}
		
		if (symbol1->name_matches(symbol2))
		{
			symbol2 = (symbol *)this->get_next_child() ;
			continue ;
		}
		else return(false) ;
	}
	
	// pattern1 has been completely stepped through. Check to see
	// whether or not there are any CONTENTS symbols left in 'this'.
	
	if (symbol2 != NIL)
	{
		if (symbol2->get_status() == CONTENTS) return(false) ;
		while (symbol2 = (symbol *)this->get_next_child())
			if (symbol2->get_status() == CONTENTS) return(false) ;
	}
	
	return(true) ;
} // sequence::contents_symbols_match

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks 'this' pattern against pre-existing patterns 
**	(in old_patterns) to see whether there is one
**	that has all and only the same CONTENTS symbols in the
**	same order. If there is, that sequence is returned. 
**	Otherwise, the method returns NIL.
**
** CALLING SEQUENCE:
**
**	sequence *sequence::check_patterns()
**
** FORMAL ARGUMENTS:
**
**	Return value:	A pre-existing pattern with all and only the
**					same CONTENTS symbols, or NIL if no such pattern
**					exists.
**
** IMPLICIT INPUTS:
**
**	patterns in old_patterns.
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

sequence *sequence::check_patterns()
{
	sequence *pattern1 ;
	list_element *el_pos1 = NIL ;
	bool pattern_found = false ;

	for (pattern1 = (sequence *)old_patterns->
		get_next_child_by_el_pos(&el_pos1);
		pattern1 != NIL ; pattern1 = (sequence *)old_patterns->
			get_next_child_by_el_pos(&el_pos1))
	{
		if (this->contents_symbols_match(pattern1))
		{
			pattern_found = true ;
			break ;
		}
	}

	if (pattern_found)
	{
		if (audit_trail)
		{
			fprintf(output_file, "Contents-symbol match found between ") ;
			this->wto(PRINT_NO_FREQUENCIES) ;
			fprintf(output_file, "%s%d%s", 
				" and ID",
				pattern1->get_ID(),
				": ") ;
			pattern1->wto(PRINT_NO_FREQUENCIES) ;
			fprintf(output_file, "\n\n") ;
		}
		return(pattern1) ;
	}

	return(NIL) ;
} // sequence::check_patterns

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Calculates the length of an alignment.
**
** CALLING SEQUENCE:
**
**	int sequence::compute_character_length()
**
** FORMAL ARGUMENTS:
**
**	Return value:	The length of the alignment in characters, including
**					The white spaces between symbols.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

int sequence::compute_character_length()
{
	int char_length = 0 ;
	symbol *symbol1 ;
	list_element *el_pos1 = NIL ;
	char *symbol_name ;
	
	while (symbol1 = (symbol *)this->get_next_child_by_el_pos(&el_pos1))
	{
		if (use_abbreviations == true)
			symbol_name = find_symbol_abbreviation(symbol1->
			get_name()) ;
		else symbol_name = symbol1->get_name() ;
		
		char_length += strlen(symbol_name) + 1 ; // Add 1
		// to the length of the symbol name for the
		// blank space between one column and the next.
	}
	
	return(char_length - 1) ; // Subtract 1 from char_length because
				// the blank character after the last symbol is not needed.
} // sequence::compute_character_length

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes a textual description of the sequence specified to 
**	the `write file'. This is called from write_tree_object()
**	and wto().
**
** CALLING SEQUENCE:
**
**	void sequence::wto(int print_code) ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
**	print_code:		PRINT_ALL_FREQUENCIES if the frequencies 
**					of the object and all its constituents
**					are to be printed,
**	print_code:		PRINT_ALL_FREQUENCIES if the frequencies 
**					of the object and all its constituents
**					are to be printed,
**					PRINT_SEQUENCE_FREQUENCIES if symbol
**					frequencies are to be suppressed.
**					PRINT_NO_FREQUENCIES otherwise.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	output_file
**
** SIDE EFFECTS:
**
**	NONE
*/

void sequence::wto(int print_code)
{
	base_object *child;
	char buffer[STRING_LENGTH] ;
	
	write_lines(output_file, "(", CONTINUE) ;
	
	this->initialise() ;
	while (child = this->get_next_child())
	{
		child->wto(print_code) ;
		if (current_el_pos->get_next() != NIL)
			write_lines(output_file, " ", CONTINUE) ;
	}
	write_lines(output_file, ")", CONTINUE) ;
	if (print_code == PRINT_NO_FREQUENCIES) return ;
	if (print_code == PRINT_SEQUENCE_FREQUENCIES && frequency != 1)
	{
		sprintf(buffer, "*%d", frequency) ;
		write_lines(output_file, buffer, CONTINUE) ;
		return ;
	}
	if (print_code == PRINT_ALL_FREQUENCIES && frequency != 1)
	{
		sprintf(buffer, "*%d", frequency) ;
		write_lines(output_file, buffer, CONTINUE) ;
	}
} /* sequence::wto */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes out IDs of members.
**
** CALLING SEQUENCE:
**
**	void group::write_IDs()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void group::write_IDs()
{
	int counter = 0 ;
	sequence *pattern1 ;
	list_for(pattern1, sequence, current_alignments_from_parsing)
	{
		fprintf(output_file, "%s%d",
			"ID",
			pattern1->get_ID()) ;

		if (current_alignments_from_parsing->this_is_last_child(pattern1))
		{
			fprintf(output_file, "\n\n") ;
			break ;
		}

		if (++counter == 10)
			fprintf(output_file, "\n") ;
		else fprintf(output_file, ", ") ;
	}
} // group::write_IDs

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes out the patterns in a set of patterns, together with
**	an introductory heading.
**
** CALLING SEQUENCE:
**
**	void group::write_patterns(char *heading, int origin)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	heading:			The heading for the set.
**	origin:				Only patterns with the specified origin
**						are to be printed. If origin is NULL_VALUE,
**						all the patterns in the set_of_patterns
**						are to be written.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Output to output_file.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void group::write_patterns(char *heading, int origin)
{
	sequence *pattern1 ;
	list_element *el_pos1 = NIL ;

	fprintf(output_file, "%s%s", heading, "\n\n") ;
	
	for (pattern1 = (sequence *)this->
		get_next_child_by_el_pos(&el_pos1); pattern1 != NIL; 
			pattern1 = (sequence *)this->
				get_next_child_by_el_pos(&el_pos1))
	{
		if (origin == NULL_VALUE || origin == pattern1->get_origin())
		{
			pattern1->print_ID() ;
			fprintf(output_file, ": ") ;
			pattern1->write_tree_object(PRINT_SEQUENCE_FREQUENCIES) ;
		}
	}
} // group::write_patterns

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Write New or Old. Called from write_new_and_old.
**
** CALLING SEQUENCE:
**
**	void group::write_patterns_with_details()
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void group::write_patterns_with_details()
{
	sequence *pattern1 ;

	this->initialise() ;
	while (pattern1 = (sequence *)this->get_next_child())
	{
		pattern1->print_ID() ;
		
		if (this == old_patterns)
			fprintf(output_file, "%s%1.2f%s",
			" Coding cost: ", 
			pattern1->get_encoding_cost(),
			", ") ;
		else if (this == new_patterns)
			fprintf(output_file, " No coding cost ") ;
		pattern1->write_tree_object(PRINT_NO_FREQUENCIES) ;
		pattern1->write_tree_object(PRINT_ALL_FREQUENCIES) ;

		pattern1->write_with_details() ;
	}
} // group::write_patterns_with_details

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Sorts a group of patterns in descending order of their 
**	compression_difference scores.
**
** CALLING SEQUENCE:
**
**	void group::sort_by_compression_difference()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void group::sort_by_compression_difference()
{
	group *temp_group = new group ;
	sequence *pattern1, *best_pattern ;
	double highest_compression_difference, 
		pattern_compression_difference ;

	while (pattern1 = (sequence *)this->extract_first_child())
		temp_group->add_child(pattern1) ;

	while(temp_group->is_empty() == false)
	{
		highest_compression_difference = LOW_VALUE ;
		list_for(pattern1, sequence, temp_group)
		{
			pattern_compression_difference =
				pattern1->get_compression_difference() ;
			if (pattern_compression_difference > 
				highest_compression_difference)
			{
				highest_compression_difference = 
					pattern_compression_difference ;
				best_pattern = pattern1 ;
			}
		}
		temp_group->extract_child(best_pattern) ;
		this->add_child(best_pattern) ;
	}

	delete temp_group ;
} // group::sort_by_compression_difference

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Sorts a group of patterns in ascending order of ID number. 
**
** CALLING SEQUENCE:
**
**	void group::sort_by_ID()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	The sorted group
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void group::sort_by_ID()
{
	group *temp_group = new group ;
	sequence *pattern1, *best_pattern ;
	int lowest_ID, pattern_ID ;

	while (pattern1 = (sequence *)this->extract_first_child())
		temp_group->add_child(pattern1) ;

	while(temp_group->is_empty() == false)
	{
		lowest_ID = HIGH_VALUE ;
		list_for(pattern1, sequence, temp_group)
		{
			pattern_ID = pattern1->get_ID() ;
			if (pattern_ID < lowest_ID)
			{
				lowest_ID = pattern_ID ;
				best_pattern = pattern1 ;
			}
		}
		temp_group->extract_child(best_pattern) ;
		this->add_child(best_pattern) ;
	}

	delete temp_group ;
} // group::sort_by_ID

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Sorts a group of symbols in ascending order of length of symbols. 
**
** CALLING SEQUENCE:
**
**	void group::sort_by_symbol_length()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	The sorted group
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void group::sort_by_symbol_length()
{
	group *temp_group = new group ;
	symbol *symbol1, *best_symbol ;
	int shortest_length, symbol_length ;

	while (symbol1 = (symbol *)this->extract_first_child())
		temp_group->add_child(symbol1) ;

	while(temp_group->is_empty() == false)
	{
		shortest_length = HIGH_VALUE ;
		list_for(symbol1, symbol, temp_group)
		{
			symbol_length = strlen(symbol1->get_name()) ;
			if (symbol_length < shortest_length)
			{
				shortest_length = symbol_length ;
				best_symbol = symbol1 ;
			}
		}
		temp_group->extract_child(best_symbol) ;
		this->add_child(best_symbol) ;
	}

	delete temp_group ;
} // group::sort_by_symbol_length

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compiles a list of the different symbol types in the patterns
**	in the group 'this' and measures the size of this alphabet.
**
** CALLING SEQUENCE:
**
**	group *group::compile_alphabet(int *alphabet_size)
**
** FORMAL ARGUMENTS:
**
**	Return value:	A group containing the list of symbols.
**
**	alphabet_size:	A pointer to a variable to hold the size of
**					the alphabet.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

group *group::compile_alphabet(int *alphabet_size)
{
	group *symbol_set = this->symbol_set() ;
	symbol_set->sort_by_name() ; // Puts symbols 
		// in the list of original_symbols_in_corpus in 
		// alpha-numeric order of their name fields.
	*alphabet_size = symbol_set->count_number_of_children() ;
	return(symbol_set) ;
} // compile_alphabet

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Scans a list of patterns in 'this' and, for each symbol in each
**	pattern, adds the frequency of the pattern to the frequency of
**	the symbol type, stored in symbol_set.
**
** CALLING SEQUENCE:
**
**	void group::compile_frequencies(group *symbol_set)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	symbol_set:		The alphabet of symbol types in the group.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void group::compile_frequencies(group *symbol_set)
{
	sequence *pattern1 ;
	symbol *symbol1, *symbol_type ;

	this->initialise() ;
	while (pattern1 = (sequence *)this->get_next_child())
	{
		pattern1->initialise() ;
		while (symbol1 = (symbol *)pattern1->get_next_child())
		{
			/* Scan through symbol_set to find the
			symbol there (representing a symbol type)
			which matches the current symbol. When it
			is found, update the frequency. */
			
			symbol_set->initialise() ;
			while (symbol_type = (symbol *)
				symbol_set->get_next_child())
			{
				if (strcmp(symbol1->get_name(),
					symbol_type->get_name()) == 0)
				{
					symbol_type->
						add_to_frequency(pattern1->get_frequency()) ;
					break ;
				}
			}
			if (symbol_type == NIL)
				abort_run("symbol_type not found in \
					group::compile_frequencies") ;
		}
	}
} // group::compile_frequencies

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes a textual description of the group specified to 
**	the `write file'. This is called from write_tree_object()
**	and wto().
**
** CALLING SEQUENCE:
**
**	void group::wto(int print_code) ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
**	print_code:		PRINT_ALL_FREQUENCIES if the frequencies 
**					of the object and all its constituents
**					are to be printed,
**					PRINT_SEQUENCE_FREQUENCIES if symbol
**					frequencies are to be suppressed.
**					PRINT_NO_FREQUENCIES otherwise.
**
** IMPLICIT INPUTS:
**
** IMPLICIT OUTPUTS:
**
**	output_file
**
** SIDE EFFECTS:
**
**	NONE
*/

void group::wto(int print_code)
{
	base_object *child ;
	char buffer[STRING_LENGTH] ;
	
	write_lines(output_file, "[", CONTINUE) ;
	
	this->initialise() ;
	while (child = this->get_next_child())
	{
		child->wto(print_code) ;
		if (current_el_pos != NIL)
			write_lines(output_file, " ", CONTINUE) ;
	}
	write_lines(output_file, "]", CONTINUE) ;
	if (print_code == PRINT_NO_FREQUENCIES) return ;
	if (print_code == PRINT_GROUP_FREQUENCIES && frequency != 1)
	{
		sprintf(buffer, "*%d", frequency) ;
		write_lines(output_file, buffer, CONTINUE) ;
		return ;
	}
	if (print_code == PRINT_ALL_FREQUENCIES && frequency != 1)
	{
		sprintf(buffer, "*%d", frequency) ;
		write_lines(output_file, buffer, CONTINUE) ;
	}
} /* group::wto */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes a textual description of the group specified to 
**	the `write file'. This is called from write_tree_object()
**	and wto().
**
** CALLING SEQUENCE:
**
**	void symbol::wto(int print_code) ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
**	print_code:		PRINT_ALL_FREQUENCIES if the frequencies 
**					of the object and all its constituents
**					are to be printed,
**					PRINT_SEQUENCE_FREQUENCIES if symbol
**					frequencies are to be suppressed.
**					PRINT_NO_FREQUENCIES otherwise.
**
** IMPLICIT INPUTS:
**
** IMPLICIT OUTPUTS:
**
**	output_file
**
** SIDE EFFECTS:
**
**	NONE
*/

void symbol::wto(int print_code)
{
	char buffer[STRING_LENGTH] ;
	
	write_lines(output_file, name, CONTINUE) ;
	
	if (print_code == PRINT_NO_FREQUENCIES) return ;
	
	if (print_code == PRINT_ALL_FREQUENCIES && frequency != 1)
	{
		sprintf(buffer, "*%d", frequency) ;
		write_lines(output_file, buffer, CONTINUE) ;
	}
} /* symbol::wto */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes a base_object to a file. Using wto(),
**	it keeps track of the numbers of characters in a line and
**	takes a new line whenever that is needed to avoid splitting
**	a symbol or a frequency value. 
**
** CALLING SEQUENCE:
**
**	void tree_object::write_tree_object(int print_code)
**
** FORMAL ARGUMENTS:
**
**	Return value:	NONE
**
**	print_code:		PRINT_ALL_FREQUENCIES if the frequencies 
**					of the object and all its constituents
**					are to be printed,
**					PRINT_SEQUENCE_FREQUENCIES if SYMBOL frequencies are
**					to be suppressed.
**					PRINT_NO_FREQUENCIES otherwise.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Text in the write file. 
**
** SIDE EFFECTS:
**
**	NONE
*/

void tree_object::write_tree_object(int print_code)
{
	write_lines(output_file, "", START) ;
	this->wto(print_code) ;
	fprintf(output_file, "\n\n") ;
	fflush(output_file) ;
} /* tree_object::write_tree_object */

/***********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks whether a set contains a symbol which
**	matches a parameter symbol.
**
** CALLING SEQUENCE:
**
**	symbol *tree_object::contains_copy_of(symbol *symbol1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The first item in the set which matches symbol1 or
**					NIL if no matching item is found.
**
**	symbol1:		The data node to be matched to items in the set.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

symbol *tree_object::contains_copy_of(symbol *symbol1)
{
	symbol *symbol_in_set ;
	
	/* Run over set until symbol1 is found or end of set reached */
	
	this->initialise() ;
	while (symbol_in_set = (symbol *)this->get_next_child())
		if (symbol1->name_matches(symbol_in_set))
			return(symbol_in_set) ;
		
	return(NIL) ;
} /* tree_object::contains_copy_of */

/***********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Returns the number of child nodes in a composite data node and
**	set that value in number_of_children.
**
** CALLING SEQUENCE:
**
**	int tree_object::count_number_of_children() ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	The number of child nodes in the given object.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

int tree_object::count_number_of_children()
{
	register int no_of_children = 0 ;
	base_object *child ;
	
	this->initialise() ;
	child = this->get_next_child() ;
	while (child != NIL)
	{
		no_of_children++ ;
		child = this->get_next_child() ;
	}
	
	number_of_children = no_of_children ;
	return(no_of_children) ;
} /* tree_object::count_count_number_of_children */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	In an symbol, finds an al_el by its position.
**
** CALLING SEQUENCE:
**
**	list_element *tree_objectect::get_el_pos_by_int_pos(int position)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The el corresponding to position.
**
**	position:		The position of the el to be found.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

list_element *tree_object::get_el_pos_by_int_pos(int position)
{
	list_element *el1 ;
	
	if (position < -1)
		abort_run("Invalid position in get_el_pos_by_int_pos") ;
	
	if (position == -1) return(NIL) ; // The position -1 corresponds
	// to the position just before the first actual position.
	
	for (el1 = this->get_first_element() ; el1 != NIL;
		el1 = el1->get_next())
			if (el1->get_position() == position) break ;
		
	if (el1 == NIL) abort_run("el1 not found in \
			get_el_pos_by_int_pos") ;
		
	return(el1) ;
} // tree_object::get_el_pos_by_int_pos

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	In an symbol, finds an al_el by its symbol.
**
** CALLING SEQUENCE:
**
**	list_element *tree_object::get_el_pos_by_child(base_object *obj1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The el corresponding to position.
**
**	obj1:			The object in the el to be found.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

list_element *tree_object::get_el_pos_by_child(base_object *obj1)
{
	list_element *el1 ;
	
	for (el1 = this->get_first_element() ; el1 != NIL;
		el1 = el1->get_next())
			if (el1->get_el_obj() == obj1) break ;
		
	if (el1 == NIL) abort_run("el1 not found in \
			get_el_pos_by_child") ;
		
	return(el1) ;
} // tree_object::get_el_pos_by_child

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	A method to create a set of (copies of) the symbols in an
**	object, without any duplicates.
**
** CALLING SEQUENCE:
**
**	group *tree_object::symbol_set()
**
** FORMAL ARGUMENTS:
**
**	Return value:	A group containing copies of symbols in an object,
**					without duplicates.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

group *tree_object::symbol_set()
{
	group *set ;
	
	set = new group ;
	this->symbol_set_recursion(set) ;
	return(set) ;
} // tree_object::symbol_set

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Calculates values for EC and total_cost for any pattern. EC is 
**	the sum of the bit_cost values of the symbols in the pattern that are
**	marked as IDENTIFICATION symbols. total_cost is the sum of the
**	bit_costs of all the symbols in the pattern.
**
**	The computed value is stored in the pattern.
**
** CALLING SEQUENCE:
**
**	void sequence::compute_costs()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	The computed value is stored in the pattern.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void sequence::compute_costs()
{
	symbol *symbol1 ;
	double cumulative_EC = 0, cumulative_cost = 0, symbol_bit_cost ;
	
	list_for(symbol1, symbol, this)
	{
		if (symbol1->get_status() != IDENTIFICATION) continue ;
		symbol_bit_cost = symbol1->get_bit_cost() ;
		cumulative_cost += symbol_bit_cost ;
		cumulative_EC += symbol_bit_cost ;
	}

	this->set_encoding_cost(cumulative_EC) ;
	this->set_total_cost(cumulative_cost) ;
} // sequence::compute_costs

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Makes a code from an alignment ('this'), assuming that patterns
**	in the alignment have code symbols. If they do not have code
**	symbols, then this method does not make a code.
**
**	The method is simply to traverse the alignment, making copies of
**	CODE_SYMBOLs that appear in a column by themselves. The copies
**	exclude the row information in the originals.
**
**	At the same time as making the code, this method computes values
**	for the 'actual cost' of the New hit symbols in the alignment
**	(NSC), the 'minimal cost' of the code symbols in the code (EC),
**	and the 'compression difference' (CD), the 'compression ratio' 
**	(CR) and the abs_P of the alignment.
**
**	The code and the calculated values are not printed out - so that
**	the arrangements and formats of these values can be tailored
**	to specific contexts.
**
** CALLING SEQUENCE:
**
**	sequence *sequence::make_code(bool return_code)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The new code.
**
**	return_code:	If true, the function returns an explicit code sequence.
**					Otherwise it returns NIL.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NSC, EC, CD, CR and abs_P for the 'this'.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

sequence *sequence::make_code(bool return_code)
{
	symbol *symbol1, *symbol2, *symbol3 ;
	int i, row ;
	double NSC, EC ;

	struct row_record
	{
		sequence *row_pattern ; // The pattern in this row.
		int previous_new_int_pos ; // Records the position in New
			// of the previous hit with row_pattern. If there has
			// not yet been a hit with New, this variable has a
			// NULL_VALUE.
		int previous_old_int_pos ; // Records the position in 
			// row_pattern of the previous hit between another pattern
			// and row_pattern. If there has not yet been a hit on
			// row_pattern, this variable has a NULL_VALUE.
		bool row_pattern_fully_recognised ; // true if all the
			// CONTENTS symbols in this row have hits with New and if
			// the corresponding hit symbols in New form a coherent
			// sequence.
	} ;

	struct row_record *rows_array ;

	rows_array = new row_record[sequence_depth] ;

	// Initialise variables in rows_array.

	for (i = 0; i < sequence_depth; i++)
	{
		rows_array[i].row_pattern = NIL ;
		rows_array[i].previous_new_int_pos = NULL_VALUE ;
		rows_array[i].previous_old_int_pos = NULL_VALUE ;
		rows_array[i].row_pattern_fully_recognised = true ;
	}

	NSC = EC = 0 ;
	symbol *symbol_new ;
	list_element *el_pos1 = NIL ;

	sequence *new_code ;
	if (return_code) new_code = new sequence(CODE_PATTERN) ;

	list_for_el_pos(symbol1, symbol, this, el_pos1)
	{
		// Identify columns in the alignment that contain hit
		// symbols from current_new_pattern. Add up the 
		// 'actual costs' of these symbols to arrive at the
		// new_symbols_cost for the alignment.

		if (symbol1->is_a_hit())
		{
			symbol_new = symbol1->get_row_symbol(0) ;
			if (symbol_new != NIL)
			{
				NSC += symbol_new->get_bit_cost() ;
			}
			continue ; // We are now only interested in non-hit
				// CODE_SYMBOLs with status IDENTIFICATION.
		}

		// Find the single symbol in this column and determine its
		// status.

		for (row = 1; row < sequence_depth; row++)
		{
			symbol3 = symbol1->get_row_symbol(row) ;
			if (symbol3 != NIL) break ;
		}

		if (symbol3->get_status() != IDENTIFICATION) continue ;

		// Make a 'shallow' copy of symbol1 (excluding information about
		// rows) and add it to the code. At the same time, add the
		// bit_cost of symbol1 to EC.

		if (return_code)
		{
			symbol2 = symbol1->shallow_copy() ;
			new_code->add_child(symbol2) ;
		}

		EC += symbol1->get_bit_cost() ;
	}

	this->set_new_symbols_cost(NSC) ;
	this->set_encoding_cost(EC) ;
	this->set_compression_difference(NSC - EC) ;
	this->set_compression_ratio(NSC / EC) ;

	double two_power_s = pow(2, EC) ;
	this->set_abs_P(1 / two_power_s) ;

	delete[] rows_array ;

	if (return_code) return(new_code) ;
	else return(NIL) ;

} // sequence::make_code

/***********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Recursive part of symbol_set.  Recursively adds all the different
**	symbols in `object' to the object set `set'.  If a SYMBOL is already
**	in `set' then it's frequency is updated every time it is encountered
**	in `object' rather than adding it again.
**
**	In the initial call the `symbol_set_recursion', `set' should be an empty
**	group.
**
** CALLING SEQUENCE:
**
**	void symbol::symbol_set_recursion(group *set)
**
** FORMAL ARGUMENTS:
**
**	Return value:			None.
**
**	set:					The object to receive the set which is 
**							built up.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	The symbols in `set' become altered by this function.
**
** SIDE EFFECTS:
**
**	NONE
*/

void symbol::symbol_set_recursion(group *set)
{
	symbol *item_in_set, *new_set_item ;
	
	// Check whether set contains a symbol that matches 'this'.
	
	item_in_set = (symbol *)set->contains_copy_of(this) ;
	if (item_in_set == NIL)
	{
		// Item is not in the set - add a copy of it.

		new_set_item = new symbol(*this) ;
		new_set_item->set_frequency(1) ;
		set->add_child(new_set_item) ;
	}
	else
	{
		// The symbol is already in the set: 
		// update it's frequency in the set.
		
		item_in_set->increment_frequency() ;
	}
} /* symbol::symbol_set_recursion */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes out the fields of a hit node (normally for diagnostic
**	purposes).
**
** CALLING SEQUENCE:
**
**	void hit_node::write_hit_node()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void hit_node::write_hit_node()
{
	fprintf(output_file, "%s%d%s%d%s%d%s%s%s%d%s%d%s%s%s",
		"Hit node ID",
		ID,
		":\n\ndriving_int_pos = ",
		driving_int_pos,
		", driving pattern ID",
		driving_pattern->get_ID(),
		", driving_symbol ",
		driving_symbol->get_name(),
		"\ntarget_int_pos = ",
		target_int_pos,
		", target pattern ID",
		target_pattern->get_ID(),
		", target_symbol ",
		target_symbol->get_name(),
		"\n\n") ;
} // hit_node::write_hit_node

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Consructor for the class hit_node.
**
** CALLING SEQUENCE:
**
**	hit_node::hit_node() : tree_object()
**
** FORMAL ARGUMENTS:
**
**	Return value:	None
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

hit_node::hit_node() : cost_tree_object()
{
	ID = hit_node_ID_number++ ;
	hn_parent = NIL ;
	driving_pattern = NIL ;
	driving_symbol = NIL ;
	driving_int_pos = NULL_VALUE ; 
	target_pattern = NIL ;
	target_symbol = NIL ;
	target_int_pos = NULL_VALUE ;
	leaf_entry = NIL ;
} // hit_node::hit_node

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Copy constructor for the class hit_node.
**
** CALLING SEQUENCE:
**
**	hit_node::hit_node(hit_node &hn) : tree_object(hn)
**
** FORMAL ARGUMENTS:
**
**	Return value:	None
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

hit_node::hit_node(hit_node &hn) : cost_tree_object(hn)
{
	ID = hit_node_ID_number++ ;
	hn_parent = hn.get_hn_parent() ;
	driving_pattern = hn.get_driving_pattern() ;
	driving_symbol = hn.get_driving_symbol() ;
	driving_int_pos = hn.get_driving_int_pos() ; 
	target_pattern = hn.get_target_pattern() ;
	target_symbol = hn.get_target_symbol() ;
	target_int_pos = hn.get_target_int_pos() ;
	leaf_entry = hn.get_leaf_entry() ;
} // hit_node::hit_node

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compares two alignments to see whether they match or not.
**
** CALLING SEQUENCE:
**
**	bool sequence::alignment_matches(sequence *al1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if the two alignments match, false otherwise.
**
**	al1:			The second of the two alignments to be compared.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

bool sequence::alignment_matches(sequence *al1)
{
	symbol *al_col1, *al_col2, *symbol1, *symbol2 ;
	alignment_element *al_el1, *al_el2 ;
	sequence *orig_patt1, *orig_patt2 ;
	int al1_length, al1_depth, row1, row2 ;
	bool match_found, al1_rows[MEDIUM_SCRATCH_ARRAY_SIZE] ;
	
	if (al1 == NIL) abort_run("NIL value for al1 in \
		sequence::alignment_matches") ;
	
	/* Compare the lengths and depths of the sequences. */
	
	al1_length = al1->get_number_of_children() ;
	
	if (number_of_children != al1_length) return(false) ;
	
	al1_depth = al1->get_sequence_depth() ;
	
	if (sequence_depth != al1_depth) return(false) ;
	
	// Check for a match in the basic sequence of columns.

	symbol2 = (symbol *)al1->get_first_child() ;
	for (symbol1 = (symbol *)this->get_first_child(); symbol1 != NIL ;
		symbol1 = (symbol *)this->get_next_child())
	{
		if (symbol1->name_matches(symbol2) == false) return(false) ;
		symbol2 = (symbol *)al1->get_next_child() ;
		if (symbol2 == NIL)
		{
			symbol1 = (symbol *)this->get_next_child() ;
			break ;
		}
	}

	if (symbol1 != NIL || symbol2 != NIL)
		return(false) ;
	
	/* Now check if actual sequences match. Two sequences can
	be the same even though the rows are in a different order.
	However, if they are to be recognised as the same, for every
	row in one sequence, there must be a row in the other sequence
	which matches it exactly. */
	
	/* Initialise al1_rows[] which is used to avoid unnecessary
	re-checking of rows in al1 which have already been found to
	match a row in *this. */
	
	for (row2 = 0; row2 < al1_depth; row2++) al1_rows[row2] = false ;
	
	for (row1 = 0; row1 < sequence_depth; row1++)
	{
		match_found = false ;
		for (row2 = 0; row2 < al1_depth; row2++)
		{
			if (al1_rows[row2] == true) continue ;
			
			al_col1 = (symbol *)this->get_first_child() ;
			this->set_current_el_pos(this->get_first_el_pos()) ;
			al_el1 = al_col1->get_al_el(row1) ;
			orig_patt1 = al_el1->get_original_pattern() ;
			
			al_col2 = (symbol *)al1->get_first_child() ;
			al1->set_current_el_pos(al1->get_first_el_pos()) ;
			al_el2 = al_col2->get_al_el(row2) ;
			orig_patt2 = al_el2->get_original_pattern() ;
			
			if (orig_patt2 != orig_patt1) continue ;
			
			symbol1 = (symbol *)al_el1->get_el_obj() ;
			symbol2 = (symbol *)al_el2->get_el_obj() ;
			if (symbol1 != symbol2) continue ;
			
			while (al_col1 = (symbol *)this->get_next_child())
			{
				symbol1 = al_col1->get_row_symbol(row1) ;
				al_col2 = (symbol *)al1->get_next_child() ;
				symbol2 = al_col2->get_row_symbol(row2) ;
				if (symbol1 != symbol2) goto L1 ;
			}
			match_found = true ;
			al1_rows[row2] = true ;
			break ;
			L1: ;
		}
		if (!match_found) return(false) ;
	}
	
	return(true) ;
} /* sequence::alignment_matches */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compares two sequences to see whether they match or not. Only the
**	names of symbols are matched, not the rows in the sequences (if any).
**
** CALLING SEQUENCE:
**
**	bool sequence::sequence_matches(sequence *seq1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if the two sequences match, false otherwise.
**
**	seq1:			The second of the two sequences to be compared.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

bool sequence::sequence_matches(sequence *seq1)
{
	symbol *al_col1, *al_col2 ;
	int seq1_length ;
	
	if (seq1 == NIL) abort_run("NIL value for seq1 in \
		sequence::sequence_matches") ;
	
	/* Compare the lengths of the sequences. */
	
	seq1_length = seq1->get_number_of_children() ;
	
	if (number_of_children != seq1_length) return(false) ;
	
	/* Now check if the sequences of symbols match. */
	
	this->initialise() ;
	seq1->initialise() ;
	while (al_col1 = (symbol *) this->get_next_child())
	{
		al_col2 = (symbol *)seq1->get_next_child() ;
		if (al_col2 == NIL) return(false) ;
		
		if (al_col1->name_matches(al_col2)) continue ;
		else return(false) ;
	}
	
	// We have reached the end of 'this'. Check to see whether
	// we have also reached the end of seq1. If we have, we have
	// an exact match between the two sequences, otherwise the
	// two sequences are not the same.
	
	al_col2 = (symbol *)seq1->get_next_child() ;
	
	if (al_col2 == NIL) return(true) ;
	else return(false) ;
	
} /* sequence::sequence_matches */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For diagnostic purposes.
**
** CALLING SEQUENCE:
**
**	void sequence::show_al_structure(hit_node *h_node)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	h_node:			The leaf node for the given alignment. The
**					value may be NIL, in which case, information
**					about driving pattern, target_pattern and
**					h_node itself are not printed out.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Printed output
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void sequence::show_al_structure(hit_node *h_node)
{
	symbol *al_col1 ;
	int i ;
	
	fprintf(output_file, "Alignment ") ;
	this->print_ID() ;
	fprintf(output_file, " : ") ;

	if (h_node != NIL)
	{
		(h_node->get_driving_pattern())->print_ID() ;
		fprintf(output_file, " : ") ;
		(h_node->get_target_pattern())->print_ID() ;
		fprintf(output_file, " : ") ;
		h_node->print_ID() ;
	}
	else fprintf(output_file, "NIL hit_node") ;

	fprintf(output_file, "%s%1.2f%s%1.2f%s%1.2f%s%d%s%d%s",
		", NSC = ", new_symbols_cost,
		", EC = ", encoding_cost,
		", CR = ", compression_ratio,
		",\nsequence_depth = ", sequence_depth,
		", number of columns = ", number_of_children,
		"\n\n") ;
	this->initialise();
	i = 0 ;
	while (al_col1 = (symbol *)this->get_next_child())
	{
		fprintf(output_file, "%s%d%s", "Column ", i, ":\n") ;
		al_col1->show_al_structure() ;
		i++ ;
	}
	fflush(output_file) ;
	
	fprintf(output_file, "Alignment ") ;
	this->print_ID() ;
	fprintf(output_file, "\n\n") ;
	this->write_alignment(output_file, write_section_chars_length, 
		NULL_VALUE, alignment_format) ;
	fprintf(output_file, "End of alignment ") ;
	this->print_ID() ;
	fprintf(output_file, "\n\n") ;
	fflush(output_file) ;
} // sequence::show_al_structure

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Fetches terminal nodes of the hit structure in sorted order and
**	sets the value of an external pointer to the corresponding sequence.
**
** CALLING SEQUENCE:
**
**	hit_node *get_leaf_nodes_in_order(int *row, sequence **alignm,
**		int *leaf_index)
**
** FORMAL ARGUMENTS:
**
**	Return value:	The next node to be fetched. NIL when there are
**					no more nodes.
**
**	row:			A pointer to an integer referencing the relevant row
**					in sort_array[]. Start *row with 0.
**	alignm: 		A pointer to an external pointer to the sequence
**					corresponding to row.
**	leaf_index:		A pointer to an external integer indexing leaf_array[]
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

hit_node *get_leaf_nodes_in_order(int *row, sequence **alignm, int *leaf_index)
{
	if (*row >= fe_sort) return(NIL) ;
	*leaf_index = sort_array[*row] ;
	(*row)++ ;
	*alignm = alignments_array[*leaf_index] ;
	return(leaf_array[*leaf_index]) ;
} // get_leaf_nodes_in_order

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Cleans up memory allocated by new_patterns and
**	then exits. If this memory is not cleaned up it may clog
**	up the machine.
**
** CALLING SEQUENCE:
**
**	void exit_routine(int status)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	status: 		0 for successful termination, 1 otherwise.
**
** IMPLICIT INPUTS:
**
**	corpus, original_symbols_in_corpus, boundary_symbol,
**	duplicate_unifications, hn_master, hit_root
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void exit_routine(int status)
{
	delete_hit_sequences(hit_sequences) ;
	delete_hit_sequences(new_hit_sequences) ;
	driving_alignments->release_children() ;
	delete driving_alignments ;
	current_alignments_from_parsing->release_children() ;
	delete current_alignments_from_parsing ;
	delete cumulative_alignments_from_parsing ;
	symbols_in_new->release_children() ;
	delete symbols_in_new ;
	concatenated_new_patterns->release_children() ;
	delete concatenated_new_patterns ;

	set_of_combinations->delete_children() ;
	delete set_of_combinations ;
	delete best_combinations ;

	delete created_code_patterns ;
	delete hn_master ;
	delete brackets_index ;
	delete used_driving_patterns ;

	delete original_symbols_in_corpus ;
	delete symbols_in_old_after_learning ;

	// delete[] sort_array ;
	// delete[] leaf_array ;
	// delete[] half_list ;
	// delete[] alignments_array ;
	delete[] write_al_chars ;
	delete[] write_al_vacant_slots ;
	delete[] key_array ;

	delete corpus ;

	fclose(output_file) ;
	fclose(input_file) ;
	fclose(parameters_file) ;
	fclose(latex_file) ;
	exit(status) ;
} /* exit_routine */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**		Aborts the program after displaying an error message.
**
** CALLING SEQUENCE:
**
**	void abort_run(char *message)
**
** FORMAL ARGUMENTS:
**
**	Return value:	None; this function will never return.
**
**	message:		The string containing the error message.
**
** IMPLICIT INPUTS:
**
**	Functions: write_message
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void abort_run(char *message)
{
	write_message(message) ;
	exit_routine(1) ;
} /* abort_run */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes a string to the logfile.
**
** CALLING SEQUENCE:
**
**	void write_message(char *s)
**
** FORMAL ARGUMENTS:
**
**	Return value:	None
**
**	s:				The string to be written.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	logfile: Written to.
**
** SIDE EFFECTS:
**
**	Logfile is updated (written to).  No checks are made
**	as to whether an error occurred during writing.
*/

void write_message(char *s)
{
	fprintf(output_file, "%s%s", s, "\n\n") ;
} /* write_message */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes to a given output stream, counting the number of characters
**	in each line and taking a new line when necessary.
**
** CALLING SEQUENCE:
**
**	void void write_lines(FILE *stream, char *string, int status)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	stream:			The file to be written to.
**	string: 		The string to be written.
**	status: 		START for starting, CONTINUE for carrying on.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void write_lines(FILE *stream, char *string, int status)
{
	static column_counter = 0 ;
	
	if (status == START) column_counter = 0 ;
	
	column_counter += strlen(string) ;
	if (column_counter > MAX_WRITE_LINE)
	{
		fprintf(stream, "\n") ;
		column_counter = 0 ;
	}
	
	fprintf(stream, "%s", string) ;
} /* write_lines */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Scans over spaces and commas in the `syntax buffer'.
**
** CALLING SEQUENCE:
**
**	void scan(void) ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
** IMPLICIT INPUTS:
**
**	Variables:	syn_buf_ptr
**
** IMPLICIT OUTPUTS:
**
**	syn_buf_ptr
**
** SIDE EFFECTS:
**
**	None, but see IMPLICIT OUTPUTS above.
*/

void scan() /* Scans over spaces, commas etc between units in the input */
{
    char c;
    
    c = *syn_buf_ptr ;
    while (c == ' ' || c == ',')
        c = *(++syn_buf_ptr) ;
} /* scan */

/*********************************************************************/

tree_object *comp_object() ;

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  Reads a frequency value for an object (if any) recorded as
**  '*<integer_string>' immediately after the object in the
**  input file. If there is a value, this overwrites the default
**  value (1) in internal representation of the object.
**
** CALLING SEQUENCE:
**
**  void read_frequency(base_object *obj) ;
**
** FORMAL ARGUMENTS:
**
**  Return value:   void
**
**  obj:            The object whose frequency value is to be read.
**
** IMPLICIT INPUTS:
**
**  syn_buf_ptr
**
** IMPLICIT OUTPUTS:
**
**  syn_buf_ptr
**
** SIDE EFFECTS:
**
**  None, but see IMPLICIT OUTPUTS above.
*/

void read_frequency(base_object *obj)
{
    char *cpx, digit_array[STRING_LENGTH] ;
    int i ;
    
    cpx = syn_buf_ptr ;
    if (*syn_buf_ptr == '*') syn_buf_ptr++ ;
    else return ;
    if (!isdigit(*syn_buf_ptr))
    {
        syn_buf_ptr = cpx ;
        return ;
    }
    i = 0 ;
    while (isdigit(*syn_buf_ptr))
    {
        digit_array[i] = *syn_buf_ptr ;
        if (++i >= STRING_LENGTH)
            abort_run("Frequency digits in input too long") ;
        syn_buf_ptr++ ;
    }
    digit_array[i] = '\0' ;
    
    obj->set_frequency(atoi(digit_array)) ;
    
    scan() ;
} /* read_frequency */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  Returns true if the string given matches the
**  `syntax buffer'.  The current place in the syntax buffer is advanced
**  past the matched string if the string matches and is left unchanged
**  if it doesn't.
**
** CALLING SEQUENCE:
**
**  bool characters(char *charsx)
**
** FORMAL ARGUMENTS:
**
**  Return value:   true if a match occurred, false otherwise.
**
**  charsx:         String to match `syntax buffer' with.
**
** IMPLICIT INPUTS:
**
**  Functions:  scan
**
**  Variables:  syn_buf_ptr
**
** IMPLICIT OUTPUTS:
**
**  syn_buf_ptr
**
** SIDE EFFECTS:
**
**  None, but see IMPLICIT OUTPUTS above.
*/

bool characters(char *charsx) /* Recognises a string of characters
in the syntax buffer */
{
    char *cpx;
    
    cpx = syn_buf_ptr;
    while (*charsx != '\0')
        if (*syn_buf_ptr++ != *charsx++)
        {
            syn_buf_ptr = cpx;
            return(false) ;
        }
    scan() ;
    return(true) ;
} /* characters */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Error non-terminal for SP parser.  Writes `informative' error message
**	and returns.
**
** CALLING SEQUENCE:
**
**	void bad_syntax(void) ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
** IMPLICIT INPUTS:
**
**	Functions:	write_message
**
**	Variables:	syn_buf_ptr
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
*/

void bad_syntax()
{
	char *cpx;
	cpx = syn_buf_ptr - 20;
	
	if (cpx < syntax_buffer) cpx = syntax_buffer;
	
	write_message("Error_in_syntax:") ;
	while (cpx <= syn_buf_ptr) write_message(cpx++) ;
	write_message("\n") ;
} /* bad_syntax */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Opens file specified using buffer size specified and reads
**	the entire file into the `syntax buffer'. The file is
**	closed after reading.
**
** CALLING SEQUENCE:
**
**	void in_syntax() ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
** IMPLICIT INPUTS:
**
**	Functions:	abort_run
**
**	Variables:	syntax_buffer
**
** IMPLICIT OUTPUTS:
**
**	syntax_buffer
**
** SIDE EFFECTS:
**
**	NONE
*/

void in_syntax()
{
	int c;
	int i = 0;
	
	/* range check required */
	while (c = getc(input_file))
	{
		if (c == EOF || c == '~') break ;
		if (i >= BUFSIZE) 
			abort_run("Array overflow in 'in_syntax'") ;
		if (c == '\n' || c == '\t')
			syntax_buffer[i++] = ' ';
		else syntax_buffer[i++] = c;
	}
} /* in_syntax */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes out the key to the abbreviations of symbols used in
**	the corpus.
**
** CALLING SEQUENCE:
**
**	void write_key()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	key_array[]
**
** IMPLICIT OUTPUTS:
**
**	Output file
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void write_key()
{
	int key_index1, key_index2, i, j ;
	int sort_array[VERY_LARGE_SCRATCH_ARRAY_SIZE] ;

	if(key_array_size >= VERY_LARGE_SCRATCH_ARRAY_SIZE)
		abort_run("sort_array[] too small in write_key()") ;

	// Compile values for sort_array[]
	// in alphabetical order of abbreviated_symbol.

	// First set 'done' field of key_array[] to false.

	for (i = 0; i < fe_key_array; i++)
		key_array[i].done = false ;

	// Now sort in alphabetical order of abbreviated symbol.

	int best_i ;
	char *best_symbol ;
	j = 0 ;
	while(true)
	{
		best_i = NULL_VALUE ;
		best_symbol = "zzzzzzzzzzzz" ;
		for (i = 0; i < fe_key_array; i++)
		{
			if (key_array[i].done == true) continue ;
			if (compare_strings(key_array[i].abbreviated_symbol,
				best_symbol) < 0)
			{
				best_symbol = key_array[i].abbreviated_symbol ;
				best_i = i ;
			}
		}
		if (best_i == NULL_VALUE) break ;
		key_array[best_i].done = true ;
		sort_array[j++] = best_i ;
	}
	
	fprintf(output_file, "KEY\n\n") ;
	for (i = 0; i < fe_key_array; i++)
	{
		key_index1 = sort_array[i] ;
		fprintf(output_file, "%s%s",
			key_array[key_index1].abbreviated_symbol, " = ") ;
		fprintf(output_file, "%s%c",
			key_array[key_index1].complete_symbol, '\n') ;
	}

	// Sort by order of abbreviated core.

	// First set 'done' field of key_array[] to false.

	for (i = 0; i < fe_key_array; i++)
		key_array[i].done = false ;

	j = 0 ;
	while(true)
	{
		best_i = NULL_VALUE ;
		best_symbol = "zzzzzzzzzzzz" ;
		for (i = 0; i < fe_key_array; i++)
		{
			if (key_array[i].done == true) continue ;
			if (compare_strings(key_array[i].abbreviated_core,
				best_symbol) < 0)
			{
				best_symbol = key_array[i].abbreviated_core ;
				best_i = i ;
			}
		}
		if (best_i == NULL_VALUE) break ;
		key_array[best_i].done = true ;
		sort_array[j++] = best_i ;
	}

	fprintf(output_file, "\nABBREVIATIONS OF SYMBOL CORES\n\n") ;

	int core_symbol_array[VERY_LARGE_SCRATCH_ARRAY_SIZE] ;
	int fe_csa = 0 ;
	char *current_core_of_symbol ;
	for (i = 0; i < fe_key_array; i++)
	{
		key_index1 = sort_array[i] ;
		current_core_of_symbol = key_array[key_index1].core_of_symbol ;

		// Check to see whether current core symbol has been
		// printed before.

		for (j = 0; j < i; j++)
		{
			key_index2 = sort_array[j] ;
			if (strcmp(current_core_of_symbol,
				key_array[key_index2].core_of_symbol) == 0)
					goto L1 ;
		}

		core_symbol_array[fe_csa++] = key_index1 ;

		L1: ;
	}

	for (i = 0; i < fe_csa; i++)	
	{
		key_index1 = core_symbol_array[i] ;
		fprintf(output_file, "%s%s",
			key_array[key_index1].abbreviated_core, " = ") ;

		if (i == fe_csa - 1)
			fprintf(output_file, "%s%c",
				key_array[key_index1].core_of_symbol, '.') ;
		else fprintf(output_file, "%s%s",
			key_array[key_index1].core_of_symbol, ", ") ;
	}

	fprintf(output_file, "\n\n") ;
} // write_key

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Finds a unique abbreviation for a symbol. Called from
**	make_letter_abbreviations.
**
**	The aim is to make an abbreviation using a subset of the characters
**	in the original symbol in this order of priority:
**
**	1 Initial letters of sub-symbols divided by '-' or '_'.
**	2 Hard consonants.
**	3 Vowels.
**
** CALLING SEQUENCE:
**
**	void make_abbreviation(int last_int_pos)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	last_int_pos:	The last position in key_array[] that needs to be
**					checked in checking for the uniqueness of
**					an abbreviation.
**
** IMPLICIT INPUTS:
**
**	complete_symbol[], key_array[]
**
** IMPLICIT OUTPUTS:
**
**	new_abbreviation[]
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void make_abbreviation(int last_int_pos)
{
	char this_core_of_symbol[MEDIUM_SCRATCH_ARRAY_SIZE] ;
	char character_classes[MEDIUM_SCRATCH_ARRAY_SIZE] ;
	char scratch_array[MEDIUM_SCRATCH_ARRAY_SIZE] ;

	strcpy(this_core_of_symbol, key_array[last_int_pos].core_of_symbol) ;

	// Check whether this_core_of_symbol == '...'. If so, use
	// this as its abbreviation.

	if (strcmp(this_core_of_symbol, "...") == 0)
	{
			strcpy(new_abbreviation, "...") ;
			return ;
	}

	// Check whether an abbreviation already exists for 
	// this_core_of_symbol.

	int j ;
	for (j = 0; j < last_int_pos; j++)
	{
		if (strcmp(this_core_of_symbol, key_array[j].core_of_symbol) == 0)
		{
			strcpy(new_abbreviation, key_array[j].abbreviated_core) ;
			return ;
		}
	}

	// No existing abbreviation has been found. Now make one.

	// Classify the characters in this_core_of_symbol.

	int this_core_length = strlen(this_core_of_symbol), i ;
	char c1, c2 ;
	for (i = 0; i < this_core_length; i++)
		scratch_array[i] = ' ' ;

	j = NULL_VALUE ;
	for (i = 0; i < this_core_length; i++)
	{
		if (j == NULL_VALUE) c1 = NULL_VALUE ;
		else c1 = this_core_of_symbol[j] ;
		c2 = this_core_of_symbol[i] ;

		if (i == 0) character_classes[i] = 1 ;
		else if (c1 == '-' || c1 == '_') character_classes[i] = 2 ;
		else if (c2 == 'a' || c2 == 'e' || c2 == 'i' || c2 == 'o'
			|| c2 == 'u')
		{
			if (i == this_core_length - 1) character_classes[i] = 9 ;
			else character_classes[i] = 10 ;
		}
		else if (c2 == 'h' || c2 == 'l' || c2 == 's' || c2 == 'w'
			|| c2 == 'y' || c2 == 'r')
		{
			if (i == this_core_length - 1) character_classes[i] = 7 ;
			else character_classes[i] = 8 ;
		}
		else if (isdigit(c2))
		{
			if (i == this_core_length - 1) character_classes[i] = 3 ;
			else character_classes[i] = 4 ;
		}
		else
		{
			if (i == this_core_length - 1) character_classes[i] = 5 ;
			else character_classes[i] = 6 ;
		}
		j = i ;
	}

	int k, current_class_level = 1, m ;

	// Make an abbreviation from the first two most highly classified
	// letters.

	bool letter_found ;
	for (k = 0; k < 2; k++)
	{
		while (true)
		{
			letter_found = false ;
			for (m = 0; m < this_core_length; m++)
			{
				if (character_classes[m] == current_class_level
					&& scratch_array[m] == ' ')
				{
					scratch_array[m] = this_core_of_symbol[m] ;
					letter_found = true ;
					break ;
				}
			}
			if (letter_found) break ;
			else
			{
				current_class_level++ ;
				if (current_class_level > CHAR_CLASS_LIMIT) goto L1 ;
			}
		}
	}

	L1: ;

	// Transfer non-space characters from scratch_array[]
	// to new_abbreviation[].

	j = 0 ;
	for (i = 0; i < this_core_length; i++)
	{
		if (scratch_array[i] == ' ') continue ;
		new_abbreviation[j++] = scratch_array[i] ;
	}

	new_abbreviation[j] = '\0' ;

	// Check whether the newly-constructed abbreviation is
	// unique.
	
	bool abbreviation_is_unique = true ;
	for (j = 0; j < last_int_pos; j++)
	{
		if (strcmp(key_array[j].abbreviated_core,
			new_abbreviation) == 0)
		{
			abbreviation_is_unique = false ;
			break ;
		}
	}
	
	// If it is not unique, keep adding characters until it is.
	
	while (abbreviation_is_unique == false)
	{
		// Add the next character.
		
		while (true)
		{
			letter_found = false ;
			for (m = 0; m < this_core_length; m++)
			{
				if (character_classes[m] == current_class_level
					&& scratch_array[m] == ' ')
				{
					scratch_array[m] = this_core_of_symbol[m] ;
					letter_found = true ;
					goto L2 ;
				}
			}
			if (letter_found) break ;
			else
			{
				current_class_level++ ;
				if (current_class_level > CHAR_CLASS_LIMIT) break ;
			}
		}

		L2: ;
		
		// Transfer non-space characters from scratch_array[]
		// to new_abbreviation[].

		j = 0 ;
		for (i = 0; i < this_core_length; i++)
		{
			if (scratch_array[i] == ' ') continue ;
			new_abbreviation[j++] = scratch_array[i] ;
		}

		new_abbreviation[j] = '\0' ;

		// Check whether the newly-constructed abbreviation is
		// unique.
		
		abbreviation_is_unique = true ;
		for (j = 0; j < last_int_pos; j++)
		{
			if (strcmp(key_array[j].abbreviated_core,
				new_abbreviation) == 0)
			{
				abbreviation_is_unique = false ;
				break ;
			}
		}
	}
} // make_abbreviation

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Creates a key to the abbreviations of symbols to be used in
**	writing out alignments.
**
**	Abbreviations are only made for symbols which do not begin with
**	a digit. Apart from the first and last characters (which can be
**	a consonant or a vowel) the abbreviation for a symbol is made
**	from consonants in the symbol, starting with the first, followed
**	by the last, followed by the consonant after the first, followed
**	by the consonant before the last, and so on. Addition of
**	consonants stops when the abbreviation is unique in the set
**	of abbreviations. Initial '<' or '</' are always preserved as
**	are trailing '>'.
**
** CALLING SEQUENCE:
**
**	void make_letter_abbreviations()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	original_symbols_in_corpus
**
** IMPLICIT OUTPUTS:
**
**	Values for key_array[]
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void make_letter_abbreviations()
{
	symbol *symbol1 ;
	int i, j, symbol_length,
		number_of_symbol_types =
			original_symbols_in_corpus->count_number_of_symbols() ;

	// Sort original_symbols_in_corpus in ascending order of the
	// length of symbols.

	original_symbols_in_corpus->sort_by_symbol_length() ;

	// Create key_array[].

	key_array_size = number_of_symbol_types + 1 ;
	key_array = new struct key_entry[key_array_size] ;
	
	// Copy the name of the symbols in the corpus into
	// the 'complete_symbol' field of key_array[] and initialise
	// the other fields at the same time.
	
	fe_key_array = 0 ;
	list_for(symbol1, symbol, original_symbols_in_corpus)
	{
		symbol_length = strlen(symbol1->get_name()) + 1 ;
		if (symbol_length > MEDIUM_SCRATCH_ARRAY_SIZE)
			abort_run("Symbol is too long in \
				make_letter_abbreviations") ;
		strcpy(key_array[fe_key_array].complete_symbol, 
			symbol1->get_name()) ;
		strcpy(key_array[fe_key_array].abbreviated_core, "") ;
		strcpy(key_array[fe_key_array].abbreviated_symbol, "") ;
		strcpy(key_array[fe_key_array].start_of_symbol, "") ;
		strcpy(key_array[fe_key_array].end_of_symbol, "") ;
		strcpy(key_array[fe_key_array].core_of_symbol, "") ;
		plus_one(&fe_key_array, key_array_size,
			"Overflow of key_array[] in read_key") ;
	}
	
	// Now step through key_array[] filling in the fields start_of_symbol,
	// end_of_symbol and core_of_symbol. At the same time, find an
	// abbreviation for the core and then create an abbreviated
	// symbol from the abbreviated core and the start and end
	// symbols. Before creating an abbreviation for any core, check
	// back to see whether one has already been created. If so, use it.

	int k ;
	char c ;
	char scratch_string[MEDIUM_SCRATCH_ARRAY_SIZE] ;
	for (i = 0; i < fe_key_array; i++)
	{
		strcpy(complete_symbol, key_array[i].complete_symbol) ;

		if (complete_symbol[0] == '#')
		{
			strcpy(key_array[i].start_of_symbol, "#") ;
			j = 1 ;
			c = complete_symbol[j] ;
			k = 0 ;
			while (c != '/0')
			{
				scratch_string[k++] = c ;
				c = complete_symbol[++j] ;
			}
			scratch_string[k++] = '\0' ;
			strcpy(key_array[i].core_of_symbol, scratch_string) ;			
		}
		else if (complete_symbol[0] == '<' && complete_symbol[1] == '/')
		{
			strcpy(key_array[i].start_of_symbol, "</") ;
			strcpy(key_array[i].end_of_symbol, ">") ;
			j = 2 ;
			c = complete_symbol[j] ;
			k = 0 ;
			while (c != '>')
			{
				scratch_string[k] = c ;
				plus_one(&k, MEDIUM_SCRATCH_ARRAY_SIZE,
					"Overflow in make_letter_abbreviations()") ;
				c = complete_symbol[++j] ;
			}
			scratch_string[k] = '\0' ;
			strcpy(key_array[i].core_of_symbol, scratch_string) ;			
		}
		else if (complete_symbol[0] == '<')
		{
			strcpy(key_array[i].start_of_symbol, "<") ;
			strcpy(key_array[i].end_of_symbol, ">") ;
			j = 1 ;
			c = complete_symbol[j] ;
			k = 0 ;
			while (c != '>')
			{
				scratch_string[k] = c ;
				plus_one(&k, MEDIUM_SCRATCH_ARRAY_SIZE,
					"Overflow in make_letter_abbreviations()") ;
				c = complete_symbol[++j] ;
			}
			scratch_string[k] = '\0' ;
			strcpy(key_array[i].core_of_symbol, scratch_string) ;			
		}
		else
		{
			j = 0 ;
			c = complete_symbol[j] ;
			k = 0 ;
			while (c != '\0')
			{
				scratch_string[k] = c ;
				plus_one(&k, MEDIUM_SCRATCH_ARRAY_SIZE,
					"Overflow in make_letter_abbreviations()") ;
				c = complete_symbol[++j] ;
			}
			scratch_string[k] = '\0' ;
			strcpy(key_array[i].core_of_symbol, scratch_string) ;			
		}
		
		make_abbreviation(i) ;
		
		// Add the newly-constructed abbreviation to the key_array[]
		// and construct the complete abbreviated_symbol by adding
		// start_of_symbol and end_of_symbol.
		
		strcpy(key_array[i].abbreviated_core, new_abbreviation) ;
		sprintf(key_array[i].abbreviated_symbol, "%s%s%s",
			key_array[i].start_of_symbol,
			key_array[i].abbreviated_core,
			key_array[i].end_of_symbol) ;
	}

	write_key() ; // Writes out the key to abbreviations of symbols
	// used in the corpus.
		
} // make_letter_abbreviations

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Reads in parameters for the program.
**
** CALLING SEQUENCE:
**
**	void read_parameters()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void read_parameters()
{
	char input_string[LONG_STRING_LENGTH],
		*parameters[] =
	{
		"/*",
		"PROBABILITIES",
		"SHOW_AL_STRUCTURE",
		"SHOW_HIT_STRUCTURE",
		"SHOW_UNSELECTED_ALIGNMENTS",
		"VERBOSE",
		"USE_ABBREVIATIONS",
		"SHOW_ALL_PARSING_ALIGNMENTS",
		"ID_C_SYMBOL_CONSTRAINT",
		"RECOGNISE_BRACKETS",
		"RECOGNISE_XML_BRACKETS",
		"BRACKETS_ID_SYMBOL_CONSTRAINT",
		"MATCH_ALL_DATA_SYMBOLS",
		"MATCH_ALL_DATA_SYMBOLS_IN_COMPOSITE_ALIGNMENTS",
		"MATCH_ALL_DATA_SYMBOLS_IN_HIERARCHICAL_ALIGNMENTS",
		"MATCH_ALL_NEW_SYMBOLS",
		"COMBINE_ALIGNMENTS",
		"ONLY_ONE_NEW_PATTERN_IN_EACH_ALIGNMENT",
		"ALIGNMENT_FORMAT",
		"HIT_STRUCTURE_ROWS",
		"FAIL_SCORE",
		"COST_FACTOR",
		"KEEP_ROWS",
		"DRIVING_KEEP_ROWS",
		"BEST_ALIGNMENT_PRINT_LIMIT",
		"FIGURE_ID",
		"MAX_NEW_GAP",
		"MAX_OLD_GAP",
		"MAX_UNSUPPORTED_CYCLES",
		"MAX_ALIGNMENTS_IN_ONE_CYCLE",
		"ORIENTATION",
		"WRITE_AL_CHARS_LENGTH",
		"WRITE_SECTION_CHARS_LENGTH",
		"MINIMUM_FONT_HEIGHT",
		"MAXIMUM_FONT_HEIGHT",
		"WINDOW_SIZE",
		"COMBINATION_LIMIT",
		"AUDIT_TRAIL",
		"LAST_ITEM"
	} ;
	int i, j, parameter_index ;
	
	fscanf(parameters_file, "PARAMETERS%s", input_string) ;
	while (strcmp(input_string, "END_PARAMETERS") != 0)
	{
		parameter_index = 0 ;
		while (true)
		{
			if (strcmp(parameters[parameter_index],
				input_string) == 0)
				break ;
			if (strcmp(parameters[parameter_index],
				"LAST_ITEM") == 0)
			{
				fprintf(output_file, "%s%s", input_string,
					"\n\n") ;
				abort_run("Parameter not identified") ;
			}
			parameter_index++ ;
		}
		
		switch (parameter_index)
		{
			case 0: /* Opening characters for a comment */
			{
				fscanf(parameters_file, "%s", input_string) ;
				while (strcmp(input_string, "*/") != 0)
					fscanf(parameters_file, "%s", 
					input_string) ;
				goto L1 ;
			}
			
			case 1: /* "PROBABILITIES" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					probabilities = ON ;
				else if (strcmp(input_string, "OFF") == 0)
					probabilities = OFF ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 2: /* "SHOW_AL_STRUCTURE" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					show_al_structure = ON ;
				else if (strcmp(input_string, "OFF") == 0)
					show_al_structure = OFF ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 3: /* "SHOW_HIT_STRUCTURE" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					show_hit_structure = ON ;
				else if (strcmp(input_string, "OFF") == 0)
					show_hit_structure = OFF ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 4: /* "SHOW_UNSELECTED_ALIGNMENTS" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					show_unselected_alignments = ON ;
				else if (strcmp(input_string, "OFF") == 0)
					show_unselected_alignments = OFF ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 5: /* "VERBOSE" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					verbose = ON ;
				else if (strcmp(input_string, "OFF") == 0)
					verbose = OFF ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 6: /* "USE_ABBREVIATIONS" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					use_abbreviations = true ;
				else if (strcmp(input_string, "OFF") == 0)
					use_abbreviations = false ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 7: /* "SHOW_ALL_PARSING_ALIGNMENTS" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					show_all_parsing_alignments = true ;
				else if (strcmp(input_string, "OFF") == 0)
					show_all_parsing_alignments = false ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 8: /* "ID_C_SYMBOL_CONSTRAINT" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					id_c_symbol_constraint = ON ;
				else if (strcmp(input_string, "OFF") == 0)
					id_c_symbol_constraint = OFF ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}

			case 9: /* "RECOGNISE_BRACKETS" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					recognise_brackets = ON ;
				else if (strcmp(input_string, "OFF") == 0)
					recognise_brackets = OFF ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}

			case 10: /* "RECOGNISE_XML_BRACKETS" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					recognise_xml_brackets = ON ;
				else if (strcmp(input_string, "OFF") == 0)
					recognise_xml_brackets = OFF ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 11: /* "BRACKETS_ID_SYMBOL_CONSTRAINT" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					brackets_id_symbol_constraint = true ;
				else if (strcmp(input_string, "OFF") == 0)
					brackets_id_symbol_constraint = false ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 12: /* "MATCH_ALL_DATA_SYMBOLS" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					match_all_data_symbols = true ;
				else if (strcmp(input_string, "OFF") == 0)
					match_all_data_symbols = false ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 13: /* "MATCH_ALL_DATA_SYMBOLS_IN_COMPOSITE_ALIGNMENTS" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					match_all_data_symbols_in_composite_alignments = true ;
				else if (strcmp(input_string, "OFF") == 0)
					match_all_data_symbols_in_composite_alignments = false ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 14: /* "MATCH_ALL_DATA_SYMBOLS_IN_HIERARCHICAL_ALIGNMENTS" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					match_all_data_symbols_in_hierarchical_alignments = true ;
				else if (strcmp(input_string, "OFF") == 0)
					match_all_data_symbols_in_hierarchical_alignments = false ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 15: /* "MATCH_ALL_NEW_SYMBOLS" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					match_all_new_symbols = true ;
				else if (strcmp(input_string, "OFF") == 0)
					match_all_new_symbols = false ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 16: /* "COMBINE_ALIGNMENTS" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					combine_alignments = true ;
				else if (strcmp(input_string, "OFF") == 0)
					combine_alignments = false ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 17: /* "ONLY_ONE_NEW_PATTERN_IN_EACH_ALIGNMENT" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					only_one_new_pattern_in_each_alignment = true ;
				else if (strcmp(input_string, "OFF") == 0)
					only_one_new_pattern_in_each_alignment = false ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}

			case 18: /* "ALIGNMENT_FORMAT" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "H") == 0)
					alignment_format = 'H' ;
				else if (strcmp(input_string, "V") == 0)
					alignment_format = 'V' ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}

			case 19: /* "HIT_STRUCTURE_ROWS" */
			{
				fscanf(parameters_file, "%d", 
					&hit_structure_rows) ;
				goto L1 ;
			}
			
			case 20: /* "FAIL_SCORE" */
			{
				fscanf(parameters_file, "%d", &fail_score) ;
				goto L1 ;
			}
			
			case 21: /* "COST_FACTOR" */
			{
				fscanf(parameters_file, "%lg", &cost_factor) ;
				goto L1 ;
			}
			
			case 22: /* "KEEP_ROWS" */
			{
				fscanf(parameters_file, "%d", 
					&keep_rows) ;
				if (keep_rows > MAX_NEW_ARRAY_ROWS)
					abort_run("keep_rows too large") ;
				goto L1 ;
			}

			case 23: /* "DRIVING_KEEP_ROWS" */
			{
				fscanf(parameters_file, "%d", 
					&driving_keep_rows) ;
				goto L1 ;
			}
			
			case 24: /* "BEST_ALIGNMENT_PRINT_LIMIT" */
			{
				fscanf(parameters_file, "%d", 
					&best_alignment_print_limit) ;
				goto L1 ;
			}

			case 25: /* "FIGURE_ID" */
			{
				fscanf(parameters_file, "%s", figure_ID) ;
				goto L1 ;
			}
			
			case 26: /* "MAX_NEW_GAP" */
			{
				fscanf(parameters_file, "%d", &max_new_gap) ;
				goto L1 ;
			}
			
			case 27: /* "MAX_OLD_GAP" */
			{
				fscanf(parameters_file, "%d", &max_old_gap) ;
				goto L1 ;
			}
			
			case 28: /* "MAX_UNSUPPORTED_CYCLES" */
			{
				fscanf(parameters_file, "%d",
					&max_unsupported_cycles) ;
				goto L1 ;
			}
			
			case 29: /* "MAX_ALIGNMENTS_IN_ONE_CYCLE" */
			{
				fscanf(parameters_file, "%d",
					&max_alignments_in_one_cycle) ;
				goto L1 ;
			}
			
			case 30: /* "ORIENTATION" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "LANDSCAPE") == 0)
					orientation = LANDSCAPE ;
				else if (strcmp(input_string, "PORTRAIT") == 0)
					orientation = PORTRAIT ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}
			
			case 31: /* "WRITE_AL_CHARS_LENGTH" */
			{
				fscanf(parameters_file, "%d", 
					&write_al_chars_length) ;
				
				goto L1 ;
			}
			
			case 32: /* "WRITE_SECTION_CHARS_LENGTH" */
			{
				fscanf(parameters_file, "%d", 
					&write_section_chars_length) ;
				
				goto L1 ;
			}
			
			case 33: /* "MINIMUM_FONT_HEIGHT" */
			{
				fscanf(parameters_file, "%lg",
					&minimum_font_height) ;
				
				// Find the smallest available font size
				// which is larger than or equal to the 
				// minimum_font_height and re-set the value
				// of minimum font height to this value.
				
				for (i = 0; i < FONT_SET_SIZE; i++)
				{
					if (font_heights[i] >= 
						minimum_font_height)
					{
						j = i ;
						break ;
					}
				}
				
				if (i >= FONT_SET_SIZE)
				abort_run("MINIMUM_FONT_HEIGHT is \
													larger than the largest \
													available font size") ;
				
				if (minimum_font_height == font_heights[j])
				{
					fprintf(output_file, "%s%1.2f%s",
						"MINIMUM_FONT_HEIGHT is ",
						minimum_font_height,
						" pt\n\n") ;
				}
				else
				{
					fprintf(output_file, "%s%1.2f%s",
						"MINIMUM_FONT_HEIGHT changed from ",
						minimum_font_height,
						" pt to ") ;
					minimum_font_height = font_heights[j] ;
					fprintf(output_file, "%1.2f%s",
						minimum_font_height,
						" pt\n\n") ;
				}
				
				goto L1 ;
			}
			
			case 34: /* "MAXIMUM_FONT_HEIGHT" */
			{
				fscanf(parameters_file, "%lg",
					&maximum_font_height) ;
				
				// Find the largest available font size
				// which is smaller than or equal to the 
				// maximum_font_height and re-set the value
				// of maximum font size to this value.
				
				for (i = FONT_SET_SIZE - 1; i >= 0; i--)
				{
					if (font_heights[i] <= 
						maximum_font_height)
					{
						j = i ;
						break ;
					}
				}
				
				if (i < 0) abort_run("MAXIMUM_FONT_HEIGHT is \
					smaller than the smallest available font size") ;
				
				if (maximum_font_height == font_heights[j])
				{
					fprintf(output_file, "%s%1.2f%s",
						"MAXIMUM_FONT_HEIGHT is ",
						maximum_font_height,
						" pt\n\n") ;
				}
				else
				{
					fprintf(output_file, "%s%1.2f%s",
						"MAXIMUM_FONT_HEIGHT changed from ",
						maximum_font_height,
						" pt to ") ;
					maximum_font_height = font_heights[j] ;
					fprintf(output_file, "%1.2f%s",
						maximum_font_height,
						" pt\n\n") ;
				}
				
				goto L1 ;
			}
			
			case 35: /* "WINDOW_SIZE" */
			{
				fscanf(parameters_file, "%d", 
					&window_size) ;
				
				goto L1 ;
			}
			
			case 36: /* "COMBINATION_LIMIT" */
			{
				fscanf(parameters_file, "%d", &combination_limit) ;
				goto L1 ;
			}
			
			case 37: /* "AUDIT_TRAIL" */
			{
				fscanf(parameters_file, "%s", input_string) ;
				if (strcmp(input_string, "ON") == 0)
					audit_trail = true ;
				else if (strcmp(input_string, "OFF") == 0)
					audit_trail = false ;
				else
				{
					fprintf(output_file, "%s%s",
						input_string, "\n\n") ;
					abort_run("Invalid parameter") ;
				}
				goto L1 ;
			}

			default:
			{
				fprintf(output_file, "%d%s",
					parameter_index, "\n\n") ;
				abort_run("Default reached in read_parameters");
			}
		}
		L1: fscanf(parameters_file, "%s", input_string) ;
	}
	
	if (maximum_font_height < minimum_font_height)
		abort_run("MAXIMUM_FONT_HEIGHT is smaller than MINIMUM_FONT_HEIGHT") ;

	if (recognise_brackets && recognise_xml_brackets)
		abort_run("recognise_brackets and recognise_xml_brackets should not \
			both be true.") ;

	if (brackets_id_symbol_constraint == true
			&& recognise_brackets == false)
		abort_run("Invalid combination of brackets_id_symbol_constraint \
			and recognise_brackets.") ;

	if (driving_keep_rows > keep_rows)
		abort_run("DRIVING_KEEP_ROWS must be <= KEEP_ROWS.") ;

	if (combine_alignments == true
		&& only_one_new_pattern_in_each_alignment == false)
	{
		fprintf(output_file, "%s%s",
			"COMBINE_ALIGNMENTS has been set to OFF because\n",
			"ONLY_ONE_NEW_PATTERN_IN_EACH_ALIGNMENT is OFF.\n\n") ;

		combine_alignments = false ;
	}
} /* read_parameters */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  Returns false if the next character in the `syntax buffer' isn't an SP
**  opening bracket, otherwise the object type associated with the bracket
**  is returned.
**
** CALLING SEQUENCE:
**
**  int left_bracket(void) ;
**
** FORMAL ARGUMENTS:
**
**  Return value:   false if not an SP opening bracket, corresponding object
**                  type (GROUP, SEQUENCE) otherwise.
**
** IMPLICIT INPUTS:
**
**  Functions: characters
**
** IMPLICIT OUTPUTS:
**
**  NONE
**
** SIDE EFFECTS:
**
**  NONE
*/

int left_bracket()
{
    int b1 = false;
    
    if (characters("(")) b1 = SEQUENCE ;
    else if (characters("[")) b1 = GROUP ;
    return(b1) ;
} /* left_bracket */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  Returns true if the character in the `syntax buffer' is the closing
**  bracket of the object type specified, false otherwise.
**
** CALLING SEQUENCE:
**
**  bool right_bracket(int b1)
**
** FORMAL ARGUMENTS:
**
**  Return value:   true if corresponding closing bracket, false otherwise.
**
**  b1:         Object type.
**
** IMPLICIT INPUTS:
**
**  Functions: characters
**
** IMPLICIT OUTPUTS:
**
**  NONE
**
** SIDE EFFECTS:
**
**  NONE
*/

bool right_bracket(int b1)
{
    if (b1 == SEQUENCE && characters(")")) return(true) ;
    else if (b1 == GROUP && characters("]")) return(true) ;
    return(false) ;
} /* right_bracket */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  Make a symbol out of a string of characters in the `syntax buffer'.
**  true is returned if a valid symbol was constructed, false otherwise.  
**  A valid symbol includes "". If the
**  symbol is too long, an error is raised and the program aborts.
**
** CALLING SEQUENCE:
**
**  bool rec_symbol(char *nme, bool *symbol_is_identifier) ;
**
** FORMAL ARGUMENTS:
**
**  Return value:           true if `nme' is valid, false otherwise.
**
**  nme:                    String to receive the symbol. 
**                          NB: Call by reference - symbol value is
**                          passed back through this parameter.
**  symbol_is_identifier:   Pointer to an external variable which is set in
**                          the function to true if the given
**                          symbol is IDENTIFICATION, false otherwise.
**
** IMPLICIT INPUTS:
**
**  NONE
**
** IMPLICIT OUTPUTS:
**
**  syn_buf_ptr
**
** SIDE EFFECTS:
**
**  The string `nme' must be declared such that it is of sufficient size
**  to accept the symbol created.  Also note that `syn_buf_ptr' is
**  updated.
*/

bool rec_symbol(char *nme, bool *symbol_is_identifier) /* Makes a
    symbol out of a string of characters */
{
    char c;
    int i = 0;
    
    c = *syn_buf_ptr;
    if (c == '\"') /* A quoted string in the input */
    {
        c = *(++syn_buf_ptr) ;
        while (c != '\"')
        {
            if (i >= (LONG_STRING_LENGTH - 2))
                abort_run("Symbol too long in 'symbol'") ;
            *(nme + i++) = c;
            c = *(++syn_buf_ptr) ;
        }
        syn_buf_ptr++;
    }
    else
    {
        if (c == '!')
        {
            identification_symbols_marked = true ;
            *symbol_is_identifier = true ;
            c = *(++syn_buf_ptr) ;
        }

        while (c != ' ' && c != ',' && c != '('
            && c != ')' && c != '[' && c != ']')
        {
            if (i >= (LONG_STRING_LENGTH - 2))
                abort_run("Symbol too long in 'symbol'") ;
            nme[i++] = c ;
            c = *(++syn_buf_ptr) ;
        }
    }
    if (i == 0) return false ;
    nme[i] = '\0' ;
    scan() ;
    return(true) ;
} /* rec_symbol */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  Recognises an `object'.
**
** CALLING SEQUENCE:
**
**  base_object *find_node() ;
**
** FORMAL ARGUMENTS:
**
**  Return value:   A new object if object recognised, NIL otherwise.
**
** IMPLICIT INPUTS:
**
**  Functions:  comp_object, symbol, scan
**
**  Variables:  syn_buf_ptr
**
** IMPLICIT OUTPUTS:
**
**  syn_buf_ptr
**
** SIDE EFFECTS:
**
**  NONE?
*/

/* object

  syntax -> object | bad_syntax ;
  object -> comp_object | simple_object ;
  comp_object -> SEQUENCE | GROUP ;
  SEQUENCE -> '(' , body , ')' ;
  GROUP -> '[' , body , ']' ;
  body -> b | NIL ;
  b -> object , body ;
  simple_object -> SYMBOL ;
  SYMBOL -> alphanumeric(+) ;
  
*/

base_object *find_node()
{
    char nme[LONG_STRING_LENGTH] ;
    base_object *node ;
    symbol *symbol1 ;
    bool symbol_is_identifier = false ;
    
    if (node = comp_object())
    {
        read_frequency(node) ;
        return(node) ;
    }
    else // Object is a symbol 
    if (rec_symbol(nme, &symbol_is_identifier))
    {
        symbol1 = new symbol(nme, 1, NULL_VALUE) ;
        if (symbol_is_identifier) symbol1->set_status(IDENTIFICATION) ;
        else symbol1->set_status(CONTENTS) ;

        read_frequency(symbol1) ;
        return(symbol1) ;
    }
    else return(NIL) ;

} /* find_node */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  Recognises a tree_object.
**
** CALLING SEQUENCE:
**
**  tree_object *comp_object() ;
**
** FORMAL ARGUMENTS:
**
**  Return value:   A new node if composite object, NIL otherwise.
**
** IMPLICIT INPUTS:
**
**  syntax_buffer
**
** IMPLICIT OUTPUTS:
**
**  NONE
**
** SIDE EFFECTS:
**
**  NONE
*/

tree_object *comp_object()
{
    int b1 ;
    tree_object *new_node ;
    base_object *child ;
    
    if (b1 = left_bracket())
    {
        if (b1 == SEQUENCE) new_node = new sequence ;
        else new_node = new group ;
        while (child = find_node())
            new_node->add_child(child) ;
        if (!right_bracket(b1))
        {
            delete new_node ;
            return(NIL) ;
        }
        new_node->mark_parent_and_int_positions_non_recursive();
        // take care! This alters the value of
        // the 'parent' field.
        new_node->set_number_of_children(new_node->
            count_number_of_children()) ;
        return(new_node) ;
    }
    else return(NIL) ;
} /* comp_object */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**  Input the SP corpus.
**
** CALLING SEQUENCE:
**
**  tree_object *top_object() ;
**
** FORMAL ARGUMENTS:
**
**  Return value:   node if input is correct, false otherwise.
**
** IMPLICIT INPUTS:
**
**  NONE
**
** IMPLICIT OUTPUTS:
**
**  syn_buf_ptr
**
** SIDE EFFECTS:
**
**  None, except see IMPLICIT OUTPUTS above.
*/

tree_object *top_object()
{
    tree_object *node ;
    
    syn_buf_ptr = syntax_buffer;
    scan() ;
    if (node = (tree_object *)find_node()) return(node) ;
    else return(NIL) ;
} /* top_object */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**		Load the SP corpus from the text file specified (using
**	the `syntax buffer' size specified), converting it to
**	it's internal format during the process.
**
** CALLING SEQUENCE:
**
**	void load() ;
**
** FORMAL ARGUMENTS:
**
**	Return value:	None.
**
** IMPLICIT INPUTS:
**
**	Input to load() from parameters file.
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void load()
{
    /* Read in syntax. */
    
    syntax_buffer = new char[BUFSIZE] ;
    if (syntax_buffer == NIL)
        abort_run("Out of memory in 'load'") ;
    
    in_syntax() ;
    
    identification_symbols_marked = false ;
    
    if (corpus = (group *)top_object())
        fprintf(output_file, "Corpus loaded successfully.\n\n") ;
    else
    {
        bad_syntax() ;
        abort_run("Error in syntax in 'load'") ;
    }

    delete[] syntax_buffer ;

} /* load */

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	A function for incrementing an integer value by 1 and, at
**	the same time, checking whether it equals or exceeds a specified limit.
**	If it does exceed the limit, the program is aborted with a
**	specified message.
**
** CALLING SEQUENCE:
**
**	void plus_one(int *var1, int limit, char *message)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void.
**
**	var1:			A pointer to an external variable holding
**					the input value.
**	limit:			The ceiling which must not be equalled or exceeded.
**	message:		The message to go to write file if the ceiling is
**					breached.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void plus_one(int *var1, int limit, char *message)
{
	(*var1)++ ;
	if (*var1 >= limit) abort_run(message) ;
} // plus_one

/*********************************************************************/
